@@ -42,315 +42,482 @@ function project {
 ########################################################################
 # MathComp
 ########################################################################
-project mathcomp 'https://github.com/math-comp/math-comp' '6dc4a759117dc29092208d6f175289cf0274bae2'
+project mathcomp 'https://github.com/math-comp/math-comp' '28837b16a2358f97d31424fc332498a4780c7afe'
+# Contact @CohenCyril, @proux01 on github
 
-project fourcolor 'https://github.com/math-comp/fourcolor' 'd6127d4e02fdb141131f3aa07fa4d3d4baf71529'
+project mathcomp_1 'https://github.com/math-comp/math-comp' '26c1db2a83deffba627719c3e9b754b2bb83e83b'
+# Contact @CohenCyril, @proux01 on github
 
-project oddorder 'https://github.com/math-comp/odd-order' '833261a01fd0c62b05ccbadfc0c682e0bc16a5e9'
+project fourcolor 'https://github.com/math-comp/fourcolor' '43719c0fb5fb6cb0c8fc1c2db09efc632c23df90'
+# Contact @ybertot, @proux01 on github
 
-project mczify 'https://github.com/math-comp/mczify' '5d6abf13b137d8499e59b12e7e216f7cc12304e2'
+project oddorder 'https://github.com/math-comp/odd-order' '8dbbae0e53a6d1fcf3471c8fae4dd14c8f18bd93'
+# Contact @gares, @proux01 on github
 
-project finmap 'https://github.com/math-comp/finmap' '27563c3d10289e7118f1ae5d40fb1f3e0b8a55ff'
+project mczify 'https://github.com/math-comp/mczify' '371b29c7eb30ce408fa29feaf7b94146eed6660b'
+# Contact @pi8027 on github
+
+project finmap "https://github.com/math-comp/finmap" "cea9f088c9cddea1173bc2f7c4c7ebda35081b60"
+# put back master when Analysis master moves to MathComp 2
+# project finmap "https://github.com/math-comp/finmap" "master"
+# Contact @CohenCyril on github
 
 project bigenough 'https://github.com/math-comp/bigenough' 'ff71b25f31658d80fdae9657e8cf34e5e1052647'
+# Contact @CohenCyril on github
 
-project analysis 'https://github.com/math-comp/analysis' '38ada1e74e69e2d8a3959c7c20cc4e5d1c85c496'
+project analysis 'https://github.com/math-comp/analysis' 'b0d400f667f98cedde05414cff99853eb86642b8'
+# Contact @affeldt-aist, @CohenCyril on github
 
 ########################################################################
 # UniMath
 ########################################################################
-project unimath 'https://github.com/UniMath/UniMath' '0df0949b951e198c461e16866107a239c8bc0a1e'
+project unimath 'https://github.com/UniMath/UniMath' 'ee7af93f47daf41c752e6bd212a7aba78b6a3aaa'
+# Contact @benediktahrens, @m-lindgren, @nmvdw, @rmatthes on github
 
 ########################################################################
 # Unicoq + Mtac2
 ########################################################################
-project unicoq 'https://github.com/unicoq/unicoq' '0c33033e921c8f7f9dcc47078f0de7a4a0c99999'
+project unicoq 'https://github.com/unicoq/unicoq' 'e0486d27f687c8c34f530158fee743b3e7a3c259'
+# Contact @beta-ziliani, @Janno, @mattam82 on github
 
-project mtac2 'https://github.com/Mtac2/Mtac2' 'c7ae44d4d2af65e9d57bb80b8b70ac0b30c3fe93'
+project mtac2 'https://github.com/Mtac2/Mtac2' '71640c37a3685db5c7aab069cede807faadacc25'
+# Contact @beta-ziliani, @Janno, @mattam82 on github
 
 ########################################################################
 # Mathclasses + Corn
 ########################################################################
-project math_classes 'https://github.com/coq-community/math-classes' '754ddc768299afdf6274d03d713907d4d0282361'
+project math_classes 'https://github.com/coq-community/math-classes' 'f9712ba67b59c90446522e102c87200c7da756ec'
+# Contact @spitters on github
 
-project corn 'https://github.com/coq-community/corn' 'd4e6db77c5082e62a643f19105d6e7ee5547df29'
+project corn 'https://github.com/coq-community/corn' '79e7ee7cc878cd98077d16cfbac12b8133a13805'
+# Contact @spitters on github
 
 ########################################################################
 # Iris
 ########################################################################
 
 # NB: stdpp and Iris refs are gotten from the opam files in the Iris
-# and lambdaRust repos respectively.
+# and iris_examples repos respectively.
 project stdpp "https://gitlab.mpi-sws.org/iris/stdpp" ""
+# Contact @RalfJung, @robbertkrebbers on github
 
 project iris "https://gitlab.mpi-sws.org/iris/iris" ""
+# Contact @RalfJung, @robbertkrebbers on github
 
-project autosubst 'https://github.com/coq-community/autosubst' '495d547fa2a3e40da0f7d2a31469ac8caab058d2'
+project autosubst 'https://github.com/coq-community/autosubst' '6ba0acccef68c75f6cca8928706c726754d69791'
+# Contact @RalfJung, @co-dan on github
 
-project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' '7b79d6e2fdbf0cedd80b9346da8b4fc9e7aaa424'
+project iris_examples 'https://gitlab.mpi-sws.org/iris/examples' '7c5ce0d556f4e03f27ddc6463462bf4040f5103d'
+# Contact @RalfJung, @robbertkrebbers on github
 
 ########################################################################
 # HoTT
 ########################################################################
-project hott 'https://github.com/HoTT/HoTT' '3fcad92c7686345f1037cb833146b3963aabd9d8'
+project hott 'https://github.com/HoTT/HoTT' '191c10ed39de2fa9821e63b6578f4fba9175ab31'
+# Contact @Alizter, @jdchristensen on github
 
 ########################################################################
 # CoqHammer
 ########################################################################
-project coqhammer 'https://github.com/lukaszcz/coqhammer' '4ac485677e2a313279fc64829f1963d92e0ca154'
-
-########################################################################
-# GeoCoq
-########################################################################
-project geocoq 'https://github.com/GeoCoq/GeoCoq' '09a02dc56715b3308689843dd872209262beb5af'
+project coqhammer 'https://github.com/lukaszcz/coqhammer' '4131a8ccac570336b5ffefd26fd40d879f1e115f'
+# Contact @lukaszcz on github
 
 ########################################################################
 # Flocq
 ########################################################################
-project flocq 'https://gitlab.inria.fr/flocq/flocq' '3c92e3f9461c382d95f43e04815c0070be687335'
+project flocq 'https://gitlab.inria.fr/flocq/flocq' '3e8b2e09d79ce3c150a732dd446610d63dabc759'
+# Contact @silene on github
 
 ########################################################################
 # coq-performance-tests
 ########################################################################
-project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' '4aaef74e5742fe3ad87c5d18d735e82b1284ecc6'
+project coq_performance_tests 'https://github.com/coq-community/coq-performance-tests' 'c5e2f8d0b0057a003dcaec727ba01f38e519bf6d'
+# Contact @JasonGross on github
 
 ########################################################################
 # coq-tools
 ########################################################################
-project coq_tools 'https://github.com/JasonGross/coq-tools' '1f1c77c9e24de645418fd476493e48d6d930a69e'
+project coq_tools 'https://github.com/JasonGross/coq-tools' '32dcc6b48607c34442e64789eaaf894dac3dd020'
+# Contact @JasonGross on github
 
 ########################################################################
 # Coquelicot
 ########################################################################
-project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' '9f6f576844a53c930609883a2850f109fb76e965'
+project coquelicot 'https://gitlab.inria.fr/coquelicot/coquelicot' '218b5cca041712775999f3c5ab0cac2f99defb7f'
+# Contact @silene on github
 
 ########################################################################
 # CompCert
 ########################################################################
-project compcert "https://github.com/AbsInt/CompCert" "5be9ae2235b16239f023a36679b7d515dd774c68"
+project compcert 'https://github.com/AbsInt/CompCert' 'c7b27e30a9057c14d5a02a9f9aa3d8f230715eb0'
+# Contact @xavierleroy on github
 
 ########################################################################
 # VST
 ########################################################################
-project vst "https://github.com/PrincetonUniversity/VST" "d635e62a083ad78c16548c29904c527a74f7b1e5"
+project vst 'https://github.com/PrincetonUniversity/VST' '6d2eb96dd42a4d3ee398eef8224e6a8ccca5c750'
+# Contact @andrew-appel on github
 
 ########################################################################
 # cross-crypto
 ########################################################################
-project cross_crypto 'https://github.com/mit-plv/cross-crypto' '27d253a606e62a4746571ae026c08ce6c170efff'
+project cross_crypto 'https://github.com/mit-plv/cross-crypto' 'c46bbe27ef7a05a82f744230c6f2d5f1fdefdaaa'
+# Contact @andres-erbsen on github
 
 ########################################################################
 # rewriter
 ########################################################################
-project rewriter 'https://github.com/mit-plv/rewriter' 'de56173b42be33ea3bcf4f96a491bd5c2b30892e'
+project rewriter 'https://github.com/mit-plv/rewriter' '026f87bfbc5d235c3f7f31b24199305d616daff3'
+# Contact @JasonGross on github
 
 ########################################################################
 # fiat_parsers
 ########################################################################
-project fiat_parsers 'https://github.com/mit-plv/fiat' 'ddb9d1c56c47e684f52272a2d6cc71749a010903'
+project fiat_parsers 'https://github.com/mit-plv/fiat' '157f13408ef9731ccf0829fbcac3c11b9856fb81'
+# Contact @JasonGross on github
 
 ########################################################################
 # fiat_crypto
 ########################################################################
-project fiat_crypto "https://github.com/mit-plv/fiat-crypto" "671cf2894c3d72a145d4cc9193f8108b61a8d402"
+project fiat_crypto 'https://github.com/mit-plv/fiat-crypto' '4d9fadc163b8a09d75a6836e08f52f051fa489d4'
+# Contact @andres-erbsen, @JasonGross on github
 
 ########################################################################
 # fiat_crypto_legacy
 ########################################################################
-project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '1fce11eade067afa33d08f2cc7596d0fb68afa1a'
+project fiat_crypto_legacy 'https://github.com/mit-plv/fiat-crypto' '47058dc312a5cad924972056cdfc4de06c432a82'
+# Contact @JasonGross on github
 
 ########################################################################
 # coq_dpdgraph
 ########################################################################
-project coq_dpdgraph 'https://github.com/Karmaki/coq-dpdgraph' 'fbbf464f9a164d915d433015fee32c3ace7c257d'
+project coq_dpdgraph 'https://github.com/coq-community/coq-dpdgraph' '8452ebdebb34f66c0b87b39b757784090772fd49'
+# Contact @Karmaki, @ybertot on github
 
 ########################################################################
 # CoLoR
 ########################################################################
-project color 'https://github.com/fblanqui/color' '4ebc8c56901fa79decea4661eb698267d7e5d050'
+project color 'https://github.com/fblanqui/color' '42f4b271d41b81d59508936a375ac2f9fafe6b9c'
+# Contact @fblanqui on github
 
 ########################################################################
 # TLC
 ########################################################################
-project tlc 'https://github.com/charguer/tlc' 'f4835e72393f84892c8bbf9246cae1dd7e5cde5d'
+project tlc 'https://github.com/charguer/tlc' '1f1df8cfbdfb1838a446fa1614c8572486c0478f'
+# Contact @charguer on github
 
 ########################################################################
 # Bignums
 ########################################################################
-project bignums 'https://github.com/coq/bignums' 'def90f1bc835b938e894bbbec58758e56e8b8bea'
+project bignums 'https://github.com/coq/bignums' 'cb65c878e764793081ead5fdcaffbe018dd713c3'
+# Contact @erikmd, @proux01 on github
 
 ########################################################################
 # coqprime
 ########################################################################
-project coqprime 'https://github.com/thery/coqprime' '762c7f5e8c5fa13c85020d5b24b7910843266f00'
+project coqprime 'https://github.com/thery/coqprime' 'd5935ca3b7f3d2f738b0810a49858d17480d6a58'
+# Contact @thery on github
 
 ########################################################################
 # bbv
 ########################################################################
-project bbv 'https://github.com/mit-plv/bbv' '18c77dac3535e71a47e3cc27efc82a76169879aa'
+project bbv 'https://github.com/mit-plv/bbv' 'f4caa0504125460afa41fb92c9071dd628154d8b'
+# Contact @JasonGross, @samuelgruetter on github
 
 ########################################################################
 # bedrock2
 ########################################################################
-project bedrock2 'https://github.com/mit-plv/bedrock2' '16c1ac1b717ec833d3d32963660fa91a4169ea2f'
+project bedrock2 'https://github.com/mit-plv/bedrock2' '65988d2662e522b0a7dcd04ce66c47acb5c66df1'
+# Contact @samuelgruetter, @andres-erbsen on github
+
+########################################################################
+# Coinduction
+########################################################################
+project coinduction 'https://github.com/damien-pous/coinduction' '65d020794e817fcf0295566abe1be454ad3284e4'
+# Contact @damien-pous on github
+
+########################################################################
+# coq-lsp
+########################################################################
+project coq_lsp 'https://github.com/ejgallego/coq-lsp' '75a92e0abeb29cbc3f75bbffa60db04857789070'
+# Contact @ejgallego on github
 
 ########################################################################
 # Equations
 ########################################################################
-project equations 'https://github.com/mattam82/Coq-Equations' '80767d571837914218bcde040c3071a483f75a72'
+project equations 'https://github.com/mattam82/Coq-Equations' 'a517a3e128719b078397a415dc9ce6b856019245'
+# Contact @mattam82 on github
 
 ########################################################################
 # Elpi + Hierarchy Builder
 ########################################################################
-project elpi 'https://github.com/LPCIC/coq-elpi' '9448eb21d3f8a3b663dbe7f85e7efa7bb1722400'
+project elpi 'https://github.com/LPCIC/coq-elpi' 'b753aa43a643ad7c0b2722d9850253b9d737fec7'
+# Contact @gares on github
 
-project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'ccf9db17d0457268e7dac0f0ee17003cd95f5a77'
+project hierarchy_builder 'https://github.com/math-comp/hierarchy-builder' 'b4e8cbce83cff0f817494adc625a127bd4114b2e'
+# Contact @CohenCyril, @gares on github
 
 ########################################################################
 # Engine-Bench
 ########################################################################
 project engine_bench 'https://github.com/mit-plv/engine-bench' '8c0b15abc38b1d3f7dc606f4eeef9fba1a986b05'
+# Contact @JasonGross on github
 
 ########################################################################
 # fcsl-pcm
 ########################################################################
-project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' 'a654207feab3c08150ae79ac929883fd04ef0191'
+project fcsl_pcm 'https://github.com/imdea-software/fcsl-pcm' '35cb0a2235adb656c2696e33512f0c4555ad753b'
+# Contact @aleksnanevski, @clayrat on github
 
 ########################################################################
 # ext-lib
 ########################################################################
-project ext_lib 'https://github.com/coq-community/coq-ext-lib' 'd48e6583c4fe01479f743ea21d9b921073cb3f5b'
+project ext_lib 'https://github.com/coq-community/coq-ext-lib' '00d3f4e2a260c7c23d2c0b9cbc69516f8be4ac92'
+# Contact @gmalecha, @liyishuai on github
 
 ########################################################################
 # simple-io
 ########################################################################
-project simple_io 'https://github.com/Lysxia/coq-simple-io' '403db47e0b96222cd3c89c48ed5f8da10319da77'
+project simple_io 'https://github.com/Lysxia/coq-simple-io' '7b1e6de2c2028a35776f080dc3db117cae117836'
+# Contact @Lysxia, @liyishuai on github
 
 ########################################################################
 # quickchick
 ########################################################################
-project quickchick 'https://github.com/QuickChick/QuickChick' '81d8ffe90d6934ccbadb0c1468f8c3bc3627cc81'
+project quickchick 'https://github.com/QuickChick/QuickChick' '7c570ed650f770e29ea6a84d49687b1c0c2b6b72'
+# Contact @lemonidas, @Lysxia, @liyishuai on github
 
 ########################################################################
 # reduction-effects
 ########################################################################
-project reduction_effects 'https://github.com/coq-community/reduction-effects' 'fc926669571e8315013eb0775e27f9653798b7f0'
+project reduction_effects 'https://github.com/coq-community/reduction-effects' 'f0570f498bc8a0d25e878115b4066b140908c4b4'
+# Contact @liyishuai, @JasonGross on github
 
 ########################################################################
 # menhirlib
 ########################################################################
 # Note: menhirlib is now in subfolder coq-menhirlib of menhir
-project menhirlib "https://gitlab.inria.fr/fpottier/menhir" "20220210"
+project menhirlib 'https://gitlab.inria.fr/fpottier/menhir' 'bb68b8fdb186266dbed67599cfa69a69a3c01272'
+# Contact @fpottier, @jhjourdan on github
+
+########################################################################
+# coq-neural-net-interp
+########################################################################
+project neural_net_interp 'https://github.com/JasonGross/neural-net-coq-interp' '06509a99be412d06847de938a19e02be64f6fc4c'
+# Contact @JasonGross on github
 
 ########################################################################
 # aac_tactics
 ########################################################################
-project aac_tactics 'https://github.com/coq-community/aac-tactics' '6e02fa08c62f38164aae355fe49fe6b923f04323'
+project aac_tactics 'https://github.com/coq-community/aac-tactics' '059dd94f4fbfe23e93d5c3de800cfb71f6a0ed43'
+# Contact @palmskog on github
 
 ########################################################################
 # paco
 ########################################################################
-project paco 'https://github.com/snu-sf/paco' '4f140d2083632e069e9f0f0cdd82171b60daf0de'
+project paco 'https://github.com/snu-sf/paco' '5c5693f46c8957f36a2349a0d906e911366136de'
+# Contact @minkiminki on github
+
+########################################################################
+# coq-itree
+########################################################################
+project itree 'https://github.com/DeepSpec/InteractionTrees' 'dda104937d79e2052d1a26f6cbe89429245ff743'
+# Contact @Lysxia on github
+
+########################################################################
+# coq-itree_io
+########################################################################
+project itree_io 'https://github.com/Lysxia/coq-itree-io' 'af0326793a19f142eba800dba6044143b108ceaa'
+# Contact @Lysxia, @liyishuai on github
+
+########################################################################
+# coq-ceres
+########################################################################
+project ceres 'https://github.com/Lysxia/coq-ceres' 'f61b24d48222db0100de19f88c19151a3aeb826f'
+# Contact @Lysxia on github
+
+########################################################################
+# coq-parsec
+########################################################################
+project parsec 'https://github.com/liyishuai/coq-parsec' '2efb4437f8451dfbeb174368f860e629135c08ab'
+# Contact @liyishuai on github
+
+########################################################################
+# coq-json
+########################################################################
+project json 'https://github.com/liyishuai/coq-json' '2d1d11eb1be9ad96614d9cd224c1df7c75ec2869'
+# Contact @liyishuai on github
+
+########################################################################
+# coq-async-test
+########################################################################
+project async_test 'https://github.com/liyishuai/coq-async-test' '0637b95ae52060d8a808261ca97890d03c9a4503'
+# Contact @liyishuai on github
+
+########################################################################
+# coq-http
+########################################################################
+project http 'https://github.com/liyishuai/coq-http' 'cabde79a4a0d978d031475c7443be7fd43a711c5'
+# Contact @liyishuai on github
 
 ########################################################################
 # paramcoq
 ########################################################################
-project paramcoq 'https://github.com/coq-community/paramcoq' '308614a72e89ee62cdd0d98a28d8fc4812f58324'
+project paramcoq 'https://github.com/coq-community/paramcoq' 'a61b1b3f8cea3fce6c904a0b8da6236345ed896f'
+# Contact @proux01 on github
 
 ########################################################################
 # relation_algebra
 ########################################################################
-project relation_algebra 'https://github.com/damien-pous/relation-algebra' '649a44282b19864a4aec28bcaa5453da1b9512df'
+project relation_algebra 'https://github.com/damien-pous/relation-algebra' '38cfe670448c90406fb72ff40526ad4d08b75319'
+# Contact @damien-pous on github
 
 ########################################################################
 # StructTact + InfSeqExt + Cheerios + Verdi + Verdi Raft
 ########################################################################
-project struct_tact 'https://github.com/uwplse/StructTact' '2ba456d80a1422bfed3e46524e8f56b436b1bd4a'
+project struct_tact 'https://github.com/uwplse/StructTact' '97268e11564c8fe59aa72b062478458d7aa53e9d'
+# Contact @palmskog on github
 
-project inf_seq_ext 'https://github.com/DistributedComponents/InfSeqExt' '41f02fefeb550d453be777fd246564f300835785'
+project inf_seq_ext 'https://github.com/DistributedComponents/InfSeqExt' '601e89ec019501c48c27fcfc14b9a3c70456e408'
+# Contact @palmskog on github
 
-project cheerios 'https://github.com/uwplse/cheerios' '48bdcefec49916116a418300672b0bd028ff91db'
+project cheerios 'https://github.com/uwplse/cheerios' '5c9318c269f9cae1c1c6583a44405969ac0be0dd'
+# Contact @palmskog on github
 
-project verdi 'https://github.com/uwplse/verdi' '26f1f4de5aeaa7043978cc18b3a30834818f484c'
+project verdi 'https://github.com/uwplse/verdi' 'f082f86b2a0cb56cc0703a79da351d8b9a6e7468'
+# Contact @palmskog on github
 
-project verdi_raft 'https://github.com/uwplse/verdi-raft' 'ce31a56f33949c9ac0d88e5bc0f7d84fded1d68f'
+project verdi_raft 'https://github.com/uwplse/verdi-raft' 'a3375e867326a82225e724cc1a7b4758b029376f'
+# Contact @palmskog on github
 
 ########################################################################
 # stdlib2
 ########################################################################
-project stdlib2 'https://github.com/coq/stdlib2' '9dfc4b760ba4ec2ed08732d1fccc904b1f9189dc'
+project stdlib2 'https://github.com/coq/stdlib2' '33212e05c51efa012c9dfccd0b9e735a42f2d924'
+# Contact @maximedenes, @vbgl on github
 
 ########################################################################
 # argosy
 ########################################################################
 project argosy 'https://github.com/mit-pdos/argosy' 'a6a5aa0d3868efd4ada0b40927b5748e5d8967d3'
+# Contact @tchajed on github
+
+########################################################################
+# ATBR
+########################################################################
+project atbr 'https://github.com/coq-community/atbr' '39de8e4c654a530238390704677a672275649d12'
+# Contact @palmskog, @tchajed on github
 
 ########################################################################
 # perennial
 ########################################################################
-project perennial 'https://github.com/mit-pdos/perennial' 'aac2736b67c16b9abf71692086c19f6501ed9afa'
+project perennial 'https://github.com/mit-pdos/perennial' 'd045aca7d0bd3c43ec561179954777f160eed1e4'
+# Contact @upamanyus, @RalfJung, @tchajed on github
+# PRs to fix Perennial failures should be submitted against the Perennial
+# `master` branch. `coq/tested` is automatically updated every night to the
+# `master` branch if CI on `master` is green. This is to avoid breaking Coq CI
+# when Perennial CI breaks.
 
 ########################################################################
 # metacoq
 ########################################################################
-project metacoq 'https://github.com/MetaCoq/metacoq' '859fdccf985c348fe7ee6bf36a9ab0c34410d723'
+project metacoq 'https://github.com/MetaCoq/metacoq' '194116d4f80dc53253f8730aa822c52225731d71'
+# Contact @mattam82, @yforster on github
 
 ########################################################################
 # SF suite
 ########################################################################
-project sf 'https://github.com/DeepSpec/sf' '9f5fb840950dcf54cc4ce0884ffb75c62f9bcd6a'
+project sf 'https://github.com/DeepSpec/sf' 'd7f9ab78cd01f36ce5a2a359dfe4cfe1e160fae9'
+# Contact @bcpierce00, @liyishuai on github
 
 ########################################################################
 # Coqtail
 ########################################################################
-project coqtail 'https://github.com/whonore/Coqtail' '98c99cc950f147a877d56a590d92b357c24086b9'
+project coqtail "https://github.com/whonore/Coqtail" "ecca5017b6b27e1bded009f9725186d941d85b53"
+# Contact @whonore on github
 
 ########################################################################
 # Deriving
 ########################################################################
-project deriving 'https://github.com/arthuraa/deriving' 'c86a0a39ff7e360da52a996d1b2b600e5fd646db'
+project deriving 'https://github.com/arthuraa/deriving' '9d6aea3df446854aa1aa009e50954e2fd0c02981'
+# Contact @arthuraa on github
 
 ########################################################################
 # VsCoq
 ########################################################################
-project vscoq 'https://github.com/maximedenes/vscoq' 'ce56385daed08a2a25130f9696d8c5f03fd94653'
+project vscoq 'https://github.com/coq-community/vscoq' '84032c7f46e0a21a01da91e5364cfc7d4951e297'
+# Contact @rtetley, @gares on github
 
 ########################################################################
 # category-theory
 ########################################################################
-project category_theory 'https://github.com/jwiegley/category-theory' '17f1254d0254afb4f8cc20ea561168a07627d4ee'
+project category_theory 'https://github.com/jwiegley/category-theory' '41ab38d0caffaa5176a6f63810ba8d92816514a4'
+# Contact @jwiegley on github
 
 ########################################################################
 # itauto
 ########################################################################
-project itauto 'https://gitlab.inria.fr/fbesson/itauto' '5b291966e48bd6f2e933977ad45d4d82b0672ae3'
+project itauto 'https://gitlab.inria.fr/fbesson/itauto' 'a527a27d9dd21ac8f69821f35fd677314d5eadbe'
+# Contact @fajb on github
 
 ########################################################################
 # Mathcomp-word
 ########################################################################
-project mathcomp_word 'https://github.com/jasmin-lang/coqword' '9012f08c33a8c0d3b65e6e9494fd792522bf7b5c'
+project mathcomp_word "https://github.com/jasmin-lang/coqword" "537141ac0dbdd0a0ca2b409826d93a6f8398b0b3"
+# Contact @vbgl, @strub on github
+# go back to "main" and change dependency to MC 2 when
+# https://github.com/jasmin-lang/jasmin/pull/560 is merged
 
 ########################################################################
 # Jasmin
 ########################################################################
-project jasmin 'https://github.com/jasmin-lang/jasmin' '9935aaace97f8bc3fe5c12c787432b8308605c64'
+project jasmin 'https://github.com/jasmin-lang/jasmin' 'd195c58958771ba26bc479fef9a5f04fd219290e'
+# Contact @vbgl, @bgregoir on github
 
 ########################################################################
 # Lean Importer
 ########################################################################
-project lean_importer 'https://github.com/SkySkimmer/coq-lean-import' '9e9c13e8b27fd9ba1b5258e89f7c862d36953079'
+project lean_importer 'https://github.com/SkySkimmer/coq-lean-import' '10363695d2f9692b4d8a9e7ee29be444ebe96c80'
+# Contact @SkySkimmer on github
 
 ########################################################################
 # SerAPI
 ########################################################################
-project serapi 'https://github.com/ejgallego/coq-serapi' 'acb183d1a5023a14025cf4b3fcc5430a6194a250'
+project serapi 'https://github.com/ejgallego/coq-serapi' 'ba0bd1d3fb086b7a7c79adc68d1aa53e9d883d42'
+# Contact @ejgallego on github
 
 ########################################################################
 # SMTCoq
 ########################################################################
-project smtcoq 'https://github.com/smtcoq/smtcoq' 'a8ab8231397ce7455826204a96991935c9efc0e0'
+project smtcoq 'https://github.com/smtcoq/smtcoq' 'e24b14768d6f61c915bc25734d769bbe9f044775'
+# Contact @ckeller on github
 
 ########################################################################
 # Stalmarck
 ########################################################################
-project stalmarck 'https://github.com/coq-community/stalmarck' '99fb0fd50c2f6230d29a412f7091013c03bf8a58'
+project stalmarck 'https://github.com/coq-community/stalmarck' '22a05ddcc6a826ffa772282f1d3a0902d3921e7b'
+# Contact @palmskog on github
+
+########################################################################
+# coq-library-undecidability
+########################################################################
+project coq_library_undecidability 'https://github.com/uds-psl/coq-library-undecidability' '91165e2e95d4372d692aca05f5d53ab628e3c8b9'
+# Contact @mrhaandi, @yforster on github
+
+########################################################################
+# Tactician
+########################################################################
+project tactician 'https://github.com/coq-tactician/coq-tactician' '938ee07cba9719753f7f975addc611cd115f3504'
+# Contact @LasseBlaauwbroek on github
+
+########################################################################
+# Ltac2 compiler
+########################################################################
+project ltac2_compiler 'https://github.com/SkySkimmer/coq-ltac2-compiler' '9fd8b13fd2eb9b3ed49e76e944c5e2f211d7c52d'
+# Contact @SkySkimmer on github
+
+########################################################################
+# Waterproof
+########################################################################
+project waterproof 'https://github.com/impermeable/coq-waterproof' 'aec4fbe2a9b586b246e18b00a52c66ee7a772e79'
+# Contact @jellooo038, @jim-portegies on github
