@@ -4,16 +4,18 @@
 # it to `tee time-of-build.log`.  We have a separate script, because
 # this only works in bash, which we don't require project-wide.
 
-set -eo pipefail
+set -o pipefail
 
 CI_NAME="$1"
 CI_SCRIPT="ci-${CI_NAME}.sh"
 
 DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"
 # assume this script is in dev/ci/, cd to the root Coq directory
-cd "${DIR}/../.."
+cd "${DIR}/../.." || exit 1
 
 export TIMED=1
-"${DIR}/${CI_SCRIPT}" 2>&1 | tee time-of-build.log
+bash "${DIR}/${CI_SCRIPT}" 2>&1 | tee "$CI_NAME.log"
+code=$?
 echo 'Aggregating timing log...'
-python ./tools/make-one-time-file.py time-of-build.log
+python ./tools/make-one-time-file.py --real "$CI_NAME.log"
+exit $code
