@@ -21,13 +21,35 @@ module ESorts = struct
 
   let equal sigma s1 s2 =
     Sorts.equal (kind sigma s1) (kind sigma s2)
+
+  let is_small sigma s = Sorts.is_small (kind sigma s)
+  let is_prop sigma s = Sorts.is_prop (kind sigma s)
+  let is_sprop sigma s = Sorts.is_sprop (kind sigma s)
+  let is_set sigma s = Sorts.is_set (kind sigma s)
+
+  let prop = make Sorts.prop
+  let sprop = make Sorts.sprop
+  let set = make Sorts.set
+  let type1 = make Sorts.type1
+
+  let super sigma s =
+    make (Sorts.super (kind sigma s))
+
+  let relevance_of_sort sigma s =
+    let r = Sorts.relevance_of_sort (unsafe_to_sorts s) in
+    UState.nf_relevance (Evd.evar_universe_context sigma) r
+
+  let family sigma s = Sorts.family (kind sigma s)
+
+  let quality sigma s = Sorts.quality (kind sigma s)
+
 end
 
 module EInstance = struct
   include Evd.MiniEConstr.EInstance
 
   let equal sigma i1 i2 =
-    Univ.Instance.equal (kind sigma i1) (kind sigma i2)
+    UVars.Instance.equal (kind sigma i1) (kind sigma i2)
 end
 
 include (Evd.MiniEConstr : module type of Evd.MiniEConstr
@@ -44,10 +66,12 @@ type case = (t, t, EInstance.t) pcase
 type fixpoint = (t, t) pfixpoint
 type cofixpoint = (t, t) pcofixpoint
 type unsafe_judgment = (constr, types) Environ.punsafe_judgment
-type unsafe_type_judgment = types Environ.punsafe_type_judgment
+type unsafe_type_judgment = (types, ESorts.t) Environ.punsafe_type_judgment
 type named_declaration = (constr, types) Context.Named.Declaration.pt
 type rel_declaration = (constr, types) Context.Rel.Declaration.pt
+type compacted_declaration = (constr, types) Context.Compacted.Declaration.pt
 type named_context = (constr, types) Context.Named.pt
+type compacted_context = compacted_declaration list
 type rel_context = (constr, types) Context.Rel.pt
 
 type 'a puniverses = 'a * EInstance.t
@@ -62,23 +86,20 @@ let mkRel n = of_kind (Rel n)
 let mkVar id = of_kind (Var id)
 let mkMeta n = of_kind (Meta n)
 let mkEvar e = of_kind (Evar e)
-let mkSort s = of_kind (Sort (ESorts.make s))
+let mkSort s = of_kind (Sort s)
 let mkCast (b, k, t) = of_kind (Cast (b, k, t))
 let mkProd (na, t, u) = of_kind (Prod (na, t, u))
 let mkLambda (na, t, c) = of_kind (Lambda (na, t, c))
 let mkLetIn (na, b, t, c) = of_kind (LetIn (na, b, t, c))
 let mkApp (f, arg) = of_kind (App (f, arg))
 let mkConstU pc = of_kind (Const pc)
-let mkConst c = of_kind (Const (in_punivs c))
 let mkIndU pi = of_kind (Ind pi)
-let mkInd i = of_kind (Ind (in_punivs i))
 let mkConstructU pc = of_kind (Construct pc)
-let mkConstruct c = of_kind (Construct (in_punivs c))
 let mkConstructUi ((ind,u),i) = of_kind (Construct ((ind,i),u))
 let mkCase (ci, u, pms, c, iv, r, p) = of_kind (Case (ci, u, pms, c, iv, r, p))
 let mkFix f = of_kind (Fix f)
 let mkCoFix f = of_kind (CoFix f)
-let mkProj (p, c) = of_kind (Proj (p, c))
+let mkProj (p, r, c) = of_kind (Proj (p, r, c))
 let mkArrow t1 r t2 = of_kind (Prod (make_annot Anonymous r, t1, t2))
 let mkArrowR t1 t2 = mkArrow t1 Sorts.Relevant t2
 let mkInt i = of_kind (Int i)
@@ -93,7 +114,7 @@ let mkRef (gr,u) = let open GlobRef in match gr with
 
 let mkLEvar = Evd.MiniEConstr.mkLEvar
 
-let type1 = mkSort Sorts.type1
+let type1 = mkSort ESorts.type1
 
 let applist (f, arg) = mkApp (f, Array.of_list arg)
 let applistc f arg = mkApp (f, Array.of_list arg)
@@ -132,15 +153,19 @@ let isRef sigma c = match kind sigma c with
   | Const _ | Ind _ | Construct _ | Var _ -> true
   | _ -> false
 
-let isRefX sigma x c =
+let isRefX env sigma x c =
   let open GlobRef in
   match x, kind sigma c with
-  | ConstRef c, Const (c', _) -> Constant.CanOrd.equal c c'
-  | IndRef i, Ind (i', _) -> Ind.CanOrd.equal i i'
-  | ConstructRef i, Construct (i', _) -> Construct.CanOrd.equal i i'
+  | ConstRef c, Const (c', _) -> Environ.QConstant.equal env c c'
+  | IndRef i, Ind (i', _) -> Environ.QInd.equal env i i'
+  | ConstructRef i, Construct (i', _) -> Environ.QConstruct.equal env i i'
   | VarRef id, Var id' -> Id.equal id id'
   | _ -> false
 
+let is_lib_ref env sigma x c =
+  match Coqlib.lib_ref_opt x with
+  | Some x -> isRefX env sigma x c
+  | None -> false
 
 let destRel sigma c = match kind sigma c with
 | Rel p -> p
@@ -207,7 +232,7 @@ let destCase sigma c = match kind sigma c with
 | _ -> raise DestKO
 
 let destProj sigma c = match kind sigma c with
-| Proj (p, c) -> (p, c)
+| Proj (p, r, c) -> (p, r, c)
 | _ -> raise DestKO
 
 let destRef sigma c = let open GlobRef in match kind sigma c with
@@ -218,11 +243,16 @@ let destRef sigma c = let open GlobRef in match kind sigma c with
   | _ -> raise DestKO
 
 let decompose_app sigma c =
+  match kind sigma c with
+  | App (f,cl) -> (f, cl)
+  | _ -> (c,[||])
+
+let decompose_app_list sigma c =
   match kind sigma c with
     | App (f,cl) -> (f, Array.to_list cl)
     | _ -> (c,[])
 
-let decompose_lam sigma c =
+let decompose_lambda sigma c =
   let rec lamdec_rec l c = match kind sigma c with
     | Lambda (x,t,c) -> lamdec_rec ((x,t)::l) c
     | Cast (c,_,_)     -> lamdec_rec l c
@@ -230,7 +260,7 @@ let decompose_lam sigma c =
   in
   lamdec_rec [] c
 
-let decompose_lam_assum sigma c =
+let decompose_lambda_decls sigma c =
   let open Rel.Declaration in
   let rec lamdec_rec l c =
     match kind sigma c with
@@ -241,10 +271,10 @@ let decompose_lam_assum sigma c =
   in
   lamdec_rec Context.Rel.empty c
 
-let decompose_lam_n_assum sigma n c =
+let decompose_lambda_n_assum sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_lam_n_assum: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_lambda_n_assum: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -252,14 +282,14 @@ let decompose_lam_n_assum sigma n c =
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) n c
       | Cast (c,_,_)    -> lamdec_rec l n c
-      | c -> anomaly Pp.(str "decompose_lam_n_assum: not enough abstractions.")
+      | c -> anomaly Pp.(str "decompose_lambda_n_assum: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n c
 
-let decompose_lam_n_decls sigma n =
+let decompose_lambda_n_decls sigma n =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_lam_n_decls: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_lambda_n_decls: integer parameter must be positive.");
   let rec lamdec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -267,7 +297,7 @@ let decompose_lam_n_decls sigma n =
       | Lambda (x,t,c)  -> lamdec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> lamdec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
       | Cast (c,_,_)    -> lamdec_rec l n c
-      | c -> anomaly Pp.(str "decompose_lam_n_decls: not enough abstractions.")
+      | c -> anomaly Pp.(str "decompose_lambda_n_decls: not enough abstractions.")
   in
   lamdec_rec Context.Rel.empty n
 
@@ -288,7 +318,7 @@ let decompose_prod sigma c =
   in
   proddec_rec [] c
 
-let decompose_prod_assum sigma c =
+let decompose_prod_decls sigma c =
   let open Rel.Declaration in
   let rec proddec_rec l c =
     match kind sigma c with
@@ -299,10 +329,10 @@ let decompose_prod_assum sigma c =
   in
   proddec_rec Context.Rel.empty c
 
-let decompose_prod_n_assum sigma n c =
+let decompose_prod_n_decls sigma n c =
   let open Rel.Declaration in
   if n < 0 then
-    anomaly Pp.(str "decompose_prod_n_assum: integer parameter must be positive.");
+    anomaly Pp.(str "decompose_prod_n_decls: integer parameter must be positive.");
   let rec prodec_rec l n c =
     if Int.equal n 0 then l,c
     else
@@ -310,10 +340,12 @@ let decompose_prod_n_assum sigma n c =
       | Prod (x,t,c)    -> prodec_rec (Context.Rel.add (LocalAssum (x,t)) l) (n-1) c
       | LetIn (x,b,t,c) -> prodec_rec (Context.Rel.add (LocalDef (x,b,t)) l) (n-1) c
       | Cast (c,_,_)    -> prodec_rec l n c
-      | c -> anomaly Pp.(str "decompose_prod_n_assum: not enough declarations.")
+      | c -> anomaly Pp.(str "decompose_prod_n_decls: not enough assumptions.")
   in
   prodec_rec Context.Rel.empty n c
 
+let prod_decls sigma t = fst (decompose_prod_decls sigma t)
+
 let existential_type = Evd.existential_type
 
 let lift n c = of_constr (Vars.lift n (unsafe_to_constr c))
@@ -358,7 +390,8 @@ let map_instance sigma f evk args =
     else SList.cons c' rem'
   | [], Some _ | _ :: _, None -> assert false
   in
-  let ctx = Evd.evar_filtered_context @@ Evd.find sigma evk in
+  let EvarInfo evi = Evd.find sigma evk in
+  let ctx = Evd.evar_filtered_context evi in
   map ctx args
 
 let map sigma f c =
@@ -404,24 +437,24 @@ let expand_case env _sigma (ci, u, pms, p, iv, c, bl) =
   let iv = unsafe_to_case_invert iv in
   let c = unsafe_to_constr c in
   let bl = unsafe_to_branches bl in
-  let (ci, p, iv, c, bl) = Inductive.expand_case env (ci, u, pms, p, iv, c, bl) in
+  let (ci, (p,r), iv, c, bl) = Inductive.expand_case env (ci, u, pms, p, iv, c, bl) in
   let p = of_constr p in
   let c = of_constr c in
   let iv = of_case_invert iv in
   let bl = of_constr_array bl in
-  (ci, p, iv, c, bl)
+  (ci, (p,r), iv, c, bl)
 
 let annotate_case env sigma (ci, u, pms, p, iv, c, bl as case) =
-  let (_, p, _, _, bl) = expand_case env sigma case in
+  let (_, (p,r), _, _, bl) = expand_case env sigma case in
   let p =
     (* Too bad we need to fetch this data in the environment, should be in the
       case_info instead. *)
     let (_, mip) = Inductive.lookup_mind_specif env ci.ci_ind in
-    decompose_lam_n_decls sigma (mip.Declarations.mind_nrealdecls + 1) p
+    decompose_lambda_n_decls sigma (mip.Declarations.mind_nrealdecls + 1) p
   in
-  let mk_br c n = decompose_lam_n_decls sigma n c in
+  let mk_br c n = decompose_lambda_n_decls sigma n c in
   let bl = Array.map2 mk_br bl ci.ci_cstr_ndecls in
-  (ci, u, pms, p, iv, c, bl)
+  (ci, u, pms, (p,r), iv, c, bl)
 
 let expand_branch env _sigma u pms (ind, i) (nas, _br) =
   let open Declarations in
@@ -436,12 +469,12 @@ let expand_branch env _sigma u pms (ind, i) (nas, _br) =
   let ans : rel_context = match Evd.MiniEConstr.unsafe_eq with Refl -> ans in
   ans
 
-let contract_case env _sigma (ci, p, iv, c, bl) =
+let contract_case env _sigma (ci, (p,r), iv, c, bl) =
   let p = unsafe_to_constr p in
   let iv = unsafe_to_case_invert iv in
   let c = unsafe_to_constr c in
   let bl = unsafe_to_constr_array bl in
-  let (ci, u, pms, p, iv, c, bl) = Inductive.contract_case env (ci, p, iv, c, bl) in
+  let (ci, u, pms, p, iv, c, bl) = Inductive.contract_case env (ci, (p,r), iv, c, bl) in
   let u = EInstance.make u in
   let pms = of_constr_array pms in
   let p = of_return p in
@@ -464,10 +497,10 @@ let iter_with_full_binders env sigma g f n c =
     let l = Evd.expand_existential sigma ev in
     List.iter (fun c -> f n c) l
   | Case (ci,u,pms,p,iv,c,bl) ->
-    let (ci, _, pms, p, iv, c, bl) = annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
+    let (ci, _, pms, (p,_), iv, c, bl) = annotate_case env sigma (ci, u, pms, p, iv, c, bl) in
     let f_ctx (ctx, c) = f (List.fold_right g ctx n) c in
     Array.Fun1.iter f n pms; f_ctx p; iter_invert (f n) iv; f n c; Array.iter f_ctx bl
-  | Proj (p,c) -> f n c
+  | Proj (_,_,c) -> f n c
   | Fix (_,(lna,tl,bl)) ->
     Array.iter (f n) tl;
     let n' = Array.fold_left2_i (fun i n na t -> g (LocalAssum (na, lift i t)) n) n lna tl in
@@ -541,43 +574,29 @@ let compare_constr sigma cmp c1 c2 =
   let cmp nargs c1 c2 = cmp c1 c2 in
   compare_gen kind eq_inst eq_sorts (eq_existential cmp) cmp 0 c1 c2
 
-let compare_cumulative_instances cv_pb nargs_ok variances u u' cstrs =
-  let open UnivProblem in
-  if not nargs_ok then enforce_eq_instances_univs false u u' cstrs
-  else
-    let make u = Sorts.sort_of_univ @@ Univ.Universe.make u in
-    CArray.fold_left3
-      (fun cstrs v u u' ->
-         let open Univ.Variance in
-         match v with
-         | Irrelevant -> Set.add (UWeak (u,u')) cstrs
-         | Covariant ->
-           (match cv_pb with
-            | Reduction.CONV -> Set.add (UEq (make u, make u')) cstrs
-            | Reduction.CUMUL -> Set.add (ULe (make u, make u')) cstrs)
-         | Invariant ->
-           Set.add (UEq (make u, make u')) cstrs)
-      cstrs variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
-
 let cmp_inductives cv_pb (mind,ind as spec) nargs u1 u2 cstrs =
   let open UnivProblem in
   match mind.Declarations.mind_variance with
   | None -> enforce_eq_instances_univs false u1 u2 cstrs
   | Some variances ->
-    let num_param_arity = Reduction.inductive_cumulativity_arguments spec in
-    compare_cumulative_instances cv_pb (Int.equal num_param_arity nargs) variances u1 u2 cstrs
+    let num_param_arity = Conversion.inductive_cumulativity_arguments spec in
+    if not (Int.equal num_param_arity nargs) then enforce_eq_instances_univs false u1 u2 cstrs
+    else compare_cumulative_instances cv_pb  variances u1 u2 cstrs
 
 let cmp_constructors (mind, ind, cns as spec) nargs u1 u2 cstrs =
   let open UnivProblem in
   match mind.Declarations.mind_variance with
   | None -> enforce_eq_instances_univs false u1 u2 cstrs
   | Some _ ->
-    let num_cnstr_args = Reduction.constructor_cumulativity_arguments spec in
+    let num_cnstr_args = Conversion.constructor_cumulativity_arguments spec in
     if not (Int.equal num_cnstr_args nargs)
     then enforce_eq_instances_univs false u1 u2 cstrs
     else
+      let qs1, us1 = UVars.Instance.to_array u1
+      and qs2, us2 = UVars.Instance.to_array u2 in
+      let cstrs = enforce_eq_qualities qs1 qs2 cstrs in
       Array.fold_left2 (fun cstrs u1 u2 -> UnivProblem.(Set.add (UWeak (u1,u2)) cstrs))
-        cstrs (Univ.Instance.to_array u1) (Univ.Instance.to_array u2)
+        cstrs us1 us2
 
 let eq_universes env sigma cstrs cv_pb refargs l l' =
   if EInstance.is_empty l then (assert (EInstance.is_empty l'); true)
@@ -588,7 +607,7 @@ let eq_universes env sigma cstrs cv_pb refargs l l' =
     let open UnivProblem in
     match refargs with
     | Some (ConstRef c, 1) when Environ.is_array_type env c ->
-      cstrs := compare_cumulative_instances cv_pb true [|Univ.Variance.Irrelevant|] l l' !cstrs;
+      cstrs := compare_cumulative_instances cv_pb [|UVars.Variance.Irrelevant|] l l' !cstrs;
       true
     | None | Some (ConstRef _, _) ->
       cstrs := enforce_eq_instances_univs true l l' !cstrs; true
@@ -608,8 +627,8 @@ let test_constr_universes env sigma leq ?(nargs=0) m n =
   if m == n then Some Set.empty
   else
     let cstrs = ref Set.empty in
-    let cv_pb = if leq then Reduction.CUMUL else Reduction.CONV in
-    let eq_universes refargs l l' = eq_universes env sigma cstrs Reduction.CONV refargs l l'
+    let cv_pb = if leq then Conversion.CUMUL else Conversion.CONV in
+    let eq_universes refargs l l' = eq_universes env sigma cstrs Conversion.CONV refargs l l'
     and leq_universes refargs l l' = eq_universes env sigma cstrs cv_pb refargs l l' in
     let eq_sorts s1 s2 =
       let s1 = ESorts.kind sigma s1 in
@@ -650,8 +669,8 @@ let leq_constr_universes env sigma ?nargs m n =
 let compare_head_gen_proj env sigma equ eqs eqev eqc' nargs m n =
   let kind c = kind sigma c in
   match kind m, kind n with
-  | Proj (p, c), App (f, args)
-  | App (f, args), Proj (p, c) ->
+  | Proj (p, _, c), App (f, args)
+  | App (f, args), Proj (p, _, c) ->
       (match kind f with
       | Const (p', u) when Environ.QConstant.equal env (Projection.constant p) p' ->
           let npars = Projection.npars p in
@@ -666,7 +685,7 @@ let eq_constr_universes_proj env sigma m n =
   if m == n then Some Set.empty
   else
     let cstrs = ref Set.empty in
-    let eq_universes ref l l' = eq_universes env sigma cstrs Reduction.CONV ref l l' in
+    let eq_universes ref l l' = eq_universes env sigma cstrs Conversion.CONV ref l l' in
     let eq_sorts s1 s2 =
       let s1 = ESorts.kind sigma s1 in
       let s2 = ESorts.kind sigma s2 in
@@ -683,30 +702,77 @@ let eq_constr_universes_proj env sigma m n =
     let res = eq_constr' 0 m n in
     if res then Some !cstrs else None
 
+let add_universes_of_instance sigma (qs,us) u =
+  let u = EInstance.kind sigma u in
+  let qs', us' = UVars.Instance.levels u in
+  let qs = Sorts.Quality.(Set.fold (fun q qs -> match q with
+      | QVar q -> Sorts.QVar.Set.add q qs
+      | QConstant _ -> qs)
+      qs' qs)
+  in
+  qs, Univ.Level.Set.union us us'
+
+let fold_annot_relevance f acc na =
+  f acc na.Context.binder_relevance
+
+let fold_case_under_context_relevance f acc (nas,_) =
+  Array.fold_left (fold_annot_relevance f) acc nas
+
+let fold_rec_declaration_relevance f acc (nas,_,_) =
+  Array.fold_left (fold_annot_relevance f) acc nas
+
+let fold_constr_relevance sigma f acc c =
+  match kind sigma c with
+  | Rel _ | Var _ | Meta _ | Evar _
+  |  Sort _ | Cast _ | App _
+  | Const _ | Ind _ | Construct _ | Proj _
+  | Int _ | Float _ | Array _ -> acc
+
+  | Prod (na,_,_) | Lambda (na,_,_) | LetIn (na,_,_,_) ->
+    fold_annot_relevance f acc na
+
+  | Case (_,_u,_params,(ret,r),_iv,_v,brs) ->
+    let acc = f acc r in
+    let acc = fold_case_under_context_relevance f acc ret in
+    let acc = CArray.fold_left (fold_case_under_context_relevance f) acc brs in
+    acc
+
+  | Fix (_,data)
+  | CoFix (_,data) ->
+    fold_rec_declaration_relevance f acc data
+
+let add_relevance sigma (qs,us as v) r =
+  let open Sorts in
+  (* NB this normalizes above_prop to Relevant which makes it disappear *)
+  match UState.nf_relevance (Evd.evar_universe_context sigma) r with
+  | Irrelevant | Relevant -> v
+  | RelevanceVar q -> QVar.Set.add q qs, us
+
 let universes_of_constr sigma c =
   let open Univ in
   let rec aux s c =
+    let s = fold_constr_relevance sigma (add_relevance sigma) s c in
     match kind sigma c with
-    | Const (c, u) ->
-      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
-    | Ind ((mind,_), u) | Construct (((mind,_),_), u) ->
-      Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s
-    | Sort u ->
-      let sort = ESorts.kind sigma u in
-      if Sorts.is_small sort then s
-      else
-        Level.Set.fold Level.Set.add (Sorts.levels sort) s
+    | Const (_, u) | Ind (_, u) | Construct (_,u) -> add_universes_of_instance sigma s u
+    | Sort u -> begin match ESorts.kind sigma u with
+        | Type u ->
+          Util.on_snd (Level.Set.fold Level.Set.add (Universe.levels u)) s
+        | QSort (q, u) ->
+          let qs, us = s in
+          Sorts.QVar.Set.add q qs, Level.Set.union us (Universe.levels u)
+        | SProp | Prop | Set -> s
+      end
     | Evar (k, args) ->
-      let concl = Evd.evar_concl (Evd.find sigma k) in
+      let concl = Evd.evar_concl (Evd.find_undefined sigma k) in
       fold sigma aux (aux s concl) c
     | Array (u,_,_,_) ->
-      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
+      let s = add_universes_of_instance sigma s u in
       fold sigma aux s c
     | Case (_,u,_,_,_,_,_) ->
-      let s = Level.Set.fold Level.Set.add (Instance.levels (EInstance.kind sigma u)) s in
+      let s = add_universes_of_instance sigma s u in
       fold sigma aux s c
     | _ -> fold sigma aux s c
-  in aux Level.Set.empty c
+  in aux (Sorts.QVar.Set.empty,Level.Set.empty) c
 
 open Context
 open Environ
@@ -846,28 +912,6 @@ match unsafe_eq with
 
 end
 
-let rec isArity sigma c =
-  match kind sigma c with
-  | Prod (_,_,c)    -> isArity sigma c
-  | LetIn (_,b,_,c) -> isArity sigma (Vars.subst1 b c)
-  | Cast (c,_,_)      -> isArity sigma c
-  | Sort _          -> true
-  | _               -> false
-
-type arity = rel_context * ESorts.t
-
-let destArity sigma =
-  let open Context.Rel.Declaration in
-  let rec prodec_rec l c =
-    match kind sigma c with
-    | Prod (x,t,c)    -> prodec_rec (LocalAssum (x,t) :: l) c
-    | LetIn (x,b,t,c) -> prodec_rec (LocalDef (x,b,t) :: l) c
-    | Cast (c,_,_)      -> prodec_rec l c
-    | Sort s          -> l,s
-    | _               -> anomaly ~label:"destArity" (Pp.str "not an arity.")
-  in
-  prodec_rec []
-
 (* Constructs either [forall x:t, c] or [let x:=b:t in c] *)
 let mkProd_or_LetIn decl c =
   let open Context.Rel.Declaration in
@@ -932,6 +976,30 @@ let it_mkNamedLambda_or_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedLa
 
 let it_mkNamedProd_wo_LetIn sigma t ctx = List.fold_left (fun c d -> mkNamedProd_wo_LetIn sigma d c) t ctx
 
+let rec isArity sigma c =
+  match kind sigma c with
+  | Prod (_,_,c)    -> isArity sigma c
+  | LetIn (_,b,_,c) -> isArity sigma (Vars.subst1 b c)
+  | Cast (c,_,_)      -> isArity sigma c
+  | Sort _          -> true
+  | _               -> false
+
+type arity = rel_context * ESorts.t
+
+let mkArity (ctx, s) = it_mkProd_or_LetIn (mkSort s) ctx
+
+let destArity sigma =
+  let open Context.Rel.Declaration in
+  let rec prodec_rec l c =
+    match kind sigma c with
+    | Prod (x,t,c)    -> prodec_rec (LocalAssum (x,t) :: l) c
+    | LetIn (x,b,t,c) -> prodec_rec (LocalDef (x,b,t) :: l) c
+    | Cast (c,_,_)      -> prodec_rec l c
+    | Sort s          -> l,s
+    | _               -> anomaly ~label:"destArity" (Pp.str "not an arity.")
+  in
+  prodec_rec []
+
 let push_rel d e = push_rel (cast_rel_decl unsafe_eq d) e
 let push_rel_context d e = push_rel_context (cast_rel_context unsafe_eq d) e
 let push_rec_types d e = push_rec_types (cast_rec_decl unsafe_eq d) e
@@ -1018,3 +1086,22 @@ let to_case_invert = unsafe_to_case_invert
 
 let eq = unsafe_eq
 end
+
+module UnsafeMonomorphic = struct
+  let mkConst c = of_kind (Const (in_punivs c))
+  let mkInd i = of_kind (Ind (in_punivs i))
+  let mkConstruct c = of_kind (Construct (in_punivs c))
+end
+
+(* deprecated *)
+
+let decompose_lambda_assum = decompose_lambda_decls
+let decompose_prod_assum = decompose_prod_decls
+let decompose_prod_n_assum = decompose_prod_n_decls
+let prod_assum = prod_decls
+let decompose_lam = decompose_lambda
+let decompose_lam_n_assum = decompose_lambda_n_assum
+let decompose_lam_n_decls = decompose_lambda_n_decls
+let decompose_lam_assum = decompose_lambda_assum
+
+include UnsafeMonomorphic
