@@ -35,7 +35,7 @@ let finalize ?abort_on_undefined_evars sigma f =
   let sigma = minimize_universes sigma in
   let uvars = ref Univ.Level.Set.empty in
   let v = f (fun c ->
-      let varsc = EConstr.universes_of_constr sigma c in
+      let _, varsc = EConstr.universes_of_constr sigma c in
       let c = EConstr.to_constr ?abort_on_undefined_evars sigma c in
       uvars := Univ.Level.Set.union !uvars varsc;
       c)
@@ -65,8 +65,11 @@ let nf_evars_universes sigma t = EConstr.to_constr ~abort_on_undefined_evars:fal
 let whd_evar = EConstr.whd_evar
 
 let nf_evar sigma c =
+  let lsubst = Evd.universe_subst sigma in
   let evar_value ev = Evd.existential_opt_value0 sigma ev in
-  EConstr.of_constr @@ UnivSubst.nf_evars_and_universes_opt_subst evar_value (universe_subst sigma) (EConstr.Unsafe.to_constr c)
+  let univ_value l = UnivFlex.normalize_univ_variable lsubst l in
+  let qvar_value q = UState.nf_qvar (Evd.evar_universe_context sigma) q in
+  EConstr.of_constr @@ UnivSubst.nf_evars_and_universes_opt_subst evar_value qvar_value univ_value (EConstr.Unsafe.to_constr c)
 
 let j_nf_evar sigma j =
   { uj_val = nf_evar sigma j.uj_val;
@@ -76,6 +79,9 @@ let jv_nf_evar sigma = Array.map (j_nf_evar sigma)
 let tj_nf_evar sigma {utj_val=v;utj_type=t} =
   {utj_val=nf_evar sigma v;utj_type=t}
 
+let nf_relevance sigma r =
+  UState.nf_relevance (Evd.evar_universe_context sigma) r
+
 let nf_named_context_evar sigma ctx =
   Context.Named.map (nf_evars_universes sigma) ctx
 
@@ -90,7 +96,7 @@ let nf_env_evar sigma env =
 let nf_evar_info evc info = map_evar_info (nf_evar evc) info
 
 let nf_evar_map evm =
-  Evd.raw_map (fun _ evi -> nf_evar_info evm evi) evm
+  Evd.raw_map { map = fun _ evi -> nf_evar_info evm evi } evm
 
 let nf_evar_map_undefined evm =
   Evd.raw_map_undefined (fun _ evi -> nf_evar_info evm evi) evm
@@ -132,7 +138,7 @@ let head_evar sigma c =
     | Case (_, _, _, _, _, c, _) -> hrec c
     | App (c,_)      -> hrec c
     | Cast (c,_,_)   -> hrec c
-    | Proj (p, c)    -> hrec c
+    | Proj (_, _, c)    -> hrec c
     | _              -> raise NoHeadEvar
   in
   hrec c
@@ -245,7 +251,7 @@ let csubst_subst sigma { csubst_len = k; csubst_var = v; csubst_rel = s } c =
   | Var id ->
     begin try Id.Map.find id v with Not_found -> c end
   | Evar (evk, args) ->
-    let evi = Evd.find sigma evk in
+    let EvarInfo evi = Evd.find sigma evk in
     let args' = subst_instance n (evar_filtered_context evi) args in
     if args' == args then c else Constr.mkEvar (evk, args') (* FIXME: preserve sharing *)
   | _ -> Constr.map_with_binders succ subst n c
@@ -411,7 +417,7 @@ let next_evar_name sigma naming = match naming with
 
 (* [new_evar] declares a new existential in an env env with type typ *)
 (* Converting the env into the sign of the evar to define *)
-let new_evar ?src ?filter ?abstract_arguments ?candidates ?(naming = IntroAnonymous) ?typeclass_candidate
+let new_evar ?src ?filter ?relevance ?abstract_arguments ?candidates ?(naming = IntroAnonymous) ?typeclass_candidate
     ?principal ?hypnaming env evd typ =
   let name = next_evar_name evd naming in
   let hypnaming = match hypnaming with
@@ -425,14 +431,18 @@ let new_evar ?src ?filter ?abstract_arguments ?candidates ?(naming = IntroAnonym
     match filter with
     | None -> instance
     | Some filter -> Filter.filter_slist filter instance in
-  let relevance = Sorts.Relevant in (* FIXME: relevant_of_type not defined yet *)
+  let relevance = match relevance with
+  | Some r -> r
+  | None -> Sorts.Relevant (* FIXME: relevant_of_type not defined yet *)
+  in
   let (evd, evk) = new_pure_evar sign evd typ' ?src ?filter ~relevance ?abstract_arguments ?candidates ?name
     ?typeclass_candidate ?principal in
   (evd, EConstr.mkEvar (evk, instance))
 
 let new_type_evar ?src ?filter ?naming ?principal ?hypnaming env evd rigid =
   let (evd', s) = new_sort_variable rigid evd in
-  let (evd', e) = new_evar env evd' ?src ?filter ?naming ~typeclass_candidate:false ?principal ?hypnaming (EConstr.mkSort s) in
+  let relevance = EConstr.ESorts.relevance_of_sort evd s in
+  let (evd', e) = new_evar env evd' ?src ?filter ~relevance ?naming ~typeclass_candidate:false ?principal ?hypnaming (EConstr.mkSort s) in
   evd', (e, s)
 
 let new_Type ?(rigid=Evd.univ_flexible) evd =
@@ -459,7 +469,7 @@ let add_unification_pb ?(tail=false) pb evd =
    the de Bruijn part of the context *)
 let generalize_evar_over_rels sigma (ev,args) =
   let open EConstr in
-  let evi = Evd.find sigma ev in
+  let evi = Evd.find_undefined sigma ev in
   let args = Evd.expand_existential sigma (ev, args) in
   let sign = named_context_of_val (Evd.evar_hyps evi) in
   List.fold_left2
@@ -487,12 +497,12 @@ let filter_effective_candidates evd evi filter candidates =
   let ids = set_of_evctx (Filter.filter_list filter (evar_context evi)) in
   List.filter (fun a -> Id.Set.subset (collect_vars evd a) ids) candidates
 
-let restrict_evar evd evk filter ?src candidates =
+let restrict_evar evd evk filter candidates =
   let evar_info = Evd.find_undefined evd evk in
   let candidates = Option.map (filter_effective_candidates evd evar_info filter) candidates in
   match candidates with
   | Some [] -> raise (ClearDependencyError (*FIXME*)(Id.of_string "blah", (NoCandidatesLeft evk), None))
-  | _ -> Evd.restrict evk filter ?candidates ?src evd
+  | _ -> Evd.restrict evk filter ?candidates evd
 
 let rec check_and_clear_in_constr ~is_section_variable env evdref err ids ~global c =
   (* returns a new constr where all the evars have been 'cleaned'
@@ -627,10 +637,11 @@ let clear_hyps2_in_evi env sigma hyps t concl ids =
    goal ([advance] is used to figure if a side effect has modified the
    goal) it terminates quickly. *)
 let rec advance sigma evk =
-  let evi = Evd.find sigma evk in
-  match Evd.evar_body evi with
-  | Evar_empty -> Some evk
-  | Evar_defined v ->
+  match Evd.find_defined sigma evk with
+  | None -> Some evk
+  | Some evi ->
+    match Evd.evar_body evi with
+    | Evar_defined v ->
       match is_aliased_evar sigma evk with
       | Some evk -> advance sigma evk
       | None -> None
@@ -668,15 +679,6 @@ let undefined_evars_of_named_context evd nc =
     nc
     ~init:Evar.Set.empty
 
-let undefined_evars_of_evar_info evd evi =
-  Evar.Set.union (undefined_evars_of_term evd (Evd.evar_concl evi))
-    (Evar.Set.union
-       (match Evd.evar_body evi with
-         | Evar_empty -> Evar.Set.empty
-         | Evar_defined b -> undefined_evars_of_term evd b)
-       (undefined_evars_of_named_context evd
-          (named_context_of_val (Evd.evar_hyps evi))))
-
 type undefined_evars_cache = {
   mutable cache : (EConstr.named_declaration * Evar.Set.t) ref Id.Map.t;
 }
@@ -714,16 +716,15 @@ let cached_evar_of_hyp cache sigma decl accu = match cache with
   in
   Evar.Set.fold Evar.Set.add evs accu
 
-let filtered_undefined_evars_of_evar_info ?cache sigma evi =
+let filtered_undefined_evars_of_evar_info (type a) ?cache sigma (evi : a evar_info) =
   let evars_of_named_context cache accu nc =
     let fold decl accu = cached_evar_of_hyp cache sigma (EConstr.of_named_decl decl) accu in
     Context.Named.fold_outside fold nc ~init:accu
   in
   let accu = match Evd.evar_body evi with
-  | Evar_empty -> Evar.Set.empty
+  | Evar_empty -> undefined_evars_of_term sigma (Evd.evar_concl evi)
   | Evar_defined b -> evars_of_term sigma b
   in
-  let accu = Evar.Set.union (undefined_evars_of_term sigma (Evd.evar_concl evi)) accu in
   let ctxt = EConstr.Unsafe.to_named_context (evar_filtered_context evi) in
   evars_of_named_context cache accu ctxt
 
@@ -746,7 +747,7 @@ let occur_evar_upto sigma n c =
 let judge_of_new_Type evd =
   let open EConstr in
   let (evd', s) = new_sort_variable univ_rigid evd in
-  (evd', { uj_val = mkSort s; uj_type = mkSort (Sorts.super s) })
+  (evd', { uj_val = mkSort s; uj_type = mkSort (ESorts.super evd s) })
 
 let subterm_source evk ?where (loc,k) =
   let evk = match k with
@@ -760,14 +761,20 @@ let compare_cumulative_instances cv_pb variances u u' sigma =
   let open UnivProblem in
   let cstrs = Univ.Constraints.empty in
   let soft = Set.empty in
+  let qs, us = UVars.Instance.to_array u
+  and qs', us' = UVars.Instance.to_array u' in
+  let qcstrs = enforce_eq_qualities qs qs' Set.empty in
+  match Evd.add_universe_constraints sigma qcstrs with
+  | exception UGraph.UniverseInconsistency p -> Inr p
+  | sigma ->
   let cstrs, soft = Array.fold_left3 (fun (cstrs, soft) v u u' ->
-      let open Univ.Variance in
+      let open UVars.Variance in
       match v with
       | Irrelevant -> cstrs, Set.add (UWeak (u,u')) soft
-      | Covariant when cv_pb == Reduction.CUMUL ->
+      | Covariant when cv_pb == Conversion.CUMUL ->
         Univ.Constraints.add (u,Univ.Le,u') cstrs, soft
       | Covariant | Invariant -> Univ.Constraints.add (u,Univ.Eq,u') cstrs, soft)
-      (cstrs,soft) variances (Univ.Instance.to_array u) (Univ.Instance.to_array u')
+      (cstrs,soft) variances us us'
   in
   match Evd.add_constraints sigma cstrs with
   | sigma ->
@@ -776,11 +783,17 @@ let compare_cumulative_instances cv_pb variances u u' sigma =
 
 let compare_constructor_instances evd u u' =
   let open UnivProblem in
-  let soft =
-    Array.fold_left2 (fun cs u u' -> Set.add (UWeak (u,u')) cs)
-      Set.empty (Univ.Instance.to_array u) (Univ.Instance.to_array u')
-  in
-  Evd.add_universe_constraints evd soft
+  let qs, us = UVars.Instance.to_array u
+  and qs', us' = UVars.Instance.to_array u' in
+  let qcstrs = enforce_eq_qualities qs qs' Set.empty in
+  match Evd.add_universe_constraints evd qcstrs with
+  | exception UGraph.UniverseInconsistency p -> Inr p
+  | evd ->
+    let soft =
+      Array.fold_left2 (fun cs u u' -> Set.add (UWeak (u,u')) cs)
+        Set.empty us us'
+    in
+    Inl (Evd.add_universe_constraints evd soft)
 
 (** [eq_constr_univs_test ~evd ~extended_evd t u] tests equality of
     [t] and [u] up to existential variable instantiation and
