@@ -10,7 +10,6 @@
 
 open Pp
 open CErrors
-open Sorts
 open Util
 open Names
 open Nameops
@@ -18,11 +17,12 @@ open Constr
 open Vars
 open Environ
 
-(* module RelDecl = Context.Rel.Declaration *)
+module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
 type econstr = constr
 type etypes = types
+type esorts = Sorts.t
 
 (** Generic filters *)
 module Filter :
@@ -204,25 +204,35 @@ module Store = Store.Make ()
 
 let string_of_existential evk = "?X" ^ string_of_int (Evar.repr evk)
 
-type evar_body =
-  | Evar_empty
-  | Evar_defined of constr
+type defined = [ `defined ]
+type undefined = [ `undefined ]
 
-type evar_info = {
-  evar_concl : constr;
+type _ evar_body =
+  | Evar_empty : undefined evar_body
+  | Evar_defined : econstr -> defined evar_body
+
+type (_, 'a) when_undefined =
+| Defined : (defined, 'a) when_undefined
+| Undefined : 'a -> (undefined, 'a) when_undefined
+
+type 'a evar_info = {
+  evar_concl : ('a, constr) when_undefined;
   evar_hyps : named_context_val;
-  evar_body : evar_body;
+  evar_body : 'a evar_body;
   evar_filter : Filter.t;
-  evar_abstract_arguments : Abstraction.t;
+  evar_abstract_arguments : ('a, Abstraction.t) when_undefined;
   evar_source : Evar_kinds.t Loc.located;
-  evar_candidates : constr list option; (* if not None, list of allowed instances *)
+  evar_candidates : ('a, constr list option) when_undefined; (* if not None, list of allowed instances *)
   evar_relevance: Sorts.relevance;
 }
 
+type any_evar_info = EvarInfo : 'a evar_info -> any_evar_info
+
 let instance_mismatch () =
   anomaly (Pp.str "Signature and its instance do not match.")
 
-let evar_concl evi = evi.evar_concl
+let evar_concl evi = match evi.evar_concl with
+| Undefined c -> c
 
 let evar_filter evi = evi.evar_filter
 
@@ -233,9 +243,11 @@ let evar_context evi = named_context_of_val evi.evar_hyps
 let evar_filtered_context evi =
   Filter.filter_list (evar_filter evi) (evar_context evi)
 
-let evar_candidates evi = evi.evar_candidates
+let evar_candidates evi = match evi.evar_candidates with
+| Undefined c -> c
 
-let evar_abstract_arguments evi = evi.evar_abstract_arguments
+let evar_abstract_arguments evi = match evi.evar_abstract_arguments with
+| Undefined c -> c
 
 let evar_relevance evi = evi.evar_relevance
 
@@ -266,16 +278,20 @@ let evar_identity_subst evi =
   in
   SList.defaultn len SList.empty
 
-let map_evar_body f = function
+let map_evar_body (type a) f : a evar_body -> a evar_body = function
   | Evar_empty -> Evar_empty
   | Evar_defined d -> Evar_defined (f d)
 
+let map_when_undefined (type a b) f : (a, b) when_undefined -> (a, b) when_undefined = function
+| Defined -> Defined
+| Undefined x -> Undefined (f x)
+
 let map_evar_info f evi =
   {evi with
     evar_body = map_evar_body f evi.evar_body;
     evar_hyps = map_named_val (fun d -> NamedDecl.map_constr f d) evi.evar_hyps;
-    evar_concl = f evi.evar_concl;
-    evar_candidates = Option.map (List.map f) evi.evar_candidates }
+    evar_concl = map_when_undefined f evi.evar_concl;
+    evar_candidates = map_when_undefined (fun c -> Option.map (List.map f) c) evi.evar_candidates }
 
 (* This exception is raised by *.existential_value *)
 exception NotInstantiatedEvar
@@ -400,7 +416,7 @@ module Metamap = Int.Map
 (*************************)
 (* Unification state *)
 
-type conv_pb = Reduction.conv_pb
+type conv_pb = Conversion.conv_pb
 type evar_constraint = conv_pb * Environ.env * constr * constr
 
 module EvMap = Evar.Map
@@ -411,7 +427,7 @@ sig
 type t
 
 val empty : t
-val add_name_undefined : Id.t option -> Evar.t -> evar_info -> t -> t
+val add_name_undefined : Id.t option -> Evar.t -> 'a evar_info -> t -> t
 val remove_name_defined : Evar.t -> t -> t
 val rename : Evar.t -> Id.t -> t -> t
 val reassign_name_defined : Evar.t -> Evar.t -> t -> t
@@ -472,7 +488,9 @@ end
 type evar_flags =
   { obligation_evars : Evar.Set.t;
     aliased_evars : Evar.t Evar.Map.t;
-    typeclass_evars : Evar.Set.t }
+    typeclass_evars : Evar.Set.t;
+    impossible_case_evars : Evar.Set.t;
+  }
 
 type side_effect_role =
 | Schema of inductive * string
@@ -622,9 +640,10 @@ end
 
 type evar_map = {
   (* Existential variables *)
-  defn_evars : evar_info EvMap.t;
-  undf_evars : evar_info EvMap.t;
+  defn_evars : defined evar_info EvMap.t;
+  undf_evars : undefined evar_info EvMap.t;
   evar_names : EvNames.t;
+  candidate_evars : Evar.Set.t; (* The subset of undefined evars with a non-empty candidate list. *)
   (** Universes *)
   universes  : UState.t;
   (** Conversion problems *)
@@ -643,8 +662,8 @@ type evar_map = {
 }
 
 let find d e =
-  try EvMap.find e d.undf_evars
-  with Not_found -> EvMap.find e d.defn_evars
+  try EvarInfo (EvMap.find e d.undf_evars)
+  with Not_found -> EvarInfo (EvMap.find e d.defn_evars)
 
 let rec thin_val = function
   | [] -> []
@@ -673,7 +692,7 @@ let replace_vars sigma var_alist x =
       | exception Not_found -> c
       end
     | Constr.Evar (evk, args) ->
-      let evi = find sigma evk in
+      let EvarInfo evi = find sigma evk in
       let args' = substrec_instance n (evar_filtered_context evi) args in
       if args' == args then c
       else Constr.mkEvar (evk, args')
@@ -705,7 +724,7 @@ let instantiate_evar_array sigma info c args =
   | _ -> replace_vars sigma inst c
 
 let expand_existential sigma (evk, args) =
-  let evi = find sigma evk in
+  let EvarInfo evi = find sigma evk in
   let rec expand ctx args = match ctx, SList.view args with
   | [], None -> []
   | _ :: ctx, Some (Some c, args) -> c :: expand ctx args
@@ -725,16 +744,25 @@ let is_maybe_typeclass sigma c = Hook.get get_is_maybe_typeclass sigma c
 let rename evk id evd =
   { evd with evar_names = EvNames.rename evk id evd.evar_names }
 
-let add_with_name ?name ?(typeclass_candidate = true) d e i = match i.evar_body with
+let add_with_name (type a) ?name ?(typeclass_candidate = true) d e (i : a evar_info) = match i.evar_body with
 | Evar_empty ->
   let evar_names = EvNames.add_name_undefined name e i d.evar_names in
   let evar_flags =
-    if typeclass_candidate && is_maybe_typeclass d i.evar_concl then
+    if typeclass_candidate && is_maybe_typeclass d (evar_concl i) then
       let flags = d.evar_flags in
       { flags with typeclass_evars = Evar.Set.add e flags.typeclass_evars }
     else d.evar_flags
   in
-  { d with undf_evars = EvMap.add e i d.undf_evars; evar_names; evar_flags }
+  let evar_flags = match i.evar_source with
+    | _, ImpossibleCase ->
+      { evar_flags with impossible_case_evars = Evar.Set.add e evar_flags.impossible_case_evars }
+    | _ -> evar_flags
+  in
+  let candidate_evars = match i.evar_candidates with
+  | Undefined None -> Evar.Set.remove e d.candidate_evars
+  | Undefined (Some _) -> Evar.Set.add e d.candidate_evars
+  in
+  { d with undf_evars = EvMap.add e i d.undf_evars; evar_names; evar_flags; candidate_evars }
 | Evar_defined _ ->
   let evar_names = EvNames.remove_name_defined e d.evar_names in
   { d with defn_evars = EvMap.add e i d.defn_evars; evar_names }
@@ -766,11 +794,14 @@ let is_obligation_evar evd evk =
   let flags = evd.evar_flags in
   Evar.Set.mem evk flags.obligation_evars
 
+let get_impossible_case_evars evd = evd.evar_flags.impossible_case_evars
+
 (** Inheritance of flags: for evar-evar and restriction cases *)
 
 let inherit_evar_flags evar_flags evk evk' =
   let evk_typeclass = Evar.Set.mem evk evar_flags.typeclass_evars in
   let evk_obligation = Evar.Set.mem evk evar_flags.obligation_evars in
+  let evk_impossible = Evar.Set.mem evk evar_flags.impossible_case_evars in
   let aliased_evars = Evar.Map.add evk evk' evar_flags.aliased_evars in
   let typeclass_evars =
     if evk_typeclass then
@@ -784,15 +815,23 @@ let inherit_evar_flags evar_flags evk evk' =
       Evar.Set.add evk' obligation_evars
     else evar_flags.obligation_evars
   in
-  { obligation_evars; aliased_evars; typeclass_evars }
+  let impossible_case_evars =
+    if evk_impossible then
+      let impossible_case_evars = Evar.Set.remove evk evar_flags.impossible_case_evars in
+      Evar.Set.add evk' impossible_case_evars
+    else evar_flags.impossible_case_evars
+  in
+  { obligation_evars; aliased_evars; typeclass_evars; impossible_case_evars; }
 
 (** Removal: in all other cases of definition *)
 
 let remove_evar_flags evk evar_flags =
   { typeclass_evars = Evar.Set.remove evk evar_flags.typeclass_evars;
     obligation_evars = Evar.Set.remove evk evar_flags.obligation_evars;
+    impossible_case_evars = Evar.Set.remove evk evar_flags.impossible_case_evars;
     (* Aliasing information is kept. *)
-    aliased_evars = evar_flags.aliased_evars }
+    aliased_evars = evar_flags.aliased_evars;
+  }
 
 (** New evars *)
 
@@ -809,12 +848,21 @@ let remove d e =
   let defn_evars = EvMap.remove e d.defn_evars in
   let future_goals = FutureGoals.remove e d.future_goals in
   let evar_flags = remove_evar_flags e d.evar_flags in
+  let candidate_evars = Evar.Set.remove e d.candidate_evars in
   { d with undf_evars; defn_evars; future_goals;
-           evar_flags }
+           evar_flags; candidate_evars }
 
-let undefine sigma e =
-  let evi = find sigma e in
-  add (remove sigma e) e { evi with evar_body = Evar_empty }
+let undefine sigma e concl =
+  let EvarInfo evi = find sigma e in
+  let evi = { evi with
+    evar_body = Evar_empty;
+    evar_concl = Undefined concl;
+    evar_candidates = Undefined None;
+    evar_abstract_arguments = Undefined Abstraction.identity;
+  } in
+  add (remove sigma e) e evi
+
+let find_defined d e = EvMap.find_opt e d.defn_evars
 
 let find_undefined d e = EvMap.find e d.undf_evars
 
@@ -827,35 +875,19 @@ let drop_all_defined d = { d with defn_evars = EvMap.empty }
 (* spiwack: not clear what folding over an evar_map, for now we shall
     simply fold over the inner evar_map. *)
 let fold f d a =
+  let f evk evi accu = f evk (EvarInfo evi) accu in
   EvMap.fold f d.defn_evars (EvMap.fold f d.undf_evars a)
 
 let fold_undefined f d a = EvMap.fold f d.undf_evars a
 
+type map = { map : 'r. Evar.t -> 'r evar_info -> 'r evar_info }
+
 let raw_map f d =
-  let f evk info =
-    let ans = f evk info in
-    let () = match info.evar_body, ans.evar_body with
-    | Evar_defined _, Evar_empty
-    | Evar_empty, Evar_defined _ ->
-      anomaly (str "Unrespectful mapping function.")
-    | _ -> ()
-    in
-    ans
-  in
-  let defn_evars = EvMap.Smart.mapi f d.defn_evars in
-  let undf_evars = EvMap.Smart.mapi f d.undf_evars in
+  let defn_evars = EvMap.Smart.mapi f.map d.defn_evars in
+  let undf_evars = EvMap.Smart.mapi f.map d.undf_evars in
   { d with defn_evars; undf_evars; }
 
 let raw_map_undefined f d =
-  let f evk info =
-    let ans = f evk info in
-    let () = match ans.evar_body with
-    | Evar_defined _ ->
-      anomaly (str "Unrespectful mapping function.")
-    | _ -> ()
-    in
-    ans
-  in
   { d with undf_evars = EvMap.Smart.mapi f d.undf_evars; }
 
 let is_evar = mem
@@ -868,9 +900,8 @@ let existential_opt_value d (n, args) =
   match EvMap.find_opt n d.defn_evars with
   | None -> None
   | Some info ->
-    match evar_body info with
-    | Evar_defined c -> Some (instantiate_evar_array d info c args)
-    | Evar_empty -> None (* impossible but w/e *)
+    let Evar_defined c = evar_body info in
+    Some (instantiate_evar_array d info c args)
 
 let existential_value d ev = match existential_opt_value d ev with
   | None -> raise NotInstantiatedEvar
@@ -883,11 +914,11 @@ let existential_opt_value0 = existential_opt_value
 let existential_expand_value0 sigma (evk, args) = match existential_opt_value sigma (evk, args) with
 | None ->
   let args = expand_existential sigma (evk, args) in
-  Constr.EvarUndefined (evk, args)
-| Some c -> Constr.EvarDefined c
+  CClosure.EvarUndefined (evk, args)
+| Some c -> CClosure.EvarDefined c
 
 let mkLEvar sigma (evk, args) =
-  let evi = find sigma evk in
+  let EvarInfo evi = find sigma evk in
   let fold decl arg accu =
     if isVarId (NamedDecl.get_id decl) arg then SList.default accu
     else SList.cons arg accu
@@ -897,25 +928,40 @@ let mkLEvar sigma (evk, args) =
 
 let evar_handler sigma =
   let evar_expand ev = existential_expand_value0 sigma ev in
-  let evar_relevance (evk, _) = match find sigma evk with
-  | evi -> evi.evar_relevance
-  | exception Not_found -> Sorts.Relevant
+  let qvar_irrelevant q = match UState.nf_qvar sigma.universes q with
+  | QConstant QSProp -> true
+  | QConstant QProp | QConstant QType | QVar _ -> false
+  in
+  let evar_irrelevant (evk, _) = match find sigma evk with
+  | EvarInfo evi ->
+    begin match evi.evar_relevance with
+    | Sorts.Relevant -> false
+    | Sorts.Irrelevant -> true
+    | Sorts.RelevanceVar q -> qvar_irrelevant q
+    end
+  | exception Not_found -> true
   in
   let evar_repack ev = mkLEvar sigma ev in
-  { evar_expand; evar_relevance; evar_repack }
+  { CClosure.evar_expand; evar_irrelevant; evar_repack; qvar_irrelevant }
 
-let existential_type d (n, args) =
-  let info =
-    try find d n
-    with Not_found ->
-      anomaly (str "Evar " ++ str (string_of_existential n) ++ str " was not declared.") in
-  instantiate_evar_array d info info.evar_concl args
+let existential_type_opt d (n, args) =
+  match find_undefined d n with
+  | exception Not_found -> None
+  | info ->
+    Some (instantiate_evar_array d info (evar_concl info) args)
+
+let existential_type d n = match existential_type_opt d n with
+  | Some t -> t
+  | None -> anomaly (str "Evar " ++ str (string_of_existential (fst n)) ++ str " was not declared.")
 
 let existential_type0 = existential_type
 
 let add_constraints d c =
   { d with universes = UState.add_constraints d.universes c }
 
+let add_quconstraints d c =
+  { d with universes = UState.add_quconstraints d.universes c }
+
 let add_universe_constraints d c =
   { d with universes = UState.add_universe_constraints d.universes c }
 
@@ -942,7 +988,9 @@ let create_evar_defs sigma = { sigma with
 let empty_evar_flags =
   { obligation_evars = Evar.Set.empty;
     aliased_evars = Evar.Map.empty;
-    typeclass_evars = Evar.Set.empty }
+    typeclass_evars = Evar.Set.empty;
+    impossible_case_evars = Evar.Set.empty;
+  }
 
 let empty_side_effects = {
   seff_private = Safe_typing.empty_private_constants;
@@ -956,6 +1004,7 @@ let empty = {
   conv_pbs   = [];
   last_mods  = Evar.Set.empty;
   evar_flags = empty_evar_flags;
+  candidate_evars = Evar.Set.empty;
   metas      = Metamap.empty;
   effects    = empty_side_effects;
   evar_names = EvNames.empty; (* id<->key for undefined evars *)
@@ -995,13 +1044,16 @@ let evar_key id evd = EvNames.key id evd.evar_names
 
 let get_aliased_evars evd = evd.evar_flags.aliased_evars
 
+let max_undefined_with_candidates evd =
+  try Some (Evar.Set.max_elt evd.candidate_evars) with Not_found -> None
+
 let is_aliased_evar evd evk =
   try Some (Evar.Map.find evk evd.evar_flags.aliased_evars)
   with Not_found -> None
 
 let downcast evk ccl evd =
   let evar_info = EvMap.find evk evd.undf_evars in
-  let evar_info' = { evar_info with evar_concl = ccl } in
+  let evar_info' = { evar_info with evar_concl = Undefined ccl } in
   { evd with undf_evars = EvMap.add evk evar_info' evd.undf_evars }
 
 (* extracts conversion problems that satisfy predicate p *)
@@ -1028,10 +1080,14 @@ let extract_all_conv_pbs evd =
 
 let loc_of_conv_pb evd (pbty,env,t1,t2) =
   match kind (fst (decompose_app t1)) with
-  | Evar (evk1,_) -> fst (evar_source (find evd evk1))
+  | Evar (evk1,_) ->
+    let EvarInfo evi = find evd evk1 in
+    fst (evar_source evi)
   | _ ->
   match kind (fst (decompose_app t2)) with
-  | Evar (evk2,_) -> fst (evar_source (find evd evk2))
+  | Evar (evk2,_) ->
+    let EvarInfo evi = find evd evk2 in
+    fst (evar_source evi)
   | _             -> None
 
 (**********************************************************)
@@ -1049,6 +1105,8 @@ let evar_universe_context d = d.universes
 
 let universe_context_set d = UState.context_set d.universes
 
+let sort_context_set d = UState.sort_context_set d.universes
+
 let to_universe_context evd = UState.context evd.universes
 
 let univ_entry ~poly evd = UState.univ_entry ~poly evd.universes
@@ -1061,11 +1119,20 @@ let restrict_universe_context evd vars =
 let universe_subst evd =
   UState.subst evd.universes
 
-let merge_context_set ?loc ?(sideff=false) rigid evd ctx' =
-  {evd with universes = UState.merge ?loc ~sideff rigid evd.universes ctx'}
+let merge_context_set ?loc ?(sideff=false) rigid evd uctx' =
+  {evd with universes = UState.merge ?loc ~sideff rigid evd.universes uctx'}
 
-let with_context_set ?loc rigid d (a, ctx) =
-  (merge_context_set ?loc rigid d ctx, a)
+let merge_sort_context_set ?loc ?(sideff=false) rigid evd ctx' =
+  {evd with universes = UState.merge_sort_context ?loc ~sideff rigid evd.universes ctx'}
+
+let merge_sort_variables ?loc ?(sideff=false) evd qs =
+  { evd with universes = UState.merge_sort_variables ?loc ~sideff evd.universes qs }
+
+let with_context_set ?loc rigid evd (a, uctx) =
+  (merge_context_set ?loc rigid evd uctx, a)
+
+let with_sort_context_set ?loc rigid d (a, ctx) =
+  (merge_sort_context_set ?loc rigid d ctx, a)
 
 let new_univ_level_variable ?loc ?name rigid evd =
   let uctx', u = UState.new_univ_variable ?loc rigid name evd.universes in
@@ -1075,17 +1142,18 @@ let new_univ_variable ?loc ?name rigid evd =
   let uctx', u = UState.new_univ_variable ?loc rigid name evd.universes in
     ({evd with universes = uctx'}, Univ.Universe.make u)
 
-let new_sort_variable ?loc ?name rigid d =
-  let (d', u) = new_univ_variable ?loc rigid ?name d in
-    (d', Sorts.sort_of_univ u)
+let new_quality_variable ?loc ?name evd =
+  let uctx, q = UState.new_sort_variable ?loc ?name evd.universes in
+  {evd with universes = uctx}, q
+
+let new_sort_variable ?loc rigid sigma =
+  let (sigma, u) = new_univ_variable ?loc rigid sigma in
+  let uctx, q = UState.new_sort_variable sigma.universes in
+  ({ sigma with universes = uctx }, Sorts.qsort q u)
 
 let add_global_univ d u =
   { d with universes = UState.add_global_univ d.universes u }
 
-let make_flexible_variable evd ~algebraic u =
-  { evd with universes =
-      UState.make_flexible_variable evd.universes ~algebraic u }
-
 let make_nonalgebraic_variable evd u =
   { evd with universes = UState.make_nonalgebraic_variable evd.universes u }
 
@@ -1094,28 +1162,26 @@ let make_nonalgebraic_variable evd u =
 (****************************************)
 
 let fresh_sort_in_family ?loc ?(rigid=univ_flexible) evd s =
-  with_context_set ?loc rigid evd (UnivGen.fresh_sort_in_family s)
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_sort_in_family s)
 
 let fresh_constant_instance ?loc ?(rigid=univ_flexible) env evd c =
-  with_context_set ?loc rigid evd (UnivGen.fresh_constant_instance env c)
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_constant_instance env c)
 
 let fresh_inductive_instance ?loc ?(rigid=univ_flexible) env evd i =
-  with_context_set ?loc rigid evd (UnivGen.fresh_inductive_instance env i)
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_inductive_instance env i)
 
 let fresh_constructor_instance ?loc ?(rigid=univ_flexible) env evd c =
-  with_context_set ?loc rigid evd (UnivGen.fresh_constructor_instance env c)
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_constructor_instance env c)
 
 let fresh_array_instance ?loc ?(rigid=univ_flexible) env evd =
-  with_context_set ?loc rigid evd (UnivGen.fresh_array_instance env)
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_array_instance env)
 
 let fresh_global ?loc ?(rigid=univ_flexible) ?names env evd gr =
-  with_context_set ?loc rigid evd (UnivGen.fresh_global_instance ?loc ?names env gr)
-
-let is_sort_variable evd s = UState.is_sort_variable evd.universes s
+  with_sort_context_set ?loc rigid evd (UnivGen.fresh_global_instance ?loc ?names env gr)
 
 let is_flexible_level evd l =
   let uctx = evd.universes in
-    Univ.Level.Map.mem l (UState.subst uctx)
+  UnivFlex.mem l (UState.subst uctx)
 
 let is_eq_sort s1 s2 =
   if Sorts.equal s1 s2 then None
@@ -1124,26 +1190,16 @@ let is_eq_sort s1 s2 =
 (* Precondition: l is not defined in the substitution *)
 let universe_rigidity evd l =
   let uctx = evd.universes in
+  (* XXX why are we considering all locals to be flexible here? *)
   if Univ.Level.Set.mem l (Univ.ContextSet.levels (UState.context_set uctx)) then
-    UnivFlexible (Univ.Level.Set.mem l (UState.algebraics uctx))
+    UnivFlexible (UState.is_algebraic l uctx)
   else UnivRigid
 
-let normalize_universe evd =
-  let vars = UState.subst evd.universes in
-  let normalize = UnivSubst.normalize_universe_opt_subst vars in
-    normalize
-
 let normalize_universe_instance evd l =
-  let vars = UState.subst evd.universes in
-  let normalize = UnivSubst.level_subst_of (UnivSubst.normalize_univ_variable_opt_subst vars) in
-  UnivSubst.subst_instance normalize l
+  UState.nf_instance evd.universes l
 
 let normalize_sort evars s =
-  match s with
-  | SProp | Prop | Set -> s
-  | Type u ->
-    let u' = normalize_universe evars u in
-    if u' == u then s else Sorts.sort_of_univ u'
+  UState.nf_sort evars.universes s
 
 (* FIXME inefficient *)
 let set_eq_sort env d s1 s2 =
@@ -1178,14 +1234,22 @@ let set_leq_sort env evd s1 s2 =
      else evd
 
 let check_eq evd s s' =
-  UGraph.check_eq_sort (UState.ugraph evd.universes) s s'
+  let ustate = evd.universes in
+  UGraph.check_eq_sort (UState.ugraph ustate) (UState.nf_sort ustate s) (UState.nf_sort ustate s')
 
 let check_leq evd s s' =
-  UGraph.check_leq_sort (UState.ugraph evd.universes) s s'
+  let ustate = evd.universes in
+  UGraph.check_leq_sort (UState.ugraph ustate) (UState.nf_sort ustate s) (UState.nf_sort ustate s')
 
 let check_constraints evd csts =
   UGraph.check_constraints csts (UState.ugraph evd.universes)
 
+let check_qconstraints evd csts =
+  UState.check_qconstraints evd.universes csts
+
+let check_quconstraints evd (qcsts,ucsts) =
+  check_qconstraints evd qcsts && check_constraints evd ucsts
+
 let fix_undefined_variables evd =
   { evd with universes = UState.fix_undefined_variables evd.universes }
 
@@ -1193,14 +1257,20 @@ let nf_univ_variables evd =
   let uctx = UState.normalize_variables evd.universes in
   {evd with universes = uctx}
 
+let collapse_sort_variables evd =
+  let universes = UState.collapse_sort_variables evd.universes in
+  { evd with universes }
 
 let minimize_universes evd =
-  let uctx' = UState.normalize_variables evd.universes in
+  let uctx' = UState.collapse_sort_variables evd.universes in
+  let uctx' = UState.normalize_variables uctx' in
   let uctx' = UState.minimize uctx' in
   {evd with universes = uctx'}
 
 let universe_of_name evd s = UState.universe_of_name evd.universes s
 
+let quality_of_name evd s = UState.quality_of_name evd.universes s
+
 let universe_binders evd = UState.universe_binders evd.universes
 
 let universes evd = UState.ugraph evd.universes
@@ -1290,12 +1360,12 @@ let new_pure_evar ?(src=default_source) ?(filter = Filter.identity) ?(relevance
   ?name ?typeclass_candidate ?(principal=false) sign evd typ =
   let evi = {
     evar_hyps = sign;
-    evar_concl = typ;
+    evar_concl = Undefined typ;
     evar_body = Evar_empty;
     evar_filter = filter;
-    evar_abstract_arguments = abstract_arguments;
+    evar_abstract_arguments = Undefined abstract_arguments;
     evar_source = src;
-    evar_candidates = candidates;
+    evar_candidates = Undefined candidates;
     evar_relevance = relevance;
   }
   in
@@ -1318,7 +1388,12 @@ let define_aux def undef evk body =
         anomaly ~label:"Evd.define" (Pp.str "cannot define undeclared evar.")
   in
   let () = assert (oldinfo.evar_body == Evar_empty) in
-  let newinfo = { oldinfo with evar_body = Evar_defined body } in
+  let newinfo = { oldinfo with
+    evar_body = Evar_defined body;
+    evar_concl = Defined;
+    evar_candidates = Defined;
+    evar_abstract_arguments = Defined;
+  } in
   EvMap.add evk newinfo def, EvMap.remove evk undef
 
 (* define the existential of section path sp as the constr body *)
@@ -1331,7 +1406,8 @@ let define_gen evk body evd evar_flags =
   | _ -> Evar.Set.add evk evd.last_mods
   in
   let evar_names = EvNames.remove_name_defined evk evd.evar_names in
-  { evd with defn_evars; undf_evars; last_mods; evar_names; evar_flags }
+  let candidate_evars = Evar.Set.remove evk evd.candidate_evars in
+  { evd with defn_evars; undf_evars; last_mods; evar_names; evar_flags; candidate_evars }
 
 (** By default, the obligation and evar tag of the evar is removed *)
 let define evk body evd =
@@ -1355,7 +1431,7 @@ let restrict evk filter ?candidates ?src evd =
   let id_inst = Filter.filter_slist filter (SList.defaultn len SList.empty) in
   let evar_info' =
     { evar_info with evar_filter = filter;
-      evar_candidates = candidates;
+      evar_candidates = Undefined candidates;
       evar_source = (match src with None -> evar_info.evar_source | Some src -> src);
     } in
   let last_mods = match evd.conv_pbs with
@@ -1365,8 +1441,18 @@ let restrict evk filter ?candidates ?src evd =
   let body = mkEvar(evk',id_inst) in
   let (defn_evars, undf_evars) = define_aux evd.defn_evars evd.undf_evars evk body in
   let evar_flags = inherit_evar_flags evd.evar_flags evk evk' in
+  let evar_flags = match src with
+    | Some (_,Evar_kinds.ImpossibleCase) ->
+      { evar_flags with impossible_case_evars = Evar.Set.add evk' evar_flags.impossible_case_evars }
+    | _ -> evar_flags
+  in
+  let candidate_evars = Evar.Set.remove evk evd.candidate_evars in
+  let candidate_evars = match candidates with
+  | None -> candidate_evars
+  | Some _ -> Evar.Set.add evk' candidate_evars
+  in
   let evd = { evd with undf_evars = EvMap.add evk' evar_info' undf_evars;
-    defn_evars; last_mods; evar_names; evar_flags }
+    defn_evars; last_mods; evar_names; evar_flags; candidate_evars }
   in
   (* Mark new evar as future goal, removing previous one,
     circumventing Proofview.advance but making Proof.run_tactic catch these. *)
@@ -1376,9 +1462,8 @@ let restrict evk filter ?candidates ?src evd =
   (evd, evk')
 
 let update_source evd evk src =
-  let evar_info = EvMap.find evk evd.undf_evars in
-  let evar_info' = { evar_info with evar_source = src } in
-  { evd with undf_evars = EvMap.add evk evar_info' evd.undf_evars }
+  let modify _ info = { info with evar_source = src } in
+  { evd with undf_evars = EvMap.modify evk modify evd.undf_evars }
 
 (**********************************************************)
 (* Accessing metas *)
@@ -1392,6 +1477,7 @@ let set_metas evd metas = {
   conv_pbs = evd.conv_pbs;
   last_mods = evd.last_mods;
   evar_flags = evd.evar_flags;
+  candidate_evars = evd.candidate_evars;
   metas;
   effects = evd.effects;
   evar_names = evd.evar_names;
@@ -1490,7 +1576,7 @@ let retract_coercible_metas evd =
   !mc, set_metas evd metas
 
 let dependent_evar_ident ev evd =
-  let evi = find evd ev in
+  let EvarInfo evi = find evd ev in
   match evi.evar_source with
   | (_,Evar_kinds.VarInstance id) -> id
   | _ -> anomaly (str "Not an evar resulting of a dependent binding.")
@@ -1573,21 +1659,19 @@ module MiniEConstr = struct
   struct
     type t = Sorts.t
     let make s = s
-    let kind sigma = function
-      | Sorts.Type u -> Sorts.sort_of_univ (normalize_universe sigma u)
-      | s -> s
+    let kind = normalize_sort
     let unsafe_to_sorts s = s
   end
 
   module EInstance =
   struct
-    type t = Univ.Instance.t
+    type t = UVars.Instance.t
     let make i = i
     let kind sigma i =
-      if Univ.Instance.is_empty i then i
+      if UVars.Instance.is_empty i then i
       else normalize_universe_instance sigma i
-    let empty = Univ.Instance.empty
-    let is_empty = Univ.Instance.is_empty
+    let empty = UVars.Instance.empty
+    let is_empty = UVars.Instance.is_empty
     let unsafe_to_instance t = t
   end
 
@@ -1641,11 +1725,15 @@ module MiniEConstr = struct
         Some (mkEvar (evk, SList.of_full_list args))
       else None
     | Some info ->
-      match evar_body info with
-      | Evar_defined c -> Some (instantiate_evar_array sigma info c args)
-      | Evar_empty -> assert false
+      let Evar_defined c = evar_body info in
+      Some (instantiate_evar_array sigma info c args)
     in
-    UnivSubst.nf_evars_and_universes_opt_subst evar_value (universe_subst sigma) c
+    let lsubst = universe_subst sigma in
+    let univ_value l =
+      UnivFlex.normalize_univ_variable lsubst l
+    in
+    let qvar_value q = UState.nf_qvar sigma.universes q in
+    UnivSubst.nf_evars_and_universes_opt_subst evar_value qvar_value univ_value c
 
   let to_constr_gen sigma c =
     let saw_evar = ref false in
@@ -1654,7 +1742,12 @@ module MiniEConstr = struct
       saw_evar := !saw_evar || Option.is_empty v;
       v
     in
-    let c = UnivSubst.nf_evars_and_universes_opt_subst evar_value (universe_subst sigma) c in
+    let lsubst = universe_subst sigma in
+    let univ_value l =
+      UnivFlex.normalize_univ_variable lsubst l
+    in
+    let qvar_value q = UState.nf_qvar sigma.universes q in
+    let c = UnivSubst.nf_evars_and_universes_opt_subst evar_value qvar_value univ_value c in
     let saw_evar = if not !saw_evar then false
       else
         let exception SawEvar in
@@ -1682,7 +1775,13 @@ module MiniEConstr = struct
   let unsafe_to_named_decl d = d
   let of_rel_decl d = d
   let unsafe_to_rel_decl d = d
-  let to_rel_decl sigma d = Context.Rel.Declaration.map_constr (to_constr sigma) d
+  let to_rel_decl sigma d = match d with
+  | RelDecl.LocalAssum (na, t) ->
+    let na = UnivSubst.nf_binder_annot (fun r -> UState.nf_relevance sigma.universes r) na in
+    RelDecl.LocalAssum (na, to_constr sigma t)
+  | RelDecl.LocalDef (na, c, t) ->
+    let na = UnivSubst.nf_binder_annot (fun r -> UState.nf_relevance sigma.universes r) na in
+    RelDecl.LocalDef (na, to_constr sigma c, to_constr sigma t)
 
   let of_named_context d = d
   let of_rel_context d = d
@@ -1712,10 +1811,37 @@ let evars_of_named_context evd nc =
     nc
     ~init:Evar.Set.empty
 
-let evars_of_filtered_evar_info evd evi =
-  Evar.Set.union (evars_of_term evd evi.evar_concl)
+let evars_of_filtered_evar_info (type a) evd (evi : a evar_info) =
+  let concl = match evi.evar_concl with
+  | Undefined c -> evars_of_term evd c
+  | Defined -> Evar.Set.empty
+  in
+  Evar.Set.union concl
     (Evar.Set.union
        (match evi.evar_body with
        | Evar_empty -> Evar.Set.empty
        | Evar_defined b -> evars_of_term evd b)
        (evars_of_named_context evd (evar_filtered_context evi)))
+
+let drop_new_defined ~original sigma =
+  let to_keep, to_drop = Evar.Map.partition (fun ev _ ->
+      Evar.Map.mem ev original.defn_evars || Evar.Map.mem ev original.undf_evars)
+      sigma.defn_evars
+  in
+  let dummy = { empty with defn_evars = to_drop } in
+  let nfc c = MiniEConstr.to_constr_nocheck dummy c in
+  assert (Metamap.is_empty sigma.metas);
+  assert (List.is_empty sigma.conv_pbs);
+  let normalize_changed _ev orig evi =
+    match orig, evi with
+    | _, None -> None
+    | None, Some evi -> Some (map_evar_info nfc evi)
+    | Some orig, Some evi -> if orig == evi then None else Some (map_evar_info nfc evi)
+  in
+  let normalize_against original current =
+    let normalized = EvMap.merge normalize_changed original current in
+    EvMap.union (fun _ _ x -> Some x) current normalized
+  in
+  let to_keep = normalize_against original.defn_evars to_keep in
+  let undf_evars = normalize_against original.undf_evars sigma.undf_evars in
+  { sigma with defn_evars = to_keep; undf_evars }
