@@ -30,6 +30,7 @@ open Environ
 
 type econstr
 type etypes = econstr
+type esorts
 
 (** {5 Existential variables and unification states} *)
 
@@ -94,54 +95,59 @@ end
 
 (** {6 Evar infos} *)
 
-type evar_body =
-  | Evar_empty
-  | Evar_defined of econstr
+type defined = [ `defined ]
+type undefined = [ `undefined ]
 
-type evar_info
+type _ evar_body =
+  | Evar_empty : undefined evar_body
+  | Evar_defined : econstr -> defined evar_body
+
+type 'a evar_info
+
+type any_evar_info = EvarInfo : 'a evar_info -> any_evar_info
 
 (** {6 Projections from evar infos} *)
 
-val evar_concl : evar_info -> econstr
+val evar_concl : undefined evar_info -> econstr
 (** Type of the evar. *)
 
-val evar_context : evar_info -> (econstr, etypes) Context.Named.pt
+val evar_context : 'a evar_info -> (econstr, etypes) Context.Named.pt
 (** Context of the evar. *)
 
-val evar_hyps : evar_info -> named_context_val
+val evar_hyps : 'a evar_info -> named_context_val
 (** Context of the evar. *)
 
-val evar_body : evar_info -> evar_body
+val evar_body : 'a evar_info -> 'a evar_body
 (** Optional content of the evar. *)
 
-val evar_candidates : evar_info -> econstr list option
+val evar_candidates : undefined evar_info -> econstr list option
 (** List of possible solutions when known that it is a finite list *)
 
-val evar_source : evar_info -> Evar_kinds.t located
+val evar_source : 'a evar_info -> Evar_kinds.t located
 
-val evar_filter : evar_info -> Filter.t
+val evar_filter : 'a evar_info -> Filter.t
 (** Boolean mask over {!evar_hyps}. Should have the same length.
     When filtered out, the corresponding variable is not allowed to occur
     in the solution *)
 
-val evar_abstract_arguments : evar_info -> Abstraction.t
+val evar_abstract_arguments : undefined evar_info -> Abstraction.t
 (** Boolean information over {!evar_hyps}, telling if an hypothesis instance
     can be imitated or should stay abstract in HO unification problems
     and inversion (see [second_order_matching_with_args] for its use). *)
 
-val evar_relevance : evar_info -> Sorts.relevance
+val evar_relevance : 'a evar_info -> Sorts.relevance
 (** Relevance of the conclusion of the evar. *)
 
 (** {6 Derived projections} *)
 
-val evar_filtered_context : evar_info -> (econstr, etypes) Context.Named.pt
-val evar_filtered_hyps : evar_info -> named_context_val
-val evar_env : env -> evar_info -> env
-val evar_filtered_env : env -> evar_info -> env
-val evar_identity_subst : evar_info -> econstr SList.t
+val evar_filtered_context : 'a evar_info -> (econstr, etypes) Context.Named.pt
+val evar_filtered_hyps : 'a evar_info -> named_context_val
+val evar_env : env -> 'a evar_info -> env
+val evar_filtered_env : env -> 'a evar_info -> env
+val evar_identity_subst : 'a evar_info -> econstr SList.t
 
-val map_evar_body : (econstr -> econstr) -> evar_body -> evar_body
-val map_evar_info : (econstr -> econstr) -> evar_info -> evar_info
+val map_evar_body : (econstr -> econstr) -> 'a evar_body -> 'a evar_body
+val map_evar_info : (econstr -> econstr) -> 'a evar_info -> 'a evar_info
 
 (** {6 Unification state} **)
 
@@ -197,40 +203,44 @@ val new_pure_evar :
   @param types The type of conclusion of the evar
 *)
 
-val add : evar_map -> Evar.t -> evar_info -> evar_map
+val add : evar_map -> Evar.t -> 'a evar_info -> evar_map
 (** [add sigma ev info] adds [ev] with evar info [info] in sigma.
     Precondition: ev must not preexist in [sigma]. *)
 
-val find : evar_map -> Evar.t -> evar_info
+val find_defined : evar_map -> Evar.t -> defined evar_info option
+
+val find : evar_map -> Evar.t -> any_evar_info
 (** Recover the data associated to an evar. *)
 
-val find_undefined : evar_map -> Evar.t -> evar_info
+val find_undefined : evar_map -> Evar.t -> undefined evar_info
 (** Same as {!find} but restricted to undefined evars. For efficiency
     reasons. *)
 
 val remove : evar_map -> Evar.t -> evar_map
 (** Remove an evar from an evar map. Use with caution. *)
 
-val undefine : evar_map -> Evar.t -> evar_map [@@ocaml.deprecated]
+val undefine : evar_map -> Evar.t -> etypes -> evar_map [@@ocaml.deprecated]
 (** Remove the body of an evar. Only there for backward compat, do not use. *)
 
 val mem : evar_map -> Evar.t -> bool
 (** Whether an evar is present in an evarmap. *)
 
-val fold : (Evar.t -> evar_info -> 'a -> 'a) -> evar_map -> 'a -> 'a
+val fold : (Evar.t -> any_evar_info -> 'a -> 'a) -> evar_map -> 'a -> 'a
 (** Apply a function to all evars and their associated info in an evarmap. *)
 
-val fold_undefined : (Evar.t -> evar_info -> 'a -> 'a) -> evar_map -> 'a -> 'a
+val fold_undefined : (Evar.t -> undefined evar_info -> 'a -> 'a) -> evar_map -> 'a -> 'a
 (** Same as {!fold}, but restricted to undefined evars. For efficiency
     reasons. *)
 
-val raw_map : (Evar.t -> evar_info -> evar_info) -> evar_map -> evar_map
+type map = { map : 'r. Evar.t -> 'r evar_info -> 'r evar_info }
+
+val raw_map : map -> evar_map -> evar_map
 (** Apply the given function to all evars in the map. Beware: this function
     expects the argument function to preserve the kind of [evar_body], i.e. it
     must send [Evar_empty] to [Evar_empty] and [Evar_defined c] to some
     [Evar_defined c']. *)
 
-val raw_map_undefined : (Evar.t -> evar_info -> evar_info) -> evar_map -> evar_map
+val raw_map_undefined : (Evar.t -> undefined evar_info -> undefined evar_info) -> evar_map -> evar_map
 (** Same as {!raw_map}, but restricted to undefined evars. For efficiency
     reasons. *)
 
@@ -261,11 +271,16 @@ val is_undefined : evar_map -> Evar.t-> bool
 val add_constraints : evar_map -> Univ.Constraints.t -> evar_map
 (** Add universe constraints in an evar map. *)
 
-val undefined_map : evar_map -> evar_info Evar.Map.t
+val add_quconstraints : evar_map -> Sorts.QUConstraints.t -> evar_map
+
+val undefined_map : evar_map -> undefined evar_info Evar.Map.t
 (** Access the undefined evar mapping directly. *)
 
 val drop_all_defined : evar_map -> evar_map
 
+val drop_new_defined : original:evar_map -> evar_map -> evar_map
+(** Drop the defined evars in the second evar map which did not exist in the first. *)
+
 val is_maybe_typeclass_hook : (evar_map -> constr -> bool) Hook.t
 
 (** {6 Instantiating partial terms} *)
@@ -278,6 +293,8 @@ val existential_value : evar_map -> econstr pexistential -> econstr
 
 val existential_value0 : evar_map -> existential -> constr
 
+val existential_type_opt : evar_map -> econstr pexistential -> etypes option
+
 val existential_type : evar_map -> econstr pexistential -> etypes
 
 val existential_type0 : evar_map -> existential -> types
@@ -288,9 +305,9 @@ val existential_opt_value : evar_map -> econstr pexistential -> econstr option
 
 val existential_opt_value0 : evar_map -> existential -> constr option
 
-val evar_handler : evar_map -> constr evar_handler
+val evar_handler : evar_map -> constr CClosure.evar_handler
 
-val existential_expand_value0 : evar_map -> existential -> constr Constr.evar_expansion
+val existential_expand_value0 : evar_map -> existential -> constr CClosure.evar_expansion
 
 val expand_existential : evar_map -> econstr pexistential -> econstr list
 (** Returns the full evar instance with implicit default variables turned into
@@ -298,7 +315,7 @@ val expand_existential : evar_map -> econstr pexistential -> econstr list
 
 val expand_existential0 : evar_map -> constr pexistential -> constr list
 
-val instantiate_evar_array : evar_map -> evar_info -> econstr -> econstr SList.t -> econstr
+val instantiate_evar_array : evar_map -> 'a evar_info -> econstr -> econstr SList.t -> econstr
 
 (** {6 Misc} *)
 
@@ -318,6 +335,9 @@ val get_aliased_evars : evar_map -> Evar.t Evar.Map.t
 val is_aliased_evar : evar_map -> Evar.t -> Evar.t option
 (** Tell if an evar has been aliased to another evar, and if yes, which *)
 
+val max_undefined_with_candidates : evar_map -> Evar.t option
+(** If any, the evar with highest id with a non-empty list of candidates. *)
+
 val set_typeclass_evars : evar_map -> Evar.Set.t -> evar_map
 (** Mark the given set of evars as available for resolution.
 
@@ -339,6 +359,9 @@ val set_obligation_evar : evar_map -> Evar.t -> evar_map
 val is_obligation_evar : evar_map -> Evar.t -> bool
 (** Is the evar declared as an obligation *)
 
+val get_impossible_case_evars : evar_map -> Evar.Set.t
+(** Set of undefined evars with ImpossibleCase evar source. *)
+
 val downcast : Evar.t-> etypes -> evar_map -> evar_map
 (** Change the type of an undefined evar to a new type assumed to be a
     subtype of its current type; subtyping must be ensured by caller *)
@@ -539,7 +562,7 @@ type clbinding =
   | Clval of Name.t * (econstr freelisted * instance_status) * econstr freelisted
 
 (** Unification constraints *)
-type conv_pb = Reduction.conv_pb
+type conv_pb = Conversion.conv_pb
 type evar_constraint = conv_pb * env * econstr * econstr
 
 (** The following two functions are for internal use only,
@@ -561,7 +584,7 @@ val evars_of_term : evar_map -> econstr -> Evar.Set.t
 
 val evars_of_named_context : evar_map -> (econstr, etypes) Context.Named.pt -> Evar.Set.t
 
-val evars_of_filtered_evar_info : evar_map -> evar_info -> Evar.Set.t
+val evars_of_filtered_evar_info : evar_map -> 'a evar_info -> Evar.Set.t
 
 (** Metas *)
 val meta_list : evar_map -> clbinding Metamap.t
@@ -621,50 +644,49 @@ val restrict_universe_context : evar_map -> Univ.Level.Set.t -> evar_map
 
 (** Raises Not_found if not a name for a universe in this map. *)
 val universe_of_name : evar_map -> Id.t -> Univ.Level.t
+val quality_of_name : evar_map -> Id.t -> Sorts.QVar.t
 
 val universe_binders : evar_map -> UnivNames.universe_binders
 
 val new_univ_level_variable : ?loc:Loc.t -> ?name:Id.t -> rigid -> evar_map -> evar_map * Univ.Level.t
-val new_sort_variable : ?loc:Loc.t -> ?name:Id.t -> rigid -> evar_map -> evar_map * Sorts.t
+val new_quality_variable : ?loc:Loc.t -> ?name:Id.t -> evar_map -> evar_map * Sorts.QVar.t
+val new_sort_variable : ?loc:Loc.t -> rigid -> evar_map -> evar_map * esorts
 
 val add_global_univ : evar_map -> Univ.Level.t -> evar_map
 
 val universe_rigidity : evar_map -> Univ.Level.t -> rigid
-val make_flexible_variable : evar_map -> algebraic:bool -> Univ.Level.t -> evar_map
-(** See [UState.make_flexible_variable] *)
 
 val make_nonalgebraic_variable : evar_map -> Univ.Level.t -> evar_map
 (** See [UState.make_nonalgebraic_variable]. *)
 
-val is_sort_variable : evar_map -> Sorts.t -> Univ.Level.t option
-(** [is_sort_variable evm s] returns [Some u] or [None] if [s] is
-    not a local sort variable declared in [evm] *)
-
 val is_flexible_level : evar_map -> Univ.Level.t -> bool
 
-val normalize_universe_instance : evar_map -> Univ.Instance.t -> Univ.Instance.t
+val normalize_universe_instance : evar_map -> UVars.Instance.t -> UVars.Instance.t
 
-val set_leq_sort : env -> evar_map -> Sorts.t -> Sorts.t -> evar_map
-val set_eq_sort : env -> evar_map -> Sorts.t -> Sorts.t -> evar_map
+val set_leq_sort : env -> evar_map -> esorts -> esorts -> evar_map
+val set_eq_sort : env -> evar_map -> esorts -> esorts -> evar_map
 val set_eq_level : evar_map -> Univ.Level.t -> Univ.Level.t -> evar_map
 val set_leq_level : evar_map -> Univ.Level.t -> Univ.Level.t -> evar_map
 val set_eq_instances : ?flex:bool ->
-  evar_map -> Univ.Instance.t -> Univ.Instance.t -> evar_map
+  evar_map -> UVars.Instance.t -> UVars.Instance.t -> evar_map
 
-val check_eq : evar_map -> Sorts.t -> Sorts.t -> bool
-val check_leq : evar_map -> Sorts.t -> Sorts.t -> bool
+val check_eq : evar_map -> esorts -> esorts -> bool
+val check_leq : evar_map -> esorts -> esorts -> bool
 
 val check_constraints : evar_map -> Univ.Constraints.t -> bool
+val check_qconstraints : evar_map -> Sorts.QConstraints.t -> bool
+val check_quconstraints : evar_map -> Sorts.QUConstraints.t -> bool
 
 val evar_universe_context : evar_map -> UState.t
 val universe_context_set : evar_map -> Univ.ContextSet.t
-val universe_subst : evar_map -> UnivSubst.universe_opt_subst
+val sort_context_set : evar_map -> UnivGen.sort_context_set
+val universe_subst : evar_map -> UnivFlex.t
 val universes : evar_map -> UGraph.t
 
 (** [to_universe_context evm] extracts the local universes and
     constraints of [evm] and orders the universes the same as
     [Univ.ContextSet.to_context]. *)
-val to_universe_context : evar_map -> Univ.UContext.t
+val to_universe_context : evar_map -> UVars.UContext.t
 
 val univ_entry : poly:bool -> evar_map -> UState.named_universes_entry
 
@@ -675,10 +697,18 @@ val set_universe_context : evar_map -> UState.t -> evar_map
 
 val merge_context_set : ?loc:Loc.t -> ?sideff:bool -> rigid -> evar_map -> Univ.ContextSet.t -> evar_map
 
+val merge_sort_context_set : ?loc:Loc.t -> ?sideff:bool -> rigid -> evar_map -> UnivGen.sort_context_set -> evar_map
+
+val merge_sort_variables : ?loc:Loc.t -> ?sideff:bool -> evar_map -> Sorts.QVar.Set.t -> evar_map
+
 val with_context_set : ?loc:Loc.t -> rigid -> evar_map -> 'a Univ.in_universe_context_set -> evar_map * 'a
 
+val with_sort_context_set : ?loc:Loc.t -> rigid -> evar_map -> 'a UnivGen.in_sort_context_set -> evar_map * 'a
+
 val nf_univ_variables : evar_map -> evar_map
 
+val collapse_sort_variables : evar_map -> evar_map
+
 val fix_undefined_variables : evar_map -> evar_map
 
 (** Universe minimization *)
@@ -690,7 +720,7 @@ val update_sigma_univs : UGraph.t -> evar_map -> evar_map
 (** Polymorphic universes *)
 
 val fresh_sort_in_family : ?loc:Loc.t -> ?rigid:rigid
-  -> evar_map -> Sorts.family -> evar_map * Sorts.t
+  -> evar_map -> Sorts.family -> evar_map * esorts
 val fresh_constant_instance : ?loc:Loc.t -> ?rigid:rigid
   -> env -> evar_map -> Constant.t -> evar_map * pconstant
 val fresh_inductive_instance : ?loc:Loc.t -> ?rigid:rigid
@@ -698,9 +728,9 @@ val fresh_inductive_instance : ?loc:Loc.t -> ?rigid:rigid
 val fresh_constructor_instance : ?loc:Loc.t -> ?rigid:rigid
   -> env -> evar_map -> constructor -> evar_map * pconstructor
 val fresh_array_instance : ?loc:Loc.t -> ?rigid:rigid
-  -> env -> evar_map  -> evar_map * Univ.Instance.t
+  -> env -> evar_map  -> evar_map * UVars.Instance.t
 
-val fresh_global : ?loc:Loc.t -> ?rigid:rigid -> ?names:Univ.Instance.t -> env ->
+val fresh_global : ?loc:Loc.t -> ?rigid:rigid -> ?names:UVars.Instance.t -> env ->
   evar_map -> GlobRef.t -> evar_map * econstr
 
 (********************************************************************)
@@ -730,7 +760,7 @@ val create_evar_defs : evar_map -> evar_map
 (** Use this module only to bootstrap EConstr *)
 module MiniEConstr : sig
   module ESorts : sig
-    type t
+    type t = esorts
     val make : Sorts.t -> t
     val kind : evar_map -> t -> Sorts.t
     val unsafe_to_sorts : t -> Sorts.t
@@ -738,17 +768,17 @@ module MiniEConstr : sig
 
   module EInstance : sig
     type t
-    val make : Univ.Instance.t -> t
-    val kind : evar_map -> t -> Univ.Instance.t
+    val make : UVars.Instance.t -> t
+    val kind : evar_map -> t -> UVars.Instance.t
     val empty : t
     val is_empty : t -> bool
-    val unsafe_to_instance : t -> Univ.Instance.t
+    val unsafe_to_instance : t -> UVars.Instance.t
   end
 
   type t = econstr
 
   val kind : evar_map -> t -> (t, t, ESorts.t, EInstance.t) Constr.kind_of_term
-  val kind_upto : evar_map -> constr -> (constr, types, Sorts.t, Univ.Instance.t) Constr.kind_of_term
+  val kind_upto : evar_map -> constr -> (constr, types, Sorts.t, UVars.Instance.t) Constr.kind_of_term
 
   val whd_evar : evar_map -> t -> t
 
