@@ -10,25 +10,9 @@
 
 open Util
 open Coq
-open Ideutils
 open Interface
 open Feedback
 
-let b2c = byte_offset_to_char_offset
-
-(** variant of buffer_off_to_byte_off: convert unicode offset (used by GTK buffer)
-    to UTF-8 byte offset (used by Coq) with size caching for performance *)
-let c2b (buffer : GText.buffer) (s_byte, s_uni) uni_off =
-  if s_uni < 0 then
-    raise (invalid_arg "c2b: s_uni must be >= 0");
-  let rec cvt iter rv =
-    if iter#offset <= s_uni then
-      rv
-    else
-      cvt iter#backward_char (rv + (ulen iter#char))
-  in
-  cvt (buffer#get_iter (`OFFSET uni_off)) s_byte
-
 type flag = [ `INCOMPLETE | `UNSAFE | `PROCESSING | `ERROR of string Loc.located | `WARNING of string Loc.located ]
 type mem_flag = [ `INCOMPLETE | `UNSAFE | `PROCESSING | `ERROR | `WARNING ]
 let mem_flag_of_flag : flag -> mem_flag = function
@@ -50,19 +34,22 @@ end
 
 module SentenceId : sig
 
-  type sentence = {
+  type sentence = private {
     start : GText.mark;
     stop : GText.mark;
-    mutable byte_start : int;
     mutable flags : flag list;
     mutable tooltips : (int * int * string) list;
+    start_offset : int;
+    (** Original start offset of the sentence, to be compared when moved *)
+    start_line : int;
+    (** Original start line of the sentence, to be compared when moved *)
     edit_id : int;
     mutable index : int;
     changed_sig : (int * mem_flag list) GUtil.signal;
   }
 
   val mk_sentence :
-    start:GText.mark -> stop:GText.mark -> flag list -> sentence
+    start:GText.mark -> stop:GText.mark -> start_offset:int -> start_line:int -> flag list -> sentence
 
   val add_flag : sentence -> flag -> unit
   val has_flag : sentence -> mem_flag -> bool
@@ -73,6 +60,16 @@ module SentenceId : sig
 
   val connect : sentence -> signals
 
+  val start_iter : GText.buffer -> sentence -> GText.iter
+  val start_stop_iters : GText.buffer -> sentence -> GText.iter * GText.iter
+
+  (** [offset_compensation] Drift of a sentence over its original
+      location, as an [(offset, line)] pair. Needed to repair outdated
+      Coq locations *)
+  val offset_compensation : GText.buffer -> sentence -> int * int
+
+  val phrase : GText.buffer -> sentence -> string
+
   val dbg_to_string :
     GText.buffer -> bool -> Stateid.t option -> sentence -> Pp.t
 
@@ -81,9 +78,12 @@ end = struct
   type sentence = {
     start : GText.mark;
     stop : GText.mark;
-    mutable byte_start : int;
     mutable flags : flag list;
     mutable tooltips : (int * int * string) list;
+    start_offset : int;
+    (** Original start offset of the sentence, to be compared when moved *)
+    start_line : int;
+    (** Original start line of the sentence, to be compared when moved *)
     edit_id : int;
     mutable index : int;
     changed_sig : (int * mem_flag list) GUtil.signal;
@@ -96,13 +96,14 @@ end = struct
     end
 
   let id = ref 0
-  let mk_sentence ~start ~stop flags = decr id; {
-    start = start;
-    stop = stop;
-    byte_start = 0;
-    flags = flags;
+  let mk_sentence ~start ~stop ~start_offset ~start_line flags = decr id; {
+    start;
+    stop;
+    flags;
     edit_id = !id;
     tooltips = [];
+    start_offset;
+    start_line;
     index = -1;
     changed_sig = new GUtil.signal ();
   }
@@ -123,6 +124,17 @@ end = struct
 
   let set_index s i = s.index <- i
 
+  let start_iter buffer sentence = buffer#get_iter_at_mark sentence.start
+  let stop_iter buffer sentence = buffer#get_iter_at_mark sentence.stop
+  let start_stop_iters buffer sentence = start_iter buffer sentence, stop_iter buffer sentence
+  let offset_compensation buffer sentence =
+    let s_start = buffer#get_iter_at_mark sentence.start in
+    s_start#offset - sentence.start_offset, s_start#line - sentence.start_line
+
+  let phrase buffer sentence =
+    let start, stop = start_stop_iters buffer sentence in
+    start#get_slice ~stop
+
   let dbg_to_string (b : GText.buffer) focused id s =
     let ellipsize s =
             Str.global_replace (Str.regexp "^[\n ]*") ""
@@ -146,6 +158,54 @@ end = struct
 end
 open SentenceId
 
+(* Given a Coq loc, convert it to a pair of iterators start / end in
+   the buffer. *)
+let coq_loc_to_gtk_offset ?(line_drift=0) (buffer : GText.buffer) loc =
+  Ideutils.get_iter_at_byte buffer ~line:(loc.Loc.line_nb - 1 + line_drift) (loc.bp - loc.bol_pos),
+  Ideutils.get_iter_at_byte buffer ~line:(loc.Loc.line_nb_last - 1 + line_drift) (loc.ep - loc.bol_pos_last)
+
+(** increase [uni_off] by the number of bytes until [s_uni] This can
+    be used to convert a character offset to byte offset if we know a
+    previous point denoted by the pair [s_byte, s_uni] *)
+let c2b (buffer : GText.buffer) (s_byte, s_uni) uni_off =
+  if s_uni < 0 then
+    raise (invalid_arg "c2b: s_uni must be >= 0");
+  let rec cvt iter rv =
+    if iter#offset <= s_uni then
+      rv
+    else
+      cvt iter#backward_char (rv + (Ideutils.ulen iter#char))
+  in
+  cvt (buffer#get_iter (`OFFSET uni_off)) s_byte
+
+(* Given [sentence], returns [bp, line, bol] where [bp] is the
+   absolute offset in _bytes_ for the start of [sentence], [line] is
+   the line number for the start of sentence and [bol] is the line
+   ofsset of [sentence] in _bytes_.
+
+   This is meant to be used in the resumption of parsing, to tell Coq
+   where the input stream was left at. In this case, these 3 points
+   uniquely determine the position where to restart
+   from. [CLexer.after] can compute this from the position of the last
+   parsed token.
+
+   However, for CoqIDE, we have a huge problem in the sense that
+   CoqIDE will do its own parsing and the locations may not start at
+   the same point. Thus, we need to perform potentially very costly
+   char to offset conversion, for that we need the whole buffer.
+*)
+let coq_loc_from_gtk_offset cached buffer sentence =
+  let start = start_iter buffer sentence in
+  (* This is in chars not in bytes, thus needs conversion *)
+  let bp = c2b buffer cached start#offset in
+  (* Coq lines start at 1, GTK lines at 0 *)
+  let line = start#line + 1 in
+  (* [line_index] already returns byte offsets *)
+  let bol_pos = bp - start#line_index in
+  (*  *)
+  let new_cached = bp, start#offset in
+  bp, line, bol_pos, new_cached
+
 let log msg : unit task =
   Coq.lift (fun () -> Minilib.log msg)
 
@@ -253,6 +313,12 @@ type queue_msg =
   | MsgFeedback of Feedback.feedback
   | MsgDebug of string * Pp.t
 
+let rec flatten = function
+| [] -> []
+| (lg, rg) :: l ->
+  let inner = flatten l in
+  List.rev_append lg inner @ rg
+
 class coqops
   (_script:Wg_ScriptView.script_view)
   (_pv:Wg_ProofView.proof_view)
@@ -318,7 +384,7 @@ object(self)
     let x, y = script#window_to_buffer_coords ~tag:`WIDGET ~x ~y in
     let iter = script#get_iter_at_location ~x ~y in
     if iter#has_tag Tags.Script.tooltip then begin
-      let s =
+      let s : SentenceId.sentence =
         let rec aux iter =
           let marks = iter#marks in
           if marks = [] then aux iter#backward_char
@@ -330,7 +396,12 @@ object(self)
             try Doc.find document mem_marks
             with Not_found -> aux iter#backward_char in
         aux iter in
-      let ss = find_all_tooltips s iter#offset in
+      (* The original list of tooltips contains offset that were not
+         set up to date if the sentences moved, however the GTK model
+         has the up-to-date information in the marks, so we can
+         compare and compensate for the shift. *)
+      let offset, _line_shift = offset_compensation script#buffer s in
+      let ss = find_all_tooltips s (iter#offset - offset) in
       let msg = String.concat "\n" (CList.uniquize ss) in
       GtkBase.Tooltip.set_icon_from_stock tooltip `INFO `BUTTON;
       script#misc#set_tooltip_markup ("<tt>" ^ msg ^ "</tt>")
@@ -395,18 +466,37 @@ object(self)
         script#recenter_insert
       end
     end;
-    Coq.bind (Coq.goals ()) (function
+    let flags = { gf_mode = "full"; gf_fg = true; gf_bg = true; gf_shelved = false; gf_given_up = false } in
+    let return x = Coq.return (Good x) in
+    let (>>=) m f = Coq.bind m (function
+    | Fail x -> Coq.return (Fail x)
+    | Good v -> f v)
+    in
+    let call =
+      Coq.subgoals flags >>= begin function
+      | None -> return Wg_ProofView.NoGoals
+      | Some { fg_goals = ((_ :: _) as fg); bg_goals = bg } ->
+        let bg = flatten (List.rev bg) in
+        return (Wg_ProofView.FocusGoals { fg; bg; })
+      | Some { fg_goals = []; bg_goals = bg } ->
+        let flags = { gf_mode = "short"; gf_fg = false; gf_bg = false; gf_shelved = true; gf_given_up = true } in
+        Coq.subgoals flags >>= fun rem ->
+        let bg = flatten (List.rev bg) in
+        let shelved, given_up = match rem with
+        | None -> [], []
+        | Some goals -> goals.shelved_goals, goals.given_up_goals
+        in
+        return (Wg_ProofView.NoFocusGoals { bg; shelved; given_up })
+      end
+    in
+    Coq.bind call begin function
     | Fail x -> self#handle_failure_aux ~move_insert x
     | Good goals ->
-      Coq.bind (Coq.evars ()) (function
-        | Fail x -> self#handle_failure_aux ~move_insert x
-        | Good evs ->
-          proof#set_goals goals;
-          proof#set_evars evs;
-          proof#refresh ~force:true;
-          Coq.return ()
-        )
-      )
+      proof#set_goals goals;
+      proof#refresh ~force:true;
+      Coq.return ()
+    end
+
   method show_goals = self#show_goals_aux ()
 
   (* This method is intended to perform stateless commands *)
@@ -467,16 +557,21 @@ object(self)
     List.iter (fun t -> buffer#apply_tag t ~start ~stop) tags
    end
 
+  (* Invariant, we store the tooltips with "original" Coq locations,
+     that means we must correct the offset even the Some case will get
+     the actual offset, beware this is a big fragile, but kinda
+     inherent to the model *)
   method private attach_tooltip ?loc sentence text =
-    let start_sentence, stop_sentence, phrase = self#get_sentence sentence in
-    let pre_bytes, post_bytes = Option.cata Loc.unloc (0, String.length phrase) loc in
-    let pre = b2c phrase (pre_bytes - (sentence.byte_start)) in
-    let post = b2c phrase (post_bytes - (sentence.byte_start)) in
-    let start = start_sentence#forward_chars pre in
-    let stop = start_sentence#forward_chars post in
+    let offset, line_drift = offset_compensation buffer sentence in
+    let start, stop = match loc with
+      | None ->
+        start_stop_iters buffer sentence
+      | Some loc ->
+        coq_loc_to_gtk_offset ~line_drift buffer loc
+    in
     let markup = Glib.Markup.escape_text text in
     buffer#apply_tag Tags.Script.tooltip ~start ~stop;
-    add_tooltip sentence start#offset stop#offset markup
+    add_tooltip sentence (start#offset - offset) (stop#offset - offset) markup
 
   method private debug_prompt ~tag msg =
     match tag with
@@ -603,14 +698,13 @@ object(self)
     buffer#move_mark ~where:stop (`NAME "start_of_input");
 
   method private apply_tag_in_sentence ?loc tag sentence =
-    let start, stop, phrase = self#get_sentence sentence in
     match loc with
     | None ->
+      let start, stop = start_stop_iters buffer sentence in
       buffer#apply_tag tag ~start ~stop
     | Some loc ->
-      let start_bytes, stop_bytes = Loc.unloc loc in
-      let stop = start#forward_chars (b2c phrase (stop_bytes - (sentence.byte_start))) in
-      let start = start#forward_chars (b2c phrase (start_bytes - (sentence.byte_start))) in
+      let _offset, line_drift = offset_compensation buffer sentence in
+      let start, stop = coq_loc_to_gtk_offset ~line_drift buffer loc in
       buffer#apply_tag tag ~start ~stop
 
   method private position_tag_at_sentence ?loc tag sentence =
@@ -618,10 +712,10 @@ object(self)
 
   method private process_interp_error ?loc queue sentence msg tip id =
     Coq.bind (Coq.return ()) (function () ->
-    let start, stop, _ = self#get_sentence sentence in
+    let start, stop = start_stop_iters buffer sentence in
     buffer#remove_tag Tags.Script.to_process ~start ~stop;
     self#discard_command_queue queue;
-    pop_info ();
+    Ideutils.pop_info ();
     if Stateid.equal id tip || Stateid.equal id Stateid.dummy then begin
       self#apply_tag_in_sentence ?loc Tags.Script.error sentence;
       buffer#place_cursor ~where:stop;
@@ -632,12 +726,6 @@ object(self)
       self#show_goals_aux ~move_insert:true ()
     )
 
-  method private get_sentence sentence =
-    let start = buffer#get_iter_at_mark sentence.start in
-    let stop = buffer#get_iter_at_mark sentence.stop in
-    let phrase = start#get_slice ~stop in
-    start, stop, phrase
-
   (** [fill_command_queue until q] fills a command queue until the [until]
       condition returns true; it is fed with the number of phrases read and the
       iters enclosing the current sentence. *)
@@ -661,10 +749,12 @@ object(self)
         end else begin
           buffer#apply_tag Tags.Script.to_process ~start ~stop;
           let sentence =
-            mk_sentence
-              ~start:(`MARK (buffer#create_mark start))
-              ~stop:(`MARK (buffer#create_mark stop))
-              [] in
+            let start_offset = start#offset in
+            let start_line = start#line in
+            let start = `MARK (buffer#create_mark start) in
+            let stop = `MARK (buffer#create_mark stop) in
+            mk_sentence ~start ~stop ~start_offset ~start_line []
+          in
           Queue.push (`Sentence sentence) queue;
           if start#offset <> prev_start && not stop#is_end then loop (succ n) stop start#offset
         end
@@ -689,7 +779,7 @@ object(self)
     let fill_queue = Coq.lift (fun () ->
       let queue = Queue.create () in
       (* Lock everything and fill the waiting queue *)
-      push_info "Coq is computing";
+      Ideutils.push_info "Coq is computing";
       messages#default_route#clear;
       script#set_editable false;
       self#fill_command_queue until queue;
@@ -699,7 +789,7 @@ object(self)
       Minilib.log "Begin command processing";
       queue) in
     let conclude topstack =
-      pop_info ();
+      Ideutils.pop_info ();
       script#recenter_insert;
       match topstack with
       | [] -> self#show_goals_aux ?move_insert ()
@@ -720,16 +810,12 @@ object(self)
         | `Sentence ({ edit_id } as sentence), [] ->
             add_flag sentence `PROCESSING;
             Doc.push document sentence;
-            let start, _, phrase = self#get_sentence sentence in
-            (* script before last_offsets is not editable because Coq is processing *)
-            let s_offs = if start#offset > (snd last_offsets) then last_offsets else (0,0) in
-            let bp = c2b buffer s_offs start#offset in
-            sentence.byte_start <- bp;
-            let line_nb = start#line + 1 in
-            let bol_uni = start#offset - start#line_offset in
-            let bol_byte = c2b buffer s_offs bol_uni in
-            let bol_pos = bp - bol_byte in
-            last_offsets <- (bol_byte, bol_uni);
+            let phrase = phrase buffer sentence in
+            (* When we retract, the cached offsets are invalid, thus we need to reset the cache *)
+            let start = start_iter buffer sentence in
+            let cached_offset = if start#offset > (snd last_offsets) then last_offsets else (0,0) in
+            let bp, line_nb, bol_pos, new_cached = coq_loc_from_gtk_offset cached_offset buffer sentence in
+            last_offsets <- new_cached;
             let coq_query = Coq.add ((((phrase,edit_id),(tip,verbose)),bp),(line_nb,bol_pos)) in
             let handle_answer = function
               | Good (id, Util.Inl (* NewTip *) ()) ->
@@ -781,9 +867,7 @@ object(self)
          let iter = begin match loc with
            | None      -> buffer#get_iter_at_mark s.start
            | Some loc ->
-             let (iter, _, phrase) = self#get_sentence s in
-             let (start, _) = Loc.unloc loc in
-             buffer#get_iter (`OFFSET (b2c phrase start))
+             fst (coq_loc_to_gtk_offset buffer loc)
          end in iter#line + 1, msg
       | _ -> assert false in
     List.rev
@@ -856,9 +940,9 @@ object(self)
     Minilib.log("backtrack_to_id "^Stateid.to_string to_id^
       " (unfocus_needed="^string_of_bool unfocus_needed^")");
     let opening () =
-      push_info "Coq is undoing" in
+      Ideutils.push_info "Coq is undoing" in
     let conclusion () =
-      pop_info ();
+      Ideutils.pop_info ();
       if move_insert then begin
         buffer#place_cursor ~where:self#get_start_of_input;
         script#recenter_insert;
@@ -962,10 +1046,10 @@ object(self)
       (* clear the views *)
       messages#default_route#clear;
       proof#clear ();
-      clear_info ();
+      Ideutils.clear_info ();
       processed <- 0;
       to_process <- 0;
-      push_info "Restarted";
+      Ideutils.push_info "Restarted";
       (* apply the initial commands to coq *)
     in
     Coq.seq (Coq.lift action) self#initialize
