@@ -13,9 +13,8 @@ open Util
 open Names
 open Nameops
 open Libnames
-open Namegen
 open Glob_term
-open Notation
+open Notationextern
 open Constrexpr
 
 (***********)
@@ -29,13 +28,21 @@ let sort_name_expr_eq c1 c2 = match c1, c2 with
   | CRawType u1, CRawType u2 -> Univ.Level.equal u1 u2
   | (CSProp|CProp|CSet|CType _|CRawType _), _ -> false
 
+let qvar_expr_eq c1 c2 = match c1, c2 with
+  | CQVar q1, CQVar q2 -> Libnames.qualid_eq q1 q2
+  | CQAnon _, CQAnon _ -> true
+  | CRawQVar q1, CRawQVar q2 -> Sorts.QVar.equal q1 q2
+  | (CQVar _ | CQAnon _ | CRawQVar _), _ -> false
+
 let univ_level_expr_eq u1 u2 =
   Glob_ops.glob_sort_gen_eq sort_name_expr_eq u1 u2
 
 let sort_expr_eq u1 u2 =
-  Glob_ops.glob_sort_gen_eq
-    (List.equal (fun (x,m) (y,n) -> sort_name_expr_eq x y && Int.equal m n))
-    u1 u2
+  let eq (q1, l1) (q2, l2) =
+    Option.equal qvar_expr_eq q1 q2 &&
+    List.equal (fun (x,m) (y,n) -> sort_name_expr_eq x y && Int.equal m n) l1 l2
+  in
+  Glob_ops.glob_sort_gen_eq eq u1 u2
 
 (***********************)
 (* For binders parsing *)
@@ -96,13 +103,14 @@ let rec cases_pattern_expr_eq p1 p2 =
       qualid_eq r1 r2 && cases_pattern_expr_eq e1 e2
     in
     List.equal equal l1 l2
-  | CPatDelimiters(s1,e1), CPatDelimiters(s2,e2) ->
-    String.equal s1 s2 && cases_pattern_expr_eq e1 e2
+  | CPatDelimiters(depth1,s1,e1), CPatDelimiters(depth2,s2,e2) ->
+    depth1 = depth2 && String.equal s1 s2 && cases_pattern_expr_eq e1 e2
   | _ -> false
 
-and cases_pattern_notation_substitution_eq (s1, n1) (s2, n2) =
+and cases_pattern_notation_substitution_eq (s1, n1, b1) (s2, n2, b2) =
   List.equal cases_pattern_expr_eq s1 s2 &&
-  List.equal (List.equal cases_pattern_expr_eq) n1 n2
+  List.equal (List.equal cases_pattern_expr_eq) n1 n2 &&
+  List.equal (eq_pair cases_pattern_expr_eq Glob_ops.binding_kind_eq) b1 b2
 
 let kinded_cases_pattern_expr_eq (p1,bk1) (p2,bk2) =
   cases_pattern_expr_eq p1 p2 && Glob_ops.binding_kind_eq bk1 bk2
@@ -228,6 +236,7 @@ module EqGen (A:sig val constr_expr_eq : constr_expr -> constr_expr -> bool end)
         constr_expr_eq t1 t2 &&
         constr_expr_eq f1 f2
       | CHole _, CHole _ -> true
+      | (CGenarg _ | CGenargGlob _), (CGenarg _ | CGenargGlob _) -> true
       | CPatVar i1, CPatVar i2 ->
         Id.equal i1 i2
       | CEvar (id1, c1), CEvar (id2, c2) ->
@@ -235,7 +244,7 @@ module EqGen (A:sig val constr_expr_eq : constr_expr -> constr_expr -> bool end)
       | CSort s1, CSort s2 ->
         sort_expr_eq s1 s2
       | CCast(c1,k1,t1), CCast(c2,k2,t2) ->
-        constr_expr_eq c1 c2 && Glob_ops.cast_kind_eq k1 k2 && constr_expr_eq t1 t2
+        constr_expr_eq c1 c2 && Option.equal Glob_ops.cast_kind_eq k1 k2 && constr_expr_eq t1 t2
       | CNotation(inscope1, n1, s1), CNotation(inscope2, n2, s2) ->
         Option.equal notation_with_optional_scope_eq inscope1 inscope2 &&
         notation_eq n1 n2 &&
@@ -245,7 +254,8 @@ module EqGen (A:sig val constr_expr_eq : constr_expr -> constr_expr -> bool end)
       | CGeneralization (bk1, e1), CGeneralization (bk2, e2) ->
         Glob_ops.binding_kind_eq bk1 bk2 &&
         constr_expr_eq e1 e2
-      | CDelimiters(s1,e1), CDelimiters(s2,e2) ->
+      | CDelimiters(depth1,s1,e1), CDelimiters(depth2,s2,e2) ->
+        depth1 = depth2 &&
         String.equal s1 s2 &&
         constr_expr_eq e1 e2
       | CArray(u1,t1,def1,ty1), CArray(u2,t2,def2,ty2) ->
@@ -254,6 +264,7 @@ module EqGen (A:sig val constr_expr_eq : constr_expr -> constr_expr -> bool end)
         eq_universes u1 u2
       | (CRef _ | CFix _ | CCoFix _ | CProdN _ | CLambdaN _ | CLetIn _ | CAppExpl _
         | CApp _ | CProj _ | CRecord _ | CCases _ | CLetTuple _ | CIf _ | CHole _
+        | CGenarg _ | CGenargGlob _
         | CPatVar _ | CEvar _ | CSort _ | CCast _ | CNotation _ | CPrim _
         | CGeneralization _ | CDelimiters _ | CArray _), _ -> false
 
@@ -301,10 +312,11 @@ let rec cases_pattern_fold_names f h nacc pt = match CAst.(pt.v) with
   | CPatCstr (_,patl1,patl2) ->
     List.fold_left (cases_pattern_fold_names f h)
       (Option.fold_left (List.fold_left (cases_pattern_fold_names f h)) nacc patl1) patl2
-  | CPatNotation (_,_,(patl,patll),patl') ->
+  | CPatNotation (_,_,(patl,patll,binderl),patl') ->
     List.fold_left (cases_pattern_fold_names f h)
-      (List.fold_left (cases_pattern_fold_names f h) nacc (patl@List.flatten patll)) patl'
-  | CPatDelimiters (_,pat) -> cases_pattern_fold_names f h nacc pat
+      (List.fold_left (cases_pattern_fold_names f h) nacc
+         (patl@List.flatten patll@List.map fst binderl)) patl'
+  | CPatDelimiters (_,_,pat) -> cases_pattern_fold_names f h nacc pat
   | CPatAtom (Some qid)
       when qualid_is_ident qid && not (is_constructor @@ qualid_basename qid) ->
       let (n, acc) = nacc in
@@ -352,11 +364,9 @@ let fold_constr_expr_with_binders g f n acc = CAst.with_val (function
       (* The following is an approximation: we don't know exactly if
          an ident is binding nor to which subterms bindings apply *)
       let acc = List.fold_left (f n) acc (l@List.flatten ll) in
-      List.fold_left (fun acc bl -> fold_local_binders g f n acc (CAst.make @@ CHole (None,IntroAnonymous,None)) bl) acc bll
+      List.fold_left (fun acc bl -> fold_local_binders g f n acc (CAst.make @@ CHole (None)) bl) acc bll
     | CGeneralization (_,c) -> f n acc c
-    | CDelimiters (_,a) -> f n acc a
-    | CHole _ | CEvar _ | CPatVar _ | CSort _ | CPrim _ | CRef _ ->
-      acc
+    | CDelimiters (_,_,a) -> f n acc a
     | CRecord l -> List.fold_left (fun acc (id, c) -> f n acc c) acc l
     | CCases (sty,rtnpo,al,bl) ->
       let ids = ids_of_cases_tomatch al in
@@ -380,6 +390,8 @@ let fold_constr_expr_with_binders g f n acc = CAst.with_val (function
     | CCoFix (_,_) ->
       Feedback.msg_warning (strbrk "Capture check in multiple binders not done"); acc
     | CArray (_u,t,def,ty) -> f n (f n (Array.fold_left (f n) acc t) def) ty
+    | CHole _ | CGenarg _ | CGenargGlob _ | CEvar _ | CPatVar _ | CSort _ | CPrim _ | CRef _ ->
+      acc
   )
 
 let free_vars_of_constr_expr c =
@@ -415,14 +427,15 @@ let rec fold_map_cases_pattern f h acc (CAst.{v=pt;loc} as p) = match pt with
     let acc, patl1 = Option.fold_left_map (List.fold_left_map (fold_map_cases_pattern f h)) acc patl1 in
     let acc, patl2 = List.fold_left_map (fold_map_cases_pattern f h) acc patl2 in
     acc, CAst.make ?loc (CPatCstr (c,patl1,patl2))
-  | CPatNotation (sc,ntn,(patl,patll),patl') ->
+  | CPatNotation (sc,ntn,(patl,patll,binderl),patl') ->
     let acc, patl = List.fold_left_map (fold_map_cases_pattern f h) acc patl in
     let acc, patll = List.fold_left_map (List.fold_left_map (fold_map_cases_pattern f h)) acc patll in
+    let acc, binderl' = List.fold_left_map (fold_fst (fold_map_cases_pattern f h)) acc binderl in
     let acc, patl' = List.fold_left_map (fold_map_cases_pattern f h) acc patl' in
-    acc, CAst.make ?loc (CPatNotation (sc,ntn,(patl,patll),patl'))
-  | CPatDelimiters (d,pat) ->
+    acc, CAst.make ?loc (CPatNotation (sc,ntn,(patl,patll,binderl'),patl'))
+  | CPatDelimiters (depth,d,pat) ->
     let acc, p = fold_map_cases_pattern f h acc pat in
-    acc, CAst.make ?loc (CPatDelimiters (d,pat))
+    acc, CAst.make ?loc (CPatDelimiters (depth,d,pat))
   | CPatAtom (Some qid)
       when qualid_is_ident qid && not (is_constructor @@ qualid_basename qid) ->
     f (qualid_basename qid) acc, p
@@ -467,9 +480,7 @@ let map_constr_expr_with_binders g f e = CAst.map (function
       CNotation (inscope,n,(List.map (f e) l,List.map (List.map (f e)) ll, bl,
                     List.map (fun bl -> snd (fold_map_local_binders f g e bl)) bll))
     | CGeneralization (b,c) -> CGeneralization (b,f e c)
-    | CDelimiters (s,a) -> CDelimiters (s,f e a)
-    | CHole _ | CEvar _ | CPatVar _ | CSort _
-    | CPrim _ | CRef _ as x -> x
+    | CDelimiters (depth,s,a) -> CDelimiters (depth,s,f e a)
     | CRecord l -> CRecord (List.map (fun (id, c) -> (id, f e c)) l)
     | CCases (sty,rtnpo,a,bl) ->
       let bl = List.map (fun {CAst.v=(patl,rhs);loc} ->
@@ -502,6 +513,8 @@ let map_constr_expr_with_binders g f e = CAst.map (function
           (id,bl',t',d')) dl)
     | CArray (u, t, def, ty) ->
       CArray (u, Array.map (f e) t, f e def, f e ty)
+    | CHole _ | CGenarg _ | CGenargGlob _ | CEvar _ | CPatVar _ | CSort _
+    | CPrim _ | CRef _ as x -> x
   )
 
 (* Used in constrintern *)
@@ -531,53 +544,13 @@ let ntn_loc ?loc (args,argslist,binders,binderslist) =
      List.map (fun (x,_) -> cases_pattern_expr_loc x) binders@
      List.map local_binders_loc binderslist)
 
-let patntn_loc ?loc (args,argslist) =
+let patntn_loc ?loc (args,argslist,binders) =
   locs_of_notation ?loc
-    (List.map cases_pattern_expr_loc (args@List.flatten argslist))
+    (List.map cases_pattern_expr_loc (args@List.flatten argslist@List.map fst binders))
 
 let error_invalid_pattern_notation ?loc () =
   CErrors.user_err ?loc  (str "Invalid notation for pattern.")
 
-(* Interpret the index of a recursion order annotation *)
-let split_at_annot bl na =
-  let open CAst in
-  let names = List.map (fun { v } -> v) (names_of_local_assums bl) in
-  match na with
-  | None ->
-    begin match names with
-      | [] -> CErrors.user_err (Pp.str "A fixpoint needs at least one parameter.")
-      | _ -> ([], bl)
-    end
-  | Some { loc; v = id } ->
-    let rec aux acc = function
-      | CLocalAssum (bls, k, t) as x :: rest ->
-        let test { CAst.v = na } = match na with
-          | Name id' -> Id.equal id id'
-          | Anonymous -> false
-        in
-        let l, r = List.split_when test bls in
-        begin match r with
-          | [] -> aux (x :: acc) rest
-          | _ ->
-            let ans = match l with
-              | [] -> acc
-              | _ -> CLocalAssum (l, k, t) :: acc
-            in
-            (List.rev ans, CLocalAssum (r, k, t) :: rest)
-        end
-      | CLocalDef ({ CAst.v = na },_,_) as x :: rest ->
-        if Name.equal (Name id) na then
-          CErrors.user_err ?loc
-            (Id.print id ++ str" must be a proper parameter and not a local definition.")
-        else
-          aux (x :: acc) rest
-      | CLocalPattern _ :: rest ->
-        Loc.raise ?loc (Gramlib.Stream.Error "pattern with quote not allowed after fix")
-      | [] ->
-        CErrors.user_err ?loc
-          (str "No parameter named " ++ Id.print id ++ str".")
-    in aux [] bl
-
 (** Pseudo-constructors *)
 
 let mkIdentC id   = CAst.make @@ CRef (qualid_of_ident id,None)
@@ -622,7 +595,7 @@ let coerce_to_id = function
 let coerce_to_name = function
   | { CAst.loc; v = CRef (qid,None) } when qualid_is_ident qid ->
     CAst.make ?loc @@ Name (qualid_basename qid)
-  | { CAst.loc; v = CHole (None,IntroAnonymous,None) } -> CAst.make ?loc Anonymous
+  | { CAst.loc; v = CHole (None) } -> CAst.make ?loc Anonymous
   | { CAst.loc; _ } -> CErrors.user_err ?loc
                          (str "This expression should be a name.")
 
@@ -645,7 +618,7 @@ let mkAppPattern ?loc p lp =
 let rec coerce_to_cases_pattern_expr c = CAst.map_with_loc (fun ?loc -> function
   | CRef (r,None) ->
      CPatAtom (Some r)
-  | CHole (None,IntroAnonymous,None) ->
+  | CHole (None) ->
      CPatAtom None
   | CLetIn ({CAst.loc;v=Name id},b,None,{ CAst.v = CRef (qid,None) })
       when qualid_is_ident qid && Id.equal id (qualid_basename qid) ->
@@ -654,17 +627,29 @@ let rec coerce_to_cases_pattern_expr c = CAst.map_with_loc (fun ?loc -> function
      (mkAppPattern (coerce_to_cases_pattern_expr p) (List.map (fun (a,_) -> coerce_to_cases_pattern_expr a) args)).CAst.v
   | CAppExpl ((r,i),args) ->
      CPatCstr (r,Some (List.map coerce_to_cases_pattern_expr args),[])
-  | CNotation (inscope,ntn,(c,cl,[],[])) ->
+  | CNotation (inscope,ntn,(c,cl,b,[])) ->
      CPatNotation (inscope,ntn,(List.map coerce_to_cases_pattern_expr c,
-                        List.map (List.map coerce_to_cases_pattern_expr) cl),[])
+                                List.map (List.map coerce_to_cases_pattern_expr) cl,
+                                b),[])
   | CPrim p ->
      CPatPrim p
   | CRecord l ->
      CPatRecord (List.map (fun (r,p) -> (r,coerce_to_cases_pattern_expr p)) l)
-  | CDelimiters (s,p) ->
-     CPatDelimiters (s,coerce_to_cases_pattern_expr p)
-  | CCast (p,Constr.DEFAULTcast, t) ->
-     CPatCast (coerce_to_cases_pattern_expr p,t)
-  | _ ->
-     CErrors.user_err ?loc
-                      (str "This expression should be coercible to a pattern.")) c
+  | CDelimiters (depth,s,p) ->
+     CPatDelimiters (depth,s,coerce_to_cases_pattern_expr p)
+  | CCast (p,Some Constr.DEFAULTcast, t) ->
+    CPatCast (coerce_to_cases_pattern_expr p,t)
+  | CLambdaN _ | CProdN _ | CSort _ | CLetIn _ | CGeneralization _
+  | CRef (_, Some _) | CCast (_, (Some (VMcast|NATIVEcast) | None), _)
+  | CFix _ | CCoFix _ | CApp _ | CProj _ | CCases _ | CLetTuple _ | CIf _
+  | CPatVar _ | CEvar _
+  | CNotation (_,_,(_,_,_,_::_))
+  | CHole (Some _) | CGenarg _ | CGenargGlob _ ->
+    CErrors.user_err ?loc
+                      (str "This expression should be coercible to a pattern.")
+  | CArray _ ->
+    CErrors.user_err ?loc
+                      (str "Arrays in patterns not supported.")) c
+
+let isCSort a =
+  match a.CAst.v with Constrexpr.CSort _ -> true | _ -> false
