@@ -14,7 +14,9 @@ open Constrexpr
 open Constrintern
 
 type module_internalization_error =
-  | NotAModuleNorModtype of string
+  | NotAModuleNorModtype of qualid
+  | NotAModuleType of qualid
+  | NotAModule of qualid
   | IncorrectWithInModule
   | IncorrectModuleApplication
 
@@ -22,18 +24,16 @@ exception ModuleInternalizationError of module_internalization_error
 
 type module_kind = Module | ModType | ModAny
 
+type module_struct_expr = (universe_decl_expr option * constr_expr) Declarations.module_alg_expr
+
 let error_not_a_module_loc ~info kind loc qid =
-  let s = string_of_qualid qid in
   let e = match kind with
-    | Module -> Modops.ModuleTypingError (Modops.NotAModule s)
-    | ModType -> Modops.ModuleTypingError (Modops.NotAModuleType s)
-    | ModAny -> ModuleInternalizationError (NotAModuleNorModtype s)
+    | Module -> NotAModule qid
+    | ModType -> NotAModuleType qid
+    | ModAny -> NotAModuleNorModtype qid
   in
   let info = Option.cata (Loc.add_loc info) info loc in
-  Exninfo.iraise (e,info)
-
-let error_application_to_not_path loc me =
-  Loc.raise ?loc (Modops.ModuleTypingError (Modops.ApplicationToNotPath me))
+  Exninfo.iraise (ModuleInternalizationError e,info)
 
 let error_incorrect_with_in_module loc =
   Loc.raise ?loc (ModuleInternalizationError IncorrectWithInModule)
@@ -85,7 +85,7 @@ let lookup_polymorphism env base kind fqid =
         | true, SFBmind mind -> Some (Declareops.inductive_is_polymorphic mind)
         | true, SFBconst const -> Some (Declareops.constant_is_polymorphic const)
       in
-      (try CList.find_map test m with Not_found -> false (* error later *))
+      (match CList.find_map test m with Some v -> v | None -> false (* error later *))
     | id::rem ->
       let next = function
         | MoreFunctor _ -> false (* error later *)
@@ -98,54 +98,65 @@ let lookup_polymorphism env base kind fqid =
         | true, SFBmodtype body ->  (* XXX is this valid? If not error later *)
           Some (next body.mod_type)
       in
-      (try CList.find_map test m with Not_found -> false (* error later *))
+      (match CList.find_map test m with Some v -> v | None -> false (* error later *))
   in
   aux (defunctor m) fqid
 
-let transl_with_decl env base kind = function
+let intern_with_decl = function
   | CWith_Module ({CAst.v=fqid},qid) ->
-      WithMod (fqid,lookup_module qid), Univ.ContextSet.empty
+    WithMod (fqid,lookup_module qid)
   | CWith_Definition ({CAst.v=fqid},udecl,c) ->
-    let sigma, udecl = Constrexpr_ops.interp_univ_decl_opt env udecl in
-    let c, ectx = interp_constr env sigma c in
-    let poly = lookup_polymorphism env base kind fqid in
-    begin match UState.check_univ_decl ~poly ectx udecl with
-      | Entries.Polymorphic_entry (nas, ctx) ->
-        let inst, ctx = Univ.abstract_universes nas ctx in
-        let c = EConstr.Vars.subst_univs_level_constr (Univ.make_instance_subst inst) c in
-        let c = EConstr.to_constr sigma c in
-        WithDef (fqid,(c, Some ctx)), Univ.ContextSet.empty
-      | Entries.Monomorphic_entry ctx ->
-        let c = EConstr.to_constr sigma c in
-        WithDef (fqid,(c, None)), ctx
-    end
+    WithDef (fqid,(udecl,c))
 
 let loc_of_module l = l.CAst.loc
 
 (* Invariant : the returned kind is never ModAny, and it is
    equal to the input kind when this one isn't ModAny. *)
 
-let rec interp_module_ast env kind m cst = match m with
+let rec intern_module_ast kind m = match m with
   | {CAst.loc;v=CMident qid} ->
       let (mp,kind) = lookup_module_or_modtype kind qid in
-      (MEident mp, mp, kind, cst)
+      (MEident mp, mp, kind)
   | {CAst.loc;v=CMapply (me1,me2)} ->
-      let me1', base, kind1, cst = interp_module_ast env kind me1 cst in
-      let me2', _, kind2, cst = interp_module_ast env ModAny me2 cst in
-      let mp2 = match me2' with
-        | MEident mp -> mp
-        | _ -> error_application_to_not_path (loc_of_module me2) me2'
-      in
+      let me1', base, kind1 = intern_module_ast kind me1 in
+      let mp2, kind2 = lookup_module_or_modtype ModAny me2 in
       if kind2 == ModType then
         error_application_to_module_type (loc_of_module me2);
-      (MEapply (me1',mp2), base, kind1, cst)
+      (MEapply (me1',mp2), base, kind1)
   | {CAst.loc;v=CMwith (me,decl)} ->
-      let me,base,kind,cst = interp_module_ast env kind me cst in
+      let me,base,kind = intern_module_ast kind me in
       if kind == Module then error_incorrect_with_in_module m.CAst.loc;
-      let decl, cst' = transl_with_decl env base kind decl in
-      let cst = Univ.ContextSet.union cst cst' in
-      (MEwith(me,decl), base, kind, cst)
+      let decl = intern_with_decl decl in
+      (MEwith(me,decl), base, kind)
 
-let interp_module_ast env kind m =
-  let me, _, kind, cst = interp_module_ast env kind m Univ.ContextSet.empty in
-  me, kind, cst
+let interp_with_decl env base kind = function
+  | WithMod (fqid,mp) -> WithMod (fqid,mp), Univ.ContextSet.empty
+  | WithDef(fqid,(udecl,c)) ->
+    let sigma, udecl = interp_univ_decl_opt env udecl in
+    let c, ectx = interp_constr env sigma c in
+    let poly = lookup_polymorphism env base kind fqid in
+    begin match fst (UState.check_univ_decl ~poly ectx udecl) with
+      | UState.Polymorphic_entry ctx ->
+        let inst, ctx = UVars.abstract_universes ctx in
+        let c = EConstr.Vars.subst_univs_level_constr (UVars.make_instance_subst inst) c in
+        let c = EConstr.to_constr sigma c in
+        WithDef (fqid,(c, Some ctx)), Univ.ContextSet.empty
+      | UState.Monomorphic_entry ctx ->
+        let c = EConstr.to_constr sigma c in
+        WithDef (fqid,(c, None)), ctx
+    end
+
+let rec interp_module_ast env kind base m cst = match m with
+| MEident mp ->
+  MEident mp, cst
+| MEapply (me,mp) ->
+  let me', cst = interp_module_ast env kind base me cst in
+  MEapply(me',mp), cst
+| MEwith(me,decl) ->
+  let me, cst = interp_module_ast env kind base me cst in
+  let decl, cst' = interp_with_decl env base kind decl in
+  let cst = Univ.ContextSet.union cst cst' in
+  MEwith(me,decl), cst
+
+let interp_module_ast env kind base m =
+  interp_module_ast env kind base m Univ.ContextSet.empty
