@@ -73,9 +73,9 @@ sp is a local copy of the global variable extern_sp. */
 #ifdef THREADED_CODE
 #  define Instruct(name) coq_lbl_##name:
 #  if defined(ARCH_SIXTYFOUR) && !defined(ARCH_CODE32)
-#    define coq_Jumptbl_base ((char *) &&coq_lbl_ACC0)
+#    define coq_Jumptbl_base &&coq_lbl_ACC0
 #  else
-#    define coq_Jumptbl_base ((char *) 0)
+#    define coq_Jumptbl_base 0
 #    define coq_jumptbl_base ((char *) 0)
 #  endif
 #  ifdef DEBUG
@@ -91,9 +91,9 @@ sp is a local copy of the global variable extern_sp. */
 /* #define _COQ_DEBUG_ */
 
 #ifdef _COQ_DEBUG_
-#   define print_instr(s) /*if (drawinstr)*/ printf("%s\n",s)
-#   define print_int(i)   /*if (drawinstr)*/ printf("%d\n",i)
-#   define print_lint(i)  /*if (drawinstr)*/ printf("%ld\n",i)
+#   define print_instr(s) printf("%s\n",s)
+#   define print_int(i)   printf("%d\n",i)
+#   define print_lint(i)  printf("%ld\n",i)
 # else
 #   define print_instr(s)
 #   define print_int(i)
@@ -125,6 +125,13 @@ if (sp - num_args < coq_stack_threshold) {                                     \
 #define Coq_alloc_small(result, wosize, tag) Alloc_small(result, wosize, tag)
 #endif
 
+/* Beware: Alloc_small can only invoked for allocations smaller than
+   Max_young_wosize, which is usually 256. For larger allocations,
+   caml_alloc_shr needs to be used instead, and the block needs to be filled
+   using caml_initialize. Note that these two functions never trigger a GC,
+   so calling Setup_for_gc is not needed.
+*/
+
 /* If there is an asynchronous exception, we reset the vm */
 #define Handle_potential_exception(res) \
   if (Is_exception_result(res)) {       \
@@ -301,8 +308,7 @@ value coq_interprete
   if (coq_pc == NULL) {           /* Interpreter is initializing */
     print_instr("Interpreter is initializing");
 #ifdef THREADED_CODE
-    coq_instr_table = (char **) coq_jumptable;
-    coq_instr_base = coq_Jumptbl_base;
+    coq_init_thread_code(coq_jumptable, coq_Jumptbl_base);
 #endif
     CAMLreturn(Val_unit);
   }
@@ -572,20 +578,26 @@ value coq_interprete
       /* We also check for signals */
 #if OCAML_VERSION >= 50000
       if (Caml_check_gc_interrupt(Caml_state)) {
+        Setup_for_gc;
         value res = caml_process_pending_actions_exn();
         Handle_potential_exception(res);
+        Restore_after_gc;
       }
 #elif OCAML_VERSION >= 41000
       if (caml_something_to_do) {
+        Setup_for_gc;
         value res = caml_process_pending_actions_exn();
         Handle_potential_exception(res);
+        Restore_after_gc;
       }
 #else
       if (caml_signals_are_pending) {
         /* If there's a Ctrl-C, we reset the vm */
         intnat sigint = caml_pending_signals[SIGINT];
         if (sigint) { coq_sp = coq_stack_high; }
+        Setup_for_gc;
         caml_process_pending_signals();
+        Restore_after_gc;
         if (sigint) {
           caml_failwith("Coq VM: Fatal error: SIGINT signal detected "
                         "but no exception was raised");
@@ -760,8 +772,6 @@ value coq_interprete
               for (i = 3; i < sz; ++i)
                 Field(block, i) = *sp++;
             } else {
-              // too large for Alloc_small, so use caml_alloc_shr instead
-              // it never triggers a GC, so no need for Setup_for_gc
               block = caml_alloc_shr(sz, Closure_tag);
               caml_initialize(&Field(block, 2), accu);
               for (i = 3; i < sz; ++i)
@@ -930,16 +940,21 @@ value coq_interprete
         print_instr("MAKEBLOCK, tag=");
         if (wosize == 0) {
           accu = Atom(tag);
-        } else {
+        } else if (wosize <= Max_young_wosize) {
           Coq_alloc_small(block, wosize, tag);
           Field(block, 0) = accu;
           for (i = 1; i < wosize; i++) Field(block, i) = *sp++;
           accu = block;
+        } else {
+          block = caml_alloc_shr(wosize, tag);
+          caml_initialize(&Field(block, 0), accu);
+          for (i = 1; i < wosize; i++) caml_initialize(&Field(block, i), *sp++);
+          accu = block;
         }
         Next;
       }
-      Instruct(MAKEBLOCK1) {
 
+      Instruct(MAKEBLOCK1) {
         tag_t tag = *pc++;
         value block;
         print_instr("MAKEBLOCK1, tag=");
@@ -1162,8 +1177,6 @@ value coq_interprete
           for (i = size; i < sz; ++i)
             Field(accu, i) = *sp++;
         } else {
-          // too large for Alloc_small, so use caml_alloc_shr instead
-          // it never triggers a GC, so no need for Setup_for_gc
           accu = caml_alloc_shr(sz, Closure_tag);
           for (i = 0; i < size; ++i)
             caml_initialize(&Field(accu, i), Field(coq_env, i));
@@ -1230,13 +1243,20 @@ value coq_interprete
             *--sp=Field(coq_global_data, annot);
             /* We save the stack */
             if (sz == 0) accu = Atom(0);
-            else {
+            else if (sz <= Max_young_wosize) {
               Coq_alloc_small(accu, sz, Default_tag);
               if (Is_tailrec_switch(*sp)) {
                 for (i = 0; i < sz; i++) Field(accu, i) = sp[i+2];
               }else{
                 for (i = 0; i < sz; i++) Field(accu, i) = sp[i+5];
               }
+            } else {
+              accu = caml_alloc_shr(sz, Default_tag);
+              if (Is_tailrec_switch(*sp)) {
+                for (i = 0; i < sz; i++) caml_initialize(&Field(accu, i), sp[i+2]);
+              }else{
+                for (i = 0; i < sz; i++) caml_initialize(&Field(accu, i), sp[i+5]);
+              }
             }
             *--sp = accu;
             /* Create bytecode wrappers */
@@ -1282,8 +1302,6 @@ value coq_interprete
           for (i = 3; i < sz; ++i)
             Field(accu, i) = *sp++;
         } else {
-          // too large for Alloc_small, so use caml_alloc_shr instead
-          // it never triggers a GC, so no need for Setup_for_gc
           accu = caml_alloc_shr(sz, Closure_tag);
           caml_initialize(&Field(accu, 2), Field(coq_atom_tbl, *pc));
           for (i = 3; i < sz; ++i)
@@ -2007,8 +2025,9 @@ value  coq_interprete_ml(value tcode, value a, value t, value g, value e, value
   // Registering the other arguments w.r.t. the OCaml GC is done by coq_interprete
   CAMLparam1(tcode);
   print_instr("coq_interprete");
-  CAMLreturn (coq_interprete(Code_val(tcode), a, t, g, e, Long_val(ea)));
+  value res = coq_interprete(Code_val(tcode), a, t, g, e, Long_val(ea));
   print_instr("end coq_interprete");
+  CAMLreturn(res);
 }
 
 value coq_interprete_byte(value* argv, int argn){
