@@ -10,79 +10,11 @@
 
 open Names
 open Constr
-open Declarations
 open Environ
 open Esubst
-
-(** {6 ... } *)
-(** Delta implies all consts (both global (= by
-  [kernel_name]) and local (= by [Rel] or [Var])), all evars, and letin's.
-  Rem: reduction of a Rel/Var bound to a term is Delta, but reduction of
-  a LetIn expression is Letin reduction *)
-
-(** Sets of reduction kinds. *)
-module RedFlags : sig
-  type reds
-  type red_kind
-
-  (** {7 The different kinds of reduction } *)
-
-  val fBETA : red_kind
-  val fDELTA : red_kind
-  val fMATCH : red_kind
-  val fFIX : red_kind
-  val fCOFIX : red_kind
-  val fZETA : red_kind
-  val fCONST : Constant.t -> red_kind
-  val fVAR : Id.t -> red_kind
-
-  (** No reduction at all *)
-  val no_red : reds
-
-  (** Adds a reduction kind to a set *)
-  val red_add : reds -> red_kind -> reds
-
-  (** Removes a reduction kind to a set *)
-  val red_sub : reds -> red_kind -> reds
-
-  (** Adds a reduction kind to a set *)
-  val red_add_transparent : reds -> TransparentState.t -> reds
-
-  (** Retrieve the transparent state of the reduction flags *)
-  val red_transparent : reds -> TransparentState.t
-
-  (** Build a reduction set from scratch = iter [red_add] on [no_red] *)
-  val mkflags : red_kind list -> reds
-
-  (** Tests if a reduction kind is set *)
-  val red_set : reds -> red_kind -> bool
-
-  (** This tests if the projection is in unfolded state already or
-      is unfodable due to delta. *)
-  val red_projection : reds -> Projection.t -> bool
-end
-
 open RedFlags
 
-(* These flags do not contain eta *)
-val all               : reds
-val allnolet          : reds
-val beta              : reds
-val betadeltazeta     : reds
-val betaiota          : reds
-val betaiotazeta      : reds
-val betazeta          : reds
-val delta             : reds
-val zeta              : reds
-val nored             : reds
-
-(***********************************************************************)
-type table_key = Constant.t Univ.puniverses tableKey
-
-module KeyTable : Hashtbl.S with type key = table_key
-
-(***********************************************************************
-  s Lazy reduction. *)
+(** Lazy reduction. *)
 
 (** [fconstr] is the type of frozen constr *)
 
@@ -92,10 +24,14 @@ type fconstr
 
 type finvert
 
-type 'a usubs = 'a subs Univ.puniverses
-
 type evar_repack
 
+type usubs = fconstr subs UVars.puniverses
+
+type table_key = Constant.t UVars.puniverses tableKey
+
+(** Relevances (eg in binder_annot or case_info) have NOT been substituted
+    when there is a usubs field *)
 type fterm =
   | FRel of int
   | FAtom of constr (** Metas and Sorts *)
@@ -103,20 +39,20 @@ type fterm =
   | FInd of pinductive
   | FConstruct of pconstructor
   | FApp of fconstr * fconstr array
-  | FProj of Projection.t * fconstr
-  | FFix of fixpoint * fconstr usubs
-  | FCoFix of cofixpoint * fconstr usubs
-  | FCaseT of case_info * Univ.Instance.t * constr array * case_return * fconstr * case_branch array * fconstr usubs (* predicate and branches are closures *)
-  | FCaseInvert of case_info * Univ.Instance.t * constr array * case_return * finvert * fconstr * case_branch array * fconstr usubs
-  | FLambda of int * (Name.t Context.binder_annot * constr) list * constr * fconstr usubs
-  | FProd of Name.t Context.binder_annot * fconstr * constr * fconstr usubs
-  | FLetIn of Name.t Context.binder_annot * fconstr * fconstr * constr * fconstr usubs
-  | FEvar of Evar.t * constr list * fconstr usubs * evar_repack
+  | FProj of Projection.t * Sorts.relevance * fconstr
+  | FFix of fixpoint * usubs
+  | FCoFix of cofixpoint * usubs
+  | FCaseT of case_info * UVars.Instance.t * constr array * case_return * fconstr * case_branch array * usubs (* predicate and branches are closures *)
+  | FCaseInvert of case_info * UVars.Instance.t * constr array * case_return * finvert * fconstr * case_branch array * usubs
+  | FLambda of int * (Name.t Context.binder_annot * constr) list * constr * usubs
+  | FProd of Name.t Context.binder_annot * fconstr * constr * usubs
+  | FLetIn of Name.t Context.binder_annot * fconstr * fconstr * constr * usubs
+  | FEvar of Evar.t * constr list * usubs * evar_repack
   | FInt of Uint63.t
   | FFloat of Float64.t
-  | FArray of Univ.Instance.t * fconstr Parray.t * fconstr
+  | FArray of UVars.Instance.t * fconstr Parray.t * fconstr
   | FLIFT of int * fconstr
-  | FCLOS of constr * fconstr usubs
+  | FCLOS of constr * usubs
   | FIrrelevant
   | FLOCKED
 
@@ -127,8 +63,8 @@ type 'a next_native_args = (CPrimitives.arg_kind * 'a) list
 
 type stack_member =
   | Zapp of fconstr array
-  | ZcaseT of case_info * Univ.Instance.t * constr array * case_return * case_branch array * fconstr usubs
-  | Zproj of Projection.Repr.t
+  | ZcaseT of case_info * UVars.Instance.t * constr array * case_return * case_branch array * usubs
+  | Zproj of Projection.Repr.t * Sorts.relevance
   | Zfix of fconstr * stack
   | Zprimitive of CPrimitives.t * pconstant * fconstr list * fconstr next_native_args
        (* operator, constr def, arguments already seen (in rev order), next arguments *)
@@ -145,20 +81,20 @@ val get_native_args1 : CPrimitives.t -> pconstant -> stack ->
   fconstr list * fconstr * fconstr next_native_args * stack
 
 val stack_args_size : stack -> int
-val eta_expand_stack : Name.t Context.binder_annot -> stack -> stack
-val skip_irrelevant_stack : stack -> stack
 
 val inductive_subst : Declarations.mutual_inductive_body
-  -> Univ.Instance.t
+  -> UVars.Instance.t
   -> fconstr array
-  -> fconstr usubs
+  -> usubs
 
-val usubs_lift : 'a usubs -> 'a usubs
-val usubs_liftn : int -> 'a usubs -> 'a usubs
-val usubs_cons : 'a -> 'a usubs -> 'a usubs
+val usubs_lift : usubs -> usubs
+val usubs_liftn : int -> usubs -> usubs
+val usubs_cons : fconstr -> usubs -> usubs
 
 (** identity if the first instance is empty *)
-val usubst_instance : 'a Univ.puniverses -> Univ.Instance.t -> Univ.Instance.t
+val usubst_instance : 'a UVars.puniverses -> UVars.Instance.t -> UVars.Instance.t
+
+val usubst_binder : _ UVars.puniverses -> 'a Context.binder_annot -> 'a Context.binder_annot
 
 (** To lazy reduce a constr, create a [clos_infos] with
    [create_clos_infos], inject the term to reduce with [inject]; then use
@@ -175,11 +111,24 @@ val mk_red : fterm -> fconstr
 val fterm_of : fconstr -> fterm
 val term_of_fconstr : fconstr -> constr
 val destFLambda :
-  (fconstr usubs -> constr -> fconstr) -> fconstr -> Name.t Context.binder_annot * fconstr * fconstr
+  (usubs -> constr -> fconstr) -> fconstr -> Name.t Context.binder_annot * fconstr * fconstr
 
 (** Global and local constant cache *)
 type clos_infos
 type clos_tab
+
+type 'a evar_expansion =
+| EvarDefined of 'a
+| EvarUndefined of Evar.t * 'a list
+
+type 'constr evar_handler = {
+  evar_expand : 'constr pexistential -> 'constr evar_expansion;
+  evar_repack : Evar.t * 'constr list -> 'constr;
+  evar_irrelevant : 'constr pexistential -> bool;
+  qvar_irrelevant : Sorts.QVar.t -> bool;
+}
+
+val default_evar_handler : env -> 'constr evar_handler
 val create_conv_infos :
   ?univs:UGraph.t -> ?evars:constr evar_handler -> reds -> env -> clos_infos
 val create_clos_infos :
@@ -191,7 +140,7 @@ val create_tab : unit -> clos_tab
 val info_env : clos_infos -> env
 val info_flags: clos_infos -> reds
 val info_univs : clos_infos -> UGraph.t
-val unfold_projection : clos_infos -> Projection.t -> stack_member option
+val unfold_projection : clos_infos -> Projection.t -> Sorts.relevance -> stack_member option
 
 val push_relevance : clos_infos -> 'b Context.binder_annot -> clos_infos
 val push_relevances : clos_infos -> 'b Context.binder_annot array -> clos_infos
@@ -199,6 +148,8 @@ val set_info_relevances : clos_infos -> Sorts.relevance Range.t -> clos_infos
 
 val info_relevances : clos_infos -> Sorts.relevance Range.t
 
+val is_irrelevant : clos_infos -> Sorts.relevance -> bool
+
 val infos_with_reds : clos_infos -> reds -> clos_infos
 
 (** Reduction function *)
@@ -207,7 +158,7 @@ val infos_with_reds : clos_infos -> reds -> clos_infos
 val norm_val : clos_infos -> clos_tab -> fconstr -> constr
 
 (** Same as [norm_val] but for terms *)
-val norm_term : clos_infos -> clos_tab -> fconstr usubs -> Constr.constr -> Constr.constr
+val norm_term : clos_infos -> clos_tab -> usubs -> Constr.constr -> Constr.constr
 
 (** [whd_val] is for weak head normalization *)
 val whd_val : clos_infos -> clos_tab -> fconstr -> constr
@@ -217,6 +168,10 @@ val whd_val : clos_infos -> clos_tab -> fconstr -> constr
 val whd_stack :
   clos_infos -> clos_tab -> fconstr -> stack -> fconstr * stack
 
+val skip_irrelevant_stack : clos_infos -> stack -> stack
+
+val eta_expand_stack : clos_infos -> Name.t Context.binder_annot -> stack -> stack
+
 (** [eta_expand_ind_stack env ind c s t] computes stacks corresponding
     to the conversion of the eta expansion of t, considered as an inhabitant
     of ind, and the Constructor c of this inductive type applied to arguments
@@ -226,16 +181,11 @@ val whd_stack :
     @raise Not_found if the inductive is not a primitive record, or if the
     constructor is partially applied.
  *)
-val eta_expand_ind_stack : env -> inductive -> fconstr -> stack ->
+val eta_expand_ind_stack : env -> pinductive -> fconstr -> stack ->
    (fconstr * stack) -> stack * stack
 
 (** Conversion auxiliary functions to do step by step normalisation *)
 
-(** [unfold_reference] unfolds references in a [fconstr]. Produces a
-    [FIrrelevant] when the reference is irrelevant. *)
-val unfold_reference : Environ.env -> TransparentState.t ->
-  clos_tab -> table_key -> (fconstr, Util.Empty.t) constant_def
-
 (** Like [unfold_reference], but handles primitives: if there are not
     enough arguments, return [None]. Otherwise return [Some] with
     [ZPrimitive] added to the stack. Produces a [FIrrelevant] when the
@@ -257,8 +207,8 @@ val set_conv : (clos_infos -> clos_tab -> fconstr -> fconstr -> bool) -> unit
 val lift_fconstr      : int -> fconstr -> fconstr
 val lift_fconstr_vect : int -> fconstr array -> fconstr array
 
-val mk_clos      : fconstr usubs -> constr -> fconstr
-val mk_clos_vect : fconstr usubs -> constr array -> fconstr array
+val mk_clos      : usubs -> constr -> fconstr
+val mk_clos_vect : usubs -> constr array -> fconstr array
 
 val kni: clos_infos -> clos_tab -> fconstr -> stack -> fconstr * stack
 val knr: clos_infos -> clos_tab -> fconstr -> stack -> fconstr * stack
@@ -268,6 +218,6 @@ val zip : fconstr -> stack -> fconstr
 
 val term_of_process : fconstr -> stack -> constr
 
-val to_constr : lift Univ.puniverses -> fconstr -> constr
+val to_constr : lift UVars.puniverses -> fconstr -> constr
 
 (** End of cbn debug section i*)
