@@ -27,7 +27,7 @@
 
 open Util
 open Names
-open Univ
+open UVars
 open Context
 
 type existential_key = Evar.t
@@ -59,7 +59,6 @@ type case_info =
                                        in addition to the parameters of the related inductive type
                                        NOTE: "lets" are therefore excluded from the count
                                        NOTE: parameters of the inductive type are also excluded from the count *)
-    ci_relevance : Sorts.relevance;
     ci_pp_info    : case_printing   (* not interpreted by the kernel *)
   }
 
@@ -76,7 +75,7 @@ type ('constr, 'types) pfixpoint =
     (int array * int) * ('constr, 'types) prec_declaration
 type ('constr, 'types) pcofixpoint =
     int * ('constr, 'types) prec_declaration
-type 'a puniverses = 'a Univ.puniverses
+type 'a puniverses = 'a UVars.puniverses
 type pconstant = Constant.t puniverses
 type pinductive = inductive puniverses
 type pconstructor = constructor puniverses
@@ -86,7 +85,7 @@ type 'constr pcase_invert =
   | CaseInvert of { indices : 'constr array }
 
 type 'constr pcase_branch = Name.t Context.binder_annot array * 'constr
-type 'types pcase_return = Name.t Context.binder_annot array * 'types
+type 'types pcase_return = (Name.t Context.binder_annot array * 'types) * Sorts.relevance
 
 type ('constr, 'types, 'univs) pcase =
   case_info * 'univs * 'constr array * 'types pcase_return * 'constr pcase_invert * 'constr * 'constr pcase_branch array
@@ -110,7 +109,7 @@ type ('constr, 'types, 'sort, 'univs) kind_of_term =
   | Case      of case_info * 'univs * 'constr array * 'types pcase_return * 'constr pcase_invert * 'constr * 'constr pcase_branch array
   | Fix       of ('constr, 'types) pfixpoint
   | CoFix     of ('constr, 'types) pcofixpoint
-  | Proj      of Projection.t * 'constr
+  | Proj      of Projection.t * Sorts.relevance * 'constr
   | Int       of Uint63.t
   | Float     of Float64.t
   | Array     of 'univs * 'constr array * 'constr * 'types
@@ -191,7 +190,7 @@ let mkSort   = function
   | Sorts.SProp -> mkSProp
   | Sorts.Prop -> mkProp (* Easy sharing *)
   | Sorts.Set -> mkSet
-  | Sorts.Type _ as s -> of_kind @@ Sort s
+  | (Sorts.Type _ | Sorts.QSort _) as s -> of_kind @@ Sort s
 
 (* Constructs the product (x:t1)t2 *)
 let mkProd (x,t1,t2) = of_kind @@ Prod (x,t1,t2)
@@ -203,14 +202,14 @@ let mkLambda (x,t1,t2) = of_kind @@ Lambda (x,t1,t2)
 let mkLetIn (x,c1,t,c2) = of_kind @@ LetIn (x,c1,t,c2)
 
 let map_puniverses f (x,u) = (f x, u)
-let in_punivs a = (a, Univ.Instance.empty)
+let in_punivs a = (a, UVars.Instance.empty)
 
 (* Constructs a constant *)
 let mkConst c = of_kind @@ Const (in_punivs c)
 let mkConstU c = of_kind @@ Const c
 
 (* Constructs an applied projection *)
-let mkProj (p,c) = of_kind @@ Proj (p,c)
+let mkProj (p,r,c) = of_kind @@ Proj (p,r,c)
 
 (* Constructs an existential variable *)
 let mkEvar e = of_kind @@ Evar e
@@ -285,6 +284,12 @@ let mkArray (u,t,def,ty) = of_kind @@ Array (u,t,def,ty)
 (* Constructs a primitive float number *)
 let mkFloat f = of_kind @@ Float f
 
+module UnsafeMonomorphic = struct
+  let mkConst = mkConst
+  let mkInd = mkInd
+  let mkConstruct = mkConstruct
+end
+
 (**********************************************************************)
 (*          Non primitive term destructors                            *)
 (**********************************************************************)
@@ -434,7 +439,7 @@ let destCase c = match kind c with
   | _ -> raise DestKO
 
 let destProj c = match kind c with
-  | Proj (p, c) -> (p, c)
+  | Proj (p, r, c) -> (p, r, c)
   | _ -> raise DestKO
 
 let destFix c = match kind c with
@@ -446,7 +451,7 @@ let destCoFix c = match kind c with
   | _ -> raise DestKO
 
 let destRef c = let open GlobRef in match kind c with
-  | Var x -> VarRef x, Univ.Instance.empty
+  | Var x -> VarRef x, UVars.Instance.empty
   | Const (c,u) -> ConstRef c, u
   | Ind (ind,u) -> IndRef ind, u
   | Construct (c,u) -> ConstructRef c, u
@@ -456,12 +461,12 @@ let destRef c = let open GlobRef in match kind c with
 (* Flattening and unflattening of embedded applications and casts *)
 (******************************************************************)
 
-let decompose_app c =
+let decompose_app_list c =
   match kind c with
     | App (f,cl) -> (f, Array.to_list cl)
     | _ -> (c,[])
 
-let decompose_appvect c =
+let decompose_app c =
   match kind c with
     | App (f,cl) -> (f, cl)
     | _ -> (c,[||])
@@ -487,9 +492,9 @@ let fold f acc c = match kind c with
   | Lambda (_,t,c) -> f (f acc t) c
   | LetIn (_,b,t,c) -> f (f (f acc b) t) c
   | App (c,l) -> Array.fold_left f (f acc c) l
-  | Proj (_p,c) -> f acc c
+  | Proj (_p,_r,c) -> f acc c
   | Evar (_,l) -> SList.Skip.fold f acc l
-  | Case (_,_,pms,(_,p),iv,c,bl) ->
+  | Case (_,_,pms,((_,p),_),iv,c,bl) ->
     Array.fold_left (fun acc (_, b) -> f acc b) (f (fold_invert f (f (Array.fold_left f acc pms) p) iv) c) bl
   | Fix (_,(_lna,tl,bl)) ->
     Array.fold_left2 (fun acc t b -> f (f acc t) b) acc tl bl
@@ -515,10 +520,10 @@ let iter f c = match kind c with
   | Lambda (_,t,c) -> f t; f c
   | LetIn (_,b,t,c) -> f b; f t; f c
   | App (c,l) -> f c; Array.iter f l
-  | Proj (_p,c) -> f c
+  | Proj (_p,_r,c) -> f c
   | Evar (_,l) -> SList.Skip.iter f l
   | Case (_,_,pms,p,iv,c,bl) ->
-    Array.iter f pms; f (snd p); iter_invert f iv; f c; Array.iter (fun (_, b) -> f b) bl
+    Array.iter f pms; f (snd @@ fst p); iter_invert f iv; f c; Array.iter (fun (_, b) -> f b) bl
   | Fix (_,(_,tl,bl)) -> Array.iter f tl; Array.iter f bl
   | CoFix (_,(_,tl,bl)) -> Array.iter f tl; Array.iter f bl
   | Array(_u,t,def,ty) -> Array.iter f t; f def; f ty
@@ -538,13 +543,13 @@ let iter_with_binders g f n c = match kind c with
   | LetIn (_,b,t,c) -> f n b; f n t; f (g n) c
   | App (c,l) -> f n c; Array.Fun1.iter f n l
   | Evar (_,l) -> SList.Skip.iter (fun c -> f n c) l
-  | Case (_,_,pms,p,iv,c,bl) ->
+  | Case (_,_,pms,(p,_),iv,c,bl) ->
     Array.Fun1.iter f n pms;
     f (iterate g (Array.length (fst p)) n) (snd p);
     iter_invert (f n) iv;
     f n c;
     Array.Fun1.iter (fun n (ctx, b) -> f (iterate g (Array.length ctx) n) b) n bl
-  | Proj (_p,c) -> f n c
+  | Proj (_p,_r,c) -> f n c
   | Fix (_,(_,tl,bl)) ->
       Array.Fun1.iter f n tl;
       Array.Fun1.iter f (iterate g (Array.length tl) n) bl
@@ -570,9 +575,9 @@ let fold_constr_with_binders g f n acc c =
   | Lambda (_na,t,c) -> f (g  n) (f n acc t) c
   | LetIn (_na,b,t,c) -> f (g  n) (f n (f n acc b) t) c
   | App (c,l) -> Array.fold_left (f n) (f n acc c) l
-  | Proj (_p,c) -> f n acc c
+  | Proj (_p,_r,c) -> f n acc c
   | Evar (_,l) -> SList.Skip.fold (f n) acc l
-  | Case (_,_,pms,p,iv,c,bl) ->
+  | Case (_,_,pms,(p,_),iv,c,bl) ->
     let fold_ctx n accu (nas, c) =
       f (iterate g (Array.length nas) n) accu c
     in
@@ -601,8 +606,9 @@ let map_branches f bl =
   let bl' = Array.map (map_under_context f) bl in
   if Array.for_all2 (==) bl' bl then bl else bl'
 
-let map_return_predicate f p =
-  map_under_context f p
+let map_return_predicate f (p,r as v) =
+  let p' = map_under_context f p in
+  if p == p' then v else p', r
 
 let map_under_context_with_binders g f l d =
   let (nas, p) = d in
@@ -614,8 +620,9 @@ let map_branches_with_binders g f l bl =
   let bl' = Array.map (map_under_context_with_binders g f l) bl in
   if Array.for_all2 (==) bl' bl then bl else bl'
 
-let map_return_predicate_with_binders g f l p =
-  map_under_context_with_binders g f l p
+let map_return_predicate_with_binders g f l (p,r as v) =
+  let p' = map_under_context_with_binders g f l p in
+  if p == p' then v else p',r
 
 let map_invert f = function
   | NoInvert -> NoInvert
@@ -653,10 +660,10 @@ let map f c = match kind c with
       let l' = Array.Smart.map f l in
       if b'==b && l'==l then c
       else mkApp (b', l')
-  | Proj (p,t) ->
+  | Proj (p,r,t) ->
       let t' = f t in
       if t' == t then c
-      else mkProj (p, t')
+      else mkProj (p, r, t')
   | Evar (e,l) ->
       let l' = SList.Smart.map f l in
       if l'==l then c
@@ -704,8 +711,10 @@ let fold_map_branches f accu bl =
   let accu, bl' = Array.Smart.fold_left_map (fold_map_under_context f) accu bl in
   if Array.for_all2 (==) bl' bl then accu, bl else accu, bl'
 
-let fold_map_return_predicate f accu p =
-  fold_map_under_context f accu p
+let fold_map_return_predicate f accu (p,r as v) =
+  let accu, p' = fold_map_under_context f accu p in
+  let v = if p == p' then v else p', r in
+  accu, v
 
 let fold_map f accu c = match kind c with
   | (Rel _ | Meta _ | Var _   | Sort _ | Const _ | Ind _
@@ -736,10 +745,10 @@ let fold_map f accu c = match kind c with
       let accu, l' = Array.Smart.fold_left_map f accu l in
       if b'==b && l'==l then accu, c
       else accu, mkApp (b', l')
-  | Proj (p,t) ->
+  | Proj (p,r,t) ->
       let accu, t' = f accu t in
       if t' == t then accu, c
-      else accu, mkProj (p, t')
+      else accu, mkProj (p, r, t')
   | Evar (e,l) ->
       let accu, l' = SList.Smart.fold_left_map f accu l in
       if l'==l then accu, c
@@ -804,10 +813,10 @@ let map_with_binders g f l c0 = match kind c0 with
     let al' = Array.Fun1.Smart.map f l al in
     if c' == c && al' == al then c0
     else mkApp (c', al')
-  | Proj (p, t) ->
+  | Proj (p, r, t) ->
     let t' = f l t in
     if t' == t then c0
-    else mkProj (p, t')
+    else mkProj (p, r, t')
   | Evar (e, al) ->
     let al' = SList.Smart.map (fun c -> f l c) al in
     if al' == al then c0
@@ -903,7 +912,8 @@ let compare_head_gen_leq_with kind1 kind2 leq_universes leq_sorts eq_evars eq le
     let len = Array.length l1 in
     Int.equal len (Array.length l2) &&
     leq (nargs+len) c1 c2 && Array.equal_norefl (eq 0) l1 l2
-  | Proj (p1,c1), Proj (p2,c2) -> Projection.CanOrd.equal p1 p2 && eq 0 c1 c2
+  | Proj (p1,r1,c1), Proj (p2,r2,c2) ->
+    Projection.CanOrd.equal p1 p2 && Sorts.relevance_equal r1 r2 && eq 0 c1 c2
   | Evar (e1,l1), Evar (e2,l2) -> eq_evars (e1, l1) (e2, l2)
   | Const (c1,u1), Const (c2,u2) ->
     (* The args length currently isn't used but may as well pass it. *)
@@ -911,7 +921,8 @@ let compare_head_gen_leq_with kind1 kind2 leq_universes leq_sorts eq_evars eq le
   | Ind (c1,u1), Ind (c2,u2) -> Ind.CanOrd.equal c1 c2 && leq_universes (Some (GlobRef.IndRef c1, nargs)) u1 u2
   | Construct (c1,u1), Construct (c2,u2) ->
     Construct.CanOrd.equal c1 c2 && leq_universes (Some (GlobRef.ConstructRef c1, nargs)) u1 u2
-  | Case (ci1,u1,pms1,p1,iv1,c1,bl1), Case (ci2,u2,pms2,p2,iv2,c2,bl2) ->
+  | Case (ci1,u1,pms1,(p1,_r1),iv1,c1,bl1), Case (ci2,u2,pms2,(p2,_r2),iv2,c2,bl2) ->
+    (* Ignore _r1/_r2: implied by comparing p1/p2 *)
     (** FIXME: what are we doing with u1 = u2 ? *)
     Ind.CanOrd.equal ci1.ci_ind ci2.ci_ind && leq_universes (Some (GlobRef.IndRef ci1.ci_ind, 0)) u1 u2 &&
     Array.equal (eq 0) pms1 pms2 && eq_under_context (eq 0) p1 p2 &&
@@ -954,7 +965,7 @@ let compare_head_gen_with kind1 kind2 eq_universes eq_sorts eq_evars eq t1 t2 =
 let compare_head_gen eq_universes eq_sorts eq_evars eq t1 t2 =
   compare_head_gen_leq eq_universes eq_sorts eq_evars eq eq t1 t2
 
-let compare_head = compare_head_gen (fun _ -> Univ.Instance.equal) Sorts.equal
+let compare_head = compare_head_gen (fun _ -> UVars.Instance.equal) Sorts.equal
 
 (*******************************)
 (*  alpha conversion functions *)
@@ -1007,18 +1018,11 @@ let compare_invert f iv1 iv2 =
     Array.compare f iv1.indices iv2.indices
 
 let constr_ord_int f t1 t2 =
-  let (=?) f g i1 i2 j1 j2=
-    let c = f i1 i2 in
-    if Int.equal c 0 then g j1 j2 else c in
-  let (==?) fg h i1 i2 j1 j2 k1 k2=
-    let c=fg i1 i2 j1 j2 in
-    if Int.equal c 0 then h k1 k2 else c in
+  let open! Compare in
   let fix_cmp (a1, i1) (a2, i2) =
-    ((Array.compare Int.compare) =? Int.compare) a1 a2 i1 i2
-  in
-  let ctx_cmp f (_n1, p1) (_n2, p2) =
-    f p1 p2
+    compare [(Array.compare Int.compare,a1,a2); (Int.compare,i1,i2)]
   in
+  let ctx_cmp f (_n1, p1) (_n2, p2) = f p1 p2 in
   match kind t1, kind t2 with
     | Cast (c1,_,_), _ -> f c1 t2
     | _, Cast (c2,_,_) -> f t1 c2
@@ -1032,19 +1036,17 @@ let constr_ord_int f t1 t2 =
     | Meta m1, Meta m2 -> Int.compare m1 m2
     | Meta _, _ -> -1 | _, Meta _ -> 1
     | Evar (e1,l1), Evar (e2,l2) ->
-        (Evar.compare =? (SList.compare f)) e1 e2 l1 l2
+      compare [(Evar.compare, e1, e2); (SList.compare f, l1, l2)]
     | Evar _, _ -> -1 | _, Evar _ -> 1
     | Sort s1, Sort s2 -> Sorts.compare s1 s2
     | Sort _, _ -> -1 | _, Sort _ -> 1
     | Prod (_,t1,c1), Prod (_,t2,c2)
-    | Lambda (_,t1,c1), Lambda (_,t2,c2) ->
-        (f =? f) t1 t2 c1 c2
+    | Lambda (_,t1,c1), Lambda (_,t2,c2) -> compare [(f,t1,t2); (f,c1,c2)]
     | Prod _, _ -> -1 | _, Prod _ -> 1
     | Lambda _, _ -> -1 | _, Lambda _ -> 1
-    | LetIn (_,b1,t1,c1), LetIn (_,b2,t2,c2) ->
-        ((f =? f) ==? f) b1 b2 t1 t2 c1 c2
+    | LetIn (_,b1,t1,c1), LetIn (_,b2,t2,c2) -> compare [(f,b1,b2); (f,t1,t2); (f,c1,c2)]
     | LetIn _, _ -> -1 | _, LetIn _ -> 1
-    | App (c1,l1), App (c2,l2) -> (f =? (Array.compare f)) c1 c2 l1 l2
+    | App (c1,l1), App (c2,l2) -> compare [(f,c1,c2); (Array.compare f, l1, l2)]
     | App _, _ -> -1 | _, App _ -> 1
     | Const (c1,_u1), Const (c2,_u2) -> Constant.CanOrd.compare c1 c2
     | Const _, _ -> -1 | _, Const _ -> 1
@@ -1052,29 +1054,28 @@ let constr_ord_int f t1 t2 =
     | Ind _, _ -> -1 | _, Ind _ -> 1
     | Construct (ct1,_u1), Construct (ct2,_u2) -> Construct.CanOrd.compare ct1 ct2
     | Construct _, _ -> -1 | _, Construct _ -> 1
-    | Case (_,_u1,pms1,p1,iv1,c1,bl1), Case (_,_u2,pms2,p2,iv2,c2,bl2) ->
-      let c = Array.compare f pms1 pms2 in
-      if Int.equal c 0 then let c = ctx_cmp f p1 p2 in
-      if Int.equal c 0 then let c = compare_invert f iv1 iv2 in
-      if Int.equal c 0 then let c = f c1 c2 in
-      if Int.equal c 0 then Array.compare (ctx_cmp f) bl1 bl2
-      else c else c else c else c
+    | Case (_,_u1,pms1,(p1,_r1),iv1,c1,bl1), Case (_,_u2,pms2,(p2,_r2),iv2,c2,bl2) ->
+      compare [
+        (Array.compare f, pms1, pms2);
+        (ctx_cmp f, p1, p2);
+        (compare_invert f, iv1, iv2);
+        (f, c1, c2);
+        (Array.compare (ctx_cmp f), bl1, bl2);
+      ]
     | Case _, _ -> -1 | _, Case _ -> 1
     | Fix (ln1,(_,tl1,bl1)), Fix (ln2,(_,tl2,bl2)) ->
-        ((fix_cmp =? (Array.compare f)) ==? (Array.compare f))
-        ln1 ln2 tl1 tl2 bl1 bl2
+      compare [(fix_cmp, ln1, ln2); (Array.compare f, tl1, tl2); (Array.compare f, bl1, bl2)]
     | Fix _, _ -> -1 | _, Fix _ -> 1
     | CoFix(ln1,(_,tl1,bl1)), CoFix(ln2,(_,tl2,bl2)) ->
-        ((Int.compare =? (Array.compare f)) ==? (Array.compare f))
-        ln1 ln2 tl1 tl2 bl1 bl2
+      compare [(Int.compare, ln1, ln2); (Array.compare f, tl1, tl2); (Array.compare f, bl1, bl2)]
     | CoFix _, _ -> -1 | _, CoFix _ -> 1
-    | Proj (p1,c1), Proj (p2,c2) -> (Projection.CanOrd.compare =? f) p1 p2 c1 c2
+    | Proj (p1,_r1,c1), Proj (p2,_r2,c2) -> compare [(Projection.CanOrd.compare, p1, p2); (f, c1, c2)]
     | Proj _, _ -> -1 | _, Proj _ -> 1
     | Int i1, Int i2 -> Uint63.compare i1 i2
     | Int _, _ -> -1 | _, Int _ -> 1
     | Float f1, Float f2 -> Float64.total_compare f1 f2
     | Array(_u1,t1,def1,ty1), Array(_u2,t2,def2,ty2) ->
-      (((Array.compare f) =? f) ==? f) t1 t2 def1 def2 ty1 ty2
+      compare [(Array.compare f, t1, t2); (f, def1, def2); (f, ty1, ty2)]
     | Array _, _ -> -1 | _, Array _ -> 1
 
 let rec compare m n=
@@ -1151,15 +1152,15 @@ let hasheq t1 t2 =
     | LetIn (n1,b1,t1,c1), LetIn (n2,b2,t2,c2) ->
       n1 == n2 && b1 == b2 && t1 == t2 && c1 == c2
     | App (c1,l1), App (c2,l2) -> c1 == c2 && array_eqeq l1 l2
-    | Proj (p1,c1), Proj(p2,c2) -> p1 == p2 && c1 == c2
+    | Proj (p1,r1,c1), Proj(p2,r2,c2) -> p1 == p2 && r1 == r2 && c1 == c2
     | Evar (e1,l1), Evar (e2,l2) -> e1 == e2 && SList.equal (==) l1 l2
     | Const (c1,u1), Const (c2,u2) -> c1 == c2 && u1 == u2
     | Ind (ind1,u1), Ind (ind2,u2) -> ind1 == ind2 && u1 == u2
     | Construct (cstr1,u1), Construct (cstr2,u2) -> cstr1 == cstr2 && u1 == u2
-    | Case (ci1,u1,pms1,p1,iv1,c1,bl1), Case (ci2,u2,pms2,p2,iv2,c2,bl2) ->
+    | Case (ci1,u1,pms1,(p1,r1),iv1,c1,bl1), Case (ci2,u2,pms2,(p2,r2),iv2,c2,bl2) ->
       (** FIXME: use deeper equality for contexts *)
       u1 == u2 && array_eqeq pms1 pms2 &&
-      ci1 == ci2 && hasheq_ctx p1 p2 &&
+      ci1 == ci2 && hasheq_ctx p1 p2 && r1 == r2 &&
       invert_eqeq iv1 iv2 && c1 == c2 && Array.equal hasheq_ctx bl1 bl2
     | Fix ((ln1, i1),(lna1,tl1,bl1)), Fix ((ln2, i2),(lna2,tl2,bl2)) ->
       Int.equal i1 i2
@@ -1230,16 +1231,17 @@ let rec hash t =
       combinesmall 10 (combine (Ind.CanOrd.hash ind) (Instance.hash u))
     | Construct (c,u) ->
       combinesmall 11 (combine (Construct.CanOrd.hash c) (Instance.hash u))
-    | Case (_ , u, pms, p, iv, c, bl) ->
-      combinesmall 12 (combine (combine (hash c) (combine (hash_invert iv) (combine (hash_term_array pms) (combine (Instance.hash u) (hash_under_context p))))) (hash_branches bl))
+    | Case (_ , u, pms, (p,r), iv, c, bl) ->
+      combinesmall 12 (combine5 (hash c) (hash_invert iv) (hash_term_array pms) (Instance.hash u)
+                         (combine3 (hash_under_context p) (Sorts.relevance_hash r) (hash_branches bl)))
     | Fix (_ln ,(_, tl, bl)) ->
       combinesmall 13 (combine (hash_term_array bl) (hash_term_array tl))
     | CoFix(_ln, (_, tl, bl)) ->
        combinesmall 14 (combine (hash_term_array bl) (hash_term_array tl))
     | Meta n -> combinesmall 15 n
     | Rel n -> combinesmall 16 n
-    | Proj (p,c) ->
-      combinesmall 17 (combine (Projection.CanOrd.hash p) (hash c))
+    | Proj (p,r, c) ->
+      combinesmall 17 (combine3 (Projection.CanOrd.hash p) (Sorts.relevance_hash r) (hash c))
     | Int i -> combinesmall 18 (Uint63.hash i)
     | Float f -> combinesmall 19 (Float64.hash f)
     | Array(u,t,def,ty) ->
@@ -1272,7 +1274,6 @@ struct
     info1.style == info2.style
   let eq ci ci' =
     ci.ci_ind == ci'.ci_ind &&
-    ci.ci_relevance == ci'.ci_relevance &&
     Int.equal ci.ci_npar ci'.ci_npar &&
     Array.equal Int.equal ci.ci_cstr_ndecls ci'.ci_cstr_ndecls && (* we use [Array.equal] on purpose *)
     Array.equal Int.equal ci.ci_cstr_nargs ci'.ci_cstr_nargs && (* we use [Array.equal] on purpose *)
@@ -1296,7 +1297,7 @@ struct
     let h3 = Array.fold_left combine 0 ci.ci_cstr_ndecls in
     let h4 = Array.fold_left combine 0 ci.ci_cstr_nargs in
     let h5 = hash_pp_info ci.ci_pp_info in
-    combinesmall (Sorts.relevance_hash ci.ci_relevance) (combine5 h1 h2 h3 h4 h5)
+    combine5 h1 h2 h3 h4 h5
 end
 
 module Hcaseinfo = Hashcons.Make(CaseinfoHash)
@@ -1359,7 +1360,7 @@ let rec hash_term (t : t) =
     let u', hu = Instance.share u in
     (Construct (hcons_construct c, u'),
      combinesmall 11 (combine (Construct.SyntacticOrd.hash c) hu))
-  | Case (ci,u,pms,p,iv,c,bl) ->
+  | Case (ci,u,pms,(p,r),iv,c,bl) ->
     (** FIXME: use a dedicated hashconsing structure *)
     let hcons_ctx (lna, c) =
       let () = Array.iteri (fun i x -> Array.unsafe_set lna i (hcons_annot x)) lna in
@@ -1379,7 +1380,7 @@ let rec hash_term (t : t) =
     in
     let hbl, bl = Array.fold_left_map fold 0 bl in
     let hbl = combine (combine hc (combine hiv (combine hpms (combine hu hp)))) hbl in
-    (Case (hcons_caseinfo ci, u, pms, p, iv, c, bl), combinesmall 12 hbl)
+    (Case (hcons_caseinfo ci, u, pms, (p,r), iv, c, bl), combinesmall 12 hbl)
   | Fix (ln,(lna,tl,bl)) ->
     let bl,hbl = hash_term_array bl in
     let tl,htl = hash_term_array tl in
@@ -1400,10 +1401,10 @@ let rec hash_term (t : t) =
     (t, combinesmall 15 n)
   | Rel n as t ->
     (t, combinesmall 16 n)
-  | Proj (p,c) ->
+  | Proj (p,r,c) ->
     let c, hc = sh_rec c in
     let p' = Projection.hcons p in
-    (Proj (p', c), combinesmall 17 (combine (Projection.SyntacticOrd.hash p') hc))
+    (Proj (p', r, c), combinesmall 17 (combine (Projection.SyntacticOrd.hash p') hc))
   | Int i as t ->
     let (h,l) = Uint63.to_int2 i in
     (t, combinesmall 18 (combine h l))
@@ -1464,22 +1465,6 @@ type rel_context = rel_declaration list
 type named_context = named_declaration list
 type compacted_context = compacted_declaration list
 
-type 'a evar_expansion =
-| EvarDefined of 'a
-| EvarUndefined of Evar.t * 'a list
-
-type 'constr evar_handler = {
-  evar_expand : 'constr pexistential -> 'constr evar_expansion;
-  evar_repack : Evar.t * 'constr list -> 'constr;
-  evar_relevance : 'constr pexistential -> Sorts.relevance;
-}
-
-let default_evar_handler = {
-  evar_expand = (fun _ -> assert false);
-  evar_repack = (fun _ -> assert false);
-  evar_relevance = (fun _ -> assert false);
-}
-
 (** Minimalistic constr printer, typically for debugging *)
 
 let debug_print_fix pr_constr ((t,i),(lna,tl,bl)) =
@@ -1493,8 +1478,8 @@ let debug_print_fix pr_constr ((t,i),(lna,tl,bl)) =
          str"}")
 
 let pr_puniverses p u =
-  if Univ.Instance.is_empty u then p
-  else Pp.(p ++ str"(*" ++ Univ.Instance.pr Univ.Level.pr u ++ str"*)")
+  if UVars.Instance.is_empty u then p
+  else Pp.(p ++ str"(*" ++ UVars.Instance.pr Sorts.QVar.raw_pr Univ.Level.raw_pr u ++ str"*)")
 
 let rec debug_print c =
   let open Pp in
@@ -1531,8 +1516,10 @@ let rec debug_print c =
   | Ind ((sp,i),u) -> str"Ind(" ++ pr_puniverses (MutInd.print sp ++ str"," ++ int i) u ++ str")"
   | Construct (((sp,i),j),u) ->
       str"Constr(" ++ pr_puniverses (MutInd.print sp ++ str"," ++ int i ++ str"," ++ int j) u ++ str")"
-  | Proj (p,c) -> str"Proj(" ++ Constant.debug_print (Projection.constant p) ++ str"," ++ bool (Projection.unfolded p) ++ str"," ++ debug_print c ++ str")"
-  | Case (_ci,_u,pms,p,iv,c,bl) ->
+  | Proj (p,_r,c) ->
+    str"Proj(" ++ Constant.debug_print (Projection.constant p) ++ str"," ++
+    bool (Projection.unfolded p) ++ str"," ++ debug_print c ++ str")"
+  | Case (_ci,_u,pms,(p,_),iv,c,bl) ->
     let pr_ctx (nas, c) =
       hov 2 (hov 0 (prvect (fun na -> Name.print na.binder_name ++ spc ()) nas ++ str "|-") ++ spc () ++
         debug_print c)
@@ -1555,7 +1542,7 @@ let rec debug_print c =
   | Float i -> str"Float("++str (Float64.to_string i) ++ str")"
   | Array(u,t,def,ty) -> str"Array(" ++ prlist_with_sep pr_comma debug_print (Array.to_list t) ++ str" | "
       ++ debug_print def ++ str " : " ++ debug_print ty
-      ++ str")@{" ++ Univ.Instance.pr Univ.Level.pr u ++ str"}"
+      ++ str")@{" ++ UVars.Instance.pr Sorts.QVar.raw_pr Univ.Level.raw_pr u ++ str"}"
 
 and debug_invert = let open Pp in function
   | NoInvert -> mt()
