@@ -1,7 +1,7 @@
 (************************************************************************)
 (*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -82,11 +82,20 @@ struct
 
     let get_relevance x = (get_annot x).binder_relevance
 
+    let set_annot x d =
+      if get_annot d == x then d else match d with
+      | LocalAssum (_,ty) -> LocalAssum (x, ty)
+      | LocalDef (_,v,ty) -> LocalDef (x, v, ty)
+
     (** Set the name that is bound by a given declaration. *)
     let set_name na = function
       | LocalAssum (x,ty) -> LocalAssum ({x with binder_name=na}, ty)
       | LocalDef (x,v,ty) -> LocalDef ({x with binder_name=na}, v, ty)
 
+    let set_relevance r = function
+      | LocalAssum (x,ty) -> LocalAssum ({x with binder_relevance=r}, ty)
+      | LocalDef (x,v,ty) -> LocalDef ({x with binder_relevance=r}, v, ty)
+
     (** Set the type of the bound variable in a given declaration. *)
     let set_type ty = function
       | LocalAssum (na,_) -> LocalAssum (na, ty)
@@ -128,6 +137,11 @@ struct
       let na' = f na in
       if na == na' then x else set_name na' x
 
+    let map_relevance f x =
+      let r = get_relevance x in
+      let r' = f r in
+      if r == r' then x else set_relevance r' x
+
     (** For local assumptions, this function returns the original local assumptions.
         For local definitions, this function maps the value in the local definition. *)
     let map_value f = function
@@ -196,12 +210,10 @@ struct
   (** Return a new rel-context enriched by with a given inner-most declaration. *)
   let add d ctx = d :: ctx
 
-  (** Return the number of {e local declarations} in a given context. *)
+  (** Return the number of {e local declarations} in a given rel-context. *)
   let length = List.length
 
-  (** [extended_rel_list n Γ] builds an instance [args] such that [Γ,Δ ⊢ args:Γ]
-      with n = |Δ| and with the local definitions of [Γ] skipped in
-      [args]. Example: for [x:T,y:=c,z:U] and [n]=2, it gives [Rel 5, Rel 3]. *)
+  (** Return the number of {e local assumptions} in a given rel-context. *)
   let nhyps ctx =
     let open Declaration in
     let rec nhyps acc = function
@@ -225,6 +237,17 @@ struct
   (** Map all terms in a given rel-context. *)
   let map f = List.Smart.map (Declaration.map_constr f)
 
+  (** Map all terms in a given rel-context. *)
+  let map_with_binders f ctx =
+    let rec aux k = function
+      | decl :: ctx as l ->
+        let decl' = Declaration.map_constr (f k) decl in
+        let ctx' = aux (k-1) ctx in
+        if decl == decl' && ctx == ctx' then l else decl' :: ctx'
+      | [] -> []
+    in
+    aux (length ctx) ctx
+
   (** Perform a given action on every declaration in a given rel-context. *)
   let iter f = List.iter (Declaration.iter_constr f)
 
@@ -236,6 +259,14 @@ struct
       Outermost declarations are processed first. *)
   let fold_outside f l ~init = List.fold_right f l init
 
+  (** Return the set of all named variables bound in a given rel-context. *)
+  let to_vars l =
+    List.fold_left (fun accu decl ->
+        match Declaration.get_name decl with
+        | Name id -> Id.Set.add id accu
+        | Anonymous -> accu)
+      Id.Set.empty l
+
   (** Map a given rel-context to a list where each {e local assumption} is mapped to [true]
       and each {e local definition} is mapped to [false]. *)
   let to_tags l =
@@ -247,6 +278,16 @@ struct
 
   let drop_bodies l = List.Smart.map Declaration.drop_body l
 
+  (** Split a context so that the second part contains [n]
+      [LocalAssum], keeping all [LocalDef] in the middle in the first part *)
+  let chop_nhyps n l =
+    let rec aux l' = function
+      | (0, l) -> (List.rev l', l)
+      | (n, (Declaration.LocalDef _ as h) :: l) -> aux (h::l') (n, l)
+      | (n, (Declaration.LocalAssum _ as h) :: l) -> aux (h::l') (n-1, l)
+      | (_, []) -> CErrors.anomaly (Pp.str "chop_nhyps: not enough hypotheses.")
+    in aux [] (n,l)
+
   (** [extended_list n Γ] builds an instance [args] such that [Γ,Δ ⊢ args:Γ]
       with n = |Δ| and with the {e local definitions} of [Γ] skipped in
       [args]. Example: for [x:T, y:=c, z:U] and [n]=2, it gives [Rel 5, Rel 3]. *)
@@ -260,6 +301,10 @@ struct
 
   (** [extended_vect n Γ] does the same, returning instead an array. *)
   let to_extended_vect mk n hyps = Array.of_list (to_extended_list mk n hyps)
+
+  (** Consistency with terminology in Named *)
+  let instance = to_extended_vect
+  let instance_list = to_extended_list
 end
 
 (** This module represents contexts that can capture non-anonymous variables.
@@ -367,6 +412,15 @@ struct
           let ty' = f ty in
           if v == v' && ty == ty' then decl else LocalDef (id, v', ty')
 
+    let map_constr_het f = function
+      | LocalAssum (id, ty) ->
+          let ty' = f ty in
+          LocalAssum (id, ty')
+      | LocalDef (id, v, ty) ->
+          let v' = f v in
+          let ty' = f ty in
+          LocalDef (id, v', ty')
+
     (** Perform a given action on all terms in a given declaration. *)
     let iter_constr f = function
       | LocalAssum (_, ty) -> f ty
@@ -413,7 +467,7 @@ struct
   (** empty named-context *)
   let empty = []
 
-  (** empty named-context *)
+  (** Return a new named-context enriched by with a given inner-most declaration. *)
   let add d ctx = d :: ctx
 
   (** Return the number of {e local declarations} in a given named-context. *)
@@ -449,7 +503,7 @@ struct
 
   let drop_bodies l = List.Smart.map Declaration.drop_body l
 
-  (** [instance_from_named_context Ω] builds an instance [args] such
+  (** [to_instance Ω] builds an instance [args] in reverse order such
       that [Ω ⊢ args:Ω] where [Ω] is a named context and with the local
       definitions of [Ω] skipped. Example: for [id1:T,id2:=c,id3:U], it
       gives [Var id1, Var id3]. All [idj] are supposed distinct. *)
@@ -459,6 +513,20 @@ struct
       | _ -> None
     in
     List.map_filter filter l
+
+  (** [instance Ω] builds an instance [args] such
+      that [Ω ⊢ args:Ω] where [Ω] is a named context and with the local
+      definitions of [Ω] skipped. Example: for [id1:T,id2:=c,id3:U], it
+      gives [Var id1, Var id3]. All [idj] are supposed distinct. *)
+  let instance_list mk l =
+    let filter = function
+      | Declaration.LocalAssum (id, _) -> Some (mk id.binder_name)
+      | _ -> None
+    in
+    List.rev (List.map_filter filter l)
+
+  let instance mk l =
+    Array.of_list (instance_list mk l)
 end
 
 module Compacted =
