@@ -23,7 +23,7 @@ val is_functor : ('ty,'a) functorize -> bool
 
 val destr_functor : ('ty,'a) functorize -> MBId.t * 'ty * ('ty,'a) functorize
 
-val destr_nofunctor : ('ty,'a) functorize -> 'a
+val destr_nofunctor : ModPath.t -> ('ty,'a) functorize -> 'a
 
 (** Conversions between [module_body] and [module_type_body] *)
 
@@ -32,11 +32,16 @@ val module_body_of_type : ModPath.t -> module_type_body -> module_body
 
 val check_modpath_equiv : env -> ModPath.t -> ModPath.t -> unit
 
-val implem_smartmap :
-  (module_signature -> module_signature) ->
+val implem_smart_map :
+  (structure_body -> structure_body) ->
   (module_expression -> module_expression) ->
   (module_implementation -> module_implementation)
 
+val annotate_module_expression : module_expression -> module_signature ->
+  (module_type_body, (constr * UVars.AbstractContext.t option) module_alg_expr) functorize
+
+val annotate_struct_body : structure_body -> module_signature -> module_signature
+
 (** {6 Substitutions } *)
 
 val subst_signature : substitution -> module_signature -> module_signature
@@ -63,15 +68,17 @@ val add_retroknowledge : module_implementation module_retroknowledge -> env -> e
 
 val strengthen : module_type_body -> ModPath.t -> module_type_body
 
+val strengthen_and_subst_module_body : module_body -> ModPath.t -> bool -> module_body
+
+val subst_modtype_signature_and_resolver : ModPath.t -> ModPath.t ->
+  module_signature -> delta_resolver -> module_signature * delta_resolver
+
+(** {6 Building map of constants to inline } *)
+
 val inline_delta_resolver :
   env -> inline -> ModPath.t -> MBId.t -> module_type_body ->
   delta_resolver -> delta_resolver
 
-val strengthen_and_subst_mb : module_body -> ModPath.t -> bool -> module_body
-
-val subst_modtype_and_resolver : module_type_body -> ModPath.t ->
-  module_type_body
-
 (** {6 Cleaning a module expression from bounded parts }
 
      For instance:
@@ -82,11 +89,6 @@ val subst_modtype_and_resolver : module_type_body -> ModPath.t ->
 
 val clean_bounded_mod_expr : module_signature -> module_signature
 
-(** {6 Stm machinery } *)
-
-val join_structure :
-  Future.UUIDSet.t -> Opaqueproof.opaquetab -> structure_body -> unit
-
 (** {6 Errors } *)
 
 type signature_mismatch_error =
@@ -108,24 +110,24 @@ type signature_mismatch_error =
   | RecordFieldExpected of bool
   | RecordProjectionsExpected of Name.t list
   | NotEqualInductiveAliases
-  | IncompatibleUniverses of Univ.univ_inconsistency
+  | IncompatibleUniverses of UGraph.univ_inconsistency
   | IncompatiblePolymorphism of env * types * types
-  | IncompatibleConstraints of { got : Univ.AUContext.t; expect : Univ.AUContext.t }
+  | IncompatibleConstraints of { got : UVars.AbstractContext.t; expect : UVars.AbstractContext.t }
   | IncompatibleVariance
 
+type subtyping_trace_elt =
+  | Submodule of Label.t
+  | FunctorArgument of int
+
 type module_typing_error =
-  | SignatureMismatch of
-      Label.t * structure_field_body * signature_mismatch_error
+  | SignatureMismatch of subtyping_trace_elt list * Label.t * signature_mismatch_error
   | LabelAlreadyDeclared of Label.t
-  | ApplicationToNotPath of module_struct_entry
   | NotAFunctor
-  | IsAFunctor
+  | IsAFunctor of ModPath.t
   | IncompatibleModuleTypes of module_type_body * module_type_body
   | NotEqualModulePaths of ModPath.t * ModPath.t
-  | NoSuchLabel of Label.t
-  | IncompatibleLabels of Label.t * Label.t
-  | NotAModule of string
-  | NotAModuleType of string
+  | NoSuchLabel of Label.t * ModPath.t
+  | NotAModuleLabel of Label.t
   | NotAConstant of Label.t
   | IncorrectWithConstraint of Label.t
   | GenerativeModuleExpected of Label.t
@@ -140,13 +142,11 @@ val error_incompatible_modtypes :
   module_type_body -> module_type_body -> 'a
 
 val error_signature_mismatch :
-  Label.t -> structure_field_body -> signature_mismatch_error -> 'a
+  subtyping_trace_elt list -> Label.t -> signature_mismatch_error -> 'a
 
-val error_incompatible_labels : Label.t -> Label.t -> 'a
+val error_no_such_label : Label.t -> ModPath.t -> 'a
 
-val error_no_such_label : Label.t -> 'a
-
-val error_not_a_module : string -> 'a
+val error_not_a_module_label : Label.t -> 'a
 
 val error_not_a_constant : Label.t -> 'a
 
