@@ -12,14 +12,13 @@ open CErrors
 open Util
 open Names
 open Univ
-open Sorts
 open Term
 open Constr
 open Context
 open Vars
 open Declarations
 open Environ
-open Reduction
+open Conversion
 open Inductive
 open Type_errors
 
@@ -43,9 +42,13 @@ let check_constraints cst env =
   if Environ.check_constraints cst env then ()
   else error_unsatisfied_constraints env cst
 
+let check_qconstraints qcst env =
+  if Sorts.QConstraints.trivial qcst then ()
+  else error_unsatisfied_qconstraints env qcst
+
 (* This should be a type (a priori without intention to be an assumption) *)
 let check_type env c t =
-  match kind(whd_all env t) with
+  match kind(Reduction.whd_all env t) with
   | Sort s -> s
   | _ -> error_not_type env (make_judge c t)
 
@@ -54,26 +57,52 @@ let check_type env c t =
 let infer_assumption env t ty =
   try
     let s = check_type env t ty in
-    (match s with Sorts.SProp -> Irrelevant | _ -> Relevant)
+    Sorts.relevance_of_sort s
   with TypeError _ ->
     error_assumption env (make_judge t ty)
 
-let warn_bad_relevance_name = "bad-relevance"
-let warn_bad_relevance =
-  CWarnings.create ~name:warn_bad_relevance_name ~category:"debug" ~default:CWarnings.Disabled
-    Pp.(function
-        | None ->  str "Bad relevance in case annotation."
-        | Some x -> str "Bad relevance for binder " ++ Name.print x.binder_name ++ str ".")
+type ('constr,'types) bad_relevance =
+| BadRelevanceBinder of Sorts.relevance * ('constr,'types) Context.Rel.Declaration.pt
+| BadRelevanceCase of Sorts.relevance * 'constr
 
-let warn_bad_relevance_ci ?loc () = warn_bad_relevance ?loc None
-let warn_bad_relevance ?loc x = warn_bad_relevance ?loc (Some x)
+let warn_bad_relevance_name = "bad-relevance"
+
+let bad_relevance_warning =
+  CWarnings.create_warning ~name:warn_bad_relevance_name ~default:CWarnings.AsError ()
+
+let bad_relevance_msg = CWarnings.create_msg bad_relevance_warning ()
+
+let default_print_bad_relevance = function
+  | BadRelevanceCase _ -> Pp.str "Bad relevance in case annotation."
+  | BadRelevanceBinder (_, na) ->
+    Pp.(str "Bad relevance for binder " ++ Name.print (RelDecl.get_name na) ++ str ".")
+
+(* used eg in the checker *)
+let () = CWarnings.register_printer bad_relevance_msg
+    (fun (_env,b) -> default_print_bad_relevance b)
+
+let warn_bad_relevance_case ?loc env rlv case =
+  match CWarnings.warning_status bad_relevance_warning with
+| CWarnings.Disabled | CWarnings.Enabled ->
+  CWarnings.warn bad_relevance_msg ?loc (env, BadRelevanceCase (rlv, mkCase case))
+| CWarnings.AsError ->
+  error_bad_case_relevance env rlv case
+
+let warn_bad_relevance_binder ?loc env rlv bnd =
+  match CWarnings.warning_status bad_relevance_warning with
+| CWarnings.Disabled | CWarnings.Enabled ->
+  CWarnings.warn bad_relevance_msg ?loc (env, BadRelevanceBinder (rlv, bnd))
+| CWarnings.AsError ->
+  error_bad_binder_relevance env rlv bnd
 
 let check_assumption env x t ty =
   let r = x.binder_relevance in
   let r' = infer_assumption env t ty in
-  let x = if Sorts.relevance_equal r r'
-    then x
-    else (warn_bad_relevance x; {x with binder_relevance = r'})
+  let x =
+    if Sorts.relevance_equal r r' then x
+    else
+      let () = warn_bad_relevance_binder env r' (RelDecl.LocalAssum (x, t)) in
+      {x with binder_relevance = r'}
   in
   x
 
@@ -88,17 +117,22 @@ exception ArgumentsMismatch
 
 let instantiate_context u subst nas ctx =
   let open Context.Rel.Declaration in
+  let instantiate_relevance na =
+    { na with binder_relevance = UVars.subst_instance_relevance u na.binder_relevance }
+  in
   let rec instantiate i ctx = match ctx with
   | [] -> if 0 <= i then raise ArgumentsMismatch else []
   | LocalAssum (na, ty) :: ctx ->
     let ctx = instantiate (pred i) ctx in
     let subst = Esubst.subs_liftn i subst in
+    let na = instantiate_relevance na in
     let ty = esubst u subst ty in
     let () = check_binding_relevance na nas.(i) in
     LocalAssum (nas.(i), ty) :: ctx
   | LocalDef (na, ty, bdy) :: ctx ->
     let ctx = instantiate (pred i) ctx in
     let subst = Esubst.subs_liftn i subst in
+    let na = instantiate_relevance na in
     let ty = esubst u subst ty in
     let bdy = esubst u subst bdy in
     let () = check_binding_relevance na nas.(i) in
@@ -125,6 +159,7 @@ let type_of_type u =
 let type_of_sort = function
   | SProp | Prop | Set -> type1
   | Type u -> type_of_type u
+  | QSort (_, u) -> type_of_type u
 
 (*s Type of a de Bruijn index. *)
 
@@ -213,7 +248,7 @@ let rec check_empty_stack = function
 let type_of_apply env func funt argsv argstv =
   let open CClosure in
   let len = Array.length argsv in
-  let infos = create_clos_infos all env in
+  let infos = create_clos_infos RedFlags.all env in
   let tab = create_tab () in
   let rec apply_rec i typ =
     if Int.equal i len then term_of_fconstr typ
@@ -266,14 +301,14 @@ let type_of_parameters env ctx u argsv argstv =
 (* Type of primitive constructs *)
 let type_of_prim_type _env u (type a) (prim : a CPrimitives.prim_type) = match prim with
   | CPrimitives.PT_int63 ->
-    assert (Univ.Instance.is_empty u);
+    assert (UVars.Instance.is_empty u);
     Constr.mkSet
   | CPrimitives.PT_float64 ->
-    assert (Univ.Instance.is_empty u);
+    assert (UVars.Instance.is_empty u);
     Constr.mkSet
   | CPrimitives.PT_array ->
-    begin match Univ.Instance.to_array u with
-    | [|u|] ->
+    begin match UVars.Instance.to_array u with
+    | [||], [|u|] ->
       let ty = Constr.mkType (Univ.Universe.make u) in
       Constr.mkProd(Context.anonR, ty , ty)
     | _ -> anomaly Pp.(str"universe instance for array type should have length 1")
@@ -281,16 +316,16 @@ let type_of_prim_type _env u (type a) (prim : a CPrimitives.prim_type) = match p
 
 let type_of_int env =
   match env.retroknowledge.Retroknowledge.retro_int63 with
-  | Some c -> mkConst c
+  | Some c -> UnsafeMonomorphic.mkConst c
   | None -> CErrors.user_err Pp.(str"The type int must be registered before this construction can be typechecked.")
 
 let type_of_float env =
   match env.retroknowledge.Retroknowledge.retro_float64 with
-  | Some c -> mkConst c
+  | Some c -> UnsafeMonomorphic.mkConst c
   | None -> CErrors.user_err Pp.(str"The type float must be registered before this construction can be typechecked.")
 
 let type_of_array env u =
-  assert (Univ.Instance.length u = 1);
+  assert (UVars.Instance.length u = (0,1));
   match env.retroknowledge.Retroknowledge.retro_array with
   | Some c -> mkConstU (c,u)
   | None -> CErrors.user_err Pp.(str"The type array must be registered before this construction can be typechecked.")
@@ -305,7 +340,7 @@ let sort_of_product env domsort rangsort =
     (* Product rule (Prop/Set,Set,Set) *)
     | ((Prop | Set),  Set) -> rangsort
     (* Product rule (Type,Set,?) *)
-    | (Type u1, Set) ->
+    | ((Type u1 | QSort (_, u1)), Set) ->
         if is_impredicative_set env then
           (* Rule is (Type,Set,Set) in the Set-impredicative calculus *)
           rangsort
@@ -314,10 +349,14 @@ let sort_of_product env domsort rangsort =
           Sorts.sort_of_univ (Universe.sup Universe.type0 u1)
     (* Product rule (Prop,Type_i,Type_i) *)
     | (Set,  Type u2)  -> Sorts.sort_of_univ (Universe.sup Universe.type0 u2)
+    | (Set,  QSort (q, u2))  ->
+      Sorts.qsort q (Universe.sup Universe.type0 u2)
     (* Product rule (Prop,Type_i,Type_i) *)
-    | (Prop, Type _)  -> rangsort
+    | (Prop, (Type _ | QSort _))  -> rangsort
     (* Product rule (Type_i,Type_i,Type_i) *)
-    | (Type u1, Type u2) -> Sorts.sort_of_univ (Universe.sup u1 u2)
+    | ((Type u1 | QSort (_, u1)), Type u2) -> Sorts.sort_of_univ (Universe.sup u1 u2)
+    | ((Type u1 | QSort (_, u1)), (QSort (q, u2))) ->
+      Sorts.qsort q (Universe.sup u1 u2)
 
 (* [judge_of_product env name (typ1,s1) (typ2,s2)] implements the rule
 
@@ -348,7 +387,7 @@ let check_cast env c ct k expected_type =
     | DEFAULTcast ->
       default_conv CUMUL env ct expected_type
     | NATIVEcast ->
-      let sigma = Genlambda.empty_evars in
+      let sigma = Genlambda.empty_evars env in
       Nativeconv.native_conv CUMUL sigma env ct expected_type
   with NotConvertible ->
     error_actual_type env (make_judge c ct) expected_type
@@ -379,11 +418,25 @@ let judge_of_array env u tj defj =
    the App case of execute; from this constraints, the expected
    dynamic constraints of the form u<=v are enforced *)
 
-let type_of_inductive_knowing_parameters env (ind,u) args =
+let make_param_univs env indu spec args argtys =
+  Array.to_list @@ Array.mapi (fun i argt ~expected ->
+      match (snd (Reduction.dest_arity env argt)) with
+      | SProp | exception Reduction.NotArity ->
+        Type_errors.error_cant_apply_bad_type env
+          (i+1, mkType (Universe.make expected), argt)
+          (make_judge (mkIndU indu) (Inductive.type_of_inductive (spec, snd indu)))
+          (make_judgev args argtys)
+      | Prop -> TemplateProp
+      | Set -> TemplateUniv Universe.type0
+      | Type u -> TemplateUniv u
+      | QSort _ -> assert false)
+    argtys
+
+let type_of_inductive_knowing_parameters env (ind,u as indu) args argst =
   let (mib,_mip) as spec = lookup_mind_specif env ind in
   check_hyps_inclusion env (GlobRef.IndRef ind) mib.mind_hyps;
   let t,cst = Inductive.constrained_type_of_inductive_knowing_parameters
-      (spec,u) (Inductive.make_param_univs env args)
+      (spec,u) (make_param_univs env indu spec args argst)
   in
   check_constraints cst env;
   t
@@ -440,10 +493,10 @@ let check_branch_types env (_mib, mip) ci u pms c _ct lft (pctx, p) =
     let expbrt = it_mkLambda_or_LetIn expbrt brctx in
     error_ill_formed_branch env c ((ci.ci_ind, i + 1), u) brt expbrt
 
-let should_invert_case env ci =
-  ci.ci_relevance == Sorts.Relevant &&
+let should_invert_case env r ci =
+  Sorts.relevance_equal r Sorts.Relevant &&
   let mib,mip = lookup_mind_specif env ci.ci_ind in
-  mip.mind_relevance == Sorts.Irrelevant &&
+  Sorts.relevance_equal mip.mind_relevance Sorts.Irrelevant &&
   (* NB: it's possible to have 2 ctors or arguments to 1 ctor by unsetting univ checks
      but we don't do special reduction in such cases
 
@@ -456,64 +509,70 @@ let should_invert_case env ci =
 let type_case_scrutinee env (mib, _mip) (u', largs) u pms (pctx, p) c =
   let (params, realargs) = List.chop mib.mind_nparams largs in
   (* Check that the type of the scrutinee is <= the expected argument type *)
-  let () = Array.iter2 (fun p1 p2 -> Reduction.conv ~l2r:true env p1 p2) (Array.of_list params) pms in
+  let () = try Array.iter2 (fun p1 p2 -> Conversion.conv ~l2r:true env p1 p2) (Array.of_list params) pms
+    with NotConvertible -> raise Type_errors.(TypeError (env,IllFormedCaseParams))
+  in
   (* We use l2r:true for compat with old versions which used CONV with arguments
      flipped. It is relevant for performance eg in bedrock / Kami. *)
-  let cst = match mib.mind_variance with
-  | None -> Univ.enforce_eq_instances u u' Univ.Constraints.empty
-  | Some variance -> Univ.enforce_leq_variance_instances variance u' u Univ.Constraints.empty
+  let qcst, ucst = match mib.mind_variance with
+  | None -> UVars.enforce_eq_instances u u' Sorts.QUConstraints.empty
+  | Some variance -> UVars.enforce_leq_variance_instances variance u' u Sorts.QUConstraints.empty
   in
-  let () = check_constraints cst env in
+  let () = check_qconstraints qcst env in
+  let () = check_constraints ucst env in
   let subst = Vars.subst_of_rel_context_instance_list pctx (realargs @ [c]) in
   Vars.substl subst p
 
-let type_of_case env (mib, mip) ci u pms (pctx, p, pt) iv c ct _lf lft =
+let type_of_case env (mib, mip as specif) ci u pms (pctx, pnas, p, rp, pt) iv c ct lf lft =
   let ((ind, u'), largs) =
     try find_rectype env ct
     with Not_found -> error_case_not_inductive env (make_judge c ct) in
   (* Various well-formedness conditions *)
-  let sp = match destSort (whd_all (push_rel_context pctx env) pt) with
+  let () = if Inductive.is_private specif then error_case_on_private_ind env ind in
+  let sp = match destSort (Reduction.whd_all (push_rel_context pctx env) pt) with
   | sp -> sp
   | exception DestKO ->
     error_elim_arity env (ind, u') c None
   in
-  let rp = Sorts.relevance_of_sort sp in
-  let ci = if ci.ci_relevance == rp then ci
-    else (warn_bad_relevance_ci (); {ci with ci_relevance=rp})
+  let rp =
+    let expected = Sorts.relevance_of_sort sp in
+    if Sorts.relevance_equal rp expected then rp
+    else
+      let () = warn_bad_relevance_case env expected (ci, u, pms, ((pnas, p), rp), iv, c, lf) in
+      expected
   in
-  let () = check_case_info env (ind, u') rp ci in
+  let () = check_case_info env (ind, u') ci in
   let () =
     let is_inversion = match iv with
       | NoInvert -> false
       | CaseInvert _ -> true (* contents already checked *)
     in
-    if not (is_inversion = should_invert_case env ci)
+    if not (is_inversion = should_invert_case env rp ci)
     then error_bad_invert env
   in
-  let () =
-    let ksort = Sorts.family sp in
-    if not (Sorts.family_leq ksort mip.mind_kelim) then
-      let s = inductive_sort_family mip in
-      let pj = make_judge (it_mkLambda_or_LetIn p pctx) (it_mkProd_or_LetIn pt pctx) in
-      let kinds = Some (pj, mip.mind_kelim, ksort, s) in
-      error_elim_arity env (ind, u') c kinds
+  let () = if not (is_allowed_elimination (specif,u) sp) then begin
+    let pj = make_judge (it_mkLambda_or_LetIn p pctx) (it_mkProd_or_LetIn pt pctx) in
+    let kinds = Some (pj, sp) in
+    error_elim_arity env (ind, u') c kinds
+  end
   in
   (* Check that the scrutinee has the right type *)
   let rslty = type_case_scrutinee env (mib, mip) (u', largs) u pms (pctx, p) c in
   (* We return the "higher" inductive universe instance from the predicate,
      the branches must be typeable using these universes. *)
   let () = check_branch_types env (mib, mip) ci u pms c ct lft (pctx, p) in
-  ci, rslty
+  rp, rslty
 
 let type_of_projection env p c ct =
-  let pty = lookup_projection p env in
+  let pr, pty = lookup_projection p env in
   let (ind,u), args =
     try find_rectype env ct
     with Not_found -> error_case_not_inductive env (make_judge c ct)
   in
   assert(Ind.CanOrd.equal (Projection.inductive p) ind);
+  let pr = UVars.subst_instance_relevance u pr in
   let ty = Vars.subst_instance_constr u pty in
-  substl (c :: CList.rev args) ty
+  pr, substl (c :: CList.rev args) ty
 
 
 (* Fixpoints. *)
@@ -534,7 +593,7 @@ let check_fixpoint env lna lar vdef vdeft =
 let type_of_global_in_context env r =
   let open Names.GlobRef in
   match r with
-  | VarRef id -> Environ.named_type id env, Univ.AbstractContext.empty
+  | VarRef id -> Environ.named_type id env, UVars.AbstractContext.empty
   | ConstRef c ->
     let cb = Environ.lookup_constant c env in
     let univs = Declareops.constant_polymorphic_context cb in
@@ -542,25 +601,36 @@ let type_of_global_in_context env r =
   | IndRef ind ->
     let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
     let univs = Declareops.inductive_polymorphic_context mib in
-    let inst = Univ.make_abstract_instance univs in
+    let inst = UVars.make_abstract_instance univs in
     Inductive.type_of_inductive (specif, inst), univs
   | ConstructRef cstr ->
     let (mib,_ as specif) =
       Inductive.lookup_mind_specif env (inductive_of_constructor cstr)
     in
     let univs = Declareops.inductive_polymorphic_context mib in
-    let inst = Univ.make_abstract_instance univs in
+    let inst = UVars.make_abstract_instance univs in
     Inductive.type_of_constructor (cstr,inst) specif, univs
 
 (************************************************************************)
 (************************************************************************)
 
-let check_binder_annot s x =
+let check_assum_annot env s x t =
   let r = x.binder_relevance in
   let r' = Sorts.relevance_of_sort s in
-  if r' == r
+  if Sorts.relevance_equal r' r
   then x
-  else (warn_bad_relevance x; {x with binder_relevance = r'})
+  else
+    let () = warn_bad_relevance_binder env r' (RelDecl.LocalAssum (x, t)) in
+    {x with binder_relevance = r'}
+
+let check_let_annot env s x c t =
+  let r = x.binder_relevance in
+  let r' = Sorts.relevance_of_sort s in
+  if Sorts.relevance_equal r' r
+  then x
+  else
+    let () = warn_bad_relevance_binder env r' (RelDecl.LocalDef (x, c, t)) in
+    {x with binder_relevance = r'}
 
 (* The typing machine. *)
     (* ATTENTION : faudra faire le typage du contexte des Const,
@@ -571,9 +641,10 @@ let rec execute env cstr =
   match kind cstr with
     (* Atomic terms *)
     | Sort s ->
-      (match s with
-       | SProp -> if not (Environ.sprop_allowed env) then error_disallowed_sprop env
-       | _ -> ());
+      let () = match s with
+      | SProp -> if not (Environ.sprop_allowed env) then error_disallowed_sprop env
+      | QSort _ | Prop | Set | Type _ -> ()
+      in
       cstr, type_of_sort s
 
     | Rel n ->
@@ -585,10 +656,12 @@ let rec execute env cstr =
     | Const c ->
       cstr, type_of_constant env c
 
-    | Proj (p, c) ->
+    | Proj (p, r, c) ->
       let c', ct = execute env c in
-      let cstr = if c == c' then cstr else mkProj (p,c') in
-      cstr, type_of_projection env p c' ct
+      let r', ty = type_of_projection env p c' ct in
+      assert (Sorts.relevance_equal r r');
+      let cstr = if c == c' then cstr else mkProj (p,r,c') in
+      cstr, ty
 
     (* Lambda calculus operators *)
     | App (f,args) ->
@@ -596,7 +669,7 @@ let rec execute env cstr =
         let f', ft =
           match kind f with
           | Ind ind when Environ.template_polymorphic_pind ind env ->
-            f, type_of_inductive_knowing_parameters env ind argst
+            f, type_of_inductive_knowing_parameters env ind args' argst
           | _ ->
             (* No template polymorphism *)
             execute env f
@@ -606,7 +679,7 @@ let rec execute env cstr =
 
     | Lambda (name,c1,c2) ->
       let c1', s = execute_is_type env c1 in
-      let name' = check_binder_annot s name in
+      let name' = check_assum_annot env s name c1' in
       let env1 = push_rel (LocalAssum (name',c1')) env in
       let c2', c2t = execute env1 c2 in
       let cstr = if name == name' && c1 == c1' && c2 == c2' then cstr else mkLambda(name',c1',c2') in
@@ -614,7 +687,7 @@ let rec execute env cstr =
 
     | Prod (name,c1,c2) ->
       let c1', vars = execute_is_type env c1 in
-      let name' = check_binder_annot vars name in
+      let name' = check_assum_annot env vars name c1' in
       let env1 = push_rel (LocalAssum (name',c1')) env in
       let c2', vars' = execute_is_type env1 c2 in
       let cstr = if name == name' && c1 == c1' && c2 == c2' then cstr else mkProd(name',c1',c2') in
@@ -623,7 +696,7 @@ let rec execute env cstr =
     | LetIn (name,c1,c2,c3) ->
       let c1', c1t = execute env c1 in
       let c2', c2s = execute_is_type env c2 in
-      let name' = check_binder_annot c2s name in
+      let name' = check_let_annot env c2s name c1' c2' in
       let () = check_cast env c1' c1t DEFAULTcast c2' in
       let env1 = push_rel (LocalDef (name',c1',c2')) env in
       let c3', c3t = execute env1 c3 in
@@ -646,7 +719,7 @@ let rec execute env cstr =
     | Construct c ->
       cstr, type_of_constructor env c
 
-    | Case (ci, u, pms, p, iv, c, lf) ->
+    | Case (ci, u, pms, (p,rp), iv, c, lf) ->
         let c', ct = execute env c in
         let iv' = match iv with
           | NoInvert -> NoInvert
@@ -655,7 +728,7 @@ let rec execute env cstr =
             let ct' = mkApp (mkIndU (ci.ci_ind,u), args) in
             let (ct', _) : constr * Sorts.t = execute_is_type env ct' in
             let () = conv_leq env ct ct' in
-            let _, args' = decompose_appvect ct' in
+            let _, args' = decompose_app ct' in
             if args == args' then iv
             else CaseInvert {indices=Array.sub args' (Array.length pms) (Array.length indices)}
         in
@@ -672,7 +745,7 @@ let rec execute env cstr =
           let realdecls, _ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
           let self =
             let args = Context.Rel.instance mkRel 0 mip.mind_arity_ctxt in
-            let inst = Instance.of_array (Array.init (Instance.length u) Level.var) in
+            let inst = UVars.Instance.(abstract_instance (length u)) in
             mkApp (mkIndU (ci.ci_ind, inst), args)
           in
           let realdecls = LocalAssum (Context.make_annot Anonymous mip.mind_relevance, self) :: realdecls in
@@ -706,10 +779,10 @@ let rec execute env cstr =
           if br == br' then b else (nas, br')
         in
         let lf' = Array.Smart.map_i build_one_branch lf in
-        let ci', t = type_of_case env (mib, mip) ci u pms' (pctx, p', pt) iv' c' ct lf' lft in
+        let rp', t = type_of_case env (mib, mip) ci u pms' (pctx, fst p, p', rp, pt) iv' c' ct lf' lft in
         let eqbr (_, br1) (_, br2) = br1 == br2 in
-        let cstr = if ci == ci' && pms == pms' && c == c' && snd p == p' && iv == iv' && Array.equal eqbr lf lf' then cstr
-          else mkCase (ci', u, pms', (fst p, p'), iv', c', lf')
+        let cstr = if rp == rp' && pms == pms' && c == c' && snd p == p' && iv == iv' && Array.equal eqbr lf lf' then cstr
+          else mkCase (ci, u, pms', ((fst p, p'), rp'), iv', c', lf')
         in
         cstr, t
 
@@ -732,8 +805,8 @@ let rec execute env cstr =
     | Float _ -> cstr, type_of_float env
     | Array(u,t,def,ty) ->
       (* ty : Type@{u} and all of t,def : ty *)
-      let ulev = match Univ.Instance.to_array u with
-        | [|u|] -> u
+      let ulev = match UVars.Instance.to_array u with
+        | [||], [|u|] -> u
         | _ -> assert false
       in
       let ty',tyty = execute env ty in
@@ -773,14 +846,27 @@ and execute_array env cs =
   let cs = Array.Smart.map_i (fun i c -> let c, ty = execute env c in tys.(i) <- ty; c) cs in
   cs, tys
 
+let execute env c =
+  NewProfile.profile "Typeops.infer" (fun () -> execute env c) ()
+
 (* Derived functions *)
 
+let check_declared_qualities env qualities =
+  let module S = Sorts.QVar.Set in
+  let unknown = S.diff qualities env.env_qualities in
+  if S.is_empty unknown then ()
+  else error_undeclared_qualities env unknown
+
 let check_wellformed_universes env c =
-  let univs = universes_of_constr c in
+  let qualities, univs = sort_and_universes_of_constr c in
+  check_declared_qualities env qualities;
   try UGraph.check_declared_universes (universes env) univs
   with UGraph.UndeclaredLevel u ->
     error_undeclared_universe env u
 
+let check_wellformed_universes env c =
+  NewProfile.profile "check-wf-univs" (fun () -> check_wellformed_universes env c) ()
+
 let infer env constr =
   let () = check_wellformed_universes env constr in
   let constr, t = execute env constr in
@@ -807,13 +893,13 @@ let check_context env rels =
     match d with
       | LocalAssum (x,ty) ->
         let jty = infer_type env ty in
-        let x = check_binder_annot jty.utj_type x in
+        let x = check_assum_annot env jty.utj_type x jty.utj_val in
         push_rel d env, LocalAssum (x,jty.utj_val) :: rels
       | LocalDef (x,bd,ty) ->
         let j1 = infer env bd in
         let jty = infer_type env ty in
         conv_leq env j1.uj_type ty;
-        let x = check_binder_annot jty.utj_type x in
+        let x = check_let_annot env jty.utj_type x j1.uj_val jty.utj_val in
         push_rel d env, LocalDef (x,j1.uj_val,jty.utj_val) :: rels)
     rels ~init:(env,[])
 
@@ -828,7 +914,8 @@ let judge_of_variable env x = make_judge (mkVar x) (type_of_variable env x)
 let judge_of_constant env cst = make_judge (mkConstU cst) (type_of_constant env cst)
 
 let judge_of_projection env p cj =
-  make_judge (mkProj (p,cj.uj_val)) (type_of_projection env p cj.uj_val cj.uj_type)
+  let r, ty = type_of_projection env p cj.uj_val cj.uj_type in
+  make_judge (mkProj (p,r,cj.uj_val)) ty
 
 let dest_judgev v =
   Array.map j_val v, Array.map j_type v
@@ -869,37 +956,38 @@ let type_of_prim_const env _u c =
     int_ty ()
 
 let type_of_prim env u t =
+  let module UM = UnsafeMonomorphic in
   let int_ty () = type_of_int env in
   let float_ty () = type_of_float env in
   let array_ty u a = mkApp(type_of_array env u, [|a|]) in
   let bool_ty () =
     match env.retroknowledge.Retroknowledge.retro_bool with
-    | Some ((ind,_),_) -> Constr.mkInd ind
+    | Some ((ind,_),_) -> UM.mkInd ind
     | None -> CErrors.user_err Pp.(str"The type bool must be registered before this primitive.")
   in
   let compare_ty () =
     match env.retroknowledge.Retroknowledge.retro_cmp with
-    | Some ((ind,_),_,_) -> Constr.mkInd ind
+    | Some ((ind,_),_,_) -> UM.mkInd ind
     | None -> CErrors.user_err Pp.(str"The type compare must be registered before this primitive.")
   in
   let f_compare_ty () =
     match env.retroknowledge.Retroknowledge.retro_f_cmp with
-    | Some ((ind,_),_,_,_) -> Constr.mkInd ind
+    | Some ((ind,_),_,_,_) -> UM.mkInd ind
     | None -> CErrors.user_err Pp.(str"The type float_comparison must be registered before this primitive.")
   in
   let f_class_ty () =
     match env.retroknowledge.Retroknowledge.retro_f_class with
-    | Some ((ind,_),_,_,_,_,_,_,_,_) -> Constr.mkInd ind
+    | Some ((ind,_),_,_,_,_,_,_,_,_) -> UM.mkInd ind
     | None -> CErrors.user_err Pp.(str"The type float_class must be registered before this primitive.")
   in
   let pair_ty fst_ty snd_ty =
     match env.retroknowledge.Retroknowledge.retro_pair with
-    | Some (ind,_) -> Constr.mkApp(Constr.mkInd ind, [|fst_ty;snd_ty|])
+    | Some (ind,_) -> Constr.mkApp(UM.mkInd ind, [|fst_ty;snd_ty|])
     | None -> CErrors.user_err Pp.(str"The type pair must be registered before this primitive.")
   in
   let carry_ty int_ty =
     match env.retroknowledge.Retroknowledge.retro_carry with
-    | Some ((ind,_),_) -> Constr.mkApp(Constr.mkInd ind, [|int_ty|])
+    | Some ((ind,_),_) -> Constr.mkApp(UM.mkInd ind, [|int_ty|])
     | None -> CErrors.user_err Pp.(str"The type carry must be registered before this primitive.")
   in
   let open CPrimitives in
@@ -928,7 +1016,7 @@ let type_of_prim env u t =
                        tr_type n arg_ty, nary_op (n + 1) ret_ty r)
   in
   let params, args_ty, ret_ty = types t in
-  assert (AbstractContext.size (univs t) = Instance.length u);
+  assert (UVars.AbstractContext.size (univs t) = UVars.Instance.length u);
   Vars.subst_instance_constr u
     (Term.it_mkProd_or_LetIn (nary_op 0 ret_ty args_ty) params)
 
