@@ -37,8 +37,11 @@ val cofix_evaluated_tag : tag
 type structured_constant =
   | Const_sort of Sorts.t
   | Const_ind of inductive
+  | Const_evar of Evar.t
   | Const_b0 of tag
+  | Const_quality of Sorts.Quality.t
   | Const_univ_level of Univ.Level.t
+  | Const_univ_instance of UVars.Instance.t
   | Const_val of structured_values
   | Const_uint of Uint63.t
   | Const_float of Float64.t
@@ -143,7 +146,7 @@ external val_of_annot_switch : annot_switch -> values = "%identity"
 (** Destructors *)
 
 val whd_val : values -> kind
-val uni_lvl_val : values -> Univ.Level.t
+val uni_instance : values -> UVars.Instance.t
 
 (** Arguments *)
 
