@@ -22,7 +22,7 @@
 type pp_tag = string
 
 type block_type =
-  | Pp_hbox   of int
+  | Pp_hbox
   | Pp_vbox   of int
   | Pp_hvbox  of int
   | Pp_hovbox of int
@@ -114,6 +114,7 @@ let spc   () = Ppcmd_print_break (1,0)
 let cut   () = Ppcmd_print_break (0,0)
 let align () = Ppcmd_print_break (0,0)
 let int   n  = str (string_of_int n)
+let int64 n  = str (Int64.to_string n)
 let real  r  = str (string_of_float r)
 let bool  b  = str (string_of_bool b)
 
@@ -131,7 +132,7 @@ let strbrk s =
 let ismt = function | Ppcmd_empty -> true | _ -> false
 
 (* boxing commands *)
-let h   n s = Ppcmd_box(Pp_hbox n,s)
+let h     s = Ppcmd_box(Pp_hbox,s)
 let v   n s = Ppcmd_box(Pp_vbox n,s)
 let hv  n s = Ppcmd_box(Pp_hvbox n,s)
 let hov n s = Ppcmd_box(Pp_hovbox n,s)
@@ -139,19 +140,9 @@ let hov n s = Ppcmd_box(Pp_hovbox n,s)
 (* Opening and closing of tags *)
 let tag t s = Ppcmd_tag(t,s)
 
-(* In new syntax only double quote char is escaped by repeating it *)
-let escape_string s =
-  let rec escape_at s i =
-    if i<0 then s
-    else if s.[i] == '"' then
-      let s' = String.sub s 0 i^"\""^String.sub s i (String.length s - i) in
-      escape_at s' (i-1)
-    else escape_at s (i-1) in
-  escape_at s (String.length s - 1)
-
-let qstring s = str "\"" ++ str (escape_string s) ++ str "\""
+let qstring s = str (CString.quote_coq_string s)
 let qs = qstring
-let quote s = h 0 (str "\"" ++ s ++ str "\"")
+let quote s = h (str "\"" ++ s ++ str "\"")
 
 let rec pr_com ft s =
   let (s1,os) =
@@ -181,7 +172,7 @@ let split_tag tag =
 (* pretty printing functions *)
 let pp_with ft pp =
   let cpp_open_box = function
-    | Pp_hbox n   -> Format.pp_open_hbox ft ()
+    | Pp_hbox     -> Format.pp_open_hbox ft ()
     | Pp_vbox n   -> Format.pp_open_vbox ft n
     | Pp_hvbox n  -> Format.pp_open_hvbox ft n
     | Pp_hovbox n -> Format.pp_open_hovbox ft n
@@ -197,9 +188,9 @@ let pp_with ft pp =
     | Ppcmd_print_break(m,n)  -> pp_print_break ft m n
     | Ppcmd_force_newline     -> pp_force_newline ft ()
     | Ppcmd_comment coms      -> List.iter (pr_com ft) coms
-    | Ppcmd_tag(tag, s)       -> pp_open_tag  ft tag [@warning "-3"];
+    | Ppcmd_tag(tag, s)       -> pp_open_stag ft (String_tag tag);
                                  pp_cmd s;
-                                 pp_close_tag ft () [@warning "-3"]
+                                 pp_close_stag ft ()
   in
   pp_cmd pp
 
@@ -222,6 +213,8 @@ let pr_arg pr x = spc () ++ pr x
 let pr_non_empty_arg pr x = let pp = pr x in if ismt pp then mt () else spc () ++ pr x
 let pr_opt pr = function None -> mt () | Some x -> pr_arg pr x
 let pr_opt_no_spc pr = function None -> mt () | Some x -> pr x
+let pr_opt_default prdf pr = function None -> prdf () | Some x -> pr_arg pr x
+let pr_opt_no_spc_default prdf pr = function None -> prdf () | Some x -> pr x
 
 (** TODO: merge with CString.ordinal *)
 let pr_nth n =
@@ -272,6 +265,8 @@ let pr_sequence pr l = prlist_sep_lastsep true spc spc pr l
    [pr a ++ str "," ++ pr b ++ str "," ++ ... ++ str "and" ++ pr c] *)
 let pr_enum pr l = prlist_sep_lastsep true pr_comma (fun () -> str " and" ++ spc ()) pr l
 
+let pr_choice pr l = prlist_sep_lastsep true pr_comma (fun () -> str " or" ++ spc ()) pr l
+
 let pr_vertical_list pr = function
   | [] -> str "none" ++ fnl ()
   | l -> fnl () ++ str "  " ++ hov 0 (prlist_with_sep fnl pr l) ++ fnl ()
@@ -309,12 +304,14 @@ let db_print_pp fmt pp =
   let block_type fmt btype =
     let (bt, v) =
       match btype with
-      | Pp_hbox v -> ("Pp_hbox", v)
-      | Pp_vbox v -> ("Pp_vbox", v)
-      | Pp_hvbox v -> ("Pp_hvbox", v)
-      | Pp_hovbox v -> ("Pp_hovbox", v)
+      | Pp_hbox -> ("Pp_hbox", None)
+      | Pp_vbox v -> ("Pp_vbox", Some v)
+      | Pp_hvbox v -> ("Pp_hvbox", Some v)
+      | Pp_hovbox v -> ("Pp_hovbox", Some v)
     in
-    fprintf fmt "%s %d" bt v
+    match v with
+    | None -> fprintf fmt "%s" bt
+    | Some v -> fprintf fmt "%s %d" bt v
   in
   let rec db_print_pp_r indent pp =
     let ind () = fprintf fmt "%s" (String.make (2 * indent) ' ') in
