@@ -26,9 +26,7 @@ end
     the standard OCaml marshalling function. *)
 type 'a summary_declaration = {
   stage : Stage.t;
-  freeze_function : marshallable:bool -> 'a;
-  (** freeze_function [true] is for marshalling to disk.
-   *  e.g. lazy must be forced *)
+  freeze_function : unit -> 'a;
   unfreeze_function : 'a -> unit;
   init_function : unit -> unit }
 
@@ -43,36 +41,31 @@ type 'a summary_declaration = {
     the responsibility of plugins to initialize themselves properly.
 *)
 
-val declare_summary : string -> 'a summary_declaration -> unit
+val declare_summary : string -> ?make_marshallable:('a -> 'a) -> 'a summary_declaration -> unit
 
 (** We provide safe projection from the summary to the types stored in
    it.*)
 module Dyn : Dyn.S
 
-val declare_summary_tag : string -> 'a summary_declaration -> 'a Dyn.tag
+val declare_summary_tag : string -> ?make_marshallable:('a -> 'a) -> 'a summary_declaration -> 'a Dyn.tag
 
 (** All-in-one reference declaration + summary registration.
     It behaves just as OCaml's standard [ref] function, except
     that a [declare_summary] is done, with [name] as string.
     The [init_function] restores the reference to its initial value.
-    The [freeze_function] can be overridden
     The [stage] argument defaults to [Interp] and should be changed to [Synterp]
     for references which are read from and written to during the syntactic
-    interpretation. *)
+    interpretation.
 
-val ref : ?stage:Stage.t -> ?freeze:(marshallable:bool -> 'a -> 'a) -> name:string -> 'a -> 'a ref
-val ref_tag : ?stage:Stage.t -> ?freeze:(marshallable:bool -> 'a -> 'a) -> name:string -> 'a -> 'a ref * 'a Dyn.tag
+    When [local:true] the value is local to the process, i.e. not sent to proof workers.
+    Consequently it doesn't need to be of a marshallable type.
+    It is useful to implement a local cache for example.
 
-(* As [ref] but the value is local to a process, i.e. not sent to, say, proof
- * workers.  It is useful to implement a local cache for example. *)
-module Local : sig
+    [ref_tag] is never local.
+*)
 
-  type 'a local_ref
-  val ref : ?stage:Stage.t -> name:string -> 'a -> 'a local_ref
-  val (:=) : 'a local_ref -> 'a -> unit
-  val (!) : 'a local_ref -> 'a
-
-end
+val ref : ?stage:Stage.t -> ?local:bool -> name:string -> 'a -> 'a ref
+val ref_tag : ?stage:Stage.t -> name:string -> 'a -> 'a ref * 'a Dyn.tag
 
 (** Special summary for ML modules.  This summary entry is special
     because its unfreeze may load ML code and hence add summary
@@ -88,22 +81,33 @@ val declare_ml_modules_summary : (string option * string) list summary_declarati
 
 val nop : unit -> unit
 
-(** The type [frozen] is a snapshot of the states of all the registered
-    tables of the system. *)
+module type FrozenStage = sig
 
-type frozen
+  (** The type [frozen] is a snapshot of the states of all the registered
+      tables of the system. *)
 
-val empty_frozen : frozen
-val freeze_summaries : marshallable:bool -> frozen
-val freeze_staged_summaries : Stage.t -> marshallable:bool -> frozen
-val unfreeze_summaries : ?partial:bool -> frozen -> unit
-val init_summaries : unit -> unit
+  type frozen
 
-(** Typed projection of the summary. Experimental API, use with CARE *)
+  val empty_frozen : frozen
+  val freeze_summaries : unit -> frozen
+  val make_marshallable : frozen -> frozen
+  val unfreeze_summaries : ?partial:bool -> frozen -> unit
+  val init_summaries : unit -> unit
 
-val modify_summary : frozen -> 'a Dyn.tag -> 'a -> frozen
-val project_from_summary : frozen -> 'a Dyn.tag -> 'a
-val remove_from_summary : frozen -> 'a Dyn.tag -> frozen
+end
+
+module Synterp : FrozenStage
+module Interp : sig
+
+  include FrozenStage
+
+  (** Typed projection of the summary. Experimental API, use with CARE *)
+
+  val modify_summary : frozen -> 'a Dyn.tag -> 'a -> frozen
+  val project_from_summary : frozen -> 'a Dyn.tag -> 'a
+  val remove_from_summary : frozen -> 'a Dyn.tag -> frozen
+
+end
 
 (** {6 Debug} *)
 val dump : unit -> (int * string) list
