@@ -28,7 +28,7 @@ let keywords =
       "error"; "delay"; "force"; "_"; "__"]
     Id.Set.empty
 
-let pp_comment s = str";; "++h 0 s++fnl ()
+let pp_comment s = str ";; " ++ h s ++ fnl ()
 
 let pp_header_comment = function
   | None -> mt ()
@@ -133,6 +133,8 @@ let rec pp_expr env args =
       paren (str "Prelude.error \"EXTRACTION OF UINT NOT IMPLEMENTED\"")
     | MLfloat _ ->
       paren (str "Prelude.error \"EXTRACTION OF FLOAT NOT IMPLEMENTED\"")
+    | MLparray _ ->
+            paren (str "Prelude.error \"EXTRACTION OF PARRAY NOT IMPLEMENTED\"")
 
 and pp_cons_args env = function
   | MLcons (_,r,args) when is_coinductive r ->
