@@ -28,7 +28,7 @@ let mkGCases (rto, l, brl) = DAst.make @@ GCases (RegularStyle, rto, l, brl)
 
 let mkGHole () =
   DAst.make
-  @@ GHole (Evar_kinds.BinderType Anonymous, Namegen.IntroAnonymous, None)
+  @@ GHole (GBinderType Anonymous)
 
 (*
   Some basic functions to decompose glob_constrs
@@ -106,7 +106,6 @@ let change_vars =
             , change_vars mapping lhs
             , change_vars mapping rhs )
         | GRec _ -> user_err ?loc Pp.(str "Local (co)fixes are not supported")
-        | GSort _ as x -> x | GHole _ as x -> x | GInt _ as x -> x
         | GFloat _ as x -> x
         | GCast (b, k, c) ->
           GCast (change_vars mapping b, k, change_vars mapping c)
@@ -115,7 +114,8 @@ let change_vars =
             ( u
             , Array.map (change_vars mapping) t
             , change_vars mapping def
-            , change_vars mapping ty ))
+            , change_vars mapping ty )
+        | GSort _ | GHole _ | GGenarg _ | GInt _ as x -> x)
       rt
   and change_vars_br mapping ({CAst.loc; v = idl, patl, res} as br) =
     let new_mapping = List.fold_right Id.Map.remove idl mapping in
@@ -283,7 +283,7 @@ let rec alpha_rt excluded rt =
         , alpha_rt excluded lhs
         , alpha_rt excluded rhs )
     | GRec _ -> user_err Pp.(str "Not handled GRec")
-    | (GSort _ | GInt _ | GFloat _ | GHole _) as rt -> rt
+    | (GSort _ | GInt _ | GFloat _ | GHole _ | GGenarg _) as rt -> rt
     | GCast (b, k, c) ->
       GCast (alpha_rt excluded b, k, alpha_rt excluded c)
     | GApp (f, args) ->
@@ -343,6 +343,7 @@ let is_free_in id =
           is_free_in cond || is_free_in br1 || is_free_in br2
         | GRec _ -> user_err Pp.(str "Not handled GRec") | GSort _ -> false
         | GHole _ -> false
+        | GGenarg _ -> false (* XXX isn't this incorrect? *)
         | GCast (b, _, t) ->
           is_free_in b || is_free_in t
         | GInt _ | GFloat _ -> false
@@ -419,7 +420,7 @@ let replace_var_by_term x_id term =
             , replace_var_by_pattern lhs
             , replace_var_by_pattern rhs )
         | GRec _ -> CErrors.user_err (Pp.str "Not handled GRec")
-        | (GSort _ | GHole _) as rt -> rt
+        | (GSort _ | GHole _ | GGenarg _) as rt -> rt (* is this correct for GGenarg? *)
         | GInt _ as rt -> rt
         | GFloat _ as rt -> rt
         | GArray (u, t, def, ty) ->
@@ -440,46 +441,46 @@ let replace_var_by_term x_id term =
 (* checking unifiability of patterns *)
 exception NotUnifiable
 
-let rec are_unifiable_aux = function
+let rec are_unifiable_aux env = function
   | [] -> ()
   | (l, r) :: eqs -> (
     match (DAst.get l, DAst.get r) with
-    | PatVar _, _ | _, PatVar _ -> are_unifiable_aux eqs
+    | PatVar _, _ | _, PatVar _ -> are_unifiable_aux env eqs
     | PatCstr (constructor1, cpl1, _), PatCstr (constructor2, cpl2, _) ->
-      if not (Construct.CanOrd.equal constructor2 constructor1) then
+      if not (Environ.QConstruct.equal env constructor2 constructor1) then
         raise NotUnifiable
       else
         let eqs' =
           try List.combine cpl1 cpl2 @ eqs
           with Invalid_argument _ -> anomaly (Pp.str "are_unifiable_aux.")
         in
-        are_unifiable_aux eqs' )
+        are_unifiable_aux env eqs' )
 
-let are_unifiable pat1 pat2 =
+let are_unifiable env pat1 pat2 =
   try
-    are_unifiable_aux [(pat1, pat2)];
+    are_unifiable_aux env [(pat1, pat2)];
     true
   with NotUnifiable -> false
 
-let rec eq_cases_pattern_aux = function
+let rec eq_cases_pattern_aux env = function
   | [] -> ()
   | (l, r) :: eqs -> (
     match (DAst.get l, DAst.get r) with
-    | PatVar _, PatVar _ -> eq_cases_pattern_aux eqs
+    | PatVar _, PatVar _ -> eq_cases_pattern_aux env eqs
     | PatCstr (constructor1, cpl1, _), PatCstr (constructor2, cpl2, _) ->
-      if not (Construct.CanOrd.equal constructor2 constructor1) then
+      if not (Environ.QConstruct.equal env constructor2 constructor1) then
         raise NotUnifiable
       else
         let eqs' =
           try List.combine cpl1 cpl2 @ eqs
           with Invalid_argument _ -> anomaly (Pp.str "eq_cases_pattern_aux.")
         in
-        eq_cases_pattern_aux eqs'
+        eq_cases_pattern_aux env eqs'
     | _ -> raise NotUnifiable )
 
-let eq_cases_pattern pat1 pat2 =
+let eq_cases_pattern env pat1 pat2 =
   try
-    eq_cases_pattern_aux [(pat1, pat2)];
+    eq_cases_pattern_aux env [(pat1, pat2)];
     true
   with NotUnifiable -> false
 
@@ -502,7 +503,7 @@ let expand_as =
   in
   let rec expand_as map =
     DAst.map (function
-      | (GRef _ | GEvar _ | GPatVar _ | GSort _ | GHole _ | GInt _ | GFloat _)
+      | (GRef _ | GEvar _ | GPatVar _ | GSort _ | GHole _ | GGenarg _ | GInt _ | GFloat _)
         as rt ->
         rt
       | GVar id as rt -> (
@@ -544,79 +545,85 @@ let expand_as =
   in
   expand_as Id.Map.empty
 
-(* [resolve_and_replace_implicits ?expected_type env sigma rt] solves implicits of [rt] w.r.t. [env] and [sigma] and then replace them by their solution
- *)
+(* [resolve_and_replace_implicits] solves implicits of its argument and replaces them by their solution *)
 
-exception Found of Evd.evar_info
-
-let resolve_and_replace_implicits ?(flags = Pretyping.all_and_fail_flags)
-    ?(expected_type = Pretyping.WithoutTypeConstraint) env sigma rt =
+let resolve_and_replace_implicits exptyp env sigma rt =
   let open Evd in
   let open Evar_kinds in
   (* we first (pseudo) understand [rt] and get back the computed evar_map *)
   (* FIXME : JF (30/03/2017) I'm not completely sure to have split understand as needed.
      If someone knows how to prevent solved existantial removal in  understand, please do not hesitate to change the computation of [ctx] here *)
-  let ctx, _, _ =
-    Pretyping.ise_pretype_gen flags env sigma Glob_ops.empty_lvar expected_type
-      rt
+  let implicit_holes = ref [] in
+  let binder_holes = ref [] in
+  let ctx =
+    let open Pretyping in
+    let open Evarutil in
+    (* Intercept the pretyper for holes and record the generated evar *)
+    let register_evar kind loc evk = match kind with
+    | GImplicitArg (grk, pk, bk) -> implicit_holes := ((loc, grk, pk, bk), evk) :: !implicit_holes
+    | GBinderType na -> binder_holes := ((loc, na), evk) :: !binder_holes
+    | _ -> ()
+    in
+    let pretype_hole self kind ?loc ~flags tycon env sigma =
+      let sigma, j = default_pretyper.pretype_hole self kind ?loc ~flags tycon env sigma in
+      (* The value is guaranteed to be an undefined evar at this point *)
+      let evk, _ = EConstr.destEvar sigma j.uj_val in
+      let () = register_evar kind loc evk  in
+      sigma, j
+    in
+    let pretype_type self c ?loc ~flags valcon env sigma =
+      let sigma, j = default_pretyper.pretype_type self c ?loc ~flags valcon env sigma in
+      let () = match DAst.get c, EConstr.kind sigma j.utj_val with
+      | GHole (kind), Evar (evk, _) -> register_evar kind c.CAst.loc evk
+      | _ -> ()
+      in
+      sigma, j
+    in
+    let flags = Pretyping.all_and_fail_flags in
+    let pretype_flags = {
+      program_mode = false;
+      use_coercions = true;
+      poly = false;
+      resolve_tc = true;
+    } in
+    let vars = Evarutil.VarSet.variables (Global.env ()) in
+    let hypnaming = RenameExistingBut vars in
+    let genv = GlobEnv.make ~hypnaming env sigma Glob_ops.empty_lvar in
+    let pretyper = { default_pretyper with pretype_hole; pretype_type } in
+    let sigma', _ = eval_pretyper pretyper ~flags:pretype_flags (Some exptyp) genv sigma rt in
+    solve_remaining_evars flags env ~initial:sigma sigma'
   in
   let ctx = Evd.minimize_universes ctx in
   let f c =
     EConstr.of_constr
       (Evarutil.nf_evars_universes ctx (EConstr.Unsafe.to_constr c))
   in
+  let expand_hole evopt default = match evopt with
+  | None -> default
+  | Some evk ->
+    (* we found the evar corresponding to this hole *)
+    let EvarInfo evi = Evd.find ctx evk in
+    match Evd.evar_body evi with
+    | Evar_defined c ->
+      (* we just have to lift the solution in glob_term *)
+      Detyping.detype Detyping.Now Id.Set.empty env ctx (f c)
+    | Evar_empty ->
+      (* the hole was not solved : we do nothing *)
+      default
+  in
   (* then we map [rt] to replace the implicit holes by their values *)
   let rec change rt =
     match DAst.get rt with
-    | GHole (ImplicitArg (grk, pk, bk), _, _) -> (
-      try
-        (* we only want to deal with implicit arguments *)
-
-        (* we scan the new evar map to find the evar corresponding to this hole (by looking the source *)
-        Evd.fold (* to simulate an iter *)
-          (fun _ evi _ ->
-            match Evd.evar_source evi with
-            | loc_evi, ImplicitArg (gr_evi, p_evi, b_evi) ->
-              if
-                GlobRef.equal grk gr_evi && pk = p_evi && bk = b_evi
-                && rt.CAst.loc = loc_evi
-              then raise (Found evi)
-            | _ -> ())
-          ctx ();
-        (* the hole was not solved : we do nothing *)
-        rt
-      with Found evi -> (
-        (* we found the evar corresponding to this hole *)
-        match Evd.evar_body evi with
-        | Evar_defined c ->
-          (* we just have to lift the solution in glob_term *)
-          Detyping.detype Detyping.Now false Id.Set.empty env ctx (f c)
-        | Evar_empty -> rt (* the hole was not solved : we do nothing *) ) )
-    | GHole (BinderType na, _, _) ->
-      (* we only want to deal with implicit arguments *)
-      let res =
-        try
-          (* we scan the new evar map to find the evar corresponding to this hole (by looking the source *)
-          Evd.fold (* to simulate an iter *)
-            (fun _ evi _ ->
-              match Evd.evar_source evi with
-              | loc_evi, BinderType na' ->
-                if Name.equal na na' && rt.CAst.loc = loc_evi then
-                  raise (Found evi)
-              | _ -> ())
-            ctx ();
-          (* the hole was not solved : we do nothing *)
-          rt
-        with Found evi -> (
-          (* we found the evar corresponding to this hole *)
-          match Evd.evar_body evi with
-          | Evar_defined c ->
-            (* we just have to lift the solution in glob_term *)
-            Detyping.detype Detyping.Now false Id.Set.empty env ctx (f c)
-          | Evar_empty -> rt )
-        (* the hole was not solved : we d when falseo nothing *)
+    | GHole (GImplicitArg (grk, pk, bk)) ->
+      let eq (loc1, gr1, p1, b1) (loc2, gr2, p2, b2) =
+        Environ.QGlobRef.equal env gr1 gr2 && p1 = p2 && b1 == (b2 : bool) && loc1 = (loc2 : Loc.t option)
       in
-      res
+      let evopt = List.assoc_f_opt eq (rt.CAst.loc, grk, pk, bk) !implicit_holes in
+      expand_hole evopt rt
+    | GHole (GBinderType na) ->
+      let eq (loc1, na1) (loc2, na2) = Name.equal na1 na2 && loc1 = (loc2 : Loc.t option) in
+      let evopt = List.assoc_f_opt eq (rt.CAst.loc, na) !binder_holes in
+      expand_hole evopt rt
     | _ -> Glob_ops.map_glob_constr change rt
   in
   change rt
