@@ -51,7 +51,7 @@ let global_of_constr_nofail sigma c =
 let mk_atom c = CClosure.mk_atom (EConstr.Unsafe.to_constr c)
 
 let rec mk_clos_but sigma f_map n t =
-  let (f, args) = Termops.decompose_app_vect sigma t in
+  let (f, args) = EConstr.decompose_app sigma t in
   match f_map (global_of_constr_nofail sigma f) with
   | Some tag ->
       let map i t = tag_arg sigma f_map n (tag i) t in
@@ -60,12 +60,12 @@ let rec mk_clos_but sigma f_map n t =
   | None -> mk_atom t
 
 and tag_arg sigma f_map n tag c = match tag with
-| Eval -> mk_clos (Esubst.subs_id n, Univ.Instance.empty) (EConstr.Unsafe.to_constr c)
+| Eval -> mk_clos (Esubst.subs_id n, UVars.Instance.empty) (EConstr.Unsafe.to_constr c)
 | Prot -> mk_atom c
 | Rec -> mk_clos_but sigma f_map n c
 
 let interp_map l t =
-  try Some(List.assoc_f GlobRef.equal t l) with Not_found -> None
+  try Some(List.assoc_f GlobRef.CanOrd.equal t l) with Not_found -> None
 
 let protect_maps : protection String.Map.t ref = ref String.Map.empty
 let add_map s m = protect_maps := String.Map.add s m !protect_maps
@@ -76,7 +76,7 @@ let lookup_map map =
 
 let protect_red map env sigma c =
   let tab = create_tab () in
-  let infos = Evarutil.create_clos_infos env sigma all in
+  let infos = Evarutil.create_clos_infos env sigma RedFlags.all in
   let map = lookup_map map sigma c in
   let rec eval n c = match EConstr.kind sigma c with
   | Prod (na, t, u) -> EConstr.mkProd (na, eval n t, eval (n + 1) u)
@@ -146,12 +146,14 @@ let decl_constant name univs c =
   let open Constr in
   let vars = CVars.universes_of_constr c in
   let univs = UState.restrict_universe_context ~lbound:(Global.universes_lbound ()) univs vars in
-  let () = DeclareUctx.declare_universe_context ~poly:false univs in
+  let () = Global.push_context_set ~strict:true univs in
   let types = (Typeops.infer (Global.env ()) c).uj_type in
   let univs = UState.Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders in
-  mkConst(declare_constant ~name
-            ~kind:Decls.(IsProof Lemma)
-            (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
+  (* UnsafeMonomorphic: we always do poly:false *)
+  UnsafeMonomorphic.mkConst
+    (declare_constant ~name
+       ~kind:Decls.(IsProof Lemma)
+       (DefinitionEntry (definition_entry ~opaque:true ~types ~univs c)))
 
 let decl_constant na suff univs c =
   let na = Namegen.next_global_ident_away (Nameops.add_suffix na suff) Id.Set.empty in
@@ -747,18 +749,18 @@ let dest_field env sigma th_spec =
   let th_typ = Retyping.get_type_of env sigma th_spec in
   match EConstr.kind sigma th_typ with
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force afield_theory) f ->
+        when isRefX env sigma (Lazy.force afield_theory) f ->
         let sigma, rth = plapp sigma af_ar
           [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (None,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;sub;opp;div;inv;req|])
-        when isRefX sigma (Lazy.force field_theory) f ->
+        when isRefX env sigma (Lazy.force field_theory) f ->
         let sigma, rth =
           plapp sigma f_r
             [|r;zero;one;add;mul;sub;opp;div;inv;req;th_spec|] in
         (Some false,r,zero,one,add,mul,Some sub,Some opp,div,inv,req,rth)
     | App(f,[|r;zero;one;add;mul;div;inv;req|])
-        when isRefX sigma (Lazy.force sfield_theory) f ->
+        when isRefX env sigma (Lazy.force sfield_theory) f ->
         let sigma, rth = plapp sigma sf_sr
           [|r;zero;one;add;mul;div;inv;req;th_spec|] in
         (Some true,r,zero,one,add,mul,None,None,div,inv,req,rth)
