@@ -47,23 +47,21 @@ val factorize_eqns : 'a cases_clauses_g -> 'a disjunctive_cases_clauses_g
    [isgoal] tells if naming must avoid global-level synonyms as intro does
    [ctx] gives the names of the free variables *)
 
-val detype_names : bool -> Id.Set.t -> names_context -> env -> evar_map -> constr -> glob_constr
-
-val detype : 'a delay -> ?lax:bool -> bool -> Id.Set.t -> env -> evar_map -> constr -> 'a glob_constr_g
+val detype : 'a delay -> ?isgoal:bool -> Id.Set.t -> env -> evar_map -> constr -> 'a glob_constr_g
 
 val detype_sort : evar_map -> Sorts.t -> glob_sort
 
-val detype_rel_context : 'a delay -> ?lax:bool -> constr option -> Id.Set.t -> (names_context * env) ->
+val detype_rel_context : 'a delay -> constr option -> Id.Set.t -> (names_context * env) ->
   evar_map -> rel_context -> 'a glob_decl_g list
 
 val share_pattern_names :
-  (Id.Set.t -> names_context -> 'c -> Pattern.constr_pattern -> 'a) -> int ->
+  (Id.Set.t -> names_context -> 'c -> 'd Pattern.constr_pattern_r -> 'a) -> int ->
   (Name.t * binding_kind * 'b option * 'a) list ->
-  Id.Set.t -> names_context -> 'c -> Pattern.constr_pattern ->
-  Pattern.constr_pattern ->
+  Id.Set.t -> names_context -> 'c -> 'd Pattern.constr_pattern_r ->
+  'd Pattern.constr_pattern_r ->
   (Name.t * binding_kind * 'b option * 'a) list * 'a * 'a
 
-val detype_closed_glob : ?lax:bool -> bool -> Id.Set.t -> env -> evar_map -> closed_glob_constr -> glob_constr
+val detype_closed_glob :  ?isgoal:bool -> Id.Set.t -> env -> evar_map -> closed_glob_constr -> glob_constr
 
 (** look for the index of a named var or a nondep var as it is renamed *)
 val lookup_name_as_displayed  : env -> evar_map -> constr -> Id.t -> int option
@@ -72,17 +70,6 @@ val lookup_index_as_renamed : env -> evar_map -> constr -> int -> int option
 val force_wildcard : unit -> bool
 val synthetize_type : unit -> bool
 
-(** Utilities to transform kernel cases to simple pattern-matching problem *)
-
-val it_destRLambda_or_LetIn_names : bool list -> glob_constr -> Name.t list * glob_constr
-val simple_cases_matrix_of_branches :
-  inductive -> (int * bool list * glob_constr) list -> cases_clauses
-val return_type_of_predicate :
-  inductive -> bool list -> glob_constr -> predicate_pattern * glob_constr option
-
-val subst_genarg_hook :
-  (substitution -> Genarg.glob_generic_argument -> Genarg.glob_generic_argument) Hook.t
-
 module PrintingInductiveMake :
   functor (Test : sig
     val encode : Environ.env -> Libnames.qualid -> Names.inductive
