@@ -11,10 +11,18 @@
 open Names
 open Glob_term
 
+val map_glob_sort_gen : ('a -> 'b) -> 'a glob_sort_gen -> 'b glob_sort_gen
+
 (** Equalities *)
 
+val glob_sort_gen_eq : ('a -> 'a -> bool) -> 'a glob_sort_gen -> 'a glob_sort_gen -> bool
+
 val glob_sort_eq : Glob_term.glob_sort -> Glob_term.glob_sort -> bool
 
+val glob_qvar_eq :  glob_qvar -> glob_qvar -> bool
+
+val glob_quality_eq :  glob_quality -> glob_quality -> bool
+
 val glob_level_eq : Glob_term.glob_level -> Glob_term.glob_level -> bool
 
 val cases_pattern_eq : 'a cases_pattern_g -> 'a cases_pattern_g -> bool
@@ -28,16 +36,10 @@ val alias_of_pat : 'a cases_pattern_g -> Name.t
 
 val set_pat_alias : Id.t -> 'a cases_pattern_g -> 'a cases_pattern_g
 
-val cast_type_eq : ('a -> 'a -> bool) ->
-  'a cast_type -> 'a cast_type -> bool
+val cast_kind_eq : Constr.cast_kind ->  Constr.cast_kind -> bool
 
 val glob_constr_eq : 'a glob_constr_g -> 'a glob_constr_g -> bool
 
-(** Mapping [cast_type] *)
-
-val map_cast_type : ('a -> 'b) -> 'a cast_type -> 'b cast_type
-val smartmap_cast_type : ('a -> 'a) -> 'a cast_type -> 'a cast_type
-
 (** Operations on [glob_constr] *)
 
 val cases_pattern_loc : 'a cases_pattern_g -> Loc.t option
@@ -57,9 +59,13 @@ val binding_kind_eq : binding_kind -> binding_kind -> bool
 val map_glob_constr_left_to_right :
   (glob_constr -> glob_constr) -> glob_constr -> glob_constr
 
+val map_glob_constr_left_to_right_with_names :
+  (glob_constr -> glob_constr) -> (Name.t -> Name.t) -> glob_constr -> glob_constr
+
 val warn_variable_collision : ?loc:Loc.t -> Id.t -> unit
 
 val mk_glob_constr_eq : (glob_constr -> glob_constr -> bool) ->
+  (Name.t -> Name.t -> glob_constr option -> glob_constr option -> bool) ->
   glob_constr -> glob_constr -> bool
 
 val fold_glob_constr : ('a -> glob_constr -> 'a) -> 'a -> glob_constr -> 'a
@@ -114,3 +120,7 @@ val add_patterns_for_params_remove_local_defs : Environ.env -> constructor ->
   'a cases_pattern_g list -> 'a cases_pattern_g list
 
 val empty_lvar : Ltac_pretype.ltac_var_map
+
+val kind_of_glob_kind : glob_evar_kind -> Evar_kinds.t
+
+val naming_of_glob_kind : glob_evar_kind -> Namegen.intro_pattern_naming_expr
