@@ -11,12 +11,10 @@
 open CErrors
 open Util
 open Names
-open Univ
 open Term
 open Constr
 open Vars
 open Context
-open Termops
 open Declarations
 open Declareops
 open Environ
@@ -29,14 +27,29 @@ open Context.Rel.Declaration
 let type_of_inductive env (ind,u) =
  let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
  Typeops.check_hyps_inclusion env (GlobRef.IndRef ind) mib.mind_hyps;
- Inductive.type_of_inductive (specif,u)
+ let t = Inductive.type_of_inductive (specif,u) in
+ Arguments_renaming.rename_type t (IndRef ind)
+
+let e_type_of_inductive env sigma (ind,u) =
+ let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
+ Reductionops.check_hyps_inclusion env sigma (GlobRef.IndRef ind) mib.mind_hyps;
+ let t = Inductive.type_of_inductive (specif, EConstr.Unsafe.to_instance u) in
+ EConstr.of_constr (Arguments_renaming.rename_type t (IndRef ind))
 
 (* Return type as quoted by the user *)
 let type_of_constructor env (cstr,u) =
  let (mib,_ as specif) =
    Inductive.lookup_mind_specif env (inductive_of_constructor cstr) in
  Typeops.check_hyps_inclusion env (GlobRef.ConstructRef cstr) mib.mind_hyps;
- Inductive.type_of_constructor (cstr,u) specif
+ let t = Inductive.type_of_constructor (cstr,u) specif in
+ Arguments_renaming.rename_type t (ConstructRef cstr)
+
+let e_type_of_constructor env sigma (cstr,u) =
+ let (mib,_ as specif) =
+   Inductive.lookup_mind_specif env (inductive_of_constructor cstr) in
+ Reductionops.check_hyps_inclusion env sigma (GlobRef.ConstructRef cstr) mib.mind_hyps;
+ let t = Inductive.type_of_constructor (cstr,EConstr.Unsafe.to_instance u) specif in
+ EConstr.of_constr (Arguments_renaming.rename_type t (ConstructRef cstr))
 
 (* Return constructor types in user form *)
 let type_of_constructors env (ind,u as indu) =
@@ -61,11 +74,13 @@ let lift_inductive_family n = liftn_inductive_family n 1
 
 let substnl_ind_family l n = map_ind_family (substnl l n)
 
-let relevance_of_inductive_family env ((ind,_),_ : inductive_family) =
+let relevance_of_inductive_family env (ind,_ : inductive_family) =
   Inductive.relevance_of_inductive env ind
 
 type inductive_type = IndType of inductive_family * EConstr.constr list
 
+let ind_of_ind_type = function IndType (((ind,_),_),_) -> ind
+
 let make_ind_type (indf, realargs) = IndType (indf,realargs)
 let dest_ind_type (IndType (indf,realargs)) = (indf,realargs)
 
@@ -101,14 +116,11 @@ let mis_is_recursive (ind,mib,mip) =
   mis_is_recursive_subset (List.interval 0 (mib.mind_ntypes - 1))
     mip.mind_recargs
 
-let mis_nf_constructor_type ((ind,u),mib,mip) j =
-  let specif = mip.mind_nf_lc
-  and ntypes = mib.mind_ntypes
-  and nconstr = Array.length mip.mind_consnames in
-  let make_Ik k = mkIndU (((fst ind),ntypes-k-1),u) in
+let mis_nf_constructor_type ((_,j),u) (mib,mip) =
+  let nconstr = Array.length mip.mind_consnames in
   if j > nconstr then user_err Pp.(str "Not enough constructors in the type.");
-  let (ctx, cty) = specif.(j - 1) in
-  substl (List.init ntypes make_Ik) (subst_instance_constr u (Term.it_mkProd_or_LetIn cty ctx))
+  let (ctx, cty) = mip.mind_nf_lc.(j - 1) in
+  subst_instance_constr u (Term.it_mkProd_or_LetIn cty ctx)
 
 (* Number of constructors *)
 
@@ -116,132 +128,95 @@ let nconstructors env ind =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   Array.length mip.mind_consnames
 
-let nconstructors_env env ind = nconstructors env ind
-[@@ocaml.deprecated "Alias for Inductiveops.nconstructors"]
-
 (* Arity of constructors excluding parameters, excluding local defs *)
 
 let constructors_nrealargs env ind =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealargs
 
-let constructors_nrealargs_env env ind = constructors_nrealargs env ind
-[@@ocaml.deprecated "Alias for Inductiveops.constructors_nrealargs"]
-
 (* Arity of constructors excluding parameters, including local defs *)
 
 let constructors_nrealdecls env ind =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealdecls
 
-let constructors_nrealdecls_env env ind = constructors_nrealdecls env ind
-[@@ocaml.deprecated "Alias for Inductiveops.constructors_nrealdecls"]
-
 (* Arity of constructors including parameters, excluding local defs *)
 
 let constructor_nallargs env (ind,j) =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealargs.(j-1) + mib.mind_nparams
 
-let constructor_nallargs_env env (indsp,j) = constructor_nallargs env (indsp,j)
-[@@ocaml.deprecated "Alias for Inductiveops.constructor_nallargs"]
-
 (* Arity of constructors including params, including local defs *)
 
 let constructor_nalldecls env (ind,j) = (* TOCHANGE en decls *)
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealdecls.(j-1) + Context.Rel.length (mib.mind_params_ctxt)
 
-let constructor_nalldecls_env env (indsp,j) = constructor_nalldecls env (indsp,j)
-[@@ocaml.deprecated "Alias for Inductiveops.constructor_nalldecls"]
-
 (* Arity of constructors excluding params, excluding local defs *)
 
 let constructor_nrealargs env (ind,j) =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealargs.(j-1)
 
-let constructor_nrealargs_env env (ind,j) = constructor_nrealargs env (ind,j)
-[@@ocaml.deprecated "Alias for Inductiveops.constructor_nrealargs"]
-
 (* Arity of constructors excluding params, including local defs *)
 
 let constructor_nrealdecls env (ind,j) = (* TOCHANGE en decls *)
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_consnrealdecls.(j-1)
 
-let constructor_nrealdecls_env env (ind,j) = constructor_nrealdecls env (ind,j)
-[@@ocaml.deprecated "Alias for Inductiveops.constructor_nrealdecls"]
-
 (* Length of arity, excluding params, excluding local defs *)
 
 let inductive_nrealargs env ind =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_nrealargs
 
-let inductive_nrealargs_env env ind = inductive_nrealargs env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nrealargs"]
-
 (* Length of arity, excluding params, including local defs *)
 
 let inductive_nrealdecls env ind =
   let (_,mip) = Inductive.lookup_mind_specif env ind in
   mip.mind_nrealdecls
 
-let inductive_nrealdecls_env env ind = inductive_nrealdecls env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nrealdecls"]
-
 (* Full length of arity (w/o local defs) *)
 
 let inductive_nallargs env ind =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   mib.mind_nparams + mip.mind_nrealargs
 
-let inductive_nallargs_env env ind = inductive_nallargs env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nallargs"]
-
 (* Length of arity (w/o local defs) *)
 
 let inductive_nparams env ind =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   mib.mind_nparams
 
-let inductive_nparams_env env ind = inductive_nparams env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nparams"]
-
 (* Length of arity (with local defs) *)
 
 let inductive_nparamdecls env ind =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   Context.Rel.length mib.mind_params_ctxt
 
-let inductive_nparamdecls_env env ind = inductive_nparamdecls env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nparamsdecls"]
-
 (* Full length of arity (with local defs) *)
 
 let inductive_nalldecls env ind =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   Context.Rel.length (mib.mind_params_ctxt) + mip.mind_nrealdecls
 
-let inductive_nalldecls_env env ind = inductive_nalldecls env ind
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_nalldecls"]
-
 (* Others *)
 
 let inductive_paramdecls env (ind,u) =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
     Inductive.inductive_paramdecls (mib,u)
 
-let inductive_paramdecls_env env (ind,u) = inductive_paramdecls env (ind,u)
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_paramsdecls"]
-
 let inductive_alldecls env (ind,u) =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
     Vars.subst_instance_context u mip.mind_arity_ctxt
 
-let inductive_alldecls_env env (ind,u) = inductive_alldecls env (ind,u)
-[@@ocaml.deprecated "Alias for Inductiveops.inductive_alldecls"]
+let inductive_alltags env ind =
+  let (mib,mip) = Inductive.lookup_mind_specif env ind in
+  Context.Rel.to_tags mip.mind_arity_ctxt
+
+let constructor_alltags env (ind,j) =
+  let (mib,mip) = Inductive.lookup_mind_specif env ind in
+  Context.Rel.to_tags (fst mip.mind_nf_lc.(j-1))
 
 let constructor_has_local_defs env (indsp,j) =
   let (mib,mip) = Inductive.lookup_mind_specif env indsp in
@@ -255,20 +230,94 @@ let inductive_has_local_defs env ind =
   let l2 = mib.mind_nparams + mip.mind_nrealargs in
   not (Int.equal l1 l2)
 
+(* XXX use above_prop from the ustate *)
+let quality_leq q q' =
+  let open Sorts.Quality in
+  match q, q' with
+  | QVar q, QVar q' -> Sorts.QVar.equal q q'
+  | QConstant q, QConstant q' ->
+    begin match q, q' with
+    | QSProp, _
+    | _, QType
+    | QProp, QProp
+      -> true
+    | (QProp|QType), _ -> false
+    end
+  | (QVar _|QConstant _), _ -> false
+
+type squash = SquashToSet | SquashToQuality of Sorts.Quality.t
+
+let is_squashed sigma ((_,mip),u) =
+  match mip.mind_arity with
+  | TemplateArity _ -> None (* template is never squashed *)
+  | RegularArity a ->
+    match mip.mind_squashed with
+    | None -> None
+    | Some squash ->
+      let indq = EConstr.ESorts.quality sigma
+          (EConstr.ESorts.make @@ UVars.subst_instance_sort u a.mind_sort)
+      in
+      match squash with
+      | AlwaysSquashed -> begin match a.mind_sort with
+          | Sorts.Set -> Some SquashToSet
+          | _ -> Some (SquashToQuality indq)
+        end
+      | SometimesSquashed squash ->
+        (* impredicative set squashes are always AlwaysSquashed,
+           so here if inds=Set it is a sort poly squash (see "foo6" in test sort_poly.v) *)
+        if Sorts.Quality.Set.for_all (fun q ->
+            let q = UVars.subst_instance_quality u q in
+            let q = UState.nf_quality (Evd.evar_universe_context sigma) q in
+            quality_leq q indq) squash
+        then None
+        else Some (SquashToQuality indq)
+
+let is_allowed_elimination sigma ((mib,_),_ as specifu) s =
+  let open Sorts in
+  match mib.mind_record with
+  | PrimRecord _ -> true
+  | NotRecord | FakeRecord ->
+    match is_squashed sigma specifu with
+    | None -> true
+    | Some SquashToSet ->
+      begin match EConstr.ESorts.kind sigma s with
+      | SProp|Prop|Set -> true
+      | QSort _ | Type _ ->
+        (* XXX in [Type u] case, should we check [u == set] in the ugraph? *)
+        false
+      end
+    | Some (SquashToQuality indq) -> quality_leq (EConstr.ESorts.quality sigma s) indq
+
+let elim_sort (_,mip) =
+  if Option.is_empty mip.mind_squashed then Sorts.InType
+  else Inductive.inductive_sort_family mip
+
 let top_allowed_sort env (kn,i as ind) =
-  let (mib,mip) = Inductive.lookup_mind_specif env ind in
-  mip.mind_kelim
+  let specif = Inductive.lookup_mind_specif env ind in
+  elim_sort specif
 
 let sorts_below top =
-  List.filter (fun s -> Sorts.family_leq s top) Sorts.[InSProp;InProp;InSet;InType]
+  List.filter (fun s ->
+      Sorts.family_equal s top
+      || match s, top with
+      | InQSort, _ -> assert false
+      | _, InQSort -> false
+      | InSProp, _ -> true
+      | InProp, InSet -> true
+      | _, InType -> true
+      | (InProp|InSet|InType), _ -> false)
+    Sorts.[InSProp;InProp;InSet;InType]
 
-let has_dependent_elim mib =
+let sorts_for_schemes specif =
+  sorts_below (elim_sort specif)
+
+let has_dependent_elim (mib,mip) =
   match mib.mind_record with
-  | PrimRecord _ -> mib.mind_finite == BiFinite
+  | PrimRecord _ -> mib.mind_finite == BiFinite || mip.mind_relevance == Irrelevant
   | NotRecord | FakeRecord -> true
 
 (* Annotation for cases *)
-let make_case_info env ind r style =
+let make_case_info env ind style =
   let (mib,mip) = Inductive.lookup_mind_specif env ind in
   let ind_tags =
     Context.Rel.to_tags (List.firstn mip.mind_nrealdecls mip.mind_arity_ctxt) in
@@ -281,7 +330,6 @@ let make_case_info env ind r style =
     ci_npar    = mib.mind_nparams;
     ci_cstr_ndecls = mip.mind_consnrealdecls;
     ci_cstr_nargs = mip.mind_consnrealargs;
-    ci_relevance = r;
     ci_pp_info = print_info }
 
 (*s Useful functions *)
@@ -298,7 +346,7 @@ let lift_constructor n cs = {
   cs_cstr = cs.cs_cstr;
   cs_params = List.map (lift n) cs.cs_params;
   cs_nargs = cs.cs_nargs;
-  cs_args = lift_rel_context n cs.cs_args;
+  cs_args = Vars.lift_rel_context n cs.cs_args;
   cs_concl_realargs = Array.map (liftn n (cs.cs_nargs+1)) cs.cs_concl_realargs
 }
 
@@ -306,18 +354,21 @@ let lift_constructor n cs = {
 let instantiate_params t params sign =
   let nnonrecpar = Context.Rel.nhyps sign - List.length params in
   (* Adjust the signature if recursively non-uniform parameters are not here *)
-  let _,sign = context_chop nnonrecpar sign in
-  let _,t = decompose_prod_n_assum (Context.Rel.length sign) t in
-  let subst = subst_of_rel_context_instance sign params in
+  let _,sign = Context.Rel.chop_nhyps nnonrecpar sign in
+  let _,t = decompose_prod_n_decls (Context.Rel.length sign) t in
+  let subst = subst_of_rel_context_instance_list sign params in
   substl subst t
 
-let get_constructor ((ind,u as indu),mib,mip,params) j =
-  assert (j <= Array.length mip.mind_consnames);
-  let typi = mis_nf_constructor_type (indu,mib,mip) j in
+let instantiate_constructor_params (_,u as cstru) (mib,_ as mind_specif) params =
+  let typi = mis_nf_constructor_type cstru mind_specif in
   let ctx = Vars.subst_instance_context u mib.mind_params_ctxt in
-  let typi = instantiate_params typi params ctx in
-  let (args,ccl) = decompose_prod_assum typi in
-  let (_,allargs) = decompose_app ccl in
+  instantiate_params typi params ctx
+
+let get_constructor ((ind,u),mib,mip,params) j =
+  assert (j <= Array.length mip.mind_consnames);
+  let typi = instantiate_constructor_params ((ind,j),u) (mib,mip) params in
+  let (args,ccl) = decompose_prod_decls typi in
+  let (_,allargs) = decompose_app_list ccl in
   let vargs = List.skipn (List.length params) allargs in
   { cs_cstr = (ith_constructor_of_inductive ind j,u);
     cs_params = params;
@@ -332,37 +383,81 @@ let get_constructors env (ind,params) =
 
 let get_projections = Environ.get_projections
 
-let make_case_or_project env sigma indf ci pred c branches =
+let make_case_invert env (IndType (((ind,u),params),indices)) ~case_relevance:r ci =
+  if Typeops.should_invert_case env r ci
+  then CaseInvert {indices=Array.of_list indices}
+  else NoInvert
+
+let make_project env sigma ind pred c branches ps =
   let open EConstr in
-  let projs = get_projections env (fst (fst indf)) in
+  assert(Array.length branches == 1);
+  let na, ty, t = destLambda sigma pred in
+  let mib, mip as specif = Inductive.lookup_mind_specif env ind in
+  let () =
+    if (* dependent *) not (Vars.noccurn sigma 1 t) &&
+         not (has_dependent_elim specif) then
+      user_err
+        Pp.(str"Dependent case analysis not allowed" ++
+              str" on inductive type " ++ Termops.pr_global_env env (IndRef ind) ++ str ".")
+  in
+  let branch = branches.(0) in
+  let ctx, br = decompose_lambda_n_decls sigma mip.mind_consnrealdecls.(0) branch in
+  let _, u = destInd sigma (fst (decompose_app sigma ty)) in
+  let u = Unsafe.to_instance u in
+  let mkProj i c =
+    let p, r = ps.(i) in
+    let r = UVars.subst_instance_relevance u r in
+    mkProj (Projection.make p true, r, c)
+  in
+  let proj = match EConstr.destRel sigma br with
+    | exception DestKO -> None
+    | i ->
+      begin match List.skipn (i-1) ctx with
+      | exception Failure _ -> None
+      | ctx -> match ctx with
+        | [] -> None
+        | LocalDef _ :: _ ->
+          (* XXX Maybe we should produce the applied constant for this letin pseudoprojection?
+             We would have to get the params etc*)
+          None
+        | LocalAssum _ :: ctx ->
+          (* This match is just a projection *)
+          Some (mkProj (Context.Rel.nhyps ctx) c)
+      end
+  in
+  match proj with
+  | Some proj -> proj
+  | None ->
+  let n, len, ctx =
+    List.fold_right
+      (fun decl (i, j, ctx) ->
+         match decl with
+         | LocalAssum (na, ty) ->
+           let t = mkProj i (mkRel j) in
+           (i + 1, j + 1, LocalDef (na, t, Vars.liftn 1 j ty) :: ctx)
+         | LocalDef (na, b, ty) ->
+           (i, j + 1, LocalDef (na, Vars.liftn 1 j b, Vars.liftn 1 j ty) :: ctx))
+      ctx (0, 1, [])
+  in
+  mkLetIn (na, c, ty, it_mkLambda_or_LetIn (Vars.liftn 1 (mip.mind_consnrealdecls.(0) + 1) br) ctx)
+
+let simple_make_case_or_project env sigma ci pred invert c branches =
+  let open EConstr in
+  let ind = ci.ci_ind in
+  let projs = get_projections env ind in
   match projs with
-  | None -> (mkCase (ci, pred, c, branches))
-  | Some ps ->
-     assert(Array.length branches == 1);
-     let na, ty, t = destLambda sigma pred in
-     let () =
-       let (ind, _), _ = dest_ind_family indf in
-       let mib, _ = Inductive.lookup_mind_specif env ind in
-       if (* dependent *) not (Vars.noccurn sigma 1 t) &&
-          not (has_dependent_elim mib) then
-       user_err ~hdr:"make_case_or_project"
-                    Pp.(str"Dependent case analysis not allowed" ++
-                     str" on inductive type " ++ Termops.Internal.print_constr_env env sigma (mkInd ind))
-     in
-     let branch = branches.(0) in
-     let ctx, br = decompose_lam_n_assum sigma (Array.length ps) branch in
-     let n, len, ctx =
-       List.fold_right
-         (fun decl (i, j, ctx) ->
-          match decl with
-          | LocalAssum (na, ty) ->
-             let t = mkProj (Projection.make ps.(i) true, mkRel j) in
-             (i + 1, j + 1, LocalDef (na, t, Vars.liftn 1 j ty) :: ctx)
-          | LocalDef (na, b, ty) ->
-             (i, j + 1, LocalDef (na, Vars.liftn 1 j b, Vars.liftn 1 j ty) :: ctx))
-         ctx (0, 1, [])
-     in
-     mkLetIn (na, c, ty, it_mkLambda_or_LetIn (Vars.liftn 1 (Array.length ps + 1) br) ctx)
+  | None -> mkCase (EConstr.contract_case env sigma (ci, pred, invert, c, branches))
+  | Some ps -> make_project env sigma ind (fst pred) c branches ps
+
+let make_case_or_project env sigma indt ci pred c branches =
+  let open EConstr in
+  let IndType (((ind,_),_),_) = indt in
+  let projs = get_projections env ind in
+  match projs with
+  | None ->
+     let invert = make_case_invert env indt ~case_relevance:(snd pred) ci in
+     mkCase (EConstr.contract_case env sigma (ci, pred, invert, c, branches))
+  | Some ps -> make_project env sigma ind (fst pred) c branches ps
 
 (* substitution in a signature *)
 
@@ -382,38 +477,36 @@ let get_arity env ((ind,u),params) =
        parameters *)
     let nparams = List.length params in
     if Int.equal nparams mib.mind_nparams then
-      mib.mind_params_ctxt
+      Inductive.inductive_paramdecls (mib,u)
     else begin
       assert (Int.equal nparams mib.mind_nparams_rec);
-      let nnonrecparamdecls = mib.mind_nparams - mib.mind_nparams_rec in
-      snd (Termops.context_chop nnonrecparamdecls mib.mind_params_ctxt)
+      snd (Inductive.inductive_nonrec_rec_paramdecls (mib,u))
     end in
-  let parsign = Vars.subst_instance_context u parsign in
   let arproperlength = List.length mip.mind_arity_ctxt - List.length parsign in
   let arsign,_ = List.chop arproperlength mip.mind_arity_ctxt in
-  let subst = subst_of_rel_context_instance parsign params in
+  let subst = subst_of_rel_context_instance_list parsign params in
   let arsign = Vars.subst_instance_context u arsign in
-  (substl_rel_context subst arsign, Inductive.inductive_sort_family mip)
+  substl_rel_context subst arsign
 
 (* Functions to build standard types related to inductive *)
 let build_dependent_constructor cs =
   applist
     (mkConstructU cs.cs_cstr,
      (List.map (lift cs.cs_nargs) cs.cs_params)
-      @(Context.Rel.to_extended_list mkRel 0 cs.cs_args))
+      @(Context.Rel.instance_list mkRel 0 cs.cs_args))
 
 let build_dependent_inductive env ((ind, params) as indf) =
-  let arsign,_ = get_arity env indf in
+  let arsign = get_arity env indf in
   let nrealargs = List.length arsign in
   applist
     (mkIndU ind,
-     (List.map (lift nrealargs) params)@(Context.Rel.to_extended_list mkRel 0 arsign))
+     (List.map (lift nrealargs) params)@(Context.Rel.instance_list mkRel 0 arsign))
 
 (* builds the arity of an elimination predicate in sort [s] *)
 
-let make_arity_signature env sigma dep indf =
-  let (arsign,s) = get_arity env indf in
-  let r = Sorts.relevance_of_sort_family s in
+let make_arity_signature env sigma dep (ind, _ as indf) =
+  let arsign = get_arity env indf in
+  let r = Inductive.relevance_of_inductive env ind in
   let anon = make_annot Anonymous r in
   let arsign = List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) arsign in
   if dep then
@@ -429,16 +522,6 @@ let make_arity env sigma dep indf s =
   let open EConstr in
   it_mkProd_or_LetIn (mkSort s) (make_arity_signature env sigma dep indf)
 
-(* [p] is the predicate and [cs] a constructor summary *)
-let build_branch_type env sigma dep p cs =
-  let base = appvect (lift cs.cs_nargs p, cs.cs_concl_realargs) in
-  if dep then
-    EConstr.Unsafe.to_constr (Namegen.it_mkProd_or_LetIn_name env sigma
-      (EConstr.of_constr (applist (base,[build_dependent_constructor cs])))
-      (List.map (fun d -> Termops.map_rel_decl EConstr.of_constr d) cs.cs_args))
-  else
-    Term.it_mkProd_or_LetIn base cs.cs_args
-
 (**************************************************)
 
 (** From a rel context describing the constructor arguments,
@@ -448,7 +531,7 @@ let build_branch_type env sigma dep p cs =
 let compute_projections env (kn, i as ind) =
   let open Term in
   let mib = Environ.lookup_mind kn env in
-  let u = make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
+  let u = UVars.make_abstract_instance (Declareops.inductive_polymorphic_context mib) in
   let x = match mib.mind_record with
   | NotRecord | FakeRecord ->
     anomaly Pp.(str "Trying to build primitive projections for a non-primitive record")
@@ -458,16 +541,14 @@ let compute_projections env (kn, i as ind) =
   in
   let pkt = mib.mind_packets.(i) in
   let { mind_nparams = nparamargs; mind_params_ctxt = params } = mib in
-  let subst = List.init mib.mind_ntypes (fun i -> mkIndU ((kn, mib.mind_ntypes - i - 1), u)) in
-  let ctx, cty = pkt.mind_nf_lc.(0) in
-  let rctx, _ = decompose_prod_assum (substl subst (Term.it_mkProd_or_LetIn cty ctx)) in
-  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) rctx in
+  let ctx, _ = pkt.mind_nf_lc.(0) in
+  let ctx, paramslet = List.chop pkt.mind_consnrealdecls.(0) ctx in
   (* We build a substitution smashing the lets in the record parameters so
      that typechecking projections requires just a substitution and not
      matching with a parameter context. *)
   let indty =
     (* [ty] = [Ind inst] is typed in context [params] *)
-    let inst = Context.Rel.to_extended_vect mkRel 0 paramslet in
+    let inst = Context.Rel.instance mkRel 0 paramslet in
     let indu = mkIndU (ind, u) in
     let ty = mkApp (indu, inst) in
     (* [Ind inst] is typed in context [params-wo-let] *)
@@ -491,6 +572,7 @@ let compute_projections env (kn, i as ind) =
       match na.binder_name with
       | Name id ->
         let lab = Label.of_id id in
+        let proj_relevant = na.binder_relevance in
         let kn = Projection.Repr.make ind ~proj_npars:mib.mind_nparams ~proj_arg lab in
         (* from [params, field1,..,fieldj |- t(params,field1,..,fieldj)]
            to [params, x:I, field1,..,fieldj |- t(params,field1,..,fieldj] *)
@@ -500,8 +582,8 @@ let compute_projections env (kn, i as ind) =
         (* from [params, x:I, field1,..,fieldj |- t(field1,..,fieldj)]
            to [params, x:I |- t(proj1 x,..,projj x)] *)
         let ty = substl subst t in
-        let term = mkProj (Projection.make kn true, mkRel 1) in
-        let fterm = mkProj (Projection.make kn false, mkRel 1) in
+        let term = mkProj (Projection.make kn true, proj_relevant, mkRel 1) in
+        let fterm = mkProj (Projection.make kn false, proj_relevant, mkRel 1) in
         let etab = it_mkLambda_or_LetIn (mkLambda (x, indty, term)) params in
         let etat = it_mkProd_or_LetIn (mkProd (x, indty, ty)) params in
         let body = (etab, etat) in
@@ -518,13 +600,13 @@ let compute_projections env (kn, i as ind) =
 
 let extract_mrectype sigma t =
   let open EConstr in
-  let (t, l) = decompose_app sigma t in
+  let (t, l) = decompose_app_list sigma t in
   match EConstr.kind sigma t with
     | Ind ind -> (ind, l)
     | _ -> raise Not_found
 
 let find_mrectype_vect env sigma c =
-  let (t, l) = Termops.decompose_app_vect sigma (whd_all env sigma c) in
+  let (t, l) = EConstr.decompose_app sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind -> (ind, l)
     | _ -> raise Not_found
@@ -534,7 +616,7 @@ let find_mrectype env sigma c =
 
 let find_rectype env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind (ind,u) ->
         let (mib,mip) = Inductive.lookup_mind_specif env ind in
@@ -547,7 +629,7 @@ let find_rectype env sigma c =
 
 let find_inductive env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite <> CoFinite ->
@@ -557,7 +639,7 @@ let find_inductive env sigma c =
 
 let find_coinductive env sigma c =
   let open EConstr in
-  let (t, l) = decompose_app sigma (whd_all env sigma c) in
+  let (t, l) = decompose_app_list sigma (whd_all env sigma c) in
   match EConstr.kind sigma t with
     | Ind ind
         when (fst (Inductive.lookup_mind_specif env (fst ind))).mind_finite == CoFinite ->
@@ -566,83 +648,10 @@ let find_coinductive env sigma c =
     | _ -> raise Not_found
 
 
-(***********************************************)
-(* find appropriate names for pattern variables. Useful in the Case
-   and Inversion (case_then_using et case_nodep_then_using) tactics. *)
-
-let is_predicate_explicitly_dep env sigma pred arsign =
-  let rec srec env pval arsign =
-    let pv' = whd_all env sigma pval in
-    match EConstr.kind sigma pv', arsign with
-      | Lambda (na,t,b), (LocalAssum _)::arsign ->
-          srec (push_rel_assum (na, t) env) b arsign
-      | Lambda (na,_,t), _ ->
-
-       (* The following code has an impact on the introduction names
-          given by the tactics "case" and "inversion": when the
-          elimination is not dependent, "case" uses Anonymous for
-          inductive types in Prop and names created by mkProd_name for
-          inductive types in Set/Type while "inversion" uses anonymous
-          for inductive types both in Prop and Set/Type !!
-
-          Previously, whether names were created or not relied on
-          whether the predicate created in Indrec.make_case_com had a
-          dependent arity or not. To avoid different predicates
-          printed the same in v8, all predicates built in indrec.ml
-          got a dependent arity (Aug 2004). The new way to decide
-          whether names have to be created or not is to use an
-          Anonymous or Named variable to enforce the expected
-          dependency status (of course, Anonymous implies non
-          dependent, but not conversely).
-
-          From Coq > 8.2, using or not the effective dependency of
-          the predicate is parametrable! *)
-
-          begin match na.binder_name with
-          | Anonymous -> false
-          | Name _ -> true
-          end
-
-      | _ -> anomaly (Pp.str "Non eta-expanded dep-expanded \"match\" predicate.")
-  in
-  srec env (EConstr.of_constr pred) arsign
-
-let is_elim_predicate_explicitly_dependent env sigma pred indf =
-  let arsign,_ = get_arity env indf in
-  is_predicate_explicitly_dep env sigma pred arsign
-
-let set_names env sigma n brty =
-  let open EConstr in
-  let (ctxt,cl) = decompose_prod_n_assum sigma n brty in
-  Namegen.it_mkProd_or_LetIn_name env sigma cl ctxt
-
-let set_pattern_names env sigma ind brv =
-  let (mib,mip) = Inductive.lookup_mind_specif env ind in
-  let arities =
-    Array.map
-      (fun (d, _) -> List.length d - mib.mind_nparams)
-      mip.mind_nf_lc in
-  Array.map2 (set_names env sigma) arities brv
-
-let type_case_branches_with_names env sigma indspec p c =
-  let (ind,args) = indspec in
-  let args = List.map EConstr.Unsafe.to_constr args in
-  let (mib,mip as specif) = Inductive.lookup_mind_specif env (fst ind) in
-  let nparams = mib.mind_nparams in
-  let (params,realargs) = List.chop nparams args in
-  let lbrty = Inductive.build_branches_type ind specif params p in
-  let lbrty = Array.map EConstr.of_constr lbrty in
-  (* Build case type *)
-  let conclty = lambda_appvect_assum (mip.mind_nrealdecls+1) p (Array.of_list (realargs@[c])) in
-  (* Adjust names *)
-  if is_elim_predicate_explicitly_dependent env sigma p (ind,params) then
-    (set_pattern_names env sigma (fst ind) lbrty, conclty)
-  else (lbrty, conclty)
-
 (* Type of Case predicates *)
 let arity_of_case_predicate env (ind,params) dep k =
-  let arsign,s = get_arity env (ind,params) in
-  let r = Sorts.relevance_of_sort_family s in
+  let arsign = get_arity env (ind,params) in
+  let r = Inductive.relevance_of_inductive env ind in
   let mind = build_dependent_inductive env (ind,params) in
   let concl = if dep then mkArrow mind r (mkSort k) else mkSort k in
   Term.it_mkProd_or_LetIn concl arsign
@@ -651,6 +660,10 @@ let arity_of_case_predicate env (ind,params) dep k =
 (* Inferring the sort of parameters of a polymorphic inductive type
    knowing the sort of the conclusion *)
 
+let univ_level_mem l s = match s with
+| Prop | Set | SProp -> false
+| Type u -> Univ.univ_level_mem l u
+| QSort (_, u) -> assert false (* template cannot contain sort variables *)
 
 (* Compute the inductive argument types: replace the sorts
    that appear in the type of the inductive by the sort of the
@@ -671,9 +684,10 @@ let rec instantiate_universes env evdref scl is = function
         else
           (* unconstrained sort: replace by fresh universe *)
           let evm, s = Evd.new_sort_variable Evd.univ_flexible !evdref in
-          let evm = Evd.set_leq_sort env evm s (Sorts.sort_of_univ u) in
+          let evm = Evd.set_leq_sort env evm s (EConstr.ESorts.make (Sorts.sort_of_univ u)) in
             evdref := evm; s
       in
+      let s = EConstr.ESorts.kind !evdref s in
       (LocalAssum (na,mkArity(ctx,s))) :: instantiate_universes env evdref scl is (sign, exp)
   | sign, [] -> sign (* Uniform parameters are exhausted *)
   | [], _ -> assert false
@@ -687,16 +701,16 @@ let type_of_inductive_knowing_conclusion env sigma ((mib,mip),u) conclty =
     | Some t -> t
     in
     let _,scl = splay_arity env sigma conclty in
-    let scl = EConstr.ESorts.kind sigma scl in
     let ctx = List.rev mip.mind_arity_ctxt in
     let evdref = ref sigma in
     let ctx =
       instantiate_universes
         env evdref scl ar.template_level (ctx,templ.template_param_levels) in
+    let scl = EConstr.ESorts.kind !evdref scl in
       !evdref, EConstr.of_constr (mkArity (List.rev ctx,scl))
 
 let type_of_projection_constant env (p,u) =
-  let pty = lookup_projection p env in
+  let _, pty = lookup_projection p env in
   Vars.subst_instance_constr u pty
 
 let type_of_projection_knowing_arg env sigma p c ty =
@@ -729,6 +743,8 @@ let control_only_guard env sigma c =
   in
   let rec iter env c =
     check_fix_cofix env c;
-    EConstr.iter_with_full_binders sigma EConstr.push_rel iter env c
+    EConstr.iter_with_full_binders env sigma EConstr.push_rel iter env c
   in
-  iter env c
+  try iter env c
+  with Type_errors.TypeError (env, e) ->
+    raise (Pretype_errors.PretypeError (env, sigma, TypingError (Type_errors.map_ptype_error EConstr.of_constr e)))
