@@ -29,19 +29,19 @@ exception Find_at of int
 
 (* timing *)
 
-let get_timing_enabled =
+let { Goptions.get = get_timing_enabled } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["NativeCompute"; "Timing"]
     ~value:false
+    ()
 
 (* profiling *)
 
-let get_profiling_enabled =
+let { Goptions.get = get_profiling_enabled } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["NativeCompute"; "Profiling"]
     ~value:false
+    ()
 
 (* for supported platforms, filename for profiler results *)
 
@@ -55,11 +55,11 @@ let profiler_platform () =
   | "Win32" -> "Windows (Win32)"
   | "Cygwin" -> "Windows (Cygwin)"
 
-let get_profile_filename =
+let { Goptions.get = get_profile_filename } =
   Goptions.declare_string_option_and_ref
-    ~depr:false
     ~key:["NativeCompute"; "Profile"; "Filename"]
     ~value:"native_compute_profile.data"
+    ()
 
 (* find unused profile filename *)
 let get_available_profile_filename () =
@@ -94,11 +94,11 @@ let decompose_prod env t =
   in
   (name,dom,codom)
 
-let e_whd_all = Reductionops.clos_whd_flags CClosure.all
+let e_whd_all = Reductionops.clos_whd_flags RedFlags.all
 
 let app_type env sigma c =
   let t = e_whd_all env sigma c in
-  decompose_appvect (EConstr.Unsafe.to_constr t)
+  decompose_app (EConstr.Unsafe.to_constr t)
 
 let find_rectype_a env sigma c =
   let (t, l) = app_type env sigma c in
@@ -117,8 +117,8 @@ let construct_of_constr_notnative const env tag (ind,u) allargs =
   (mkApp(mkConstructU((ind,i),u), params), ctyp)
 
 let construct_of_constr const env sigma tag typ =
-  let typ = Reductionops.clos_whd_flags CClosure.all env sigma (EConstr.of_constr typ) in
-  let t, l = decompose_appvect (EConstr.Unsafe.to_constr typ) in
+  let typ = Reductionops.clos_whd_flags RedFlags.all env sigma (EConstr.of_constr typ) in
+  let t, l = decompose_app (EConstr.Unsafe.to_constr typ) in
   match Constr.kind t with
   | Ind indu ->
       construct_of_constr_notnative const env tag indu l
@@ -142,7 +142,7 @@ let build_branches_type env sigma mib mip (ind,u) params (pctx, p) =
   let p = it_mkLambda_or_LetIn p pctx in (* TODO: prevent useless cut? *)
   let build_one_branch i (ctx, _) =
     let typi = Inductiveops.instantiate_constructor_params ((ind,i+1),u) (mib,mip) paramsl in
-    let decl,indapp = Reductionops.splay_prod env sigma (EConstr.of_constr typi) in
+    let decl,indapp = Reductionops.whd_decompose_prod env sigma (EConstr.of_constr typi) in
     let decl = List.map (on_snd EConstr.Unsafe.to_constr) decl in
     let ind,cargs = find_rectype_a env sigma indapp in
     let nparams = Array.length params in
@@ -185,8 +185,8 @@ let branch_of_switch lvl ans bs =
   Array.init (Array.length tbl) branch
 
 let get_proj env (ind, proj_arg) =
-  let p = Environ.get_projection env ind ~proj_arg in
-  Projection.make p true
+  let p, r = Environ.get_projection env ind ~proj_arg in
+  Projection.make p true, r
 
 let rec nf_val env sigma v typ =
   match kind_of_value v with
@@ -260,7 +260,7 @@ and nf_args env sigma args t =
     let c = nf_val env sigma arg dom in
     (subst1 c codom, c::l)
   in
-  let t,l = Array.fold_right aux args (t,[]) in
+  let t,l = List.fold_right aux args (t,[]) in
   t, List.rev l
 
 and nf_bargs env sigma b t =
@@ -294,9 +294,11 @@ and nf_atom env sigma atom =
   | Asort s -> mkSort s
   | Avar id -> mkVar id
   | Aproj (p, c) ->
-      let c = nf_accu env sigma c in
-      let p = get_proj env p in
-      mkProj(p, c)
+      let c, cty = nf_accu_type env sigma c in
+      let p, r = get_proj env p in
+      let (_, u), _ = find_rectype_a env sigma (EConstr.of_constr cty) in
+      let r = UVars.subst_instance_relevance u r in
+      mkProj(p, r, c)
   | _ -> fst (nf_atom_type env sigma atom)
 
 and nf_atom_type env sigma atom =
@@ -314,21 +316,18 @@ and nf_atom_type env sigma atom =
       mkVar id, Typeops.type_of_variable env id
   | Acase(ans,accu,p,bs) ->
       let a,ta = nf_accu_type env sigma accu in
-      let ((mind,_),u as ind),allargs = find_rectype_a env sigma (EConstr.of_constr ta) in
-      let (mib,mip) = Inductive.lookup_mind_specif env (fst ind) in
+      let ((mind, _ as ind), u),allargs = find_rectype_a env sigma (EConstr.of_constr ta) in
+      let (mib,mip) = Inductive.lookup_mind_specif env ind in
       let nparams = mib.mind_nparams in
       let params,realargs = Array.chop nparams allargs in
-      let iv = if Typeops.should_invert_case env ans.asw_ci then
-          CaseInvert {indices=realargs}
-        else NoInvert
+      let pctx =
+        let realdecls, _ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
+        let nas = List.rev_map get_annot realdecls @ [nameR (Id.of_string "c")] in
+        expand_arity (mib, mip) (ind, u) params (Array.of_list nas)
       in
-      let nparamdecls = Context.Rel.length (Inductive.inductive_paramdecls (mib,u)) in
-      let pT =
-        hnf_prod_applist_assum env nparamdecls
-          (Inductiveops.type_of_inductive env ind) (Array.to_list params) in
-      let pctx, p = nf_predicate env sigma ind mip params [] p pT in
+      let p, relevance = nf_predicate env sigma ind mip params p pctx in
       (* Calcul du type des branches *)
-      let btypes = build_branches_type env sigma mib mip ind params (pctx, p) in
+      let btypes = build_branches_type env sigma mib mip (ind, u) params (pctx, p) in
       (* calcul des branches *)
       let bsw = branch_of_switch (nb_rel env) ans bs in
       let mkbranch i v =
@@ -339,8 +338,12 @@ and nf_atom_type env sigma atom =
       let branchs = Array.mapi mkbranch bsw in
       let tcase = build_case_type (pctx, p) realargs a in
       let p = (get_case_annot pctx, p) in
-      let ci = ans.asw_ci in
-      mkCase (ci, u, params, p, iv, a, branchs), tcase
+      let ci = Inductiveops.make_case_info env ind RegularStyle in
+      let iv = if Typeops.should_invert_case env relevance ci then
+          CaseInvert {indices=realargs}
+        else NoInvert
+      in
+      mkCase (ci, u, params, (p,relevance), iv, a, branchs), tcase
   | Afix(tt,ft,rp,s) ->
       let tt = Array.map (fun t -> nf_type_sort env sigma t) tt in
       let tt = Array.map fst tt and rt = Array.map snd tt in
@@ -371,45 +374,27 @@ and nf_atom_type env sigma atom =
   | Aproj(p,c) ->
       let c,tc = nf_accu_type env sigma c in
       let cj = make_judge c tc in
-      let p = get_proj env p in
+      let p, _ = get_proj env p in
       let uj = Typeops.judge_of_projection env p cj in
       uj.uj_val, uj.uj_type
 
 
-and nf_predicate env sigma ind mip params pctx v pT =
-  match kind (whd_allnolet env pT) with
-  | LetIn (name,b,t,pT) ->
-    let decl = LocalDef (name, b, t) in
-    nf_predicate (push_rel decl env) sigma ind mip params (decl :: pctx) v pT
-  | Prod (name,dom,codom) -> begin
+and nf_predicate env sigma ind mip params v pctx =
+  let fold decl (k, v) = match decl with
+  | LocalDef _ -> (k + 1, v)
+  | LocalAssum _ ->
     match kind_of_value v with
-    | Vfun f ->
-      let k = nb_rel env in
-      let vb = f (mk_rel_accu k) in
-      let decl = LocalAssum (name, dom) in
-      nf_predicate (push_rel decl env) sigma ind mip params (decl :: pctx) vb codom
-    | _ -> assert false
-    end
-  | _ ->
-    match kind_of_value v with
-    | Vfun f ->
-      let k = nb_rel env in
-      let vb = f (mk_rel_accu k) in
-      let name = Name (Id.of_string "c") in
-      let n = mip.mind_nrealargs in
-      let rargs = Array.init n (fun i -> mkRel (n-i)) in
-      let params = if Int.equal n 0 then params else Array.map (lift n) params in
-      let dom = mkApp(mkIndU ind,Array.append params rargs) in
-      let r = Inductive.relevance_of_inductive env (fst ind) in
-      let name = make_annot name r in
-      let decl = LocalAssum (name, dom) in
-      let env = push_rel decl env in
-      let body = nf_type env sigma vb in
-      decl :: pctx, body
+    | Vfun f -> (k + 1, f (mk_rel_accu k))
     | _ -> assert false
+  in
+  let (_, v) = List.fold_right fold pctx (nb_rel env, v) in
+  let env = push_rel_context pctx env in
+  let body = nf_type env sigma v in
+  let rel = Retyping.relevance_of_type env sigma (EConstr.of_constr body) in
+  body, rel
 
 and nf_evar env sigma evk args =
-  let evi = try Evd.find sigma evk with Not_found -> assert false in
+  let evi = try Evd.find_undefined sigma evk with Not_found -> assert false in
   let hyps = EConstr.named_context_of_val (Evd.evar_filtered_hyps evi) in
   if List.is_empty hyps then begin
     assert (Array.is_empty args);
@@ -424,7 +409,7 @@ and nf_evar env sigma evk args =
     let fold accu d = EConstr.mkNamedProd_or_LetIn sigma d accu in
     let t = List.fold_left fold ty hyps in
     let t = EConstr.to_constr ~abort_on_undefined_evars:false sigma t in
-    let ty, args = nf_args env sigma args t in
+    let ty, args = nf_args env sigma (Array.to_list args) t in
     (* nf_args takes arguments in the reverse order but produces them
        in the correct one, so we have to reverse them again for the
        evar node *)
@@ -528,6 +513,6 @@ let native_norm env sigma c ty =
     EConstr.of_constr res
 
 let native_norm env sigma c ty =
-  if not (Flags.get_native_compiler ()) then
+  if not (Environ.typing_flags env).enable_native_compiler then
     user_err Pp.(str "Native_compute reduction has been disabled.");
   native_norm env sigma c ty
