@@ -59,9 +59,10 @@ type pretype_error =
   | UnexpectedType of constr * constr * unification_error
   | NotProduct of constr
   | TypingError of type_error
+  | CantApplyBadTypeExplained of (constr,types) pcant_apply_bad_type * unification_error
   | CannotUnifyOccurrences of subterm_unification_error
   | UnsatisfiableConstraints of
-    (Evar.t * Evar_kinds.t) option * Evar.Set.t option
+    (Evar.t * Evar_kinds.t) option * Evar.Set.t
   | DisallowedSProp
 
 exception PretypeError of env * Evd.evar_map * pretype_error
@@ -94,10 +95,14 @@ let error_cant_apply_not_functional ?loc env sigma rator randl =
   raise_type_error ?loc
     (env, sigma, CantApplyNonFunctional (rator, randl))
 
-let error_cant_apply_bad_type ?loc env sigma (n,c,t) rator randl =
-  raise_type_error ?loc
-    (env, sigma,
-       CantApplyBadType ((n,c,t), rator, randl))
+let error_cant_apply_bad_type ?loc env sigma ?error (n,c,t) rator randl =
+  let v = ((n,c,t), rator, randl) in
+  match error with
+  | None ->
+    raise_type_error ?loc
+      (env, sigma,
+       CantApplyBadType v)
+  | Some e -> raise_pretype_error ?loc (env,sigma, CantApplyBadTypeExplained (v, e))
 
 let error_ill_formed_branch ?loc env sigma c i actty expty =
   raise_type_error
@@ -114,6 +119,8 @@ let error_ill_typed_rec_body ?loc env sigma i na jl tys =
     (env, sigma, IllTypedRecBody (i, na, jl, tys))
 
 let error_elim_arity ?loc env sigma pi c a =
+  (* XXX type_errors should have a 'sort type parameter *)
+  let a = Option.map (fun (x,s) -> x, EConstr.Unsafe.to_sorts s) a in
   raise_type_error ?loc
     (env, sigma, ElimArity (pi, c, a))
 
@@ -188,7 +195,7 @@ let unsatisfiable_constraints env evd ev comp =
     let err = UnsatisfiableConstraints (None, comp) in
     raise (PretypeError (env,evd,err))
   | Some ev ->
-    let loc, kind = Evd.evar_source (Evd.find evd ev) in
+    let loc, kind = Evd.evar_source (Evd.find_undefined evd ev) in
     let err = UnsatisfiableConstraints (Some (ev, kind), comp) in
     Loc.raise ?loc (PretypeError (env,evd,err))
 
