@@ -93,6 +93,9 @@ let search_guard ?loc env possible_indexes fixdefs =
     indexes
   else
     (* we now search recursively among all combinations *)
+    let combinations = List.combinations possible_indexes in
+    if List.is_empty combinations then
+      user_err ?loc (Pp.str "A fixpoint needs at least one parameter.");
     (try
        List.iter
          (fun l ->
@@ -109,7 +112,7 @@ let search_guard ?loc env possible_indexes fixdefs =
               let env = Environ.set_typing_flags flags env in
               check_fix env fix; raise (Found indexes)
             with TypeError _ -> ())
-         (List.combinations possible_indexes);
+          combinations;
        let errmsg = "Cannot guess decreasing argument of fix." in
          user_err ?loc (Pp.str errmsg)
      with Found indexes -> indexes)
@@ -122,11 +125,11 @@ let esearch_guard ?loc env sigma indexes fix =
 
 (* To force universe name declaration before use *)
 
-let is_strict_universe_declarations =
+let { Goptions.get = is_strict_universe_declarations } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Strict";"Universe";"Declaration"]
     ~value:true
+    ()
 
 (** Miscellaneous interpretation functions *)
 
@@ -149,12 +152,52 @@ let level_name sigma = function
     let sigma, u = universe_level_name sigma l in
     Some (sigma, u)
 
-let sort_info ?loc sigma l = match l with
+let glob_level ?loc evd : glob_level -> _ = function
+  | UAnonymous {rigid} ->
+    assert (rigid <> UnivFlexible true);
+    new_univ_level_variable ?loc rigid evd
+  | UNamed s ->
+    match level_name evd s with
+    | None ->
+      user_err ?loc
+        (str "Universe instances cannot contain non-Set small levels," ++ spc() ++
+         str "polymorphic universe instances must be greater or equal to Set.");
+    | Some r -> r
+
+let glob_qvar ?loc evd : glob_qvar -> _ = function
+  | GQVar q -> evd, q
+  | GLocalQVar {v=Anonymous} ->
+    let evd, q = new_quality_variable ?loc evd in
+    evd, q
+  | GRawQVar q ->
+    let evd = Evd.merge_sort_variables ~sideff:true evd (Sorts.QVar.Set.singleton q) in
+    evd, q
+  | GLocalQVar {v=Name id; loc} ->
+    try evd, (Evd.quality_of_name evd id)
+    with Not_found ->
+      if not (is_strict_universe_declarations()) then
+        let evd, q = new_quality_variable ?loc evd in
+        evd, q
+      else user_err ?loc Pp.(str "Undeclared quality: " ++ Id.print id ++ str".")
+
+let glob_quality ?loc evd = let open Sorts.Quality in function
+  | GQConstant q -> evd, QConstant q
+  | GQualVar (GQVar _ | GLocalQVar _ | GRawQVar _ as q) ->
+    let evd, q = glob_qvar ?loc evd q in
+    evd, QVar q
+
+let sort_info ?loc sigma q l = match l with
 | [] -> assert false
-| [GSProp, 0] -> sigma, Sorts.sprop
-| [GProp, 0] -> sigma, Sorts.prop
+| [GSProp, 0] -> assert (Option.is_empty q); sigma, Sorts.sprop
+| [GProp, 0] -> assert (Option.is_empty q); sigma, Sorts.prop
 | (u, n) :: us ->
   let open Pp in
+  let sigma, q = match q with
+    | None -> sigma, None
+    | Some q ->
+      let sigma, q = glob_qvar ?loc sigma q in
+      sigma, Some q
+  in
   let get_level sigma u n = match level_name sigma u with
   | None ->
     user_err ?loc
@@ -176,7 +219,11 @@ let sort_info ?loc sigma l = match l with
   in
   let (sigma, u) = get_level sigma u n in
   let (sigma, u) = List.fold_left fold (sigma, u) us in
-  sigma, Sorts.sort_of_univ u
+  let s = match q with
+    | None -> Sorts.sort_of_univ u
+    | Some q -> Sorts.qsort q u
+  in
+  sigma, s
 
 type inference_hook = env -> evar_map -> Evar.t -> (evar_map * constr) option
 
@@ -207,34 +254,49 @@ type pretype_flags = {
    [sigma'] into those already in [sigma] or deriving from an evar in
    [sigma] by restriction, and the evars properly created in [sigma'] *)
 
-type frozen =
-| FrozenId of evar_info Evar.Map.t
-  (** No pending evars. We do not put a set here not to reallocate like crazy,
-      but the actual data of the map is not used, only keys matter. All
-      functions operating on this type must have the same behaviour on
-      [FrozenId map] and [FrozenProgress (Evar.Map.domain map, Evar.Set.empty)] *)
-| FrozenProgress of (Evar.Set.t * Evar.Set.t) Lazy.t
-  (** Proper partition of the evar map as described above. *)
+type frozen_and_pending =
+  Frz :
+    'a Evar.Map.t
+    (* Undefined from [sigma']. This is used only as a set,
+       guaranteed by the existential type 'a, but we do not use
+       Evar.Set to avoid reallocating. *)
+    * Evar.Set.t Lazy.t option
+    (* Undefined evars in [sigma'] which are neither in [sigma] or aliases thereof.
+       [None] means empty.*)
+    -> frozen_and_pending
 
 let frozen_and_pending_holes (sigma, sigma') =
   let undefined0 = Option.cata Evd.undefined_map Evar.Map.empty sigma in
-  (* Fast path when the undefined evars where not modified *)
-  if undefined0 == Evd.undefined_map sigma' then
-    FrozenId undefined0
-  else
-    let data = lazy begin
-    let add_derivative_of evk evi acc =
-      match advance sigma' evk with None -> acc | Some evk' -> Evar.Set.add evk' acc in
-    let frozen = Evar.Map.fold add_derivative_of undefined0 Evar.Set.empty in
-    let fold evk _ accu = if not (Evar.Set.mem evk frozen) then Evar.Set.add evk accu else accu in
-    let pending = Evd.fold_undefined fold sigma' Evar.Set.empty in
-    (frozen, pending)
-    end in
-    FrozenProgress data
+  let pending =
+    if undefined0 == Evd.undefined_map sigma'
+    then None
+    else
+      Some (lazy begin
+        let pending, aliases =
+          Evar.Map.symmetric_diff_fold (fun ev v v' (pending,aliases as acc) -> match v, v' with
+              | None, None -> assert false
+              | Some _, None ->
+                (* ev got defined in sigma', but is it an alias? *)
+                begin match advance sigma' ev with
+                | None -> acc
+                | Some ev -> pending, Evar.Set.add ev aliases
+                end
+              | None, Some _ ->
+                (* ev is new in sigma' *)
+                Evar.Set.add ev pending, aliases
+              | Some _, Some _ -> (* ev is still undefined in sigma' *) acc)
+            undefined0
+            (Evd.undefined_map sigma')
+            (Evar.Set.empty, Evar.Set.empty)
+        in
+        Evar.Set.diff pending aliases;
+      end)
+  in
+  Frz (Evd.undefined_map sigma', pending)
 
 let filter_frozen frozen = match frozen with
-  | FrozenId map -> fun evk -> Evar.Map.mem evk map
-  | FrozenProgress (lazy (frozen, _)) -> fun evk -> Evar.Set.mem evk frozen
+  | Frz (undf, None) -> fun evk -> Evar.Map.mem evk undf
+  | Frz (undf, Some (lazy pending)) -> fun evk -> not (Evar.Set.mem evk pending) && Evar.Map.mem evk undf
 
 let typeclasses_filter ~program_mode frozen =
   if program_mode
@@ -244,16 +306,16 @@ let typeclasses_filter ~program_mode frozen =
 let apply_typeclasses ~program_mode ~fail_evar env sigma frozen =
   let sigma = Typeclasses.resolve_typeclasses
       ~filter:(typeclasses_filter ~program_mode frozen)
-      ~split:true ~fail:fail_evar env sigma in
+      ~fail:fail_evar env sigma in
   let sigma = if program_mode then (* Try optionally solving the obligations *)
       Typeclasses.resolve_typeclasses
-        ~filter:(fun evk evi -> Typeclasses.all_evars evk evi && not (filter_frozen frozen evk)) ~split:true ~fail:false env sigma
+        ~filter:(fun evk evi -> Typeclasses.all_evars evk evi && not (filter_frozen frozen evk)) ~fail:false env sigma
     else sigma in
   sigma
 
 let apply_inference_hook (hook : inference_hook) env sigma frozen = match frozen with
-| FrozenId _ -> sigma
-| FrozenProgress (lazy (_, pending)) ->
+| Frz (_, None) -> sigma
+| Frz (_, Some (lazy pending)) ->
   Evar.Set.fold (fun evk sigma ->
     if Evd.is_undefined sigma evk (* in particular not defined by side-effect *)
     then
@@ -276,16 +338,16 @@ let check_typeclasses_instances_are_solved ~program_mode env sigma frozen =
       sigma
   in
   if not (Evar.Set.is_empty tcs) then begin
-    Typeclasses.error_unresolvable env sigma (Some tcs)
+    Typeclasses.error_unresolvable env sigma tcs
   end
 
 let check_extra_evars_are_solved env current_sigma frozen = match frozen with
-| FrozenId _ -> ()
-| FrozenProgress (lazy (_, pending)) ->
+| Frz (_, None) -> ()
+| Frz (_, Some (lazy pending)) ->
   Evar.Set.iter
     (fun evk ->
       if not (Evd.is_defined current_sigma evk) then
-        let (loc,k) = evar_source (Evd.find current_sigma evk) in
+        let (loc,k) = evar_source (Evd.find_undefined current_sigma evk) in
         match k with
         | Evar_kinds.ImplicitArg (gr, (i, id), false) -> ()
         | _ ->
@@ -300,7 +362,8 @@ let check_evars env ?initial sigma c =
       (match initial with
        | Some initial when Evd.mem initial evk -> ()
        | _ ->
-         let (loc,k) = evar_source (Evd.find sigma evk) in
+        let EvarInfo evi = Evd.find sigma evk in
+         let (loc,k) = evar_source evi in
          begin match k with
            | Evar_kinds.ImplicitArg (gr, (i, id), false) -> ()
            | _ -> Pretype_errors.error_unsolvable_implicit ?loc env sigma evk None
@@ -347,20 +410,13 @@ let process_inference_flags flags env initial (sigma,c,cty) =
 let adjust_evar_source sigma na c =
   match na, kind sigma c with
   | Name id, Evar (evk,args) ->
-     let evi = Evd.find sigma evk in
+     let evi = Evd.find_undefined sigma evk in
      begin match Evd.evar_source evi with
-     | loc, Evar_kinds.QuestionMark {
-         Evar_kinds.qm_obligation=b;
-         Evar_kinds.qm_name=Anonymous;
-         Evar_kinds.qm_record_field=recfieldname;
-        } ->
-        let src = (loc,Evar_kinds.QuestionMark {
-            Evar_kinds.qm_obligation=b;
-            Evar_kinds.qm_name=na;
-            Evar_kinds.qm_record_field=recfieldname;
-        }) in
-        let (sigma, evk') = restrict_evar sigma evk (evar_filter evi) ~src None in
-        sigma, mkEvar (evk',args)
+     | loc, Evar_kinds.QuestionMark ({ Evar_kinds.qm_name=Anonymous } as qm) ->
+       let src = (loc,Evar_kinds.QuestionMark { qm with Evar_kinds.qm_name=na }) in
+       (* Evd.update_source doesn't work for some reason, cf test bug_18260_1.v *)
+       let (sigma, evk') = Evd.restrict evk (evar_filter evi) ~src sigma in
+       sigma, mkEvar (evk',args)
      | _ -> sigma, c
      end
   | _, _ -> sigma, c
@@ -405,25 +461,22 @@ let pretype_id pretype loc env sigma id =
 (*************************************************************************)
 (* Main pretyping function                                               *)
 
-let glob_level ?loc evd : glob_level -> _ = function
-  | UAnonymous {rigid} -> new_univ_level_variable ?loc (if rigid then univ_rigid else univ_flexible) evd
-  | UNamed s ->
-    match level_name evd s with
-    | None ->
-      user_err ?loc
-        (str "Universe instances cannot contain non-Set small levels, polymorphic" ++
-        str " universe instances must be greater or equal to Set.");
-    | Some r -> r
-
-let instance ?loc evd l =
-  let evd, l' =
+let instance ?loc evd (ql,ul) =
+  let evd, ql' =
+    List.fold_left
+      (fun (evd, quals) l ->
+         let evd, l = glob_quality ?loc evd l in
+         (evd, l :: quals)) (evd, [])
+      ql
+  in
+  let evd, ul' =
     List.fold_left
       (fun (evd, univs) l ->
          let evd, l = glob_level ?loc evd l in
          (evd, l :: univs)) (evd, [])
-      l
+      ul
   in
-  evd, Some (Univ.Instance.of_array (Array.of_list (List.rev l')))
+  evd, Some (UVars.Instance.of_array (Array.rev_of_list ql', Array.rev_of_list ul'))
 
 let pretype_global ?loc rigid env evd gr us =
   let evd, instance =
@@ -440,12 +493,12 @@ let pretype_ref ?loc sigma env ref us =
     (try
        let ty = NamedDecl.get_type (lookup_named id !!env) in
        (match us with
-        | None | Some [] -> ()
-        | Some us ->
+        | None | Some ([],[]) -> ()
+        | Some (qs,us) ->
             let open UnivGen in
             Loc.raise ?loc (UniverseLengthMismatch {
-              actual = List.length us;
-              expect = 0;
+              actual = List.length qs, List.length us;
+              expect = 0, 0;
             }));
        sigma, make_judge (mkVar id) ty
        with Not_found ->
@@ -460,21 +513,23 @@ let pretype_ref ?loc sigma env ref us =
 
 let sort ?loc evd : glob_sort -> _ = function
   | UAnonymous {rigid} ->
-    let evd, l = new_univ_level_variable ?loc (if rigid then univ_rigid else univ_flexible) evd in
-    evd, Sorts.sort_of_univ (Univ.Universe.make l)
-  | UNamed l -> sort_info ?loc evd l
+    let evd, l = new_univ_level_variable ?loc rigid evd in
+    evd, ESorts.make (Sorts.sort_of_univ (Univ.Universe.make l))
+  | UNamed (q, l) ->
+    let evd, s = sort_info ?loc evd q l in
+    evd, ESorts.make s
 
 let judge_of_sort ?loc evd s =
   let judge =
-    { uj_val = mkSort s; uj_type = mkSort (Sorts.super s) }
+    { uj_val = mkSort s; uj_type = mkSort (ESorts.super evd s) }
   in
     evd, judge
 
 let pretype_sort ?loc sigma s =
   match s with
-  | UNamed [GSProp,0] -> sigma, judge_of_sprop
-  | UNamed [GProp,0] -> sigma, judge_of_prop
-  | UNamed [GSet,0] -> sigma, judge_of_set
+  | UNamed (None, [GSProp, 0]) -> sigma, judge_of_sprop
+  | UNamed (None, [GProp, 0]) -> sigma, judge_of_prop
+  | UNamed (None, [GSet, 0]) -> sigma, judge_of_set
   | _ ->
   let sigma, s = sort ?loc sigma s in
   judge_of_sort ?loc sigma s
@@ -503,12 +558,12 @@ let mark_obligation_evar sigma k evc =
 type 'a pretype_fun = ?loc:Loc.t -> flags:pretype_flags -> type_constraint -> GlobEnv.t -> evar_map -> evar_map * 'a
 
 type pretyper = {
-  pretype_ref : pretyper -> GlobRef.t * glob_level list option -> unsafe_judgment pretype_fun;
+  pretype_ref : pretyper -> GlobRef.t * glob_instance option -> unsafe_judgment pretype_fun;
   pretype_var : pretyper -> Id.t -> unsafe_judgment pretype_fun;
   pretype_evar : pretyper -> existential_name CAst.t * (lident * glob_constr) list -> unsafe_judgment pretype_fun;
   pretype_patvar : pretyper -> Evar_kinds.matching_var_kind -> unsafe_judgment pretype_fun;
   pretype_app : pretyper -> glob_constr * glob_constr list -> unsafe_judgment pretype_fun;
-  pretype_proj : pretyper -> (Constant.t * glob_level list option) * glob_constr list * glob_constr -> unsafe_judgment pretype_fun;
+  pretype_proj : pretyper -> (Constant.t * glob_instance option) * glob_constr list * glob_constr -> unsafe_judgment pretype_fun;
   pretype_lambda : pretyper -> Name.t * binding_kind * glob_constr * glob_constr -> unsafe_judgment pretype_fun;
   pretype_prod : pretyper -> Name.t * binding_kind * glob_constr * glob_constr -> unsafe_judgment pretype_fun;
   pretype_letin : pretyper -> Name.t * glob_constr * glob_constr option * glob_constr -> unsafe_judgment pretype_fun;
@@ -517,11 +572,12 @@ type pretyper = {
   pretype_if : pretyper -> glob_constr * (Name.t * glob_constr option) * glob_constr * glob_constr -> unsafe_judgment pretype_fun;
   pretype_rec : pretyper -> glob_fix_kind * Id.t array * glob_decl list array * glob_constr array * glob_constr array -> unsafe_judgment pretype_fun;
   pretype_sort : pretyper -> glob_sort -> unsafe_judgment pretype_fun;
-  pretype_hole : pretyper -> Evar_kinds.t * Namegen.intro_pattern_naming_expr * Genarg.glob_generic_argument option -> unsafe_judgment pretype_fun;
-  pretype_cast : pretyper -> glob_constr * cast_kind * glob_constr -> unsafe_judgment pretype_fun;
+  pretype_hole : pretyper -> Evar_kinds.glob_evar_kind -> unsafe_judgment pretype_fun;
+  pretype_genarg : pretyper -> Genarg.glob_generic_argument -> unsafe_judgment pretype_fun;
+  pretype_cast : pretyper -> glob_constr * cast_kind option * glob_constr -> unsafe_judgment pretype_fun;
   pretype_int : pretyper -> Uint63.t -> unsafe_judgment pretype_fun;
   pretype_float : pretyper -> Float64.t -> unsafe_judgment pretype_fun;
-  pretype_array : pretyper -> glob_level list option * glob_constr array * glob_constr * glob_constr -> unsafe_judgment pretype_fun;
+  pretype_array : pretyper -> glob_instance option * glob_constr array * glob_constr * glob_constr -> unsafe_judgment pretype_fun;
   pretype_type : pretyper -> glob_constr -> unsafe_type_judgment pretype_fun;
 }
 
@@ -557,8 +613,10 @@ let eval_pretyper self ~flags tycon env sigma t =
     self.pretype_rec self (knd, nas, decl, c, t) ?loc ~flags tycon env sigma
   | GSort s ->
     self.pretype_sort self s ?loc ~flags tycon env sigma
-  | GHole (knd, nam, arg) ->
-    self.pretype_hole self (knd, nam, arg) ?loc ~flags tycon env sigma
+  | GHole knd ->
+    self.pretype_hole self knd ?loc ~flags tycon env sigma
+  | GGenarg arg ->
+    self.pretype_genarg self arg ?loc ~flags tycon env sigma
   | GCast (c, k, t) ->
     self.pretype_cast self (c, k, t) ?loc ~flags tycon env sigma
   | GInt n ->
@@ -650,7 +708,8 @@ struct
       let evk =
         try Evd.evar_key id sigma
         with Not_found -> error_evar_not_found ?loc:locid !!env sigma id in
-      let hyps = evar_filtered_context (Evd.find sigma evk) in
+      let EvarInfo evi = Evd.find sigma evk in
+      let hyps = evar_filtered_context evi in
       let sigma, args = pretype_instance self ~flags env sigma loc hyps evk inst in
       let c = mkLEvar sigma (evk, args) in
       let j = Retyping.get_judgment_of !!env sigma c in
@@ -661,22 +720,21 @@ struct
     let sigma, uj_val, uj_type = new_typed_evar env sigma ~src:(loc,k) tycon in
     sigma, { uj_val; uj_type }
 
-  let pretype_hole self (k, naming, ext) =
-    fun ?loc ~flags tycon env sigma ->
-    match ext with
-    | None ->
-      let open Namegen in
-      let naming = match naming with
-        | IntroIdentifier id -> IntroIdentifier (interp_ltac_id env id)
-        | IntroAnonymous -> IntroAnonymous
-        | IntroFresh id -> IntroFresh (interp_ltac_id env id) in
-      let sigma, uj_val, uj_type = new_typed_evar env sigma ~src:(loc,k) ~naming tycon in
-      let sigma = if flags.program_mode then mark_obligation_evar sigma k uj_val else sigma in
-      sigma, { uj_val; uj_type }
+  let pretype_hole self k ?loc ~flags tycon env sigma =
+    let open Namegen in
+    let naming = naming_of_glob_kind k in
+    let naming = match naming with
+      | IntroIdentifier id -> IntroIdentifier (interp_ltac_id env id)
+      | IntroAnonymous -> IntroAnonymous
+      | IntroFresh id -> IntroFresh (interp_ltac_id env id) in
+    let k = kind_of_glob_kind k in
+    let sigma, uj_val, uj_type = new_typed_evar env sigma ~src:(loc,k) ~naming tycon in
+    let sigma = if flags.program_mode then mark_obligation_evar sigma k uj_val else sigma in
+    sigma, { uj_val; uj_type }
 
-    | Some arg ->
-      let j, sigma = GlobEnv.interp_glob_genarg ?loc ~poly:flags.poly env sigma tycon arg in
-      sigma, j
+  let pretype_genarg self arg ?loc ~flags tycon env sigma =
+    let j, sigma = GlobEnv.interp_glob_genarg ?loc ~poly:flags.poly env sigma tycon arg in
+    sigma, j
 
   let pretype_rec self (fixkind, names, bl, lar, vdef) =
     fun ?loc ~flags tycon env sigma ->
@@ -689,13 +747,13 @@ struct
       | [] -> sigma, ctxt
       | (na,bk,None,ty)::bl ->
         let sigma, ty' = pretype_type empty_valcon env sigma ty in
-        let rty' = Sorts.relevance_of_sort ty'.utj_type in
+        let rty' = ESorts.relevance_of_sort sigma ty'.utj_type in
         let dcl = LocalAssum (make_annot na rty', ty'.utj_val) in
         let dcl', env = push_rel ~hypnaming sigma dcl env in
         type_bl env sigma (Context.Rel.add dcl' ctxt) bl
       | (na,bk,Some bd,ty)::bl ->
         let sigma, ty' = pretype_type empty_valcon env sigma ty in
-        let rty' = Sorts.relevance_of_sort ty'.utj_type in
+        let rty' = ESorts.relevance_of_sort sigma ty'.utj_type in
         let sigma, bd' = pretype (mk_tycon ty'.utj_val) env sigma bd in
         let dcl = LocalDef (make_annot na rty', bd'.uj_val, ty'.utj_val) in
         let dcl', env = push_rel ~hypnaming sigma dcl env in
@@ -735,7 +793,7 @@ struct
            (* we lift nbfix times the type in tycon, because of
             * the nbfix variables pushed to newenv *)
            let (ctxt,ty) =
-             decompose_prod_n_assum sigma (Context.Rel.length ctxt)
+             decompose_prod_n_decls sigma (Context.Rel.length ctxt)
                (lift nbfix ftys.(i)) in
            let ctxt,nenv = push_rel_context ~hypnaming sigma ctxt newenv in
            let sigma, j = pretype (mk_tycon ty) nenv sigma def in
@@ -758,7 +816,7 @@ struct
             Array.to_list (Array.mapi
                              (fun i annot -> match annot with
                              | Some n -> [n]
-                             | None -> List.map_i (fun i _ -> i) 0 ctxtv.(i))
+                             | None -> List.interval 0 (Context.Rel.nhyps ctxtv.(i) - 1))
            vn)
           in
           let fixdecls = (names,ftys,fdefs) in
@@ -817,7 +875,7 @@ struct
             try
               let IndType (indf, args) = find_rectype !!env sigma ty in
               let ((ind',u'),pars) = dest_ind_family indf in
-              if Ind.CanOrd.equal ind ind' then List.map EConstr.of_constr pars
+              if QInd.equal !!env ind ind' then List.map EConstr.of_constr pars
               else (* Let the usual code throw an error *) []
             with Not_found -> []
       else []
@@ -955,8 +1013,8 @@ struct
              are impredicative (typically Prop) we want to keep the
              information when typing the body. *)
           let s = Retyping.get_sort_of !!env sigma ty in
-          if Environ.is_impredicative_sort !!env s
-             || Evd.check_leq sigma Sorts.type1 s
+          if Environ.is_impredicative_sort !!env (ESorts.kind sigma s)
+             || Evd.check_leq sigma ESorts.type1 s
           then
             let sigma, prod = define_evar_as_product !!env sigma ev in
             let na,dom,rng = destProd sigma prod in
@@ -972,14 +1030,14 @@ struct
     in
     let dom_valcon = valcon_of_tycon dom in
     let sigma, j = eval_type_pretyper self ~flags dom_valcon env sigma c1 in
-    let name = {binder_name=name; binder_relevance=Sorts.relevance_of_sort j.utj_type} in
+    let name = {binder_name=name; binder_relevance=ESorts.relevance_of_sort sigma j.utj_type} in
     let var = LocalAssum (name, j.utj_val) in
     let vars = VarSet.variables (Global.env ()) in
     let hypnaming = if flags.program_mode then ProgramNaming vars else RenameExistingBut vars in
     let var',env' = push_rel ~hypnaming sigma var env in
     let sigma, j' = eval_pretyper self ~flags rng env' sigma c2 in
     let name = get_name var' in
-    let resj = judge_of_abstraction !!env (orelse_name name name') j j' in
+    let resj = judge_of_abstraction !!env sigma (orelse_name name name') j j' in
     discard_trace @@ inh_conv_coerce_to_tycon ?loc ~flags env sigma resj tycon
 
   let pretype_prod self (name, bk, c1, c2) =
@@ -994,7 +1052,7 @@ struct
         let sigma, j = pretype_type empty_valcon env sigma c2 in
         sigma, name, { j with utj_val = lift 1 j.utj_val }
       | Name _ ->
-        let r = Sorts.relevance_of_sort j.utj_type in
+        let r = ESorts.relevance_of_sort sigma j.utj_type in
         let var = LocalAssum (make_annot name r, j.utj_val) in
         let var, env' = push_rel ~hypnaming sigma var env in
         let sigma, c2_j = pretype_type empty_valcon env' sigma c2 in
@@ -1002,7 +1060,7 @@ struct
     in
     let resj =
       try
-        judge_of_product !!env name j j'
+        judge_of_product !!env sigma name j j'
       with TypeError _ as e ->
         let (e, info) = Exninfo.capture e in
         let info = Option.cata (Loc.add_loc info) info loc in
@@ -1066,9 +1124,9 @@ struct
           match names, l with
           | na :: names, (LocalAssum (na', t) :: l) ->
             let t = EConstr.of_constr t in
-            let proj = Projection.make ps.(cs.cs_nargs - k) true in
+            let proj = Projection.make (fst ps.(cs.cs_nargs - k)) true in
             LocalDef ({na' with binder_name = na},
-                      lift (cs.cs_nargs - n) (mkProj (proj, cj.uj_val)), t)
+                      lift (cs.cs_nargs - n) (mkProj (proj, na'.binder_relevance, cj.uj_val)), t)
             :: aux (n+1) (k + 1) names l
           | na :: names, (decl :: l) ->
             set_name na decl :: aux (n+1) k names l
@@ -1082,14 +1140,14 @@ struct
     let obj indt rci p v f =
       if not record then
         let f = it_mkLambda_or_LetIn f fsign in
-        let ci = make_case_info !!env (ind_of_ind_type indt) rci LetStyle in
-          mkCase (EConstr.contract_case !!env sigma (ci, p, make_case_invert !!env indt ci, cj.uj_val,[|f|]))
+        let ci = make_case_info !!env (ind_of_ind_type indt) LetStyle in
+          mkCase (EConstr.contract_case !!env sigma (ci, (p,rci), make_case_invert !!env indt ~case_relevance:rci ci, cj.uj_val,[|f|]))
       else it_mkLambda_or_LetIn f fsign
     in
     (* Make dependencies from arity signature impossible *)
     let arsgn, indr =
-      let arsgn,s = get_arity !!env indf in
-      List.map (set_name Anonymous) arsgn, Sorts.relevance_of_sort_family s
+      let arsgn = get_arity !!env indf in
+      List.map (set_name Anonymous) arsgn, Inductiveops.relevance_of_inductive_family !!env indf
     in
       let indt = build_dependent_inductive !!env indf in
       let psign = LocalAssum (make_annot na indr, indt) :: arsgn in (* For locating names in [po] *)
@@ -1154,9 +1212,9 @@ struct
                       (str "If is only for inductive types with two constructors.");
 
       let arsgn, indr =
-        let arsgn,s = get_arity !!env indf in
+        let arsgn = get_arity !!env indf in
         (* Make dependencies from arity signature impossible *)
-        List.map (set_name Anonymous) arsgn, Sorts.relevance_of_sort_family s
+        List.map (set_name Anonymous) arsgn, Inductiveops.relevance_of_inductive_family !!env indf
       in
       let nar = List.length arsgn in
       let indt = build_dependent_inductive !!env indf in
@@ -1199,8 +1257,11 @@ struct
         let ind,_ = dest_ind_family indf in
         let pred = nf_evar sigma pred in
         let rci = Typing.check_allowed_sort !!env sigma ind cj.uj_val pred in
-        let ci = make_case_info !!env (fst ind) rci IfStyle in
-        mkCase (EConstr.contract_case !!env sigma (ci, pred, make_case_invert !!env indty ci, cj.uj_val, [|b1;b2|]))
+        let ci = make_case_info !!env (fst ind) IfStyle in
+        mkCase (EConstr.contract_case !!env sigma
+                  (ci, (pred,rci),
+                   make_case_invert !!env indty ~case_relevance:rci ci, cj.uj_val,
+                   [|b1;b2|]))
       in
       let cj = { uj_val = v; uj_type = p } in
       discard_trace @@ inh_conv_coerce_to_tycon ?loc ~flags env sigma cj tycon
@@ -1214,7 +1275,7 @@ struct
           ~onlyalg:true ~status:Evd.univ_flexible (Some false) !!env sigma tj.utj_val in
       let tval = nf_evar sigma tval in
       let (sigma, cj), tval = match k with
-        | VMcast ->
+        | Some VMcast ->
           let sigma, cj = pretype empty_tycon env sigma c in
           let cty = nf_evar sigma cj.uj_type and tval = nf_evar sigma tval in
           begin match Reductionops.vm_infer_conv !!env sigma cty tval with
@@ -1223,7 +1284,7 @@ struct
               error_actual_type ?loc !!env sigma cj tval
                 (ConversionFailed (!!env,cty,tval))
           end
-        | NATIVEcast ->
+        | Some NATIVEcast ->
           let sigma, cj = pretype empty_tycon env sigma c in
           let cty = nf_evar sigma cj.uj_type and tval = nf_evar sigma tval in
           begin
@@ -1233,24 +1294,29 @@ struct
               error_actual_type ?loc !!env sigma cj tval
                 (ConversionFailed (!!env,cty,tval))
           end
-        | _ ->
+        | None | Some DEFAULTcast ->
           pretype (mk_tycon tval) env sigma c, tval
       in
-      let v = mkCast (cj.uj_val, k, tval) in
+      let v = match k with
+        | None -> cj.uj_val
+        | Some k -> mkCast (cj.uj_val, k, tval)
+      in
       sigma, { uj_val = v; uj_type = tval }
     in discard_trace @@ inh_conv_coerce_to_tycon ?loc ~flags env sigma cj tycon
 
 (* [pretype_type valcon env sigma c] coerces [c] into a type *)
 let pretype_type self c ?loc ~flags valcon (env : GlobEnv.t) sigma = match DAst.get c with
-  | GHole (knd, naming, None) ->
+  | GHole knd ->
       let loc = loc_of_glob_constr c in
+      let naming = naming_of_glob_kind knd in
+      let knd = kind_of_glob_kind knd in
       (match valcon with
        | Some v ->
            let sigma, s =
              let t = Retyping.get_type_of !!env sigma v in
                match EConstr.kind sigma (whd_all !!env sigma t) with
                | Sort s ->
-                 sigma, ESorts.kind sigma s
+                 sigma, s
                | Evar ev when is_Type sigma (existential_type sigma ev) ->
                  define_evar_as_sort !!env sigma ev
                | _ -> anomaly (Pp.str "Found a type constraint which is not a type.")
@@ -1305,14 +1371,14 @@ let pretype_type self c ?loc ~flags valcon (env : GlobEnv.t) sigma = match DAst.
     fun ?loc ~flags tycon env sigma ->
     let sigma, u = match u with
       | None -> sigma, None
-      | Some [u] ->
+      | Some ([],[u]) ->
         let sigma, u = glob_level ?loc sigma u in
         sigma, Some u
-      | Some u ->
+      | Some (qs,us) ->
           let open UnivGen in
           Loc.raise ?loc (UniverseLengthMismatch {
-            actual = List.length u;
-            expect = 1;
+            actual = List.length qs, List.length us;
+            expect = 0, 1;
           })
     in
     let sigma, tycon' = split_as_array !!env sigma tycon in
@@ -1321,11 +1387,11 @@ let pretype_type self c ?loc ~flags valcon (env : GlobEnv.t) sigma = match DAst.
     | Some u -> sigma, u
     | None -> Evd.new_univ_level_variable UState.univ_flexible sigma
     in
-    let sigma = Evd.set_leq_sort !!env sigma jty.utj_type (Sorts.sort_of_univ (Univ.Universe.make u)) in
+    let sigma = Evd.set_leq_sort !!env sigma jty.utj_type (ESorts.make (Sorts.sort_of_univ (Univ.Universe.make u))) in
     let sigma, jdef = eval_pretyper self ~flags (mk_tycon jty.utj_val) env sigma def in
     let pretype_elem = eval_pretyper self ~flags (mk_tycon jty.utj_val) env in
     let sigma, jt = Array.fold_left_map pretype_elem sigma t in
-    let u = Univ.Instance.of_array [| u |] in
+    let u = UVars.Instance.of_array ([||],[| u |]) in
     let ta = EConstr.of_constr @@ Typeops.type_of_array !!env u in
     let j = {
       uj_val = EConstr.mkArray(EInstance.make u, Array.map (fun j -> j.uj_val) jt, jdef.uj_val, jty.utj_val);
@@ -1357,6 +1423,7 @@ let default_pretyper =
     pretype_rec = pretype_rec;
     pretype_sort = pretype_sort;
     pretype_hole = pretype_hole;
+    pretype_genarg = pretype_genarg;
     pretype_cast = pretype_cast;
     pretype_int = pretype_int;
     pretype_float = pretype_float;
@@ -1395,6 +1462,11 @@ let ise_pretype_gen (flags : inference_flags) env sigma lvar kind c =
   in
   process_inference_flags flags !!env sigma (sigma',c',c'_ty)
 
+let ise_pretype_gen flags env sigma lvar kind c : _ * _ * _ =
+  NewProfile.profile "pretyping" (fun () ->
+      ise_pretype_gen flags env sigma lvar kind c)
+    ()
+
 let default_inference_flags fail = {
   use_coercions = true;
   use_typeclasses = UseTC;
@@ -1441,6 +1513,9 @@ let understand_ltac flags env sigma lvar kind c =
   let (sigma, c, _) = ise_pretype_gen flags env sigma lvar kind c in
   (sigma, c)
 
+let understand_ltac_ty flags env sigma lvar kind c =
+  ise_pretype_gen flags env sigma lvar kind c
+
 (* Fully evaluate an untyped constr *)
 let understand_uconstr ?(flags = all_and_fail_flags)
   ?(expected_type = WithoutTypeConstraint) env sigma c =
@@ -1461,7 +1536,7 @@ let path_convertible env sigma cl p q =
   let mkGApp(rt,rtl)      = DAst.make @@ Glob_term.GApp(rt,rtl) in
   let mkGLambda(n,t,b)    = DAst.make @@ Glob_term.GLambda(n,Explicit,t,b) in
   let mkGSort u           = DAst.make @@ Glob_term.GSort u in
-  let mkGHole ()          = DAst.make @@ Glob_term.GHole(Evar_kinds.BinderType Anonymous,Namegen.IntroAnonymous,None) in
+  let mkGHole ()          = DAst.make @@ Glob_term.GHole (GBinderType Anonymous) in
   let path_to_gterm p =
     match p with
     | ic :: p' ->
@@ -1482,7 +1557,7 @@ let path_convertible env sigma cl p q =
       let params = class_nparams cl in
       let clty =
         match cl with
-        | CL_SORT -> mkGSort (Glob_term.UAnonymous {rigid=false})
+        | CL_SORT -> mkGSort (Glob_term.UAnonymous {rigid=UnivFlexible false})
         | CL_FUN -> anomaly (str "A source class must not be Funclass.")
         | CL_SECVAR v -> mkGRef (GlobRef.VarRef v)
         | CL_CONST c -> mkGRef (GlobRef.ConstRef c)
