@@ -33,20 +33,33 @@ let meta_type env evd mv =
   let ty =
     try Evd.meta_ftype evd mv
     with Not_found -> anomaly (str "unknown meta ?" ++ str (Nameops.string_of_meta mv) ++ str ".") in
-  meta_instance env (create_meta_instance_subst evd) ty
+  meta_instance env evd ty
 
-let inductive_type_knowing_parameters env sigma (ind,u) jl =
+let make_param_univs env sigma indu spec jl =
+  Array.to_list @@ Array.mapi (fun i j ~expected ->
+      match ESorts.kind sigma @@ Reductionops.sort_of_arity env sigma j.uj_type with
+      | Sorts.SProp | exception Reduction.NotArity ->
+        let indty = EConstr.of_constr @@
+          Inductive.type_of_inductive (spec, Unsafe.to_instance @@ snd indu)
+        in
+        error_cant_apply_bad_type env sigma
+          (i+1, mkType (Univ.Universe.make expected), j.uj_type)
+          (make_judge (mkIndU indu) indty)
+          jl
+      | Sorts.Prop -> TemplateProp
+      | Sorts.Set -> TemplateUniv Univ.Universe.type0
+      | Sorts.Type u | Sorts.QSort (_, u) -> TemplateUniv u)
+    jl
+
+let inductive_type_knowing_parameters env sigma (ind,u as indu) jl =
   let u = Unsafe.to_instance u in
   let mspec = lookup_mind_specif env ind in
-  let paramstyp = Array.map_to_list (fun j () ->
-      let s = Reductionops.sort_of_arity env sigma j.uj_type in
-      EConstr.ESorts.kind sigma s) jl
-  in
+  let paramstyp = make_param_univs env sigma indu mspec jl in
   Inductive.type_of_inductive_knowing_parameters (mspec,u) paramstyp
 
 let type_judgment env sigma j =
   match EConstr.kind sigma (whd_all env sigma j.uj_type) with
-    | Sort s -> sigma, {utj_val = j.uj_val; utj_type = ESorts.kind sigma s }
+    | Sort s -> sigma, {utj_val = j.uj_val; utj_type = s }
     | Evar ev ->
         let (sigma,s) = Evardefine.define_evar_as_sort env sigma ev in
         sigma, { utj_val = j.uj_val; utj_type = s }
@@ -87,8 +100,8 @@ let judge_of_apply env sigma funj argjv =
     match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
     | sigma ->
       apply_rec sigma (n+1) subs c2 restjl
-    | exception Evarconv.UnableToUnify _ ->
-      error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj argjv
+    | exception Evarconv.UnableToUnify (sigma,error) ->
+      error_cant_apply_bad_type env sigma ~error (n, c1, hj.uj_type) funj argjv
   in
   apply_rec sigma 1 (Esubst.subs_id 0) funj.uj_type (Array.to_list argjv)
 
@@ -136,8 +149,13 @@ let is_correct_arity env sigma c pj ind specif params =
           srec (push_rel (LocalAssum (na1,a1)) env) sigma t ar'
       end
     | Sort s, [] ->
-        let s = ESorts.kind sigma s in
-        if not (List.mem_f Sorts.family_equal (Sorts.family s) allowed_sorts)
+        let sigma = match ESorts.kind sigma s with
+        | QSort (_, u) ->
+          (* Arbitrarily set the return sort to Type *)
+          Evd.set_eq_sort env sigma s (ESorts.make (Sorts.sort_of_univ u))
+        | Set | Type _ | Prop | SProp -> sigma
+        in
+        if not (List.mem_f Sorts.family_equal (ESorts.family sigma s) allowed_sorts)
         then error ()
         else sigma, s
     | Evar (ev,_), [] ->
@@ -151,7 +169,7 @@ let is_correct_arity env sigma c pj ind specif params =
   in
   srec env sigma pj.uj_type (List.rev arsign)
 
-let lambda_applist_assum sigma n c l =
+let lambda_applist_decls sigma n c l =
   let rec app n subst t l =
     if Int.equal n 0 then
       if l == [] then substl subst t
@@ -162,8 +180,7 @@ let lambda_applist_assum sigma n c l =
     | _ -> anomaly (Pp.str "Not enough lambda/let's.") in
   app n [] c l
 
-let type_case_branches env sigma (ind,largs) pj c =
-  let specif = lookup_mind_specif env (fst ind) in
+let type_case_branches env sigma (ind,largs) specif pj c =
   let nparams = inductive_params specif in
   let (params,realargs) = List.chop nparams largs in
   let p = pj.uj_val in
@@ -172,18 +189,55 @@ let type_case_branches env sigma (ind,largs) pj c =
   let lc = build_branches_type ind specif params (EConstr.to_constr ~abort_on_undefined_evars:false sigma p) in
   let lc = Array.map EConstr.of_constr lc in
   let n = (snd specif).Declarations.mind_nrealdecls in
-  let ty = whd_betaiota env sigma (lambda_applist_assum sigma (n+1) p (realargs@[c])) in
-  sigma, (lc, ty, Sorts.relevance_of_sort ps)
+  let ty = whd_betaiota env sigma (lambda_applist_decls sigma (n+1) p (realargs@[c])) in
+  sigma, (lc, ty, ESorts.relevance_of_sort sigma ps)
 
-let judge_of_case env sigma case ci pj iv cj lfj =
+let unify_relevance sigma r1 r2 =
+  match Evarutil.nf_relevance sigma r1, Evarutil.nf_relevance sigma r2 with
+  | Relevant, Relevant | Irrelevant, Irrelevant -> Some sigma
+  | Relevant, Irrelevant | Irrelevant, Relevant -> None
+  | Irrelevant, RelevanceVar q | RelevanceVar q, Irrelevant ->
+    let sigma =
+      Evd.add_quconstraints sigma
+        (Sorts.QConstraints.singleton (Sorts.Quality.qsprop, Equal, QVar q),
+         Univ.Constraints.empty)
+    in
+    Some sigma
+  | Relevant, RelevanceVar q | RelevanceVar q, Relevant ->
+    let sigma =
+      Evd.add_quconstraints sigma
+        (Sorts.QConstraints.singleton (Sorts.Quality.qprop, Leq, QVar q),
+         Univ.Constraints.empty)
+    in
+    Some sigma
+  | RelevanceVar q1, RelevanceVar q2 ->
+    let sigma =
+      Evd.add_quconstraints sigma
+        (Sorts.QConstraints.singleton (QVar q1, Equal, QVar q2),
+         Univ.Constraints.empty)
+    in
+    Some sigma
+
+let judge_of_case env sigma case ci (pj,rp) iv cj lfj =
   let ((ind, u), spec) =
     try find_mrectype env sigma cj.uj_type
     with Not_found -> error_case_not_inductive env sigma cj in
+  let specif = lookup_mind_specif env ind in
+  let () = if Inductive.is_private specif then Type_errors.error_case_on_private_ind env ind in
   let indspec = ((ind, EInstance.kind sigma u), spec) in
-  let sigma, (bty,rslty,rci) = type_case_branches env sigma indspec pj cj.uj_val in
-  let () = check_case_info env (fst indspec) rci ci in
+  let sigma, (bty,rslty,rci) = type_case_branches env sigma indspec specif pj cj.uj_val in
+  (* should we have evar map aware should_invert_case? *)
+  let sigma, rp =
+    if Sorts.relevance_equal rp rci then sigma, rp
+    else match unify_relevance sigma rp rci with
+    | None ->
+      raise_type_error (env,sigma,Type_errors.BadCaseRelevance (rp, mkCase case))
+    | Some sigma -> sigma, rci
+  in
+  let () = check_case_info env (fst indspec) ci in
   let sigma = check_branch_types env sigma (fst indspec) cj (lfj,bty) in
-  let () = if (match iv with | NoInvert -> false | CaseInvert _ -> true) != should_invert_case env ci
+  let () = if (match iv with | NoInvert -> false | CaseInvert _ -> true)
+              != should_invert_case env rp ci
     then Type_errors.error_bad_invert env
   in
   sigma, { uj_val  = mkCase case;
@@ -204,17 +258,16 @@ let check_type_fixpoint ?loc env sigma lna lar vdefj =
 (* FIXME: might depend on the level of actual parameters!*)
 let check_allowed_sort env sigma ind c p =
   let specif = lookup_mind_specif env (fst ind) in
-  let sorts = elim_sort specif in
   let pj = Retyping.get_judgment_of env sigma p in
-  let _, s = splay_prod env sigma pj.uj_type in
-  let ksort = match EConstr.kind sigma s with
-  | Sort s -> Sorts.family (ESorts.kind sigma s)
-  | _ -> error_elim_arity env sigma ind c None in
-  if not (Sorts.family_leq ksort sorts) then
-    let s = inductive_sort_family (snd specif) in
-    error_elim_arity env sigma ind c (Some (pj, sorts, ksort, s))
+  let _, s = whd_decompose_prod env sigma pj.uj_type in
+  let sort =  match EConstr.kind sigma s with
+    | Sort s -> s
+    | _ -> error_elim_arity env sigma ind c None
+  in
+  if Inductiveops.is_allowed_elimination sigma (specif,(snd ind)) sort then
+    ESorts.relevance_of_sort sigma sort
   else
-    Sorts.relevance_of_sort_family ksort
+    error_elim_arity env sigma ind c (Some (pj, sort))
 
 let check_actual_type env sigma cj t =
   try Evarconv.unify_leq_delay env sigma cj.uj_type t
@@ -244,17 +297,25 @@ let judge_of_sprop =
 
 let judge_of_prop =
   { uj_val = EConstr.mkProp;
-    uj_type = EConstr.mkSort Sorts.type1 }
+    uj_type = EConstr.mkSort (ESorts.type1) }
 
 let judge_of_set =
   { uj_val = EConstr.mkSet;
-    uj_type = EConstr.mkSort Sorts.type1 }
+    uj_type = EConstr.mkSort (ESorts.type1) }
 
 let judge_of_type u =
   let uu = Univ.Universe.super u in
     { uj_val = EConstr.mkType u;
       uj_type = EConstr.mkType uu }
 
+let judge_of_sort s =
+  let open Sorts in
+  let u = match s with
+  | Prop | SProp | Set -> Univ.Universe.type1
+  | Type u | QSort (_, u) -> Univ.Universe.super u
+  in
+  { uj_val = EConstr.mkSort (ESorts.make s); uj_type = EConstr.mkType u }
+
 let judge_of_relative env v =
   Environ.on_judgment EConstr.of_constr (judge_of_relative env v)
 
@@ -265,42 +326,40 @@ let judge_of_variable env id =
   Environ.on_judgment EConstr.of_constr (judge_of_variable env id)
 
 let judge_of_projection env sigma p cj =
-  let pty = lookup_projection p env in
+  let pr, pty = lookup_projection p env in
   let (ind,u), args =
     try find_mrectype env sigma cj.uj_type
     with Not_found -> error_case_not_inductive env sigma cj
   in
   let u = EInstance.kind sigma u in
+  let pr = UVars.subst_instance_relevance u pr in
   let ty = EConstr.of_constr (CVars.subst_instance_constr u pty) in
   let ty = substl (cj.uj_val :: List.rev args) ty in
-  {uj_val = EConstr.mkProj (p,cj.uj_val);
+  {uj_val = EConstr.mkProj (p,pr,cj.uj_val);
    uj_type = ty}
 
-let judge_of_abstraction env name var j =
-  let r = Sorts.relevance_of_sort var.utj_type in
+let judge_of_abstraction env sigma name var j =
+  let r = ESorts.relevance_of_sort sigma var.utj_type in
   { uj_val = mkLambda (make_annot name r, var.utj_val, j.uj_val);
     uj_type = mkProd (make_annot name r, var.utj_val, j.uj_type) }
 
-let judge_of_product env name t1 t2 =
-  let r = Sorts.relevance_of_sort t1.utj_type in
-  let s = sort_of_product env t1.utj_type t2.utj_type in
+let judge_of_product env sigma name t1 t2 =
+  let s1 = ESorts.kind sigma t1.utj_type in
+  let s2 = ESorts.kind sigma t2.utj_type in
+  let r = Sorts.relevance_of_sort s1 in
+  let s = ESorts.make (sort_of_product env s1 s2) in
   { uj_val = mkProd (make_annot name r, t1.utj_val, t2.utj_val);
     uj_type = mkSort s }
 
-let judge_of_letin env name defj typj j =
-  let r = Sorts.relevance_of_sort typj.utj_type in
+let judge_of_letin env sigma name defj typj j =
+  let r = ESorts.relevance_of_sort sigma typj.utj_type in
   { uj_val = mkLetIn (make_annot name r, defj.uj_val, typj.utj_val, j.uj_val) ;
     uj_type = subst1 defj.uj_val j.uj_type }
 
-let check_hyps_inclusion env sigma x hyps =
-  let env = Environ.set_universes (Evd.universes sigma) env in
-  let evars = Evd.evar_handler sigma in
-  Typeops.check_hyps_inclusion env ~evars x hyps
-
 let type_of_constant env sigma (c,u) =
   let open Declarations in
   let cb = Environ.lookup_constant c env in
-  let () = check_hyps_inclusion env sigma (GR.ConstRef c) cb.const_hyps in
+  let () = Reductionops.check_hyps_inclusion env sigma (GR.ConstRef c) cb.const_hyps in
   let u = EInstance.kind sigma u in
   let ty, csts = Environ.constant_type env (c,u) in
   let sigma = Evd.add_constraints sigma csts in
@@ -309,7 +368,7 @@ let type_of_constant env sigma (c,u) =
 let type_of_inductive env sigma (ind,u) =
   let open Declarations in
   let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
-  let () = check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
+  let () = Reductionops.check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
   let u = EInstance.kind sigma u in
   let ty, csts = Inductive.constrained_type_of_inductive (specif,u) in
   let sigma = Evd.add_constraints sigma csts in
@@ -318,7 +377,7 @@ let type_of_inductive env sigma (ind,u) =
 let type_of_constructor env sigma ((ind,_ as ctor),u) =
   let open Declarations in
   let (mib,_ as specif) = Inductive.lookup_mind_specif env ind in
-  let () = check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
+  let () = Reductionops.check_hyps_inclusion env sigma (GR.IndRef ind) mib.mind_hyps in
   let u = EInstance.kind sigma u in
   let ty, csts = Inductive.constrained_type_of_constructor (ctor,u) specif in
   let sigma = Evd.add_constraints sigma csts in
@@ -331,12 +390,12 @@ let judge_of_float env v =
   Environ.on_judgment EConstr.of_constr (judge_of_float env v)
 
 let judge_of_array env sigma u tj defj tyj =
-  let ulev = match Univ.Instance.to_array u with
-    | [|u|] -> u
+  let ulev = match UVars.Instance.to_array u with
+    | [||], [|u|] -> u
     | _ -> assert false
   in
   let sigma = Evd.set_leq_sort env sigma tyj.utj_type
-      (Sorts.sort_of_univ (Univ.Universe.make ulev))
+      (ESorts.make (Sorts.sort_of_univ (Univ.Universe.make ulev)))
   in
   let check_one sigma j = check_actual_type env sigma j tyj.utj_val in
   let sigma = check_one sigma defj in
@@ -347,6 +406,53 @@ let judge_of_array env sigma u tj defj tyj =
   in
   sigma, j
 
+let bad_relevance_msg = CWarnings.create_msg Typeops.bad_relevance_warning ()
+(* no need for default printer, pretyping is always linked with himsg in practice *)
+
+let warn_bad_relevance_binder ?loc env sigma rlv bnd =
+  match CWarnings.warning_status Typeops.bad_relevance_warning with
+| CWarnings.Disabled | CWarnings.Enabled ->
+  CWarnings.warn bad_relevance_msg ?loc (env,sigma,Typeops.BadRelevanceBinder (rlv, bnd))
+| CWarnings.AsError ->
+  CErrors.anomaly (CErrors.print (PretypeError (env, sigma, TypingError (Type_errors.BadBinderRelevance (rlv, bnd)))))
+
+type relevance_preunify =
+  | Trivial
+  | Impossible
+  | DummySort of ESorts.t
+
+let check_binder_relevance env sigma s decl =
+  let preunify = match ESorts.kind sigma s, Evarutil.nf_relevance sigma (get_relevance decl) with
+    | (Prop | Set | Type _), Relevant -> Trivial
+    | (Prop | Set | Type _), Irrelevant -> Impossible
+    | SProp, Irrelevant -> Trivial
+    | SProp, Relevant -> Impossible
+    | QSort (_,l), RelevanceVar q' -> DummySort (ESorts.make (Sorts.qsort q' l))
+    | (SProp | Prop | Set), RelevanceVar q ->
+      DummySort (ESorts.make (Sorts.qsort q Univ.Universe.type0))
+    | Type l, RelevanceVar q -> DummySort (ESorts.make (Sorts.qsort q l))
+    | QSort (_,l), Relevant -> DummySort (ESorts.make (Sorts.sort_of_univ l))
+    | QSort _, Irrelevant -> DummySort ESorts.sprop
+  in
+  let unify = match preunify with
+    | Trivial -> Some sigma
+    | Impossible -> None
+    | DummySort s' ->
+      match Evd.set_leq_sort env sigma s s' with
+      | sigma -> Some sigma
+      | exception UGraph.UniverseInconsistency _ -> None
+  in
+  match unify with
+  | Some sigma -> sigma, decl
+  | None ->
+    (* TODO always anomaly *)
+    let rs = ESorts.relevance_of_sort sigma s in
+    let () =
+      if not (UGraph.type_in_type (Evd.universes sigma))
+      then warn_bad_relevance_binder env sigma rs decl
+    in
+    sigma, set_annot { (get_annot decl) with binder_relevance = rs } decl
+
 (* cstr must be in n.f. w.r.t. evars and execute returns a judgement
    where both the term and type are in n.f. *)
 let rec execute env sigma cstr =
@@ -381,7 +487,7 @@ let rec execute env sigma cstr =
 
     | Case (ci, u, pms, p, iv, c, lf) ->
         let case = (ci, u, pms, p, iv, c, lf) in
-        let (ci, p, iv, c, lf) = EConstr.expand_case env sigma case in
+        let (ci, (p,rp), iv, c, lf) = EConstr.expand_case env sigma case in
         let sigma, cj = execute env sigma c in
         let sigma, pj = execute env sigma p in
         let sigma, lfj = execute_array env sigma lf in
@@ -395,7 +501,7 @@ let rec execute env sigma cstr =
             let sigma = check_actual_type env sigma cj tj.utj_val in
             sigma
         in
-        judge_of_case env sigma case ci pj iv cj lfj
+        judge_of_case env sigma case ci (pj,rp) iv cj lfj
 
     | Fix ((vn,i as vni),recdef) ->
         let sigma, (_,tys,_ as recdef') = execute_recdef env sigma recdef in
@@ -417,9 +523,10 @@ let rec execute env sigma cstr =
         | Prop -> sigma, judge_of_prop
         | Set -> sigma, judge_of_set
         | Type u -> sigma, judge_of_type u
+        | QSort _ as s -> sigma, judge_of_sort s
       end
 
-    | Proj (p, c) ->
+    | Proj (p, _, c) ->
       let sigma, cj = execute env sigma c in
       sigma, judge_of_projection env sigma p cj
 
@@ -437,29 +544,29 @@ let rec execute env sigma cstr =
     | Lambda (name,c1,c2) ->
         let sigma, j = execute env sigma c1 in
         let sigma, var = type_judgment env sigma j in
-        let name = check_binder_annot var.utj_type name in
-        let env1 = push_rel (LocalAssum (name, var.utj_val)) env in
+        let sigma, decl = check_binder_relevance env sigma var.utj_type (LocalAssum (name, var.utj_val)) in
+        let env1 = push_rel decl env in
         let sigma, j' = execute env1 sigma c2 in
-        sigma, judge_of_abstraction env1 name.binder_name var j'
+        sigma, judge_of_abstraction env1 sigma name.binder_name var j'
 
     | Prod (name,c1,c2) ->
         let sigma, j = execute env sigma c1 in
         let sigma, varj = type_judgment env sigma j in
-        let name = check_binder_annot varj.utj_type name in
-        let env1 = push_rel (LocalAssum (name, varj.utj_val)) env in
+        let sigma, decl = check_binder_relevance env sigma varj.utj_type (LocalAssum (name, varj.utj_val)) in
+        let env1 = push_rel decl env in
         let sigma, j' = execute env1 sigma c2 in
         let sigma, varj' = type_judgment env1 sigma j' in
-        sigma, judge_of_product env name.binder_name varj varj'
+        sigma, judge_of_product env sigma name.binder_name varj varj'
 
      | LetIn (name,c1,c2,c3) ->
         let sigma, j1 = execute env sigma c1 in
         let sigma, j2 = execute env sigma c2 in
         let sigma, j2 = type_judgment env sigma j2 in
         let sigma, _ =  judge_of_cast env sigma j1 DEFAULTcast j2 in
-        let name = check_binder_annot j2.utj_type name in
-        let env1 = push_rel (LocalDef (name, j1.uj_val, j2.utj_val)) env in
+        let sigma, decl = check_binder_relevance env sigma j2.utj_type (LocalDef (name, j1.uj_val, j2.utj_val)) in
+        let env1 = push_rel decl env in
         let sigma, j3 = execute env1 sigma c3 in
-        sigma, judge_of_letin env name.binder_name j1 j2 j3
+        sigma, judge_of_letin env sigma name.binder_name j1 j2 j3
 
     | Cast (c,k,t) ->
         let sigma, cj = execute env sigma c in
@@ -568,15 +675,15 @@ let judge_of_apply_against env sigma changedf funj argjv =
         match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
         | sigma ->
           apply_rec sigma (Changed {bodyonly=lazy false}) (n+1) subs c2 restjl
-        | exception Evarconv.UnableToUnify _ ->
-          error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj (Array.map snd argjv)
+        | exception Evarconv.UnableToUnify (sigma,error) ->
+          error_cant_apply_bad_type env sigma ~error (n, c1, hj.uj_type) funj (Array.map snd argjv)
     end
     else
       match Evarconv.unify_leq_delay env sigma hj.uj_type c1 with
       | sigma ->
         apply_rec sigma changedf (n+1) subs c2 restjl
-      | exception Evarconv.UnableToUnify _ ->
-        error_cant_apply_bad_type env sigma (n, c1, hj.uj_type) funj (Array.map snd argjv)
+      | exception Evarconv.UnableToUnify (sigma,error) ->
+        error_cant_apply_bad_type env sigma ~error (n, c1, hj.uj_type) funj (Array.map snd argjv)
   in
   apply_rec sigma changedf 1 (Esubst.subs_id 0) funj.uj_type (Array.to_list argjv)
 
@@ -607,9 +714,9 @@ let rec recheck_against env sigma good c =
 
     | Case (gci, gu, gpms, gp, giv, gc, glf),
       Case (ci, u, pms, p, iv, c, lf) ->
-      let (gci, gp, giv, gc, glf) = EConstr.expand_case env sigma (gci, gu, gpms, gp, giv, gc, glf) in
+      let (gci, (gp,_), giv, gc, glf) = EConstr.expand_case env sigma (gci, gu, gpms, gp, giv, gc, glf) in
       let case = (ci, u, pms, p, iv, c, lf) in
-      let (ci, p, iv, c, lf) = EConstr.expand_case env sigma case in
+      let (ci, (p,rp), iv, c, lf) = EConstr.expand_case env sigma case in
       let sigma, changedc, cj = recheck_against env sigma gc c in
       let sigma, changedp, pj = recheck_against env sigma gp p in
       let (sigma, changedlf), lfj =
@@ -647,10 +754,10 @@ let rec recheck_against env sigma good c =
       in
       if unchanged changedc && unchanged changedp && unchanged changediv && bodyonly changedlf
       then assume_unchanged_type sigma
-      else maybe_changed (judge_of_case env sigma case ci pj iv cj lfj)
+      else maybe_changed (judge_of_case env sigma case ci (pj,rp) iv cj lfj)
 
-    | Proj (gp, gc),
-      Proj (p, c) ->
+    | Proj (gp, _, gc),
+      Proj (p, _, c) ->
       if not (QProjection.equal env gp p) then default ()
       else
         let sigma, changed, c = recheck_against env sigma gc c in
@@ -692,26 +799,26 @@ let rec recheck_against env sigma good c =
       Lambda (name, c1, c2) ->
       let sigma, changedj, j = recheck_against env sigma gc1 c1 in
       let sigma, var = type_judgment env sigma j in
-      let name = check_binder_annot var.utj_type name in
-      let env1 = push_rel (LocalAssum (name, var.utj_val)) env in
+      let sigma, decl = check_binder_relevance env sigma var.utj_type (LocalAssum (name, var.utj_val)) in
+      let env1 = push_rel decl env in
       let sigma, changedj', j' = if unchanged changedj then recheck_against env1 sigma gc2 c2
         else let sigma, j' = execute env1 sigma c2 in
           sigma, Changed {bodyonly=lazy false}, j'
       in
-      sigma, merge_changes changedj' changedj, judge_of_abstraction env1 name.binder_name var j'
+      sigma, merge_changes changedj' changedj, judge_of_abstraction env1 sigma name.binder_name var j'
 
     | Prod (_, gc1, gc2),
       Prod (name, c1, c2) ->
       let sigma, changedj, j = recheck_against env sigma gc1 c1 in
       let sigma, var = type_judgment env sigma j in
-      let name = check_binder_annot var.utj_type name in
-      let env1 = push_rel (LocalAssum (name, var.utj_val)) env in
+      let sigma, decl = check_binder_relevance env sigma var.utj_type (LocalAssum (name, var.utj_val)) in
+      let env1 = push_rel decl env in
       let sigma, changedj', j' = if unchanged changedj then recheck_against env1 sigma gc2 c2
         else let sigma, j' = execute env1 sigma c2 in
           sigma, Changed {bodyonly=lazy false}, j'
       in
       let sigma, j' = type_judgment env1 sigma j' in
-      sigma, merge_changes changedj' changedj, judge_of_product env1 name.binder_name var j'
+      sigma, merge_changes changedj' changedj, judge_of_product env1 sigma name.binder_name var j'
 
     | Cast (gc, _, gt),
       Cast (c, k, t) ->
