@@ -20,7 +20,7 @@ open EConstr
 open Vars
 open Namegen
 open Evd
-open Reduction
+open Conversion
 open Reductionops
 open Structures
 open Evarutil
@@ -39,29 +39,16 @@ type subst0 =
     metabinding list *
       ((Environ.env * int) * EConstr.existential * EConstr.t) list)
 
-module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-let is_keyed_unification =
+let { Goptions.get = is_keyed_unification } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Keyed";"Unification"]
     ~value:false
+    ()
 
 let debug_tactic_unification = CDebug.create ~name:"tactic-unification" ()
 
-(** Making this unification algorithm correct w.r.t. the evar-map abstraction
-    breaks too much stuff. So we redefine incorrect functions here. *)
-
-let unsafe_occur_meta_or_existential c =
-  let c = EConstr.Unsafe.to_constr c in
-  let rec occrec c = match Constr.kind c with
-    | Evar _ -> raise Occur
-    | Meta _ -> raise Occur
-    | _ -> Constr.iter occrec c
-  in try occrec c; false with Occur -> true
-
-
 let occur_meta_or_undefined_evar evd c =
   (* This is performance-critical. Using the evar-insensitive API changes the
      resulting heuristic. *)
@@ -69,7 +56,8 @@ let occur_meta_or_undefined_evar evd c =
   let rec occrec c = match Constr.kind c with
     | Meta _ -> raise Occur
     | Evar (ev,args) ->
-        (match evar_body (Evd.find evd ev) with
+      let EvarInfo evi = Evd.find evd ev in
+        (match evar_body evi with
         | Evar_defined c ->
             occrec (EConstr.Unsafe.to_constr c); SList.Skip.iter occrec args
         | Evar_empty -> raise Occur)
@@ -100,9 +88,7 @@ let abstract_scheme env evd c l lname_typ =
     mkLambda (map_annot (named_hd env evd a) n, a, b)
   in
   List.fold_left2
-    (fun (t,evd) (locc,a) decl ->
-       let na = RelDecl.get_annot decl in
-       let ta = RelDecl.get_type decl in
+    (fun (t,evd) (locc,a) (na,ta) ->
        let na = match EConstr.kind evd a with Var id -> {na with binder_name=Name id} | _ -> na in
 (* [occur_meta ta] test removed for support of eelim/ecase but consequences
    are unclear...
@@ -119,12 +105,12 @@ let abstract_scheme env evd c l lname_typ =
 (* Precondition: resulting abstraction is expected to be of type [typ] *)
 
 let abstract_list_all env evd typ c l =
-  let ctxt,_ = splay_prod_n env evd (List.length l) typ in
+  let ctxt,_ = whd_decompose_prod_n env evd (List.length l) typ in
   let l_with_all_occs = List.map (function a -> (LikeFirst,a)) l in
   let p,evd = abstract_scheme env evd c l_with_all_occs ctxt in
   let evd,typp =
     try
-      let c = EConstr.it_mkLambda_or_LetIn (EConstr.Vars.lift (List.length ctxt) c) ctxt in
+      let c = EConstr.it_mkLambda (EConstr.Vars.lift (List.length ctxt) c) ctxt in
       Typing.recheck_against env evd c p
     with
     | UserError _ ->
@@ -476,23 +462,23 @@ let use_metas_pattern_unification sigma flags nb l =
 
 type key =
   | IsKey of CClosure.table_key
-  | IsProj of Projection.t * EConstr.constr
+  | IsProj of Projection.t * Sorts.relevance * EConstr.constr
 
 let expand_table_key env = function
   | ConstKey cst -> constant_opt_value_in env cst
   | VarKey id -> (try named_body id env with Not_found -> None)
   | RelKey _ -> None
 
-let unfold_projection env p stk =
-  let s = Stack.Proj p in
+let unfold_projection env p r stk =
+  let s = Stack.Proj (p,r) in
   s :: stk
 
 let expand_key ts env sigma = function
   | Some (IsKey k) -> Option.map EConstr.of_constr (expand_table_key env k)
-  | Some (IsProj (p, c)) ->
+  | Some (IsProj (p, r, c)) ->
     let red = Stack.zip sigma (whd_betaiota_deltazeta_for_iota_state ts env sigma
-                               (c, unfold_projection env p []))
-    in if EConstr.eq_constr sigma (EConstr.mkProj (p, c)) red then None else Some red
+                               (c, unfold_projection env p r []))
+    in if EConstr.eq_constr sigma (EConstr.mkProj (p, r, c)) red then None else Some red
   | None -> None
 
 let isApp_or_Proj sigma c =
@@ -520,10 +506,10 @@ let key_of env sigma b flags f =
   | Var id when is_transparent env (VarKey id) &&
       TransparentState.is_transparent_variable flags.modulo_delta id ->
     Some (IsKey (VarKey id))
-  | Proj (p, c) when Names.Projection.unfolded p
+  | Proj (p, r, c) when Names.Projection.unfolded p
     || (is_transparent env (ConstKey (Projection.constant p)) &&
        (TransparentState.is_transparent_constant flags.modulo_delta (Projection.constant p))) ->
-    Some (IsProj (p, c))
+    Some (IsProj (p, r, c))
   | _ -> None
 
 
@@ -534,7 +520,7 @@ let translate_key = function
 
 let translate_key = function
   | IsKey k -> translate_key k
-  | IsProj (c, _) -> ConstKey (Projection.constant c)
+  | IsProj (c, _, _) -> ConstKey (Projection.constant c)
 
 let oracle_order env cf1 cf2 =
   match cf1 with
@@ -547,10 +533,10 @@ let oracle_order env cf1 cf2 =
       | None -> Some true
       | Some k2 ->
         match k1, k2 with
-        | IsProj (p, _), IsKey (ConstKey (p',_))
+        | IsProj (p, _, _), IsKey (ConstKey (p',_))
           when Environ.QConstant.equal env (Projection.constant p) p' ->
           Some (not (Projection.unfolded p))
-        | IsKey (ConstKey (p,_)), IsProj (p', _)
+        | IsKey (ConstKey (p,_)), IsProj (p', _, _)
           when Environ.QConstant.equal env p (Projection.constant p') ->
           Some (Projection.unfolded p')
         | _ ->
@@ -565,22 +551,11 @@ let is_rigid_head sigma flags t =
   | Fix _ | CoFix _ -> true
   | Rel _ | Var _ | Meta _ | Evar _ | Sort _ | Cast (_, _, _) | Prod _
     | Lambda _ | LetIn _ | App (_, _) | Case _
-    | Proj (_, _) -> false (* Why aren't Prod, Sort rigid heads ? *)
-
-let force_eqs c =
-  let open UnivProblem in
-  Set.fold
-    (fun c acc ->
-       let c' = match c with
-         | ULub (l, r) -> UEq (Sorts.sort_of_univ @@ Univ.Universe.make l, Sorts.sort_of_univ @@ Univ.Universe.make r)
-         | ULe _ | UEq _ | UWeak _ -> c
-       in
-        Set.add c' acc)
-    c Set.empty
+    | Proj _ -> false (* Why aren't Prod, Sort rigid heads ? *)
 
 let constr_cmp pb env sigma flags ?nargs t u =
   let cstrs =
-    if pb == Reduction.CONV then EConstr.eq_constr_universes env sigma ?nargs t u
+    if pb == Conversion.CONV then EConstr.eq_constr_universes env sigma ?nargs t u
     else EConstr.leq_constr_universes env sigma ?nargs t u
   in
   match cstrs with
@@ -589,7 +564,7 @@ let constr_cmp pb env sigma flags ?nargs t u =
       with UGraph.UniverseInconsistency _ -> None
       | Evd.UniversesDiffer ->
         if is_rigid_head sigma flags t then
-          try Some (Evd.add_universe_constraints sigma (force_eqs cstrs))
+          try Some (Evd.add_universe_constraints sigma (UnivProblem.Set.force cstrs))
           with UGraph.UniverseInconsistency _ -> None
         else None
       end
@@ -622,7 +597,9 @@ let subst_defined_metas_evars sigma (bl,el) c =
       let select (_,(evk',args'),_) = Evar.equal evk evk' && SList.equal eq args args' in
       begin match List.find select el with
       | (_, _, c) -> substrec (EConstr.Unsafe.to_constr c)
-      | exception Not_found -> Constr.map substrec c
+      | exception Not_found ->
+        let c = EConstr.(Unsafe.to_constr (whd_evar sigma (of_constr c))) in
+        Constr.map substrec c
       end
     | _ -> Constr.map substrec c
   in try Some (EConstr.of_constr (substrec c)) with Not_found -> None
@@ -654,7 +631,7 @@ let check_compatibility_ustate env pbty flags (sigma,metasubst,evarsubst : subst
     else UnivProblem.Set.empty
 
 let rec is_neutral env sigma ts t =
-  let (f, l) = decompose_app_vect sigma t in
+  let (f, l) = decompose_app sigma t in
     match EConstr.kind sigma f with
     | Const (c, u) ->
       not (Environ.evaluable_constant c env) ||
@@ -667,7 +644,7 @@ let rec is_neutral env sigma ts t =
     | Rel n -> true
     | Evar _ | Meta _ -> true
     | Case (_, _, _, _, _, c, _) -> is_neutral env sigma ts c
-    | Proj (p, c) -> is_neutral env sigma ts c
+    | Proj (p, _, c) -> is_neutral env sigma ts c
     | Lambda _ | LetIn _ | Construct _ | CoFix _ | Int _ | Float _ | Array _ -> false
     | Sort _ | Cast (_, _, _) | Prod (_, _, _) | Ind _ -> false (* Really? *)
     | Fix _ -> false (* This is an approximation *)
@@ -696,8 +673,9 @@ let eta_constructor_app env sigma f l1 term =
         let npars = mib.Declarations.mind_nparams in
         let pars, l1' = Array.chop npars l1 in
         let arg = Array.append pars [|term|] in
-        let l2 = Array.map (fun p -> mkApp (mkConstU (Projection.Repr.constant p,u), arg)) projs in
-          l1', l2
+        let l2 = Array.map (fun (p,_) ->
+            mkApp (mkConstU (Projection.Repr.constant p,u), arg)) projs in
+        l1', l2
       | _ -> assert false)
   | _ -> assert false
 
@@ -728,7 +706,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
     and cN = Evarutil.whd_head_evar sigma curn in
     let () =
       debug_tactic_unification (fun () ->
-          Termops.Internal.print_constr_env curenv sigma cM ++ str" ~= " ++
+          Termops.Internal.print_constr_env curenv sigma cM ++ strbrk" ~= " ++
           Termops.Internal.print_constr_env curenv sigma cN)
     in
       match (EConstr.kind sigma cM, EConstr.kind sigma cN) with
@@ -810,8 +788,6 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
               else error_cannot_unify_local curenv sigma (m,n,cN)
         | Sort s1, Sort s2 ->
             (try
-              let s1 = ESorts.kind sigma s1 in
-              let s2 = ESorts.kind sigma s2 in
                let sigma' =
                  if pb == CUMUL
                  then Evd.set_leq_sort curenv sigma s1 s2
@@ -830,7 +806,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
         | _, LetIn (_,a,_,c) -> unirec_rec curenvnb pb opt substn cM (subst1 a c)
 
         (* Fast path for projections. *)
-        | Proj (p1,c1), Proj (p2,c2) when Environ.QConstant.equal env
+        | Proj (p1,_,c1), Proj (p2,_,c2) when Environ.QConstant.equal env
             (Projection.constant p1) (Projection.constant p2) ->
           (try unify_same_proj curenvnb cv_pb {opt with at_top = true}
                substn c1 c2
@@ -851,9 +827,10 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
                and not a constructor. *)
             is_eta_constructor_app curenv sigma flags.modulo_delta f1 l1 cN ->
           (try
-             let l1', l2' = eta_constructor_app curenv sigma f1 l1 cN in
              let opt' = {opt with at_top = true; with_cs = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
+             let (sigma, _, _) as substn = check_type_eta_constructor_app curenvnb opt' substn cM cN in
+             let l1', l2' = eta_constructor_app curenv sigma f1 l1 cN in
+             Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
            with ex when precatchable_exception ex ->
              match EConstr.kind sigma cN with
              | App(f2,l2) when
@@ -865,9 +842,10 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
         | _, App (f2, l2) when flags.modulo_eta &&
             is_eta_constructor_app curenv sigma flags.modulo_delta f2 l2 cM ->
           (try
-             let l2', l1' = eta_constructor_app curenv sigma f2 l2 cM in
              let opt' = {opt with at_top = true; with_cs = false} in
-               Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
+             let (sigma, _, _) as substn = check_type_eta_constructor_app curenvnb opt' substn cN cM in
+             let l2', l1' = eta_constructor_app curenv sigma f2 l2 cM in
+             Array.fold_left2 (unirec_rec curenvnb CONV opt' ~nargs:0) substn l1' l2'
            with ex when precatchable_exception ex ->
              match EConstr.kind sigma cM with
              | App(f1,l1) when
@@ -876,12 +854,12 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
                unify_app_pattern true curenvnb pb opt substn cM f1 l1 cN f2 l2
              | _ -> raise ex)
 
-        | Case (ci1, u1, pms1, p1, iv1, c1, cl1), Case (ci2, u2, pms2, p2, iv2, c2, cl2) ->
+        | Case (ci1, u1, pms1, p1, iv1, c1, cl1), Case (ci2, u2, pms2, (p2,_), iv2, c2, cl2) ->
             (try
-             let () = if not (Ind.CanOrd.equal ci1.ci_ind ci2.ci_ind) then error_cannot_unify curenv sigma (cM,cN) in
+             let () = if not (QInd.equal curenv ci1.ci_ind ci2.ci_ind) then error_cannot_unify curenv sigma (cM,cN) in
              let opt' = {opt with at_top = true; with_types = false} in
              let substn = Array.fold_left2 (unirec_rec curenvnb CONV ~nargs:0 opt') substn pms1 pms2 in
-             let (ci1, _, _, p1, _, c1, cl1) = EConstr.annotate_case env sigma (ci1, u1, pms1, p1, iv1, c1, cl1) in
+             let (ci1, _, _, (p1,_), _, c1, cl1) = EConstr.annotate_case env sigma (ci1, u1, pms1, p1, iv1, c1, cl1) in
              let unif opt substn (ctx1, c1) (_, c2) =
                let curenvnb' = List.fold_right (fun decl (env, n) -> push_rel decl env, n + 1) ctx1 curenvnb in
                unirec_rec curenvnb' CONV opt' substn c1 c2
@@ -925,15 +903,37 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
         | App (f1,l1), App (f2,l2) ->
           unify_app curenvnb pb opt substn cM f1 l1 cN f2 l2
 
-        | App (f1,l1), Proj(p2,c2) ->
+        | App (f1,l1), Proj(p2,_,c2) ->
           unify_app curenvnb pb opt substn cM f1 l1 cN cN [||]
 
-        | Proj (p1,c1), App(f2,l2) ->
+        | Proj (p1,_,c1), App(f2,l2) ->
           unify_app curenvnb pb opt substn cM cM [||] cN f2 l2
 
         | _ ->
           unify_not_same_head curenvnb pb opt substn ~nargs cM cN
 
+  and check_type_eta_constructor_app (env,nb as curenvnb) opt ((sigma,metasubst,evarsubst) as substn : subst0) other term =
+    let  (((_, i as ind), j), u) =
+      EConstr.destConstruct sigma (fst (decompose_app sigma other))
+    in
+    (* ensure that we only eta expand if we are at the same inductive
+       (we accept that univs params and indices may be different) *)
+    let fail () =
+      error_cannot_unify env sigma (other, term)
+    in
+    let tterm = try Retyping.get_type_of ~lax:true env sigma term
+      with RetypeError _ -> fail ()
+    in
+    let tterm' = Reductionops.whd_all env sigma tterm in
+     match EConstr.kind sigma (fst (decompose_app sigma tterm')) with
+      | Ind (ind',_) when QInd.equal env ind ind' -> substn
+      | _ ->
+        let tother = try Retyping.get_type_of ~lax:true env sigma other
+          with RetypeError _ ->
+            fail ()
+        in
+        unirec_rec curenvnb CONV opt ~nargs:0 substn tother tterm
+
   and unify_app_pattern dir curenvnb pb opt (sigma, _, _ as substn) cM f1 l1 cN f2 l2 =
     let f, l, t = if dir then f1, l1, cN else f2, l2, cM in
       match is_unification_pattern curenvnb sigma f (Array.to_list l) t with
@@ -960,7 +960,7 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
       in
       let expand_proj c c' l =
         match EConstr.kind sigma c with
-        | Proj (p, t) when not (Projection.unfolded p) && needs_expansion p c' ->
+        | Proj (p, _, t) when not (Projection.unfolded p) && needs_expansion p c' ->
           (try destApp sigma (Retyping.expand_projection curenv sigma p t (Array.to_list l))
            with RetypeError _ -> (* Unification can be called on ill-typed terms, due
                                      to FO and eta in particular, fail gracefully in that case *)
@@ -1161,8 +1161,8 @@ let rec unify_0_with_initial_metas (sigma,ms,es as subst : subst0) conv_at_top e
   in
 
   debug_tactic_unification (fun () ->
-      str "Starting unification: " ++
-      Termops.Internal.print_constr_env env sigma (fst m) ++ str" ~= " ++
+      str "Starting unification:" ++ spc() ++
+      Termops.Internal.print_constr_env env sigma (fst m) ++ strbrk" ~= " ++
       Termops.Internal.print_constr_env env sigma (fst n));
 
   let opt = { at_top = conv_at_top; with_types = false; with_cs = true } in
@@ -1358,6 +1358,10 @@ let w_coerce env evd mv c =
   let mvty = Typing.meta_type env evd mv in
   w_coerce_to_type env evd c cty mvty
 
+let nf_meta env sigma c =
+  let cl = mk_freelisted c in
+  meta_instance env sigma { cl with rebus = cl.rebus }
+
 let unify_to_type env sigma flags c status u =
   let sigma, c = refresh_universes (Some false) env sigma c in
   let t = get_type_of env sigma (nf_meta env sigma c) in
@@ -1549,8 +1553,7 @@ let check_types env flags (sigma,_,_ as subst) m n =
 
 let try_resolve_typeclasses env evd flag m n =
   if flag then
-    Typeclasses.resolve_typeclasses ~filter:Typeclasses.no_goals ~split:false
-      ~fail:true env evd
+    Typeclasses.resolve_typeclasses ~filter:Typeclasses.no_goals ~fail:true env evd
   else evd
 
 let w_unify_core_0 env evd with_types cv_pb flags m n =
@@ -1645,12 +1648,12 @@ exception PatternNotFound
 let make_pattern_test from_prefix_of_ind is_correct_type env sigma (pending,c) =
   let flags =
     if from_prefix_of_ind then
-      let flags = default_matching_flags pending in
+      let flags = default_matching_flags (Option.default Evd.empty pending) in
       { flags with core_unify_flags = { flags.core_unify_flags with
         modulo_conv_on_closed_terms = Some TransparentState.full;
         restrict_conv_on_strict_subterms = true } }
-    else default_matching_flags pending in
-  let n = Array.length (snd (decompose_app_vect sigma c)) in
+    else default_matching_flags (Option.default Evd.empty pending) in
+  let n = Array.length (snd (decompose_app sigma c)) in
   let cgnd = if occur_meta_or_undefined_evar sigma c then NotGround else Ground in
   let matching_fun _ t =
     (* make_pattern_test is only ever called with an empty rel context *)
@@ -1660,14 +1663,14 @@ let make_pattern_test from_prefix_of_ind is_correct_type env sigma (pending,c) =
         if from_prefix_of_ind then
           (* We check for fully applied subterms of the form "u u1 .. un" *)
           (* of inductive type knowing only a prefix "u u1 .. ui" *)
-          let t,l = decompose_app sigma t in
+          let t,l = decompose_app_list sigma t in
           let l1,l2 =
             try List.chop n l with Failure _ -> raise (NotUnifiable None) in
           if not (List.for_all (fun c -> Vars.closed0 sigma c) l2) then raise (NotUnifiable None)
           else
             applist (t,l1), l2
         else t, [] in
-      let sigma = w_typed_unify env sigma Reduction.CONV flags (c, cgnd) (t', Unknown) in
+      let sigma = w_typed_unify env sigma Conversion.CONV flags (c, cgnd) (t', Unknown) in
       let ty = Retyping.get_type_of env sigma t in
       if not (is_correct_type ty) then raise (NotUnifiable None);
       Some(sigma, t, l2)
@@ -1776,7 +1779,7 @@ let make_abstraction_core name (test,out) env sigma c ty occs check_occs concl =
 type prefix_of_inductive_support_flag = bool
 
 type abstraction_request =
-| AbstractPattern of prefix_of_inductive_support_flag * (types -> bool) * Name.t * (evar_map * constr) * clause * bool
+| AbstractPattern of prefix_of_inductive_support_flag * (types -> bool) * Name.t * (evar_map option * constr) * clause * bool
 | AbstractExact of Name.t * constr * types option * clause * bool
 
 type 'r abstraction_result =
@@ -1885,14 +1888,14 @@ let rec make sigma c0 = match EConstr.kind sigma c0 with
   let c = make sigma c in
   let ald, al = make_array sigma al in
   { proj = c0; self = AApp (c, al); data = max c.data ald }
-| Proj (p, t) ->
+| Proj (p, _, t) ->
   let t = make sigma t in
   { proj = c0; self = AOther [|t|]; data = t.data }
 | Evar (e, al) ->
   (* Unification doesn't recurse on the subterms in evar instances *)
   let data = SList.Skip.fold (fun accu v -> max accu (get_max_rel sigma v)) 0 al in
   { proj = c0; self = AOther [||]; data }
-| Case (ci, u, pms, p, iv, c, bl) ->
+| Case (ci, u, pms, (p,_), iv, c, bl) ->
   let pmsd = get_max_rel_array sigma pms in
   let pd =
     let (nas, p) = p in
@@ -1952,8 +1955,8 @@ let w_unify_to_subterm env evd ?(flags=default_unify_flags ()) (op,cl) =
        if is_closed && not (isEvar evd cl) && keyed_unify env evd kop cl then
        (try
          if is_keyed_unification () then
-           let f1, l1 = decompose_app_vect evd op in
-           let f2, l2 = decompose_app_vect evd cl in
+           let f1, l1 = decompose_app evd op in
+           let f2, l2 = decompose_app evd cl in
            w_typed_unify_array env evd flags f1 l1 f2 l2,cl
          else w_typed_unify env evd CONV flags (op, opgnd) (cl, Unknown),cl
        with ex when Pretype_errors.unsatisfiable_exception ex ->
@@ -2021,7 +2024,7 @@ let w_unify_to_subterm_all env evd ?(flags=default_unify_flags ()) (op,cl) =
             | Case(_,_,_,_,_,c,lf) -> (* does not search in the predicate *)
                 bind (matchrec c) (bind_iter matchrec (Array.map snd lf))
 
-            | Proj (p,c) -> matchrec c
+            | Proj (p,_,c) -> matchrec c
 
             | LetIn(_,c1,_,c2) ->
                 bind (matchrec c1) (matchrec c2)
@@ -2064,7 +2067,7 @@ let w_unify_to_subterm_list env evd flags hdmeta oplist t =
       else
         let allow_K = flags.allow_K_in_toplevel_higher_order_unification in
         let flags =
-          if unsafe_occur_meta_or_existential op || is_keyed_unification () then
+          if is_keyed_unification () || occur_meta_or_existential evd op then
             (* This is up to delta for subterms w/o metas ... *)
             flags
           else
@@ -2159,8 +2162,8 @@ let w_unify2 env evd flags dep cv_pb ty1 ty2 =
    convertible and first-order otherwise. But if failed if e.g. the type of
    Meta(1) had meta-variables in it. *)
 let w_unify env evd cv_pb ?(flags=default_unify_flags ()) ty1 ty2 =
-  let hd1,l1 = decompose_app_vect evd (whd_nored env evd ty1) in
-  let hd2,l2 = decompose_app_vect evd (whd_nored env evd ty2) in
+  let hd1,l1 = decompose_app evd (whd_nored env evd ty1) in
+  let hd2,l2 = decompose_app evd (whd_nored env evd ty2) in
   let is_empty1 = Array.is_empty l1 in
   let is_empty2 = Array.is_empty l2 in
     match EConstr.kind evd hd1, not is_empty1, EConstr.kind evd hd2, not is_empty2 with
