@@ -27,23 +27,23 @@ module CompactedDecl = Context.Compacted.Declaration
 (* This is set on by proofgeneral proof-tree mode. But may be used for
    other purposes *)
 let print_goal_tag_opt_name = ["Printing";"Goal";"Tags"]
-let should_tag =
+let { Goptions.get = should_tag } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:print_goal_tag_opt_name
     ~value:false
+    ()
 
-let should_unfoc =
+let { Goptions.get = should_unfoc } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Unfocused"]
     ~value:false
+    ()
 
-let should_gname =
+let { Goptions.get = should_gname } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Goal";"Names"]
     ~value:false
+    ()
 
 let print_goal_names = should_gname (* for export *)
 
@@ -58,24 +58,24 @@ let print_goal_names = should_gname (* for export *)
    and only names of goal/section variables and rel names that do
    _not_ occur in the scope of the binder to be printed are avoided. *)
 
-let pr_econstr_n_env ?lax ?inctx ?scope env sigma n t =
-  pr_constr_expr_n env sigma n (extern_constr ?lax ?inctx ?scope env sigma t)
-let pr_econstr_env ?lax ?inctx ?scope env sigma t =
-  pr_constr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
-let pr_leconstr_env ?lax ?inctx ?scope env sigma t =
-  Ppconstr.pr_lconstr_expr env sigma (extern_constr ?lax ?inctx ?scope env sigma t)
+let pr_econstr_n_env ?inctx ?scope env sigma n t =
+  pr_constr_expr_n env sigma n (extern_constr ?inctx ?scope env sigma t)
+let pr_econstr_env ?inctx ?scope env sigma t =
+  pr_constr_expr env sigma (extern_constr ?inctx ?scope env sigma t)
+let pr_leconstr_env ?inctx ?scope env sigma t =
+  Ppconstr.pr_lconstr_expr env sigma (extern_constr ?inctx ?scope env sigma t)
 
-let pr_constr_n_env ?lax ?inctx ?scope env sigma n c =
-  pr_econstr_n_env ?lax ?inctx ?scope env sigma n (EConstr.of_constr c)
-let pr_constr_env ?lax ?inctx ?scope env sigma c =
-  pr_econstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
-let pr_lconstr_env ?lax ?inctx ?scope env sigma c =
-  pr_leconstr_env ?lax ?inctx ?scope env sigma (EConstr.of_constr c)
+let pr_constr_n_env ?inctx ?scope env sigma n c =
+  pr_econstr_n_env ?inctx ?scope env sigma n (EConstr.of_constr c)
+let pr_constr_env ?inctx ?scope env sigma c =
+  pr_econstr_env ?inctx ?scope env sigma (EConstr.of_constr c)
+let pr_lconstr_env ?inctx ?scope env sigma c =
+  pr_leconstr_env ?inctx ?scope env sigma (EConstr.of_constr c)
 
-let pr_open_lconstr_env ?lax ?inctx ?scope env sigma (_,c) =
-  pr_leconstr_env ?lax ?inctx ?scope env sigma c
-let pr_open_constr_env ?lax ?inctx ?scope env sigma (_,c) =
-  pr_econstr_env ?lax ?inctx ?scope env sigma c
+let pr_open_lconstr_env ?inctx ?scope env sigma (_,c) =
+  pr_leconstr_env ?inctx ?scope env sigma c
+let pr_open_constr_env ?inctx ?scope env sigma (_,c) =
+  pr_econstr_env ?inctx ?scope env sigma c
 
 let pr_constr_under_binders_env_gen pr env sigma (ids,c) =
   (* Warning: clashes can occur with variables of same name in env but *)
@@ -87,15 +87,15 @@ let pr_constr_under_binders_env_gen pr env sigma (ids,c) =
 let pr_constr_under_binders_env = pr_constr_under_binders_env_gen pr_econstr_env
 let pr_lconstr_under_binders_env = pr_constr_under_binders_env_gen pr_leconstr_env
 
-let pr_etype_env ?lax ?goal_concl_style env sigma t =
-  pr_constr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma t)
-let pr_letype_env ?lax ?goal_concl_style env sigma ?impargs t =
-  pr_lconstr_expr env sigma (extern_type ?lax ?goal_concl_style env sigma ?impargs t)
+let pr_etype_env ?goal_concl_style env sigma t =
+  pr_constr_expr env sigma (extern_type ?goal_concl_style env sigma t)
+let pr_letype_env ?goal_concl_style env sigma ?impargs t =
+  pr_lconstr_expr env sigma (extern_type ?goal_concl_style env sigma ?impargs t)
 
-let pr_type_env ?lax ?goal_concl_style env sigma c =
-  pr_etype_env ?lax ?goal_concl_style env sigma (EConstr.of_constr c)
-let pr_ltype_env ?lax ?goal_concl_style env sigma ?impargs c =
-  pr_letype_env ?lax ?goal_concl_style env sigma ?impargs (EConstr.of_constr c)
+let pr_type_env ?goal_concl_style env sigma c =
+  pr_etype_env ?goal_concl_style env sigma (EConstr.of_constr c)
+let pr_ltype_env ?goal_concl_style env sigma ?impargs c =
+  pr_letype_env ?goal_concl_style env sigma ?impargs (EConstr.of_constr c)
 
 let pr_ljudge_env env sigma j =
   (pr_leconstr_env env sigma j.uj_val, pr_leconstr_env env sigma j.uj_type)
@@ -105,10 +105,10 @@ let pr_lglob_constr_env env sigma c =
 let pr_glob_constr_env env sigma c =
   pr_constr_expr env sigma (extern_glob_constr (extern_env env sigma) c)
 
-let pr_closed_glob_n_env ?lax ?goal_concl_style ?inctx ?scope env sigma n c =
-  pr_constr_expr_n env sigma n (extern_closed_glob ?lax ?goal_concl_style ?inctx ?scope env sigma c)
-let pr_closed_glob_env ?lax ?goal_concl_style ?inctx ?scope env sigma c =
-  pr_constr_expr env sigma (extern_closed_glob ?lax ?goal_concl_style ?inctx ?scope env sigma c)
+let pr_closed_glob_n_env ?goal_concl_style ?inctx ?scope env sigma n c =
+  pr_constr_expr_n env sigma n (extern_closed_glob ?goal_concl_style ?inctx ?scope env sigma c)
+let pr_closed_glob_env ?goal_concl_style ?inctx ?scope env sigma c =
+  pr_constr_expr env sigma (extern_closed_glob ?goal_concl_style ?inctx ?scope env sigma c)
 
 let pr_lconstr_pattern_env env sigma c =
   pr_lconstr_pattern_expr env sigma (extern_constr_pattern (Termops.names_of_rel_context env) sigma c)
@@ -121,7 +121,7 @@ let pr_cases_pattern t =
 let pr_sort sigma s = pr_sort_expr (extern_sort sigma s)
 
 let () = Termops.Internal.set_print_constr
-  (fun env sigma t -> pr_lconstr_expr env sigma (extern_constr ~lax:true env sigma t))
+  (fun env sigma t -> pr_lconstr_expr env sigma (extern_constr env sigma t))
 
 let pr_in_comment x = str "(* " ++ x ++ str " *)"
 
@@ -177,41 +177,66 @@ let safe_gen f env sigma c =
 let safe_pr_lconstr_env = safe_gen pr_lconstr_env
 let safe_pr_constr_env = safe_gen pr_constr_env
 
+let q_ident = Id.of_string "α"
+
 let u_ident = Id.of_string "u"
 
 let universe_binders_with_opt_names orig names =
   let open Univ in
-  let orig = Univ.AbstractContext.names orig in
-  let orig = Array.to_list orig in
-  let udecl = match names with
+  let qorig, uorig = UVars.AbstractContext.names orig in
+  let qorig, uorig as orig = Array.to_list qorig, Array.to_list uorig in
+  let qdecl, udecl = match names with
   | None -> orig
-  | Some udecl ->
+  | Some (qdecl,udecl) ->
     try
-      List.map2 (fun orig {CAst.v = na} ->
-          match na with
-          | Anonymous -> orig
-          | Name id -> Name id) orig udecl
+      let qs =
+        List.map2 (fun orig {CAst.v = na} ->
+            match na with
+            | Anonymous -> orig
+            | Name id -> Name id) qorig qdecl
+      in
+      let us =
+        List.map2 (fun orig {CAst.v = na} ->
+            match na with
+            | Anonymous -> orig
+            | Name id -> Name id) uorig udecl
+      in
+      qs, us
     with Invalid_argument _ ->
       let open UnivGen in
       raise (UniverseLengthMismatch {
-          actual = List.length orig;
-          expect = List.length udecl;
+          actual = List.length qorig, List.length uorig;
+          expect = List.length qdecl, List.length udecl;
         })
   in
-  let fold_named i (ubind,revubind as o) = function
-    | Name id -> let ui = Level.var i in
-      Id.Map.add id ui ubind, Level.Map.add ui id revubind
+  let fold_qnamed i ((qbind,ubind),(revqbind,revubind) as o) = function
+    | Name id -> let ui = Sorts.QVar.make_var i in
+      (Id.Map.add id ui qbind, ubind), (Sorts.QVar.Map.add ui id revqbind, revubind)
     | Anonymous -> o
   in
-  let names = List.fold_left_i fold_named 0 (UnivNames.empty_binders,Level.Map.empty) udecl in
-  let fold_anons i (u_ident, (ubind, revubind) as o) = function
+  let fold_unamed i ((qbind,ubind),(revqbind,revubind) as o) = function
+    | Name id -> let ui = Level.var i in
+      (qbind, Id.Map.add id ui ubind), (revqbind, Level.Map.add ui id revubind)
+    | Anonymous -> o
+  in
+  let names = List.fold_left_i fold_qnamed 0 UnivNames.(empty_binders,empty_rev_binders) qdecl in
+  let names = List.fold_left_i fold_unamed 0 names udecl in
+  let fold_qanons i (u_ident, ((qbind,ubind), (revqbind,revubind)) as o) = function
+    | Name _ -> o
+    | Anonymous ->
+      let ui = Sorts.QVar.make_var i in
+      let id = Namegen.next_ident_away_from u_ident (fun id -> Id.Map.mem id qbind) in
+      (id, ((Id.Map.add id ui qbind, ubind), (Sorts.QVar.Map.add ui id revqbind, revubind)))
+  in
+  let fold_uanons i (u_ident, ((qbind,ubind), (revqbind,revubind)) as o) = function
     | Name _ -> o
     | Anonymous ->
       let ui = Level.var i in
       let id = Namegen.next_ident_away_from u_ident (fun id -> Id.Map.mem id ubind) in
-      (id, (Id.Map.add id ui ubind, Level.Map.add ui id revubind))
+      (id, ((qbind,Id.Map.add id ui ubind), (revqbind,Level.Map.add ui id revubind)))
   in
-  let (_, names) = List.fold_left_i fold_anons 0 (u_ident, names) udecl in
+  let (_, names) = List.fold_left_i fold_qanons 0 (q_ident, names) qdecl in
+  let (_, names) = List.fold_left_i fold_uanons 0 (u_ident, names) udecl in
   names
 
 let pr_universe_ctx_set sigma c =
@@ -221,8 +246,12 @@ let pr_universe_ctx_set sigma c =
     mt()
 
 let pr_universe_ctx sigma ?variance c =
-  if !Detyping.print_universes && not (Univ.UContext.is_empty c) then
-    fnl()++pr_in_comment (v 0 (Univ.pr_universe_context (Termops.pr_evd_level sigma) ?variance c))
+  if !Detyping.print_universes && not (UVars.UContext.is_empty c) then
+    fnl()++
+    pr_in_comment
+      (v 0
+         (UVars.pr_universe_context (Termops.pr_evd_qvar sigma) (Termops.pr_evd_level sigma)
+            ?variance c))
   else
     mt()
 
@@ -230,9 +259,10 @@ let pr_abstract_universe_ctx sigma ?variance ?priv c =
   let open Univ in
   let priv = Option.default Univ.ContextSet.empty priv in
   let has_priv = not (ContextSet.is_empty priv) in
-  if !Detyping.print_universes && (not (Univ.AbstractContext.is_empty c) || has_priv) then
+  if !Detyping.print_universes && (not (UVars.AbstractContext.is_empty c) || has_priv) then
+    let prqvar u = Termops.pr_evd_qvar sigma u in
     let prlev u = Termops.pr_evd_level sigma u in
-    let pub = (if has_priv then str "Public universes:" ++ fnl() else mt()) ++ v 0 (Univ.pr_abstract_universe_context prlev ?variance c) in
+    let pub = (if has_priv then str "Public universes:" ++ fnl() else mt()) ++ v 0 (UVars.pr_abstract_universe_context prqvar prlev ?variance c) in
     let priv = if has_priv then fnl() ++ str "Private universes:" ++ fnl() ++ v 0 (Univ.pr_universe_context_set prlev priv) else mt() in
     fnl()++pr_in_comment (pub ++ priv)
   else
@@ -250,6 +280,7 @@ let pr_global = pr_global_env Id.Set.empty
 
 let pr_universe_instance_constraints evd inst csts =
   let open Univ in
+  let prqvar = Termops.pr_evd_qvar evd in
   let prlev = Termops.pr_evd_level evd in
   let pcsts = if Constraints.is_empty csts then mt()
     else str " |= " ++
@@ -257,7 +288,7 @@ let pr_universe_instance_constraints evd inst csts =
            (fun (u,d,v) -> hov 0 (prlev u ++ pr_constraint_type d ++ prlev v))
            (Constraints.elements csts)
   in
-  str"@{" ++ Instance.pr prlev inst ++ pcsts ++ str"}"
+  str"@{" ++ UVars.Instance.pr prqvar prlev inst ++ pcsts ++ str"}"
 
 let pr_universe_instance evd inst =
   pr_universe_instance_constraints evd inst Univ.Constraints.empty
@@ -281,6 +312,13 @@ let pr_pconstructor = pr_puniverses pr_constructor
 let pr_evaluable_reference ref =
   pr_global (Tacred.global_of_evaluable_reference ref)
 
+let pr_notation_interpretation_env env sigma c =
+  Constrextern.without_symbols (pr_glob_constr_env env sigma) c
+
+let pr_notation_interpretation c =
+  let env = Global.env () in
+  pr_notation_interpretation_env env (Evd.from_env env) c
+
 (*let pr_glob_constr t =
   pr_lconstr (Constrextern.extern_glob_constr Id.Set.empty t)*)
 
@@ -313,9 +351,17 @@ let pr_compacted_decl env sigma decl =
   let ptyp = (str" : " ++ pt) in
   hov 0 (pids ++ pbody ++ ptyp)
 
+let pr_ecompacted_decl env sigma (decl:EConstr.compacted_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_compacted_decl env sigma decl
+
 let pr_named_decl env sigma decl =
   decl |> CompactedDecl.of_named_decl |> pr_compacted_decl env sigma
 
+let pr_enamed_decl env sigma (decl:EConstr.named_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_named_decl env sigma decl
+
 let pr_rel_decl env sigma decl =
   let na = RelDecl.get_name decl in
   let typ = RelDecl.get_type decl in
@@ -331,6 +377,11 @@ let pr_rel_decl env sigma decl =
   | Anonymous -> hov 0 (str"<>" ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
   | Name id -> hov 0 (pr_id id ++ spc () ++ pbody ++ str":" ++ spc () ++ ptyp)
 
+let pr_erel_decl env sigma (decl:EConstr.rel_declaration) =
+  let Refl = EConstr.Unsafe.eq in
+  pr_rel_decl env sigma decl
+
+
 
 (* Prints out an "env" in a nice format.  We print out the
  * signature,then a horizontal bar, then the debruijn environment.
@@ -343,7 +394,7 @@ let pr_named_context_of env sigma =
   hv 0 (prlist_with_sep (fun _ -> ws 2) (fun x -> x) psl)
 
 let pr_var_list_decl env sigma decl =
-  hov 0 (pr_compacted_decl env sigma decl)
+  hov 0 (pr_ecompacted_decl env sigma decl)
 
 let pr_named_context env sigma ne_context =
   hv 0 (Context.Named.fold_outside
@@ -362,9 +413,9 @@ let pr_context_unlimited env sigma =
   let sign_env =
     Context.Compacted.fold
       (fun d pps ->
-         let pidt =  pr_compacted_decl env sigma d in
+         let pidt =  pr_ecompacted_decl env sigma d in
          (pps ++ fnl () ++ pidt))
-      (Termops.compact_named_context (named_context env)) ~init:(mt ())
+      (Termops.compact_named_context sigma (EConstr.named_context env)) ~init:(mt ())
   in
   let db_env =
     fold_rel_context
@@ -383,8 +434,8 @@ let pr_ne_context_of header env sigma =
    considers as "variables": An hypothesis H:T where T:S and S<>Prop. *)
 let should_compact env sigma typ =
   get_compact_context() &&
-    let type_of_typ = Retyping.get_type_of env sigma (EConstr.of_constr typ) in
-    not (is_Prop (EConstr.to_constr sigma type_of_typ))
+    let type_of_typ = Retyping.get_type_of env sigma typ in
+    not (Termops.is_Prop sigma type_of_typ)
 
 
 (* If option Compact Contexts is set, we pack "simple" hypothesis in a
@@ -414,7 +465,8 @@ and bld_sign_env_id env sigma ctxt pps is_start =
 (* compact printing an env (variables and de Bruijn). Separator: three
    spaces between simple hyps, and newline otherwise *)
 let pr_context_limit_compact ?n env sigma =
-  let ctxt = Termops.compact_named_context (named_context env) in
+  let ctxt = EConstr.named_context env in
+  let ctxt = Termops.compact_named_context sigma ctxt in
   let lgth = List.length ctxt in
   let n_capped =
     match n with
@@ -433,8 +485,11 @@ let pr_context_limit_compact ?n env sigma =
 
 (* The number of printed hypothesis in a goal *)
 (* If [None], no limit *)
-let print_hyps_limit =
-  Goptions.declare_intopt_option_and_ref ~depr:false ~key:["Hyps";"Limit"]
+let { Goptions.get = print_hyps_limit } =
+  Goptions.declare_intopt_option_and_ref
+    ~key:["Hyps";"Limit"]
+    ~value:None
+    ()
 
 let pr_context_of env sigma = match print_hyps_limit () with
   | None -> hv 0 (pr_context_limit_compact env sigma)
@@ -464,8 +519,13 @@ let pr_transparent_state ts =
         str"CONSTANTS: " ++ pr_cpred ts.TransparentState.tr_cst ++ fnl ())
 
 let goal_repr sigma g =
-  let evi = Evd.find sigma g in
-  Evd.evar_filtered_env (Global.env ()) evi, Evd.evar_concl evi
+  let EvarInfo evi = Evd.find sigma g in
+  let env = Evd.evar_filtered_env (Global.env ()) evi in
+  let concl = match Evd.evar_body evi with
+  | Evd.Evar_empty -> Evd.evar_concl evi
+  | Evd.Evar_defined b -> Retyping.get_type_of env sigma b
+  in
+  env, concl
 
 (* display complete goal
  og_s has goal+sigma on the previous proof step for diffs
@@ -520,7 +580,7 @@ let pr_concl n ?diffs sigma g =
   header ++ str " is:" ++ cut () ++ str" "  ++ pc
 
 (* display evar type: a context and a type *)
-let pr_evgl_sign env sigma evi =
+let pr_evgl_sign env sigma (evi : undefined evar_info) =
   let env = evar_env env evi in
   let ps = pr_named_context_of env sigma in
   let _, l = match Filter.repr (evar_filter evi) with
@@ -532,14 +592,15 @@ let pr_evgl_sign env sigma evi =
     if List.is_empty ids then mt () else
       (str " (" ++ prlist_with_sep pr_comma pr_id ids ++ str " cannot be used)")
   in
-  let pc = pr_leconstr_env env sigma (Evd.evar_concl evi) in
+  let concl = Evd.evar_concl evi in
+  let pc = pr_leconstr_env env sigma concl in
   let candidates =
-    match Evd.evar_body evi, Evd.evar_candidates evi with
-    | Evar_empty, Some l ->
-       spc () ++ str "= {" ++
-         prlist_with_sep (fun () -> str "|") (pr_leconstr_env env sigma) l ++ str "}"
-    | _ ->
-       mt ()
+    begin match Evd.evar_candidates evi with
+    | None -> mt ()
+    | Some l ->
+      spc () ++ str "= {" ++
+      prlist_with_sep (fun () -> str "|") (pr_leconstr_env env sigma) l ++ str "}"
+    end
   in
   hov 0 (str"[" ++ ps ++ spc () ++ str"|- "  ++ pc ++ str"]" ++
            candidates ++ warn)
@@ -575,9 +636,11 @@ let pr_evars sigma evs =
 (* Display a list of evars given by their name, with a prefix *)
 let pr_ne_evar_set hd tl sigma l =
   if l != Evar.Set.empty then
-    let l = Evar.Set.fold (fun ev ->
-      Evar.Map.add ev (Evarutil.nf_evar_info sigma (Evd.find sigma ev)))
-      l Evar.Map.empty in
+    let l = Evar.Map.bind (fun ev ->
+        let evi = Evd.find_undefined sigma ev in
+        Evarutil.nf_evar_info sigma evi)
+        l
+    in
     hd ++ pr_evars sigma l ++ tl
   else
     mt ()
@@ -627,8 +690,8 @@ let print_evar_constraints gl sigma =
     str" " ++
       hov 2 (pr_env env ++ pr_leconstr_env env sigma t1 ++ spc () ++
              str (match pbty with
-                  | Reduction.CONV -> "=="
-                  | Reduction.CUMUL -> "<=") ++
+                  | Conversion.CONV -> "=="
+                  | Conversion.CUMUL -> "<=") ++
              spc () ++ pr_leconstr_env env sigma t2)
   in
   let pr_candidate ev evi (candidates,acc) =
@@ -649,11 +712,11 @@ let print_evar_constraints gl sigma =
         str" with candidates:" ++ fnl () ++ hov 0 ppcandidates
     else mt ()
 
-let should_print_dependent_evars =
+let { Goptions.get = should_print_dependent_evars } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Printing";"Dependent";"Evars";"Line"]
     ~value:false
+    ()
 
 let evar_nodes_of_term c =
   let rec evrec acc c =
@@ -670,10 +733,16 @@ let queue_term q is_dependent c =
   queue_set q is_dependent (evar_nodes_of_term c)
 
 let process_dependent_evar q acc evm is_dependent e =
-  let evi = Evd.find evm e in
+  let EvarInfo evi = Evd.find evm e in
   (* Queues evars appearing in the types of the goal (conclusion, then
      hypotheses), they are all dependent. *)
-  queue_term q true (Evd.evar_concl evi);
+  let () = match Evd.evar_body evi with
+  | Evar_empty ->
+    queue_term q true (Evd.evar_concl evi)
+  | Evar_defined b ->
+    let env = Evd.evar_filtered_env (Global.env ()) evi in
+    queue_term q true (Retyping.get_type_of env evm b)
+  in
   List.iter begin fun decl ->
     let open NamedDecl in
     queue_term q true (NamedDecl.get_type decl);
@@ -719,7 +788,7 @@ let gather_dependent_evars evm l =
 
 let gather_dependent_evars_goal sigma goals =
   let map evk =
-    let evi = Evd.find sigma evk in
+    let EvarInfo evi = Evd.find sigma evk in
     EConstr.mkEvar (evk, Evd.evar_identity_subst evi)
   in
   gather_dependent_evars sigma (List.map map goals)
@@ -1104,7 +1173,6 @@ let pr_typing_flags flags =
   str "check_guarded: " ++ bool flags.check_guarded ++ fnl ()
   ++ str "check_positive: " ++ bool flags.check_positive ++ fnl ()
   ++ str "check_universes: " ++ bool flags.check_universes ++ fnl ()
-  ++ str "cumulative sprop: " ++ bool flags.cumulative_sprop ++ fnl ()
   ++ str "definitional uip: " ++ bool flags.allow_uip
 
 module Debug =
