@@ -150,19 +150,10 @@ let is_done p =
   Proofview.finished (unroll_focus p.proofview p.focus_stack)
 
 (* spiwack: for compatibility with <= 8.2 proof engine *)
-let has_unresolved_evar p =
-  Evd.has_undefined (Proofview.return p.proofview)
-let has_shelved_goals p =
-  let (_goals,sigma) = Proofview.proofview p.proofview in
-  Evd.has_shelved sigma
 let has_given_up_goals p =
   let (_goals,sigma) = Proofview.proofview p.proofview in
   Evd.has_given_up sigma
 
-let is_complete p =
-  is_done p && not (has_unresolved_evar p) &&
-  not (has_shelved_goals p) && not (has_given_up_goals p)
-
 (* Returns the list of partial proofs to initial goals *)
 let partial_proof p = Proofview.partial_proof p.entry p.proofview
 
@@ -332,22 +323,12 @@ let _ = CErrors.register_handler begin function
     | _ -> None
   end
 
-let warn_remaining_shelved_goals =
-  CWarnings.create ~name:"remaining-shelved-goals" ~category:"tactics"
-    (fun () -> Pp.str"The proof has remaining shelved goals")
-
-let warn_remaining_unresolved_evars =
-  CWarnings.create ~name:"remaining-unresolved-evars" ~category:"tactics"
-    (fun () -> Pp.str"The proof has unresolved variables")
-
-let return ?pid (p : t) =
+let return ?pid ?(allow_given_up=false) (p : t) =
   if not (is_done p) then
     raise (OpenProof(pid, UnfinishedProof))
-  else if has_given_up_goals p then
+  else if not allow_given_up && has_given_up_goals p then
     raise (OpenProof(pid, HasGivenUpGoals))
   else begin
-    if has_shelved_goals p then warn_remaining_shelved_goals ()
-    else if has_unresolved_evar p then warn_remaining_unresolved_evars ();
     let p = unfocus end_of_stack_kind p () in
     Proofview.return p.proofview
   end
@@ -467,14 +448,28 @@ let pr_proof p =
     str "]"
   )
 
-let use_unification_heuristics =
+let { Goptions.get = use_unification_heuristics } =
   Goptions.declare_bool_option_and_ref
-    ~depr:false
     ~key:["Solve";"Unification";"Constraints"]
     ~value:true
+    ()
 
 exception SuggestNoSuchGoals of int * t
 
+exception FailedConstraints of exn
+
+let () = CErrors.register_handler (function
+    | FailedConstraints e ->
+      Some Pp.(CErrors.print e ++ spc() ++
+               hov 1
+                 (str "(while solving unification constraints," ++ spc() ++
+                  str "see flag \"Solve Unification Constraints\")"))
+    | _ -> None)
+
+let solve_constraints =
+  Proofview.tclOR Refine.solve_constraints
+    (fun (e,info) -> Proofview.tclZERO ~info (FailedConstraints e))
+
 let solve ?with_end_tac gi info_lvl tac pr =
     let tac = match with_end_tac with
       | None -> tac
@@ -490,7 +485,7 @@ let solve ?with_end_tac gi info_lvl tac pr =
     let tac = Goal_select.tclSELECT ~nosuchgoal gi tac in
     let tac =
       if use_unification_heuristics () then
-        Proofview.tclTHEN tac Refine.solve_constraints
+        Proofview.tclTHEN tac solve_constraints
       else tac
     in
     let env = Global.env () in
@@ -556,7 +551,7 @@ let refine_by_tactic ~name ~poly env sigma ty tac =
 let get_goal_context_gen pf i =
   let { sigma; goals } = data pf in
   let goal = try List.nth goals (i-1) with Failure _ -> raise (NoSuchGoal None) in
-  let env = Evd.evar_filtered_env (Global.env ()) (Evd.find sigma goal) in
+  let env = Evd.evar_filtered_env (Global.env ()) (Evd.find_undefined sigma goal) in
   (sigma, env)
 
 let get_proof_context p =
