@@ -43,10 +43,6 @@ type data =
   (** Entry for the proofview *)
   ; stack : (Evar.t list * Evar.t list) list
   (** A representation of the focus stack *)
-  ; shelf : Evar.t list
-  (** A representation of the shelf  *)
-  ; given_up : Evar.t list
-  (** A representation of the given up goals  *)
   ; name : Names.Id.t
   (** The name of the theorem whose proof is being constructed *)
   ; poly : bool;
@@ -59,25 +55,27 @@ val data : t -> data
 val start
   :  name:Names.Id.t
   -> poly:bool
+  -> ?typing_flags:Declarations.typing_flags
   -> Evd.evar_map -> (Environ.env * EConstr.types) list -> t
 
 val dependent_start
   :  name:Names.Id.t
   -> poly:bool
+  -> ?typing_flags:Declarations.typing_flags
   -> Proofview.telescope -> t
 
 (* Returns [true] if the considered proof is completed, that is if no goal remain
     to be considered (this does not require that all evars have been solved). *)
 val is_done : t -> bool
 
-(* Like is_done, but this time it really means done (i.e. nothing left to do) *)
-val is_complete : t -> bool
-
 (* Returns the list of partial proofs to initial goals. *)
 val partial_proof : t -> EConstr.constr list
 
 val compact : t -> t
 
+(** [update_sigma_univs] lifts [UState.update_sigma_univs] to the proof *)
+val update_sigma_univs : UGraph.t -> t -> t
+
 (* Returns the proofs (with their type) of the initial goals.
     Raises [UnfinishedProof] is some goals remain to be considered.
     Raises [HasShelvedGoals] if some goals are left on the shelf.
@@ -88,7 +86,7 @@ type open_error_reason =
 
 exception OpenProof of Names.Id.t option * open_error_reason
 
-val return : ?pid:Names.Id.t -> t -> Evd.evar_map
+val return : ?pid:Names.Id.t -> ?allow_given_up:bool -> t -> Evd.evar_map
 
 (*** Focusing actions ***)
 
@@ -131,7 +129,7 @@ val done_cond : ?loose_end:bool -> 'a focus_kind -> 'a focus_condition
 val focus : 'a focus_condition -> 'a -> int -> t -> t
 
 (* focus on goal named id *)
-val focus_id : 'aa focus_condition -> 'a -> Names.Id.t -> t -> t
+val focus_id : 'a focus_condition -> 'a -> Names.Id.t -> t -> t
 
 exception FullyUnfocused
 exception CannotUnfocusThisWay
@@ -182,33 +180,17 @@ val maximal_unfocus : 'a focus_kind -> t -> t
    focused goals. *)
 val unshelve : t -> t
 
+(* Gives a unique identifier to each goal. The identifier is
+   guaranteed to contain no space. *)
+val goal_uid : Evar.t -> string
+
 val pr_proof : t -> Pp.t
 
-(*** Compatibility layer with <=v8.2 ***)
-module V82 : sig
-
-  (* All the subgoals of the proof, including those which are not focused. *)
-  val background_subgoals : t -> Goal.goal list Evd.sigma
-
-  val top_goal : t -> Goal.goal Evd.sigma
-
-  (* returns the existential variable used to start the proof *)
-  val top_evars : t -> Evar.t list
-
-  (* Turns the unresolved evars into goals.
-     Raises [UnfinishedProof] if there are still unsolved goals. *)
-  val grab_evars : t -> t
-
-  (* Implements the Existential command *)
-  val instantiate_evar
-    :  Environ.env
-    -> int
-    -> (Environ.env -> Evd.evar_map -> Glob_term.glob_constr)
-    -> t -> t
-end
+(* All the subgoals of the proof, including those which are not focused. *)
+val background_subgoals : t -> Evar.t list
 
 (* returns the set of all goals in the proof *)
-val all_goals : t -> Goal.Set.t
+val all_goals : t -> Evar.Set.t
 
 (** [solve (SelectNth n) tac] applies tactic [tac] to the [n]th
     subgoal of the current focused proof. [solve SelectAll
