@@ -27,7 +27,7 @@ type term_label =
 
 let compare_term_label t1 t2 = match t1, t2 with
 | GRLabel gr1, GRLabel gr2 -> GlobRef.CanOrd.compare gr1 gr2
-| _ -> pervasives_compare t1 t2 (** OK *)
+| _ -> Stdlib.compare t1 t2 (** OK *)
 
 type 'res lookup_res = 'res Dn.lookup_res = Label of 'res | Nothing | Everything
 
@@ -35,7 +35,7 @@ let eta_reduce = Reductionops.shrink_eta
 
 (* TODO: instead of doing that on patterns we should try to perform it on terms
    before translating them into patterns in Hints. *)
-let rec eta_reduce_pat p = match p with
+let rec eta_reduce_pat (p:constr_pattern) = match p with
 | PLambda (_, _, q) ->
   let f, cl = match eta_reduce_pat q with
   | PApp (f, cl) -> f, cl
@@ -54,6 +54,7 @@ let rec eta_reduce_pat p = match p with
 | PRef _ | PVar _ | PEvar _ | PRel _ | PApp _ | PSoApp _ | PProj _ | PProd _
 | PLetIn _ | PSort _ | PMeta _ | PIf _ | PCase _ | PFix _ | PCoFix _ | PInt _
 | PFloat _ | PArray _ -> p
+| PUninstantiated _ -> .
 
 let decomp_pat p =
   let rec decrec acc = function
@@ -69,11 +70,12 @@ let decomp_pat p =
 let decomp sigma t =
   let rec decrec acc c = match EConstr.kind sigma c with
     | App (f,l) -> decrec (Array.fold_right (fun a l -> a::l) l acc) f
-    | Proj (p, c) ->
+    | Proj (p, _, c) ->
       (* Hack: fake evar to generate [Everything] in the functions below *)
       let hole = mkEvar (Evar.unsafe_of_int (-1), SList.empty) in
       let params = List.make (Projection.npars p) hole in
-      (mkConst (Projection.constant p), params @ c :: acc)
+      (* UnsafeMonomorphic: universes are ignored by the only user *)
+      (UnsafeMonomorphic.mkConst (Projection.constant p), params @ c :: acc)
     | Cast (c1,_,_) -> decrec acc c1
     | _ -> (c,acc)
   in
