@@ -52,11 +52,11 @@ open Names
 open Term
 open Constr
 open Context
-open Vars
 open Declarations
 open Environ
 open Inductive
 open Termops
+open Vars
 open Namegen
 open Inductiveops
 open Ind_tables
@@ -67,17 +67,19 @@ module RelDecl = Context.Rel.Declaration
 
 let hid = Id.of_string "H"
 let xid = Id.of_string "X"
-let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType -> xid
-let fresh env id = next_global_ident_away id Id.Set.empty
+let default_id_of_sort = function InSProp | InProp | InSet -> hid | InType | InQSort -> xid
+let fresh env id avoid =
+  let freshid = next_global_ident_away id avoid in
+  freshid, Id.Set.add freshid avoid
 let with_context_set ctx (b, ctx') =
-  (b, Univ.ContextSet.union ctx ctx')
+  (b, UnivGen.sort_context_union ctx ctx')
 
 let build_dependent_inductive ind (mib,mip) =
   let realargs,_ = List.chop mip.mind_nrealdecls mip.mind_arity_ctxt in
   applist
     (mkIndU ind,
-       Context.Rel.to_extended_list mkRel mip.mind_nrealdecls mib.mind_params_ctxt
-       @ Context.Rel.to_extended_list mkRel 0 realargs)
+       Context.Rel.instance_list mkRel mip.mind_nrealdecls mib.mind_params_ctxt
+       @ Context.Rel.instance_list mkRel 0 realargs)
 
 let named_hd env t na = named_hd env (Evd.from_env env) (EConstr.of_constr t) na
 let name_assumption env = function
@@ -91,22 +93,17 @@ let name_context env hyps =
           let d' = name_assumption env d in (push_rel d' env, d' :: hyps))
        (env,[]) (List.rev hyps))
 
-let my_it_mkLambda_or_LetIn s c = it_mkLambda_or_LetIn c s
+let my_it_mkLambda_or_LetIn s c = Term.it_mkLambda_or_LetIn c s
 let my_it_mkProd_or_LetIn s c = Term.it_mkProd_or_LetIn c s
-let my_it_mkLambda_or_LetIn_name s c =
-  let env = Global.env () in
+let my_it_mkLambda_or_LetIn_name env s c =
   let mkLambda_or_LetIn_name d b = mkLambda_or_LetIn (name_assumption env d) b in
   List.fold_left (fun c d -> mkLambda_or_LetIn_name d c) c s
 
-let get_coq_eq ctx =
-  try
-    let eq = Globnames.destIndRef (Coqlib.lib_ref "core.eq.type") in
-    (* Do not force the lazy if they are not defined *)
-    let eq, ctx = with_context_set ctx
-      (UnivGen.fresh_inductive_instance (Global.env ()) eq) in
-      mkIndU eq, mkConstructUi (eq,1), ctx
-  with Not_found ->
-    user_err Pp.(str "eq not found.")
+let get_coq_eq env ctx =
+  let eq = Globnames.destIndRef (Coqlib.lib_ref "core.eq.type") in
+  let eq, ctx = with_context_set ctx
+      (UnivGen.fresh_inductive_instance env eq) in
+  mkIndU eq, mkConstructUi (eq,1), ctx
 
 let univ_of_eq env eq =
   let open EConstr in
@@ -140,7 +137,7 @@ let get_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported.";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments"; (* This can be relaxed... *)
   let params,constrargs = List.chop mib.mind_nparams constrargs in
@@ -175,7 +172,7 @@ let get_non_sym_eq_data env (ind,u) =
   if List.exists is_local_def realsign then
     error "Inductive equalities with local definitions in arity not supported";
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let _,constrargs = List.chop mib.mind_nparams constrargs in
@@ -197,39 +194,45 @@ let get_non_sym_eq_data env (ind,u) =
 (*                                                                    *)
 (**********************************************************************)
 
-let build_sym_scheme env ind =
+let build_sym_scheme env _handle ind =
   let (ind,u as indu), ctx = UnivGen.fresh_inductive_instance env ind in
   let (mib,mip as specif),nrealargs,realsign,paramsctxt,paramsctxt1 =
     get_sym_eq_data env indu in
   let cstr n =
-    mkApp (mkConstructUi(indu,1),Context.Rel.to_extended_vect mkRel n mib.mind_params_ctxt) in
-  let inds = snd (mind_arity mip) in
-  let varH = fresh env (default_id_of_sort inds) in
+    mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
+  let inds = inductive_sort_family mip in
+  let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
-  let indr = Sorts.relevance_of_sort_family inds in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
+  let p =
+    my_it_mkLambda_or_LetIn_name env
+      (lift_rel_context (nrealargs+1) realsign_ind)
+      (mkApp (mkIndU indu,
+              Array.concat
+                [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
+                 rel_vect 1 nrealargs;
+                 rel_vect (2*nrealargs+2) nrealargs]))
+  in
   let c =
   (my_it_mkLambda_or_LetIn paramsctxt
-  (my_it_mkLambda_or_LetIn_name realsign_ind
-  (mkCase (ci,
-     my_it_mkLambda_or_LetIn_name
-       (lift_rel_context (nrealargs+1) realsign_ind)
-       (mkApp (mkIndU indu,Array.concat
-          [Context.Rel.to_extended_vect mkRel (3*nrealargs+2) paramsctxt1;
-           rel_vect 1 nrealargs;
-           rel_vect (2*nrealargs+2) nrealargs])),
-     mkRel 1 (* varH *),
-       [|cstr (nrealargs+1)|]))))
-  in c, UState.of_context_set ctx
+  (my_it_mkLambda_or_LetIn_name env realsign_ind
+     (mkCase
+        (Inductive.contract_case env
+           (ci,
+            (p,rci),
+            NoInvert,
+            mkRel 1 (* varH *),
+            [|cstr (nrealargs+1)|])))))
+  in
+  c, UState.of_context_set ctx
 
 let sym_scheme_kind =
   declare_individual_scheme_object "_sym_internal"
-  (fun _ ind ->
-    let c, ctx = build_sym_scheme (Global.env() (* side-effect! *)) ind in
-      (c, ctx))
+  build_sym_scheme
 
 (**********************************************************************)
 (* Build the involutivity of symmetry for an inductive type           *)
@@ -247,63 +250,63 @@ let sym_scheme_kind =
 (*                                                                    *)
 (**********************************************************************)
 
-let const_of_scheme kind env ind ctx =
-  let sym_scheme = match lookup_scheme kind ind with Some cst -> cst | None -> assert false in
+let const_of_scheme kind env handle ind ctx =
+  let sym_scheme = match local_lookup_scheme handle kind ind with Some cst -> cst | None -> assert false in
   let sym, ctx = with_context_set ctx
-    (UnivGen.fresh_constant_instance (Global.env()) sym_scheme) in
+    (UnivGen.fresh_constant_instance env sym_scheme) in
     mkConstU sym, ctx
 
-let build_sym_involutive_scheme env ind =
+let build_sym_involutive_scheme env handle ind =
   let (ind,u as indu), ctx = UnivGen.fresh_inductive_instance env ind in
   let (mib,mip as specif),nrealargs,realsign,paramsctxt,paramsctxt1 =
     get_sym_eq_data env indu in
-  let eq,eqrefl,ctx = get_coq_eq ctx in
-  let sym, ctx = const_of_scheme sym_scheme_kind env ind ctx in
-  let cstr n = mkApp (mkConstructUi (indu,1),Context.Rel.to_extended_vect mkRel n paramsctxt) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
-  let varH = fresh env (default_id_of_sort inds) in
+  let eq,eqrefl,ctx = get_coq_eq env ctx in
+  let sym, ctx = const_of_scheme sym_scheme_kind env handle ind ctx in
+  let cstr n = mkApp (mkConstructUi (indu,1),Context.Rel.instance mkRel n paramsctxt) in
+  let inds = inductive_sort_family mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
+  let varH,_ = fresh env (default_id_of_sort inds) Id.Set.empty in
   let applied_ind = build_dependent_inductive indu specif in
   let applied_ind_C =
     mkApp
       (mkIndU indu, Array.append
-         (Context.Rel.to_extended_vect mkRel (nrealargs+1) mib.mind_params_ctxt)
+         (Context.Rel.instance mkRel (nrealargs+1) mib.mind_params_ctxt)
          (rel_vect (nrealargs+1) nrealargs)) in
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let c =
     (my_it_mkLambda_or_LetIn paramsctxt
-     (my_it_mkLambda_or_LetIn_name realsign_ind
-      (mkCase (ci,
-               my_it_mkLambda_or_LetIn_name
-               (lift_rel_context (nrealargs+1) realsign_ind)
-               (mkApp (eq,[|
-               mkApp
-               (mkIndU indu, Array.concat
-               [Context.Rel.to_extended_vect mkRel (3*nrealargs+2) paramsctxt1;
+     (my_it_mkLambda_or_LetIn_name env realsign_ind
+      (mkCase (Inductive.contract_case env (ci,
+                (my_it_mkLambda_or_LetIn_name env
+                (lift_rel_context (nrealargs+1) realsign_ind)
+                (mkApp (eq,[|
+                mkApp
+                (mkIndU indu, Array.concat
+                [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
                 rel_vect (2*nrealargs+2) nrealargs;
                 rel_vect 1 nrealargs]);
                mkApp (sym,Array.concat
-               [Context.Rel.to_extended_vect mkRel (3*nrealargs+2) paramsctxt1;
+               [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
                 rel_vect 1 nrealargs;
                 rel_vect (2*nrealargs+2) nrealargs;
                 [|mkApp (sym,Array.concat
-                [Context.Rel.to_extended_vect mkRel (3*nrealargs+2) paramsctxt1;
+                [Context.Rel.instance mkRel (3*nrealargs+2) paramsctxt1;
                  rel_vect (2*nrealargs+2) nrealargs;
                  rel_vect 1 nrealargs;
                  [|mkRel 1|]])|]]);
-               mkRel 1|])),
+               mkRel 1|])), rci),
+               NoInvert,
                mkRel 1 (* varH *),
-               [|mkApp(eqrefl,[|applied_ind_C;cstr (nrealargs+1)|])|]))))
+               [|mkApp(eqrefl,[|applied_ind_C;cstr (nrealargs+1)|])|])))))
   in (c, UState.of_context_set ctx)
 
 let sym_involutive_scheme_kind =
   declare_individual_scheme_object "_sym_involutive"
-  ~deps:(fun ind -> [SchemeIndividualDep (ind, sym_scheme_kind)])
-  (fun _ ind ->
-    build_sym_involutive_scheme (Global.env() (* side-effect! *)) ind)
+  ~deps:(fun _ ind -> [SchemeIndividualDep (ind, sym_scheme_kind)])
+  build_sym_involutive_scheme
 
 (**********************************************************************)
 (* Build the left-to-right rewriting lemma for conclusion associated  *)
@@ -365,31 +368,31 @@ let sym_involutive_scheme_kind =
 (*                                                                    *)
 (**********************************************************************)
 
-let build_l2r_rew_scheme dep env ind kind =
+let build_l2r_rew_scheme dep env handle ind kind =
   let (ind,u as indu), ctx = UnivGen.fresh_inductive_instance env ind in
   let (mib,mip as specif),nrealargs,realsign,paramsctxt,paramsctxt1 =
     get_sym_eq_data env indu in
-  let sym, ctx = const_of_scheme sym_scheme_kind env ind ctx in
-  let sym_involutive, ctx = const_of_scheme sym_involutive_scheme_kind env ind ctx in
-  let eq,eqrefl,ctx = get_coq_eq ctx in
+  let sym, ctx = const_of_scheme sym_scheme_kind env handle ind ctx in
+  let sym_involutive, ctx = const_of_scheme sym_involutive_scheme_kind env handle ind ctx in
+  let eq,eqrefl,ctx = get_coq_eq env ctx in
   let cstr n p =
     mkApp (mkConstructUi(indu,1),
-      Array.concat [Context.Rel.to_extended_vect mkRel n paramsctxt1;
+      Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
-  let varH = fresh env (default_id_of_sort inds) in
-  let varHC = fresh env (Id.of_string "HC") in
-  let varP = fresh env (Id.of_string "P") in
+  let inds = inductive_sort_family mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
+  let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
+  let varHC,avoid = fresh env (Id.of_string "HC") avoid in
+  let varP,_ = fresh env (Id.of_string "P") avoid in
   let applied_ind = build_dependent_inductive indu specif in
   let applied_ind_P =
     mkApp (mkIndU indu, Array.concat
-       [Context.Rel.to_extended_vect mkRel (3*nrealargs) paramsctxt1;
+       [Context.Rel.instance mkRel (3*nrealargs) paramsctxt1;
         rel_vect 0 nrealargs;
         rel_vect nrealargs nrealargs]) in
   let applied_ind_G =
     mkApp (mkIndU indu, Array.concat
-       [Context.Rel.to_extended_vect mkRel (3*nrealargs+3) paramsctxt1;
+       [Context.Rel.instance mkRel (3*nrealargs+3) paramsctxt1;
         rel_vect (nrealargs+3) nrealargs;
         rel_vect 0 nrealargs]) in
   let realsign_P = lift_rel_context nrealargs realsign in
@@ -400,21 +403,21 @@ let build_l2r_rew_scheme dep env ind kind =
                       lift_rel_context (nrealargs+3) realsign) in
   let applied_sym_C n =
      mkApp(sym,
-       Array.append (Context.Rel.to_extended_vect mkRel n mip.mind_arity_ctxt) [|mkVar varH|]) in
+       Array.append (Context.Rel.instance mkRel n mip.mind_arity_ctxt) [|mkVar varH|]) in
   let applied_sym_G =
      mkApp(sym,
-       Array.concat [Context.Rel.to_extended_vect mkRel (nrealargs*3+4) paramsctxt1;
+       Array.concat [Context.Rel.instance mkRel (nrealargs*3+4) paramsctxt1;
                      rel_vect (nrealargs+4) nrealargs;
                      rel_vect 1 nrealargs;
                      [|mkRel 1|]]) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
-  let cieq = make_case_info (Global.env()) (fst (destInd eq)) rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
+  let cieq = make_case_info env (fst (destInd eq)) RegularStyle in
   let applied_PC =
-    mkApp (mkVar varP,Array.append (Context.Rel.to_extended_vect mkRel 1 realsign)
+    mkApp (mkVar varP,Array.append (Context.Rel.instance mkRel 1 realsign)
            (if dep then [|cstr (2*nrealargs+1) 1|] else [||])) in
   let applied_PG =
     mkApp (mkVar varP,Array.append (rel_vect 1 nrealargs)
@@ -424,35 +427,38 @@ let build_l2r_rew_scheme dep env ind kind =
            (if dep then [|mkRel 2|] else [||])) in
   let applied_sym_sym =
          mkApp (sym,Array.concat
-           [Context.Rel.to_extended_vect mkRel (2*nrealargs+4) paramsctxt1;
+           [Context.Rel.instance mkRel (2*nrealargs+4) paramsctxt1;
             rel_vect 4 nrealargs;
             rel_vect (nrealargs+4) nrealargs;
             [|mkApp (sym,Array.concat
-              [Context.Rel.to_extended_vect mkRel (2*nrealargs+4) paramsctxt1;
+              [Context.Rel.instance mkRel (2*nrealargs+4) paramsctxt1;
                rel_vect (nrealargs+4) nrealargs;
                rel_vect 4 nrealargs;
                [|mkRel 2|]])|]]) in
   let main_body =
-    mkCase (ci,
-      my_it_mkLambda_or_LetIn_name realsign_ind_G applied_PG,
-      applied_sym_C 3,
-      [|mkVar varHC|]) in
+    mkCase (Inductive.contract_case env (ci,
+            (my_it_mkLambda_or_LetIn_name env realsign_ind_G applied_PG, rci),
+            NoInvert,
+            applied_sym_C 3,
+            [|mkVar varHC|]))
+  in
   let c =
   (my_it_mkLambda_or_LetIn paramsctxt
-  (my_it_mkLambda_or_LetIn_name realsign
+  (my_it_mkLambda_or_LetIn_name env realsign
   (mkNamedLambda (make_annot varP indr)
     (my_it_mkProd_or_LetIn (if dep then realsign_ind_P else realsign_P) s)
   (mkNamedLambda (make_annot varHC indr) applied_PC
   (mkNamedLambda (make_annot varH indr) (lift 2 applied_ind)
-  (if dep then (* we need a coercion *)
-     mkCase (cieq,
-       mkLambda (make_annot (Name varH) indr,lift 3 applied_ind,
+     (if dep then (* we need a coercion *)
+     mkCase (Inductive.contract_case env (cieq,
+       (mkLambda (make_annot (Name varH) indr,lift 3 applied_ind,
          mkLambda (make_annot Anonymous indr,
                    mkApp (eq,[|lift 4 applied_ind;applied_sym_sym;mkRel 1|]),
-                   applied_PR)),
+                   applied_PR)), rci),
+       NoInvert,
        mkApp (sym_involutive,
-         Array.append (Context.Rel.to_extended_vect mkRel 3 mip.mind_arity_ctxt) [|mkVar varH|]),
-       [|main_body|])
+         Array.append (Context.Rel.instance mkRel 3 mip.mind_arity_ctxt) [|mkVar varH|]),
+       [|main_body|]))
    else
      main_body))))))
   in (c, UState.of_context_set ctx)
@@ -489,22 +495,22 @@ let build_l2r_forward_rew_scheme dep env ind kind =
     get_sym_eq_data env indu in
   let cstr n p =
     mkApp (mkConstructUi(indu,1),
-      Array.concat [Context.Rel.to_extended_vect mkRel n paramsctxt1;
+      Array.concat [Context.Rel.instance mkRel n paramsctxt1;
                     rel_vect p nrealargs]) in
-  let inds = snd (mind_arity mip) in
-  let indr = Sorts.relevance_of_sort_family inds in
-  let varH = fresh env (default_id_of_sort inds) in
-  let varHC = fresh env (Id.of_string "HC") in
-  let varP = fresh env (Id.of_string "P") in
+  let inds = inductive_sort_family mip in
+  let indr = UVars.subst_instance_relevance u mip.mind_relevance in
+  let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
+  let varHC,avoid = fresh env (Id.of_string "HC") avoid in
+  let varP,_ = fresh env (Id.of_string "P") avoid in
   let applied_ind = build_dependent_inductive indu specif in
   let applied_ind_P =
     mkApp (mkIndU indu, Array.concat
-       [Context.Rel.to_extended_vect mkRel (4*nrealargs+2) paramsctxt1;
+       [Context.Rel.instance mkRel (4*nrealargs+2) paramsctxt1;
         rel_vect 0 nrealargs;
         rel_vect (nrealargs+1) nrealargs]) in
   let applied_ind_P' =
     mkApp (mkIndU indu, Array.concat
-       [Context.Rel.to_extended_vect mkRel (3*nrealargs+1) paramsctxt1;
+       [Context.Rel.instance mkRel (3*nrealargs+1) paramsctxt1;
         rel_vect 0 nrealargs;
         rel_vect (2*nrealargs+1) nrealargs]) in
   let realsign_P n = lift_rel_context (nrealargs*n+n) realsign in
@@ -513,10 +519,10 @@ let build_l2r_forward_rew_scheme dep env ind kind =
   let realsign_ind_P n aP =
     name_context env ((LocalAssum (make_annot (Name varH) indr,aP))::realsign_P n) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let applied_PC =
     mkApp (mkVar varP,Array.append
            (rel_vect (nrealargs*2+3) nrealargs)
@@ -531,21 +537,22 @@ let build_l2r_forward_rew_scheme dep env ind kind =
            (if dep then [|cstr (3*nrealargs+4) 3|] else [||])) in
   let c =
   (my_it_mkLambda_or_LetIn paramsctxt
-  (my_it_mkLambda_or_LetIn_name realsign
+  (my_it_mkLambda_or_LetIn_name env realsign
   (mkNamedLambda (make_annot varH indr) applied_ind
-  (mkCase (ci,
-     my_it_mkLambda_or_LetIn_name
+  (mkCase (Inductive.contract_case env (ci,
+     (my_it_mkLambda_or_LetIn_name env
        (lift_rel_context (nrealargs+1) realsign_ind)
        (mkNamedProd (make_annot varP indr)
          (my_it_mkProd_or_LetIn
            (if dep then realsign_ind_P 2 applied_ind_P else realsign_P 2) s)
-       (mkNamedProd (make_annot varHC indr) applied_PC applied_PG)),
+       (mkNamedProd (make_annot varHC indr) applied_PC applied_PG)), rci),
+     NoInvert,
      (mkVar varH),
      [|mkNamedLambda (make_annot varP indr)
         (my_it_mkProd_or_LetIn
           (if dep then realsign_ind_P 1 applied_ind_P' else realsign_P 2) s)
       (mkNamedLambda (make_annot varHC indr) applied_PC'
-        (mkVar varHC))|])))))
+        (mkVar varHC))|]))))))
   in c, UState.of_context_set ctx
 
 (**********************************************************************)
@@ -583,44 +590,45 @@ let build_r2l_forward_rew_scheme dep env ind kind =
   let ((mib,mip as specif),constrargs,realsign,paramsctxt,nrealargs) =
     get_non_sym_eq_data env indu in
   let cstr n =
-    mkApp (mkConstructUi(indu,1),Context.Rel.to_extended_vect mkRel n mib.mind_params_ctxt) in
+    mkApp (mkConstructUi(indu,1),Context.Rel.instance mkRel n mib.mind_params_ctxt) in
   let constrargs_cstr = constrargs@[cstr 0] in
-  let inds = snd (mind_arity mip) in
+  let inds = inductive_sort_family mip in
   let indr = Sorts.relevance_of_sort_family inds in
-  let varH = fresh env (default_id_of_sort inds) in
-  let varHC = fresh env (Id.of_string "HC") in
-  let varP = fresh env (Id.of_string "P") in
+  let varH,avoid = fresh env (default_id_of_sort inds) Id.Set.empty in
+  let varHC,avoid = fresh env (Id.of_string "HC") avoid in
+  let varP,_ = fresh env (Id.of_string "P") avoid in
   let applied_ind = build_dependent_inductive indu specif in
   let realsign_ind =
     name_context env ((LocalAssum (make_annot (Name varH) indr,applied_ind))::realsign) in
   let s, ctx' = UnivGen.fresh_sort_in_family kind in
-  let ctx = Univ.ContextSet.union ctx ctx' in
+  let ctx = UnivGen.sort_context_union ctx ctx' in
   let s = mkSort s in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
+  let ci = make_case_info env ind RegularStyle in
   let applied_PC =
     applist (mkVar varP,if dep then constrargs_cstr else constrargs) in
   let applied_PG =
     mkApp (mkVar varP,
-           if dep then Context.Rel.to_extended_vect mkRel 0 realsign_ind
-           else Context.Rel.to_extended_vect mkRel 1 realsign) in
+           if dep then Context.Rel.instance mkRel 0 realsign_ind
+           else Context.Rel.instance mkRel 1 realsign) in
   let c =
   (my_it_mkLambda_or_LetIn paramsctxt
-  (my_it_mkLambda_or_LetIn_name realsign_ind
+  (my_it_mkLambda_or_LetIn_name env realsign_ind
   (mkNamedLambda (make_annot varP indr)
     (my_it_mkProd_or_LetIn (lift_rel_context (nrealargs+1)
                              (if dep then realsign_ind else realsign)) s)
   (mkNamedLambda (make_annot varHC indr) (lift 1 applied_PG)
   (mkApp
-    (mkCase (ci,
-       my_it_mkLambda_or_LetIn_name
+    (mkCase (Inductive.contract_case env (ci,
+       (my_it_mkLambda_or_LetIn_name env
          (lift_rel_context (nrealargs+3) realsign_ind)
-         (mkArrow applied_PG indr (lift (2*nrealargs+5) applied_PC)),
+         (mkArrow applied_PG indr (lift (2*nrealargs+5) applied_PC)), rci),
+       NoInvert,
        mkRel 3 (* varH *),
        [|mkLambda
           (make_annot (Name varHC) indr,
            lift (nrealargs+3) applied_PC,
-           mkRel 1)|]),
+           mkRel 1)|])),
     [|mkVar varHC|]))))))
   in c, UState.of_context_set ctx
 
@@ -640,12 +648,11 @@ let build_r2l_forward_rew_scheme dep env ind kind =
 (*                                                                    *)
 (**********************************************************************)
 
-let fix_r2l_forward_rew_scheme (c, ctx') =
-  let env = Global.env () in
+let fix_r2l_forward_rew_scheme env (c, ctx') =
   let sigma = Evd.from_env env in
   let t = Retyping.get_type_of env sigma (EConstr.of_constr c) in
   let t = EConstr.Unsafe.to_constr t in
-  let ctx,_ = decompose_prod_assum t in
+  let ctx,_ = decompose_prod_decls t in
   match ctx with
   | hp :: p :: ind :: indargs ->
      let c' =
@@ -655,7 +662,7 @@ let fix_r2l_forward_rew_scheme (c, ctx') =
             (mkLambda_or_LetIn (RelDecl.map_constr (lift 2) ind)
               (EConstr.Unsafe.to_constr (Reductionops.whd_beta env sigma
                 (EConstr.of_constr (applist (c,
-                  Context.Rel.to_extended_list mkRel 3 indargs @ [mkRel 1;mkRel 3;mkRel 2]))))))))
+                  Context.Rel.instance_list mkRel 3 indargs @ [mkRel 1;mkRel 3;mkRel 2]))))))))
       in c', ctx'
   | _ -> anomaly (Pp.str "Ill-formed non-dependent left-to-right rewriting scheme.")
 
@@ -684,7 +691,8 @@ let build_r2l_rew_scheme dep env ind k =
   let sigma = Evd.from_env env in
   let (sigma, indu) = Evd.fresh_inductive_instance env sigma ind in
   let (sigma, c) = build_case_analysis_scheme env sigma indu dep k in
-    c, Evd.evar_universe_context sigma
+  let (c, _) = Indrec.eval_case_analysis c in
+  EConstr.Unsafe.to_constr c, Evd.evar_universe_context sigma
 
 (**********************************************************************)
 (* Register the rewriting schemes                                     *)
@@ -698,11 +706,11 @@ let build_r2l_rew_scheme dep env ind k =
 (**********************************************************************)
 let rew_l2r_dep_scheme_kind =
   declare_individual_scheme_object "_rew_r_dep"
-  ~deps:(fun ind -> [
+  ~deps:(fun _ ind -> [
     SchemeIndividualDep (ind, sym_scheme_kind);
     SchemeIndividualDep (ind, sym_involutive_scheme_kind);
   ])
-  (fun _ ind -> build_l2r_rew_scheme true (Global.env()) ind InType)
+  (fun env handle ind -> build_l2r_rew_scheme true env handle ind InType)
 
 (**********************************************************************)
 (* Dependent rewrite from right-to-left in conclusion                 *)
@@ -712,7 +720,7 @@ let rew_l2r_dep_scheme_kind =
 (**********************************************************************)
 let rew_r2l_dep_scheme_kind =
   declare_individual_scheme_object "_rew_dep"
-  (fun _ ind -> build_r2l_rew_scheme true (Global.env()) ind InType)
+  (fun env _ ind -> build_r2l_rew_scheme true env ind InType)
 
 (**********************************************************************)
 (* Dependent rewrite from right-to-left in hypotheses                 *)
@@ -722,7 +730,7 @@ let rew_r2l_dep_scheme_kind =
 (**********************************************************************)
 let rew_r2l_forward_dep_scheme_kind =
   declare_individual_scheme_object "_rew_fwd_dep"
-  (fun _ ind -> build_r2l_forward_rew_scheme true (Global.env()) ind InType)
+  (fun env _ ind -> build_r2l_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
 (* Dependent rewrite from left-to-right in hypotheses                 *)
@@ -732,7 +740,7 @@ let rew_r2l_forward_dep_scheme_kind =
 (**********************************************************************)
 let rew_l2r_forward_dep_scheme_kind =
   declare_individual_scheme_object "_rew_fwd_r_dep"
-  (fun _ ind -> build_l2r_forward_rew_scheme true (Global.env()) ind InType)
+  (fun env _ ind -> build_l2r_forward_rew_scheme true env ind InType)
 
 (**********************************************************************)
 (* Non-dependent rewrite from either left-to-right in conclusion or   *)
@@ -745,8 +753,8 @@ let rew_l2r_forward_dep_scheme_kind =
 (**********************************************************************)
 let rew_l2r_scheme_kind =
   declare_individual_scheme_object "_rew_r"
-  (fun _ ind -> fix_r2l_forward_rew_scheme
-     (build_r2l_forward_rew_scheme false (Global.env()) ind InType))
+  (fun env _ ind -> fix_r2l_forward_rew_scheme env
+     (build_r2l_forward_rew_scheme false env ind InType))
 
 (**********************************************************************)
 (* Non-dependent rewrite from either right-to-left in conclusion or   *)
@@ -756,7 +764,7 @@ let rew_l2r_scheme_kind =
 (**********************************************************************)
 let rew_r2l_scheme_kind =
   declare_individual_scheme_object "_rew"
-  (fun _ ind -> build_r2l_rew_scheme false (Global.env()) ind InType)
+  (fun env _ ind -> build_r2l_rew_scheme false env ind InType)
 
 (* End of rewriting schemes *)
 
@@ -795,48 +803,52 @@ let build_congr env (eq,refl,ctx) ind =
     RelDecl.get_type decl, RelDecl.get_relevance decl
   in
   let constrsign,ccl = mip.mind_nf_lc.(0) in
-  let _,constrargs = decompose_app ccl in
+  let _,constrargs = decompose_app_list ccl in
   if not (Int.equal (Context.Rel.length constrsign) (Context.Rel.length mib.mind_params_ctxt)) then
     error "Constructor must have no arguments";
   let b = List.nth constrargs (i + mib.mind_nparams - 1) in
-  let varB = fresh env (Id.of_string "B") in
-  let varH = fresh env (Id.of_string "H") in
-  let varf = fresh env (Id.of_string "f") in
+  let varB,avoid = fresh env (Id.of_string "B") Id.Set.empty in
+  let varH,avoid = fresh env (Id.of_string "H") avoid in
+  let varf,avoid = fresh env (Id.of_string "f") avoid in
   let rci = Sorts.Relevant in (* TODO relevance *)
-  let ci = make_case_info (Global.env()) ind rci RegularStyle in
-  let uni, ctx = Univ.extend_in_context_set (UnivGen.new_global_univ ()) ctx in
-  let ctx = (fst ctx, Univ.enforce_leq uni (univ_of_eq env eq) (snd ctx)) in
+  let ci = make_case_info env ind RegularStyle in
+  let uni, ctx' = UnivGen.new_global_univ () in
+  let ctx =
+    let (qs,us),csts = ctx in
+    let us, csts = Univ.ContextSet.union (us,csts) ctx' in
+    ((qs, us), UnivSubst.enforce_leq uni (univ_of_eq env eq) csts) in
   let c =
   my_it_mkLambda_or_LetIn paramsctxt
      (mkNamedLambda (make_annot varB Sorts.Relevant) (mkType uni)
      (mkNamedLambda (make_annot varf Sorts.Relevant) (mkArrow (lift 1 ty) tyr (mkVar varB))
-     (my_it_mkLambda_or_LetIn_name (lift_rel_context 2 realsign)
+     (my_it_mkLambda_or_LetIn_name env (lift_rel_context 2 realsign)
      (mkNamedLambda (make_annot varH Sorts.Relevant)
         (applist
            (mkIndU indu,
-            Context.Rel.to_extended_list mkRel (mip.mind_nrealargs+2) paramsctxt @
-            Context.Rel.to_extended_list mkRel 0 realsign))
-     (mkCase (ci,
-       my_it_mkLambda_or_LetIn_name
+            Context.Rel.instance_list mkRel (mip.mind_nrealargs+2) paramsctxt @
+            Context.Rel.instance_list mkRel 0 realsign))
+     (mkCase (Inductive.contract_case env (ci,
+       (my_it_mkLambda_or_LetIn_name env
          (lift_rel_context (mip.mind_nrealargs+3) realsign)
          (mkLambda
            (make_annot Anonymous Sorts.Relevant,
             applist
              (mkIndU indu,
-                Context.Rel.to_extended_list mkRel (2*mip.mind_nrealdecls+3)
+                Context.Rel.instance_list mkRel (2*mip.mind_nrealdecls+3)
                   paramsctxt
-                @ Context.Rel.to_extended_list mkRel 0 realsign),
+                @ Context.Rel.instance_list mkRel 0 realsign),
             mkApp (eq,
               [|mkVar varB;
                 mkApp (mkVar varf, [|lift (2*mip.mind_nrealdecls+4) b|]);
-                mkApp (mkVar varf, [|mkRel (mip.mind_nrealargs - i + 2)|])|]))),
+                mkApp (mkVar varf, [|mkRel (mip.mind_nrealargs - i + 2)|])|]))), rci),
+       NoInvert,
        mkVar varH,
        [|mkApp (refl,
           [|mkVar varB;
-            mkApp (mkVar varf, [|lift (mip.mind_nrealargs+3) b|])|])|]))))))
+            mkApp (mkVar varf, [|lift (mip.mind_nrealargs+3) b|])|])|])))))))
   in c, UState.of_context_set ctx
 
 let congr_scheme_kind = declare_individual_scheme_object "_congr"
-  (fun _ ind ->
+  (fun env _ ind ->
      (* May fail if equality is not defined *)
-   build_congr (Global.env()) (get_coq_eq Univ.ContextSet.empty) ind)
+   build_congr env (get_coq_eq env UnivGen.empty_sort_context) ind)
