@@ -14,8 +14,6 @@ open EConstr
 open Environ
 open Evd
 open Tactypes
-open Clenv
-open Pattern
 open Typeclasses
 
 (** {6 General functions. } *)
@@ -30,6 +28,8 @@ val secvars_of_hyps : ('c, 't) Context.Named.pt -> Id.Pred.t
 
 val empty_hint_info : 'a Typeclasses.hint_info_gen
 
+val hint_cat : Libobject.category
+
 (** Pre-created hint databases *)
 
 type 'a hint_ast =
@@ -37,11 +37,12 @@ type 'a hint_ast =
   | ERes_pf    of 'a (* Hint EApply *)
   | Give_exact of 'a
   | Res_pf_THEN_trivial_fail of 'a (* Hint Immediate *)
-  | Unfold_nth of evaluable_global_reference       (* Hint Unfold *)
-  | Extern     of Genarg.glob_generic_argument       (* Hint Extern *)
+  | Unfold_nth of Tacred.evaluable_global_reference       (* Hint Unfold *)
+  | Extern     of Pattern.constr_pattern option * Genarg.glob_generic_argument       (* Hint Extern *)
 
 type hint
-type raw_hint = constr * types * Univ.ContextSet.t
+
+val hint_as_term : hint -> UnivGen.sort_context_set option * constr
 
 type 'a hints_path_atom_gen =
   | PathHints of 'a list
@@ -51,31 +52,24 @@ type 'a hints_path_atom_gen =
 type hints_path_atom = GlobRef.t hints_path_atom_gen
 type hint_db_name = string
 
-type 'a with_metadata = private
-  { pri     : int
-  (** A number lower is higher priority *)
-  ; poly    : bool
-  (** Is the hint polymorpic and hence should be refreshed at each application *)
-  ; pat     : constr_pattern option
-  (** A pattern for the concl of the Goal *)
-  ; name    : hints_path_atom
-  (** A potential name to refer to the hint *)
-  ; db : string option
-  (** The database from which the hint comes *)
-  ; secvars : Id.Pred.t
-  (** The set of section variables the hint depends on *)
-  ; code    : 'a
-  (** the tactic to apply when the concl matches pat *)
-  }
+module FullHint :
+sig
+  type t
+  val priority : t -> int
+  val pattern : t -> Pattern.constr_pattern option
+  val database : t -> string option
+  val run : t -> (hint hint_ast -> 'r Proofview.tactic) -> 'r Proofview.tactic
+  val name : t -> GlobRef.t option
+  val print : env -> evar_map -> t -> Pp.t
+  val subgoals : t -> int option
 
-type full_hint = hint with_metadata
-
-type search_entry
+  (** This function is for backward compatibility only, not to use in newly
+    written code. *)
+  val repr : t -> hint hint_ast
+end
 
 (** The head may not be bound. *)
 
-type hint_entry
-
 type hint_mode =
   | ModeInput (* No evars *)
   | ModeNoHeadEvar (* No evar at the head *)
@@ -109,50 +103,44 @@ val glob_hints_path_atom :
 val glob_hints_path :
   Libnames.qualid hints_path_gen -> GlobRef.t hints_path_gen
 
+type mode_match =
+  | NoMode
+  | WithMode of hint_mode array
+
 type 'a with_mode =
-  | ModeMatch of 'a
+  | ModeMatch of mode_match * 'a
   | ModeMismatch
 
 module Hint_db :
   sig
     type t
     val empty : ?name:hint_db_name -> TransparentState.t -> bool -> t
-    val find : GlobRef.t -> t -> search_entry
 
     (** All hints which have no pattern.
      * [secvars] represent the set of section variables that
      * can be used in the hint. *)
-    val map_none : secvars:Id.Pred.t -> t -> full_hint list
+    val map_none : secvars:Id.Pred.t -> t -> FullHint.t list
 
     (** All hints associated to the reference *)
-    val map_all : secvars:Id.Pred.t -> GlobRef.t -> t -> full_hint list
-
-    (** All hints associated to the reference, respecting modes if evars appear in the
-        arguments, _not_ using the discrimination net.
-        Returns a [ModeMismatch] if there are declared modes and none matches.
-        *)
-    val map_existential : evar_map -> secvars:Id.Pred.t ->
-      (GlobRef.t * constr array) -> constr -> t -> full_hint list with_mode
+    val map_all : secvars:Id.Pred.t -> GlobRef.t -> t -> FullHint.t list
 
     (** All hints associated to the reference, respecting modes if evars appear in the
         arguments and using the discrimination net.
         Returns a [ModeMismatch] if there are declared modes and none matches. *)
-    val map_eauto : evar_map -> secvars:Id.Pred.t -> (GlobRef.t * constr array) -> constr -> t -> full_hint list with_mode
+    val map_eauto : env -> evar_map -> secvars:Id.Pred.t -> (GlobRef.t * constr array) -> constr -> t -> FullHint.t list with_mode
 
     (** All hints associated to the reference.
         Precondition: no evars should appear in the arguments, so no modes
         are checked. *)
-    val map_auto : evar_map -> secvars:Id.Pred.t ->
-       (GlobRef.t * constr array) -> constr -> t -> full_hint list
+    val map_auto : env -> evar_map -> secvars:Id.Pred.t ->
+       (GlobRef.t * constr array) -> constr -> t -> FullHint.t list
 
-    val add_one : env -> evar_map -> hint_entry -> t -> t
-    val add_list : env -> evar_map -> hint_entry list -> t -> t
-    val remove_one : GlobRef.t -> t -> t
-    val remove_list : GlobRef.t list -> t -> t
+    val remove_one : Environ.env -> GlobRef.t -> t -> t
+    val remove_list : Environ.env -> GlobRef.t list -> t -> t
     val iter : (GlobRef.t option ->
-                hint_mode array list -> full_hint list -> unit) -> t -> unit
+                hint_mode array list -> FullHint.t list -> unit) -> t -> unit
 
-    val fold : (GlobRef.t option -> hint_mode array list -> full_hint list -> 'a -> 'a) -> t -> 'a -> 'a
+    val fold : (GlobRef.t option -> hint_mode array list -> FullHint.t list -> 'a -> 'a) -> t -> 'a -> 'a
 
     val use_dn : t -> bool
     val transparent_state : t -> TransparentState.t
@@ -171,16 +159,14 @@ type hint_db = Hint_db.t
 
 type hnf = bool
 
-type hint_term =
-  | IsGlobRef of GlobRef.t
-  | IsConstr of constr * Univ.ContextSet.t [@ocaml.deprecated "Declare a hint constant instead"]
+type hint_term
 
 type hints_entry =
-  | HintsResolveEntry of (hint_info * bool * hnf * hints_path_atom * hint_term) list
-  | HintsImmediateEntry of (hints_path_atom * bool * hint_term) list
+  | HintsResolveEntry of (hint_info * hnf * hint_term) list
+  | HintsImmediateEntry of hint_term list
   | HintsCutEntry of hints_path
-  | HintsUnfoldEntry of evaluable_global_reference list
-  | HintsTransparencyEntry of evaluable_global_reference hints_transparency_target * bool
+  | HintsUnfoldEntry of Tacred.evaluable_global_reference list
+  | HintsTransparencyEntry of Tacred.evaluable_global_reference hints_transparency_target * bool
   | HintsModeEntry of GlobRef.t * hint_mode list
   | HintsExternEntry of hint_info * Genarg.glob_generic_argument
 
@@ -188,6 +174,11 @@ val searchtable_map : hint_db_name -> hint_db
 
 val searchtable_add : (hint_db_name * hint_db) -> unit
 
+type hint_locality = Local | Export | SuperGlobal
+
+val default_hint_locality : unit -> hint_locality
+(** Warns *)
+
 (** [create_hint_db local name st use_dn].
    [st] is a transparency state for unification using this db
    [use_dn] switches the use of the discrimination net for all hints
@@ -195,46 +186,18 @@ val searchtable_add : (hint_db_name * hint_db) -> unit
 
 val create_hint_db : bool -> hint_db_name -> TransparentState.t -> bool -> unit
 
-val remove_hints : bool -> hint_db_name list -> GlobRef.t list -> unit
+val remove_hints : locality:hint_locality -> hint_db_name list -> GlobRef.t list -> unit
 
 val current_db_names : unit -> String.Set.t
 
 val current_pure_db : unit -> hint_db list
 
-val add_hints : locality:Goptions.option_locality -> hint_db_name list -> hints_entry -> unit
+val add_hints : locality:hint_locality -> hint_db_name list -> hints_entry -> unit
 
-val prepare_hint : bool (* Check no remaining evars *) ->
-  env -> evar_map -> evar_map * constr -> (constr * Univ.ContextSet.t)
+val hint_globref : GlobRef.t -> hint_term
 
-(** [make_exact_entry info (c, ctyp, ctx)].
-   [c] is the term given as an exact proof to solve the goal;
-   [ctyp] is the type of [c].
-   [ctx] is its (refreshable) universe context.
-   In info:
-   [hint_priority] is the hint's desired priority, it is 0 if unspecified
-   [hint_pattern] is the hint's desired pattern, it is inferred if not specified
-*)
-
-val make_exact_entry : env -> evar_map -> hint_info -> poly:bool -> ?name:hints_path_atom ->
-  (constr * types * Univ.ContextSet.t) -> hint_entry
-
-(** [make_apply_entry (eapply,hnf,verbose) info (c,cty,ctx))].
-   [eapply] is true if this hint will be used only with EApply;
-   [hnf] should be true if we should expand the head of cty before searching for
-   products;
-   [c] is the term given as an exact proof to solve the goal;
-   [cty] is the type of [c].
-   [ctx] is its (refreshable) universe context.
-   In info:
-   [hint_priority] is the hint's desired priority, it is computed as the number of products in [cty]
-   if unspecified
-   [hint_pattern] is the hint's desired pattern, it is inferred from the conclusion of [cty]
-   if not specified
-*)
-
-val make_apply_entry :
-  env -> evar_map -> bool * bool * bool -> hint_info -> poly:bool -> ?name:hints_path_atom ->
-  (constr * types * Univ.ContextSet.t) -> hint_entry
+val hint_constr : constr * UnivGen.sort_context_set option -> hint_term
+[@ocaml.deprecated "Declare a hint constant instead"]
 
 (** A constr which is Hint'ed will be:
    - (1) used as an Exact, if it does not start with a product
@@ -243,30 +206,16 @@ val make_apply_entry :
    - (3) used as an EApply, if its HNF starts with a product, and
          has missing arguments. *)
 
-val make_resolves :
-  env -> evar_map -> bool * bool * bool -> hint_info -> poly:bool -> ?name:hints_path_atom ->
-  hint_term -> hint_entry list
+val push_resolves :
+  env -> evar_map -> GlobRef.t -> Hint_db.t -> Hint_db.t
 
-(** [make_resolve_hyp hname htyp].
+(** [push_resolve_hyp hname htyp db].
    used to add an hypothesis to the local hint database;
    Never raises a user exception;
-   If the hyp cannot be used as a Hint, the empty list is returned. *)
+   If the hyp cannot be used as a Hint, it is not added. *)
 
-val make_resolve_hyp :
-  env -> evar_map -> named_declaration -> hint_entry list
-
-(** [make_extern pri pattern tactic_expr] *)
-
-val make_extern :
-  int -> constr_pattern option -> Genarg.glob_generic_argument
-      -> hint_entry
-
-val run_hint : hint ->
-  ((raw_hint * clausenv) hint_ast -> 'r Proofview.tactic) -> 'r Proofview.tactic
-
-(** This function is for backward compatibility only, not to use in newly
-    written code. *)
-val repr_hint : hint -> (raw_hint * clausenv) hint_ast
+val push_resolve_hyp :
+  env -> evar_map -> Id.t -> Hint_db.t -> Hint_db.t
 
 (** Create a Hint database from the pairs (name, constr).
    Useful to take the current goal hypotheses as hints;
@@ -285,6 +234,11 @@ val wrap_hint_warning_fun : env -> evar_map ->
   (evar_map -> 'a * evar_map) -> 'a * evar_map
 (** Variant of the above for non-tactics *)
 
+val fresh_hint : env -> evar_map -> hint -> evar_map * constr
+
+val hint_res_pf : ?with_evars:bool -> ?with_classes:bool ->
+  ?flags:Unification.unify_flags -> hint -> unit Proofview.tactic
+
 (** Printing  hints *)
 
 val pr_searchtable : env -> evar_map -> Pp.t
@@ -292,4 +246,3 @@ val pr_applicable_hint : Proof.t -> Pp.t
 val pr_hint_ref : env -> evar_map -> GlobRef.t -> Pp.t
 val pr_hint_db_by_name : env -> evar_map -> hint_db_name -> Pp.t
 val pr_hint_db_env : env -> evar_map -> Hint_db.t -> Pp.t
-val pr_hint : env -> evar_map -> hint -> Pp.t
