@@ -16,7 +16,6 @@ open Constr
 open Context
 open EConstr
 open Vars
-open Reduction
 open Tacticals
 open Tactics
 open Pretype_errors
@@ -32,9 +31,6 @@ open Libnames
 open Proofview.Notations
 open Context.Named.Declaration
 
-module NamedDecl = Context.Named.Declaration
-(* module RelDecl = Context.Rel.Declaration *)
-
 module TC = Typeclasses
 
 (** Typeclass-based generalized rewriting. *)
@@ -143,7 +139,7 @@ type hypinfo = {
   sort : bool; (* true = Prop; false = Type *)
   c1 : constr;
   c2 : constr;
-  holes : Clenv.hole list;
+  holes : EClause.hole list;
 }
 
 let error_no_relation () = user_err Pp.(str "Cannot find a relation to rewrite.")
@@ -192,16 +188,16 @@ let decompose_applied_relation env sigma (c,l) =
   let open Context.Rel.Declaration in
   let ctype = Retyping.get_type_of env sigma c in
   let find_rel ty =
-    let sigma, cl = Clenv.make_evar_clause env sigma ty in
-    let sigma = Clenv.solve_evar_clause env sigma true cl l in
-    let { Clenv.cl_holes = holes; Clenv.cl_concl = t } = cl in
+    let sigma, cl = EClause.make_evar_clause env sigma ty in
+    let sigma = EClause.solve_evar_clause env sigma true cl l in
+    let { EClause.cl_holes = holes; EClause.cl_concl = t } = cl in
     match decompose_app_rel env sigma t with
     | None -> None
     | Some (equiv, ty1, ty2, concl, c1, c2) ->
       match evd_convertible env sigma ty1 ty2 with
       | None -> None
       | Some sigma ->
-        let args = Array.map_of_list (fun h -> h.Clenv.hole_evar) holes in
+        let args = Array.map_of_list (fun h -> h.EClause.hole_evar) holes in
         let value = mkApp (c, args) in
           Some (sigma, { prf=value;
                   car=ty1; rel = equiv; sort = Sorts.is_prop (ESorts.kind sigma concl);
@@ -210,7 +206,7 @@ let decompose_applied_relation env sigma (c,l) =
     match find_rel ctype with
     | Some c -> c
     | None ->
-      let ctx,t' = Reductionops.splay_prod env sigma ctype in (* Search for underlying eq *)
+      let ctx,t' = Reductionops.whd_decompose_prod env sigma ctype in (* Search for underlying eq *)
       let t' = it_mkProd_or_LetIn t' (List.map (fun (n,t) -> LocalAssum (n, t)) ctx) in
       match find_rel t' with
       | Some c -> c
@@ -258,22 +254,22 @@ end) = struct
 
   let proper_class =
     let r = lazy (find_reference morphisms "Proper") in
-    fun env sigma -> TC.class_info env sigma (Lazy.force r)
+    fun () -> Option.get (TC.class_info (Lazy.force r))
 
   let proper_proxy_class =
     let r = lazy (find_reference morphisms "ProperProxy") in
-    fun env sigma -> TC.class_info env sigma (Lazy.force r)
+    fun () -> Option.get (TC.class_info (Lazy.force r))
 
-  let proper_proj env sigma =
-    mkConst (Option.get (List.hd (proper_class env sigma).TC.cl_projs).TC.meth_const)
+  let proper_proj () =
+    UnsafeMonomorphic.mkConst (Option.get (List.hd (proper_class ()).TC.cl_projs).TC.meth_const)
 
   let proper_type env (sigma,cstrs) =
-    let l = (proper_class env sigma).TC.cl_impl in
+    let l = (proper_class ()).TC.cl_impl in
     let (sigma, c) = Evd.fresh_global env sigma l in
     (sigma, cstrs), c
 
   let proper_proxy_type env (sigma,cstrs) =
-    let l = (proper_proxy_class env sigma).TC.cl_impl in
+    let l = (proper_proxy_class ()).TC.cl_impl in
     let (sigma, c) = Evd.fresh_global env sigma l in
     (sigma, cstrs), c
 
@@ -373,23 +369,27 @@ end) = struct
       else (* None in Prop, use arrow *)
         (app_poly env evd arrow [| a; b |]), unfold_impl n
 
-  let rec decomp_pointwise sigma n c =
-    if Int.equal n 0 then c
+  let rec decomp_pointwise env sigma n c =
+    if Int.equal n 0 then Some c
     else
       match EConstr.kind sigma c with
-      | App (f, [| a; b; relb |]) when isRefX sigma (pointwise_relation_ref ()) f ->
-        decomp_pointwise sigma (pred n) relb
-      | App (f, [| a; b; arelb |]) when isRefX sigma (forall_relation_ref ()) f ->
-        decomp_pointwise sigma (pred n) (Reductionops.beta_applist sigma (arelb, [mkRel 1]))
-      | _ -> invalid_arg "decomp_pointwise"
+      | App (f, [| a; b; relb |]) when isRefX env sigma (pointwise_relation_ref ()) f ->
+        decomp_pointwise env sigma (pred n) relb
+      | App (f, [| a; b; arelb |]) when isRefX env sigma (forall_relation_ref ()) f ->
+        decomp_pointwise env sigma (pred n) (Reductionops.beta_applist sigma (arelb, [mkRel 1]))
+      | _ ->
+        (* cf #11347: when rewriting a commutative cut, we
+           decomp_pointwise on "c := eq (Prop -> Prop)"
+           Maybe if funext is available it's possible to do something? *)
+        None
 
-  let rec apply_pointwise sigma rel = function
+  let rec apply_pointwise env sigma rel = function
     | arg :: args ->
       (match EConstr.kind sigma rel with
-      | App (f, [| a; b; relb |]) when isRefX sigma (pointwise_relation_ref ()) f ->
-        apply_pointwise sigma relb args
-      | App (f, [| a; b; arelb |]) when isRefX sigma (forall_relation_ref ()) f ->
-        apply_pointwise sigma (Reductionops.beta_applist sigma (arelb, [arg])) args
+      | App (f, [| a; b; relb |]) when isRefX env sigma (pointwise_relation_ref ()) f ->
+        apply_pointwise env sigma relb args
+      | App (f, [| a; b; arelb |]) when isRefX env sigma (forall_relation_ref ()) f ->
+        apply_pointwise env sigma (Reductionops.beta_applist sigma (arelb, [arg])) args
       | _ -> invalid_arg "apply_pointwise")
     | [] -> rel
 
@@ -444,14 +444,14 @@ end) = struct
 
   let unlift_cstr env sigma = function
     | None -> None
-    | Some codom -> Some (decomp_pointwise (goalevars sigma) 1 codom)
+    | Some codom -> decomp_pointwise env (goalevars sigma) 1 codom
 
   (** Looking up declared rewrite relations (instances of [RewriteRelation]) *)
   let is_applied_rewrite_relation env sigma rels t =
     match EConstr.kind sigma t with
     | App (c, args) when Array.length args >= 2 ->
       let head = if isApp sigma c then fst (destApp sigma c) else c in
-        if isRefX sigma (coq_eq_ref ()) head then None
+        if isRefX env sigma (coq_eq_ref ()) head then None
         else
           (try
             let env' = push_rel_context rels env in
@@ -622,8 +622,8 @@ let solve_remaining_by env sigma holes by =
   | None -> sigma
   | Some tac ->
     let map h =
-      if h.Clenv.hole_deps then None
-      else match EConstr.kind sigma h.Clenv.hole_evar with
+      if h.EClause.hole_deps then None
+      else match EConstr.kind sigma h.EClause.hole_evar with
       | Evar (evk, _) ->
         Some evk
       | _ -> None
@@ -722,7 +722,7 @@ let unify_eqn (car, rel, prf, c1, c2, holes, sort) l2r flags env (sigma, cstrs)
     let sigma = TC.resolve_typeclasses ~filter:(no_constraints cstrs)
       ~fail:true env sigma in
     let sigma = solve_remaining_by env sigma holes by in
-    let nf c = Reductionops.nf_evar sigma (Reductionops.nf_meta env sigma c) in
+    let nf c = Reductionops.nf_evar sigma c in
     let c1 = nf c1 and c2 = nf c2
     and rew_car = nf car and rel = nf rel
     and prf = nf prf in
@@ -784,7 +784,7 @@ let poly_subrelation sort =
   if sort then PropGlobal.subrelation else TypeGlobal.subrelation
 
 let resolve_subrelation env car rel sort prf rel' res =
-  if Termops.eq_constr (fst res.rew_evars) rel rel' then res
+  if Termops.eq_constr env (fst res.rew_evars) rel rel' then res
   else
     let evars, app = app_poly_check env res.rew_evars (poly_subrelation sort) [|car; rel; rel'|] in
     let evars, subrel = new_cstr_evar evars env app in
@@ -880,7 +880,7 @@ let apply_rule unify : occurrences_count pure_strategy =
         | Some rew ->
           let b, occs = update_occurrence_counter occs in
             if not b then (occs, Fail)
-            else if Termops.eq_constr (fst rew.rew_evars) t rew.rew_to then (occs, Identity)
+            else if Termops.eq_constr env (fst rew.rew_evars) t rew.rew_to then (occs, Identity)
             else
               let res = { rew with rew_car = ty } in
               let res = Success (coerce env cstr res) in
@@ -931,11 +931,11 @@ let make_leibniz_proof env c ty r =
 
 let fold_match ?(force=false) env sigma c =
   let case = destCase sigma c in
-  let (ci, p, iv, c, brs) = EConstr.expand_case env sigma case in
+  let (ci, (p,_), iv, c, brs) = EConstr.expand_case env sigma case in
   let cty = Retyping.get_type_of env sigma c in
   let dep, pred, sk =
     let env', ctx, body =
-      let ctx, pred = decompose_lam_assum sigma p in
+      let ctx, pred = decompose_lambda_decls sigma p in
       let env' = push_rel_context ctx env in
         env', ctx, pred
     in
@@ -966,17 +966,21 @@ let fold_match ?(force=false) env sigma c =
       raise Not_found
   in
   let app =
+    let sk = if Global.is_polymorphic (ConstRef sk)
+      then CErrors.anomaly Pp.(str "Unexpected univ poly in Rewrite.fold_match")
+      else UnsafeMonomorphic.mkConst sk
+    in
     let ind, args = Inductiveops.find_mrectype env sigma cty in
     let pars, args = List.chop ci.ci_npar args in
     let meths = Array.to_list brs in
-      applist (mkConst sk, pars @ [pred] @ meths @ args @ [c])
+      applist (sk, pars @ [pred] @ meths @ args @ [c])
   in
-    sk, env, app
+    sk, app
 
 let unfold_match env sigma sk app =
   match EConstr.kind sigma app with
   | App (f', args) when QConstant.equal env (fst (destConst sigma f')) sk ->
-      let v = Environ.constant_value_in env (sk,Univ.Instance.empty)(*FIXME*) in
+      let v = Environ.constant_value_in env (sk,UVars.Instance.empty)(*FIXME*) in
       let v = EConstr.of_constr v in
         Reductionops.whd_beta env sigma (mkApp (v, args))
   | _ -> app
@@ -1067,7 +1071,7 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
                         let app = if prop then PropGlobal.apply_pointwise
                           else TypeGlobal.apply_pointwise
                         in
-                          RewPrf (app (goalevars evars) rel argsl, mkApp (prf, args))
+                          RewPrf (app env (goalevars evars) rel argsl, mkApp (prf, args))
                       | x -> x
                     in
                     let res =
@@ -1167,7 +1171,7 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
         in state, res
 
       | Case (ci, u, pms, p, iv, c, brs) ->
-        let (ci, p, iv, c, brs) = EConstr.expand_case env (goalevars evars) (ci, u, pms, p, iv, c, brs) in
+        let (ci, (p,rp), iv, c, brs) = EConstr.expand_case env (goalevars evars) (ci, u, pms, p, iv, c, brs) in
         let cty = Retyping.get_type_of env (goalevars evars) c in
         let evars', eqty = app_poly_sort prop env evars coq_eq [| cty |] in
         let cstr' = Some eqty in
@@ -1177,7 +1181,7 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
         let state, res =
           match c' with
           | Success r ->
-            let case = mkCase (EConstr.contract_case env (goalevars evars) (ci, lift 1 p, map_invert (lift 1) iv, mkRel 1, Array.map (lift 1) brs)) in
+            let case = mkCase (EConstr.contract_case env (goalevars evars) (ci, (lift 1 p,rp), map_invert (lift 1) iv, mkRel 1, Array.map (lift 1) brs)) in
             let res = make_leibniz_proof env case ty r in
               state, Success (coerce env (prop,cstr) res)
           | Fail | Identity ->
@@ -1199,13 +1203,13 @@ let subterm all flags (s : 'a pure_strategy) : 'a pure_strategy =
               in
                 match found with
                 | Some r ->
-                  let ctxc = mkCase (EConstr.contract_case env (goalevars evars) (ci, lift 1 p, map_invert (lift 1) iv, lift 1 c, Array.of_list (List.rev (brs' c')))) in
+                  let ctxc = mkCase (EConstr.contract_case env (goalevars evars) (ci, (lift 1 p, rp), map_invert (lift 1) iv, lift 1 c, Array.of_list (List.rev (brs' c')))) in
                     state, Success (make_leibniz_proof env ctxc ty r)
                 | None -> state, c'
             else
               match try Some (fold_match env (goalevars evars) t) with Not_found -> None with
               | None -> state, c'
-              | Some (cst, _, t') ->
+              | Some (cst, t') ->
                  let state, res = aux { state ; env ; unfresh ;
                                         term1 = t' ; ty1 = ty ;
                                         cstr = (prop,cstr) ; evars } in
@@ -1370,9 +1374,11 @@ module Strategies =
 
     let inj_open hint = (); fun sigma ->
       let (ctx, lemma) = Autorewrite.RewRule.rew_lemma hint in
-      (* not sure sideff:true is really needed here *)
-      let sigma = Evd.merge_context_set ~sideff:true UnivRigid sigma ctx in
-      (sigma, (EConstr.of_constr lemma, NoBindings))
+      let subst, ctx = UnivGen.fresh_universe_context_set_instance ctx in
+      let subst = Sorts.QVar.Map.empty, subst in
+      let lemma = Vars.subst_univs_level_constr subst (EConstr.of_constr lemma) in
+      let sigma = Evd.merge_context_set UnivRigid sigma ctx in
+      (sigma, (lemma, NoBindings))
 
     let old_hints (db : string) : 'a pure_strategy =
       let rules = Autorewrite.find_rewrites db in
@@ -1381,9 +1387,9 @@ module Strategies =
                                   Autorewrite.RewRule.rew_tac hint)) rules)
 
     let hints (db : string) : 'a pure_strategy = { strategy =
-      fun ({ term1 = t } as input) ->
+      fun ({ term1 = t; env } as input) ->
       let t = EConstr.Unsafe.to_constr t in
-      let rules = Autorewrite.find_matches db t in
+      let rules = Autorewrite.find_matches env db t in
       let lemma hint = (inj_open hint, Autorewrite.RewRule.rew_l2r hint,
                         Autorewrite.RewRule.rew_tac hint) in
       let lems = List.map lemma rules in
@@ -1395,7 +1401,7 @@ module Strategies =
           let rfn, ckind = Redexpr.reduction_of_red_expr env r in
           let sigma = goalevars evars in
           let (sigma, t') = rfn env sigma t in
-            if Termops.eq_constr sigma t' t then
+            if Termops.eq_constr env sigma t' t then
               state, Identity
             else
               state, Success { rew_car = ty; rew_from = t; rew_to = t';
@@ -1407,10 +1413,9 @@ module Strategies =
       fun { state ; env ; term1 = t ; ty1 = ty ; cstr ; evars } ->
 (*         let sigma, (c,_) = Tacinterp.interp_open_constr_with_bindings is env (goalevars evars) c in *)
         let sigma, c = Pretyping.understand_tcc env (goalevars evars) c in
-        let unfolded =
-          try Tacred.try_red_product env sigma c
-          with e when CErrors.noncritical e ->
-            user_err Pp.(str "fold: the term is not unfoldable!")
+        let unfolded = match Tacred.red_product env sigma c with
+        | None -> user_err Pp.(str "fold: the term is not unfoldable!")
+        | Some c -> c
         in
           try
             let sigma = Unification.w_unify env sigma CONV ~flags:(Unification.elim_flags ()) unfolded t in
@@ -1457,11 +1462,11 @@ let apply_strategy (s : strategy) env unfresh concl (prop, cstr) evars =
 let solve_constraints env (evars,cstrs) =
   let oldtcs = Evd.get_typeclass_evars evars in
   let evars' = Evd.set_typeclass_evars evars cstrs in
-  let evars' = TC.resolve_typeclasses env ~filter:TC.all_evars ~split:false ~fail:true evars' in
+  let evars' = TC.resolve_typeclasses env ~filter:TC.all_evars ~fail:true evars' in
   Evd.set_typeclass_evars evars' oldtcs
 
 let nf_zeta =
-  Reductionops.clos_norm_flags (CClosure.RedFlags.mkflags [CClosure.RedFlags.fZETA])
+  Reductionops.clos_norm_flags (RedFlags.mkflags [RedFlags.fZETA])
 
 exception RewriteFailure of Environ.env * Evd.evar_map * pretype_error
 
@@ -1472,7 +1477,7 @@ exception UnsolvedConstraints of Environ.env * Evd.evar_map * Evar.t
 let () = CErrors.register_handler begin function
 | UnsolvedConstraints (env, evars, ev) ->
   Some (str "Unsolved constraint remaining: " ++ spc () ++
-    Termops.pr_evar_info env evars (Evd.find evars ev) ++ str ".")
+    Termops.pr_evar_info env evars (Evd.find_undefined evars ev) ++ str ".")
 | _ -> None
 end
 
@@ -1482,7 +1487,7 @@ let cl_rewrite_clause_aux ?(abs=None) strat env avoid sigma concl is_hyp : resul
   let evars = (!evdref, Evar.Set.empty) in
   let evars, cstr =
     let prop, (evars, arrow) =
-      if Sorts.is_prop sort then true, app_poly_sort true env evars impl [||]
+      if ESorts.is_prop sigma sort then true, app_poly_sort true env evars impl [||]
       else false, app_poly_sort false env evars TypeGlobal.arrow [||]
     in
     match is_hyp with
@@ -1519,39 +1524,8 @@ let cl_rewrite_clause_aux ?(abs=None) strat env avoid sigma concl is_hyp : resul
     in
     Some (Some (evars, res, newt))
 
-(** Insert a declaration after the last declaration it depends on *)
-let rec insert_dependent env sigma decl accu hyps = match hyps with
-| [] -> List.rev_append accu [decl]
-| ndecl :: rem ->
-  if occur_var_in_decl env sigma (NamedDecl.get_id ndecl) decl then
-    List.rev_append accu (decl :: hyps)
-  else
-    insert_dependent env sigma decl (ndecl :: accu) rem
-
 let assert_replacing id newt tac =
-  let prf = Proofview.Goal.enter begin fun gl ->
-    let concl = Proofview.Goal.concl gl in
-    let env = Proofview.Goal.env gl in
-    let sigma = Tacmach.project gl in
-    let ctx = named_context env in
-    let after, before = List.split_when (NamedDecl.get_id %> Id.equal id) ctx in
-    let nc = match before with
-    | [] -> assert false
-    | d :: rem -> insert_dependent env sigma
-                    (LocalAssum (make_annot (NamedDecl.get_id d) Sorts.Relevant, newt)) [] after @ rem
-    in
-    let env' = Environ.reset_with_named_context (val_of_named_context nc) env in
-    Refine.refine ~typecheck:true begin fun sigma ->
-      let (sigma, ev) = Evarutil.new_evar env' sigma concl in
-      let (sigma, ev') = Evarutil.new_evar env sigma newt in
-      let map d =
-        let n = NamedDecl.get_id d in
-        if Id.equal n id then ev' else mkVar n
-      in
-      let (e, _) = destEvar sigma ev in
-      (sigma, mkLEvar sigma (e, List.map map nc))
-    end
-  end in
+  let prf = Tactics.assert_after_replacing id newt in
   Proofview.tclTHEN prf (Proofview.tclFOCUS 2 2 tac)
 
 let newfail n s =
@@ -1723,16 +1697,17 @@ let pr_ustrategy = function
 
 let paren p = str "(" ++ p ++ str ")"
 
-let rec pr_strategy prc prr = function
+let rec pr_strategy0 prc prr = function
 | StratId -> str "id"
 | StratFail -> str "fail"
 | StratRefl -> str "refl"
+| str -> paren (pr_strategy prc prr str)
+
+and pr_strategy1 prc prr = function
 | StratUnary (s, str) ->
-  pr_ustrategy s ++ spc () ++ paren (pr_strategy prc prr str)
+  pr_ustrategy s ++ spc () ++ pr_strategy1 prc prr str
 | StratNAry (Choice, strs) ->
-  str "choice" ++ spc () ++ prlist_with_sep spc (fun str -> paren (pr_strategy prc prr str)) strs
-| StratBinary (Compose, str1, str2) ->
-  pr_strategy prc prr str1 ++ str ";" ++ spc () ++ pr_strategy prc prr str2
+  str "choice" ++ brk (1,2) ++ prlist_with_sep spc (fun str -> hov 0 (pr_strategy0 prc prr str)) strs
 | StratConstr (c, true) -> prc c
 | StratConstr (c, false) -> str "<-" ++ spc () ++ prc c
 | StratTerms cl -> str "terms" ++ spc () ++ pr_sequence prc cl
@@ -1741,6 +1716,12 @@ let rec pr_strategy prc prr = function
   str cmd ++ spc () ++ str id
 | StratEval r -> str "eval" ++ spc () ++ prr r
 | StratFold c -> str "fold" ++ spc () ++ prc c
+| str -> pr_strategy0 prc prr str
+
+and pr_strategy prc prr = function
+| StratBinary (Compose, str1, str2) ->
+  pr_strategy prc prr str1 ++ str ";" ++ spc () ++ hov 0 (pr_strategy1 prc prr str2)
+| str -> hov 0 (pr_strategy1 prc prr str)
 
 let rec strategy_of_ast = function
   | StratId -> Strategies.id
@@ -1786,12 +1767,12 @@ let rec strategy_of_ast = function
                                              evars = (sigma,cstrevars evars) }) }
   | StratFold c -> Strategies.fold_glob (fst c)
 
-let proper_projection env sigma r ty =
+let proper_projection sigma r ty =
   let rel_vect n m = Array.init m (fun i -> mkRel(n+m-i)) in
-  let ctx, inst = decompose_prod_assum sigma ty in
+  let ctx, inst = decompose_prod_decls sigma ty in
   let mor, args = destApp sigma inst in
   let instarg = mkApp (r, rel_vect 0 (List.length ctx)) in
-  let app = mkApp (PropGlobal.proper_proj env sigma,
+  let app = mkApp (PropGlobal.proper_proj (),
                   Array.append args [| instarg |]) in
     it_mkLambda_or_LetIn app ctx
 
@@ -1830,7 +1811,7 @@ let default_morphism env sigma sign m =
   in
   let evars, morph = app_poly_check env evars PropGlobal.proper_type [| t; sign; m |] in
   let evars, mor = TC.resolve_one_typeclass env (goalevars evars) morph in
-    mor, proper_projection env sigma mor morph
+    mor, proper_projection sigma mor morph
 
 (** Bind to "rewrite" too *)
 
@@ -1928,9 +1909,9 @@ let setoid_proof ty fn fallback =
         try
           let rel, ty1, ty2, concl, _, _ = decompose_app_rel_error env sigma concl in
           let (sigma, t) = Typing.type_of env sigma rel in
-          let car = snd (List.hd (fst (Reductionops.splay_prod env sigma t))) in
-            (try init_relation_classes () with _ -> raise Not_found);
-            fn env sigma car rel
+          let car = snd (List.hd (fst (Reductionops.whd_decompose_prod env sigma t))) in
+          (try init_relation_classes () with e when CErrors.noncritical e -> raise Not_found);
+          fn env sigma car rel
         with e when CErrors.noncritical e ->
           (* XXX what is the right test here as to whether e can be converted ? *)
           let e, info = Exninfo.capture e in
@@ -1993,8 +1974,8 @@ let setoid_symmetry_in id =
   let env = Proofview.Goal.env gl in
   let sigma = Proofview.Goal.sigma gl in
   let ctype = Retyping.get_type_of env sigma (mkVar id) in
-  let binders,concl = decompose_prod_assum sigma ctype in
-  let (equiv, args) = decompose_app sigma concl in
+  let binders,concl = decompose_prod_decls sigma ctype in
+  let (equiv, args) = decompose_app_list sigma concl in
   let rec split_last_two = function
     | [c1;c2] -> [],(c1, c2)
     | x::y::z -> let l,res = split_last_two (y::z) in x::l, res
