@@ -31,7 +31,7 @@ let set_typeclass_transparency ~locality c b =
     (Hints.HintsTransparencyEntry (Hints.HintsReferences c, b))
 
 let warn_deprecated_tc_transparency_without_locality =
-  CWarnings.create ~name:"deprecated-typeclasses-transparency-without-locality" ~category:"deprecated"
+  CWarnings.create ~name:"deprecated-typeclasses-transparency-without-locality" ~category:Deprecation.Version.v8_15
     Pp.(fun () -> strbrk
   "The default value for Typeclasses Opaque and Typeclasses \
    Transparent locality is currently \"local\" in a section and \
@@ -44,7 +44,7 @@ let warn_deprecated_tc_transparency_without_locality =
    This is supported since Coq 8.15.")
 
 let default_tc_transparency_locality () =
-  if Global.sections_are_opened () then Hints.Local
+  if Lib.sections_are_opened () then Hints.Local
   else
     let () = warn_deprecated_tc_transparency_without_locality () in
     Hints.SuperGlobal
@@ -60,81 +60,80 @@ let set_typeclass_transparency_com ~locality refs b =
   in
   set_typeclass_transparency ~locality refs b
 
-let add_instance_hint inst path ~locality info =
+let add_instance_hint gr ~locality info =
+  let inst = Hints.hint_globref gr in
      Flags.silently (fun () ->
        Hints.add_hints ~locality [typeclasses_db]
           (Hints.HintsResolveEntry
-             [info, false, Hints.PathHints path, inst])) ()
+             [info, false, inst])) ()
 
 (* short names without opening all Hints *)
 type locality = Hints.hint_locality = Local | Export | SuperGlobal
 
-type instance_obj = {
-  inst_class : GlobRef.t;
-  inst_info: hint_info;
+type instance = {
+  class_name : GlobRef.t;
+  instance : GlobRef.t;
+  info : Typeclasses.hint_info;
   (* Sections where the instance should be redeclared,
      None for discard, Some 0 for none. *)
-  inst_global: Hints.hint_locality;
-  inst_impl: GlobRef.t;
+  locality : Hints.hint_locality;
 }
 
 let add_instance_base inst =
-  let locality = match inst.inst_global with
+  let locality = match inst.locality with
   | Local -> Local
   | SuperGlobal ->
     (* i.e. in a section, declare the hint as local since discharge is managed
        by rebuild_instance which calls again add_instance_hint; don't ask hints
        to take discharge into account itself *)
-    if Global.sections_are_opened () then Local
+    if Lib.sections_are_opened () then Local
     else SuperGlobal
   | Export ->
     (* Same as above for export *)
-    if Global.sections_are_opened () then Local
+    if Lib.sections_are_opened () then Local
     else Export
   in
-  add_instance_hint (Hints.hint_globref inst.inst_impl) [inst.inst_impl] ~locality
-    inst.inst_info
+  add_instance_hint inst.instance ~locality
+    inst.info
 
 (*
  * instances persistent object
  *)
 let perform_instance i =
-  let i = { is_class = i.inst_class; is_info = i.inst_info; is_impl = i.inst_impl } in
+  let i = { is_class = i.class_name; is_info = i.info; is_impl = i.instance } in
   Typeclasses.load_instance i
 
-let cache_instance inst = perform_instance inst
+let cache_instance inst =
+  perform_instance inst;
+  add_instance_base inst
 
-let load_instance _ inst = match inst.inst_global with
+let load_instance _ inst = match inst.locality with
 | Local -> assert false
 | SuperGlobal -> perform_instance inst
 | Export -> ()
 
-let open_instance i inst = match inst.inst_global with
+let open_instance i inst = match inst.locality with
 | Local -> assert false
 | SuperGlobal -> perform_instance inst
 | Export -> if Int.equal i 1 then perform_instance inst
 
 let subst_instance (subst, inst) =
   { inst with
-      inst_class = fst (subst_global subst inst.inst_class);
-      inst_impl = fst (subst_global subst inst.inst_impl) }
+      class_name = fst (subst_global subst inst.class_name);
+      instance = fst (subst_global subst inst.instance) }
 
 let discharge_instance inst =
-  match inst.inst_global with
+  match inst.locality with
   | Local -> None
   | SuperGlobal | Export ->
-    assert (not (isVarRef inst.inst_impl));
+    assert (not (isVarRef inst.instance));
     Some inst
 
-let rebuild_instance inst =
-  add_instance_base inst;
-  inst
-
-let classify_instance inst = match inst.inst_global with
+let classify_instance inst = match inst.locality with
 | Local -> Dispose
 | SuperGlobal | Export -> Substitute
 
-let instance_input : instance_obj -> obj =
+let instance_input : instance -> obj =
   declare_object
     { (default_object "type classes instances state") with
       cache_function = cache_instance;
@@ -142,53 +141,65 @@ let instance_input : instance_obj -> obj =
       open_function = simple_open ~cat:Hints.hint_cat open_instance;
       classify_function = classify_instance;
       discharge_function = discharge_instance;
-      rebuild_function = rebuild_instance;
       subst_function = subst_instance }
 
-let warn_deprecated_instance_without_locality =
-  let open Pp in
-  CWarnings.create ~name:"deprecated-instance-without-locality" ~category:"deprecated"
-    ~default:CWarnings.AsError
-    (fun () -> strbrk "The default value for instance locality is currently \
-    \"local\" in a section and \"global\" otherwise, but is scheduled to change \
-    in a future release. For the time being, adding instances outside of sections \
-    without specifying an explicit locality attribute is therefore deprecated. It is \
-    recommended to use \"export\" whenever possible. Use the attributes \
-    #[local], #[global] and #[export] depending on your choice. For example: \
-    \"#[export] Instance Foo : Bar := baz.\"")
-
 let default_locality () =
-  if Global.sections_are_opened () then Local
-  else
-    let () = warn_deprecated_instance_without_locality () in
-    SuperGlobal
+  if Lib.sections_are_opened () then Local
+  else Export
 
 let instance_locality =
   Attributes.hint_locality ~default:default_locality
 
+module Event = struct
+  type t =
+    | NewClass of typeclass
+    | NewInstance of instance
+end
+
+type observer = string
+
+let observers = ref CString.Map.empty
+
+let active_observers = Summary.ref ~name:"active typeclass observers" []
+
+let register_observer ~name ?(override=false) o =
+  if not override && CString.Map.mem name !observers then
+    CErrors.anomaly Pp.(str "Typeclass observer " ++ str name ++ str " already registered.");
+  observers := CString.Map.add name o !observers;
+  name
+
+let deactivate_observer name =
+  active_observers := List.remove String.equal name !active_observers
+
+let activate_observer name =
+  assert (CString.Map.mem name !observers);
+  deactivate_observer name;
+  active_observers := name :: !active_observers
+
+let observe event =
+  List.iter (fun name -> (CString.Map.get name !observers) event) !active_observers
+
 let add_instance cl info global impl =
   let () = match global with
     | Local -> ()
     | SuperGlobal ->
-      if Global.sections_are_opened () && isVarRef impl then
+      if Lib.sections_are_opened () && isVarRef impl then
         CErrors.user_err (Pp.str "Cannot set Global an instance referring to a section variable.")
     | Export ->
-      if Global.sections_are_opened () && isVarRef impl then
+      if Lib.sections_are_opened () && isVarRef impl then
         CErrors.user_err (Pp.str "The export attribute cannot be applied to an instance referring to a section variable.")
   in
   let i = {
-    inst_class = cl.cl_impl;
-    inst_info = info ;
-    inst_global = global ;
-    inst_impl = impl;
+    class_name = cl.cl_impl;
+    info = info ;
+    locality = global ;
+    instance = impl;
   } in
   Lib.add_leaf (instance_input i);
-  add_instance_base i
+  observe (Event.NewInstance { class_name = cl.cl_impl; instance = impl; info; locality = global })
 
 let warning_not_a_class =
-  let name = "not-a-class" in
-  let category = "typeclasses" in
-  CWarnings.create ~name ~category (fun (n, ty) ->
+  CWarnings.create ~name:"not-a-class" (fun (n, ty) ->
       let env = Global.env () in
       let evd = Evd.from_env env in
       Pp.(str "Ignored instance declaration for “"
@@ -271,7 +282,8 @@ let class_input : typeclass -> obj =
       subst_function = subst_class }
 
 let add_class cl =
-  Lib.add_leaf (class_input cl)
+  Lib.add_leaf (class_input cl);
+  observe (Event.NewClass cl)
 
 let intern_info {hint_priority;hint_pattern} =
   let env = Global.env() in
@@ -280,18 +292,16 @@ let intern_info {hint_priority;hint_pattern} =
   {hint_priority;hint_pattern}
 
 (** TODO: add subinstances *)
-let existing_instance glob g info =
-  let c = Nametab.global g in
+let existing_instance ?loc glob c info =
   let info = Option.default Hints.empty_hint_info info in
   let info = intern_info info in
   let env = Global.env() in
   let sigma = Evd.from_env env in
   let instance, _ = Typeops.type_of_global_in_context env c in
-  let ctx, r = Term.decompose_prod_assum instance in
+  let ctx, r = Term.decompose_prod_decls instance in
     match class_of_constr (Environ.push_rel_context ctx env) sigma (EConstr.of_constr r) with
       | Some (_, ((tc,u), _)) -> add_instance tc info glob c
-      | None -> user_err ?loc:g.CAst.loc
-                  (Pp.str "Constant does not build instances of a declared type class.")
+      | None -> user_err ?loc (Pp.str "Constant does not build instances of a declared type class.")
 
 (* Declare everything in the parameters as implicit, and the class instance as well *)
 
@@ -448,7 +458,7 @@ let do_instance_type_ctx_instance props k env' ctx' sigma ~program_mode subst =
                    Option.iter (fun x -> Dumpglob.add_glob ?loc (GlobRef.ConstRef x)) m.meth_const) k.cl_projs;
              c :: props, rest'
            with Not_found ->
-             ((CAst.make @@ CHole (None(* Some Evar_kinds.GoalEvar *), Namegen.IntroAnonymous, None)) :: props), rest
+             ((CAst.make @@ CHole (None)) :: props), rest
          else props, rest)
       ([], props) k.cl_props
   in
@@ -530,7 +540,7 @@ let interp_instance_context ~program_mode env ctx pl tclass =
   let flags = Pretyping.{ all_no_fail_flags with program_mode } in
   let sigma, (c', imps') = interp_type_evars_impls ~flags ~impls env' sigma tclass in
   let imps = imps @ imps' in
-  let ctx', c = decompose_prod_assum sigma c' in
+  let ctx', c = decompose_prod_decls sigma c' in
   let ctx'' = ctx' @ ctx in
   let (k, u), args = Typeclasses.dest_class_app (push_rel_context ctx'' env) sigma c in
   let u_s = EInstance.kind sigma u in
@@ -605,9 +615,6 @@ let refine_att =
   | None -> return false
   | Some () -> return true
 
-module Internal =
-struct
-let add_instance cl info glob r =
-  let glob = if glob then SuperGlobal else Local in
-  add_instance cl info glob r
+module Internal = struct
+  let add_instance = add_instance
 end
