@@ -13,112 +13,105 @@ open Util
 open Vars
 open Names
 open Context
-open Constrexpr_ops
 open Constrintern
 open Impargs
 open Pretyping
-open Entries
 
 module RelDecl = Context.Rel.Declaration
 (* 2| Variable/Hypothesis/Parameter/Axiom declarations *)
 
-let declare_variable is_coe ~kind typ imps impl {CAst.v=name} =
+let declare_variable is_coe ~kind typ univs imps impl {CAst.v=name} =
   let kind = Decls.IsAssumption kind in
-  let decl = Declare.SectionLocalAssum {typ; impl} in
-  let () = Declare.declare_variable ~name ~kind decl in
+  let () = Declare.declare_variable ~name ~kind ~typing_flags:None ~typ ~impl ~univs in
   let () = Declare.assumption_message name in
   let r = GlobRef.VarRef name in
   let () = maybe_declare_manual_implicits true r imps in
   let env = Global.env () in
   let sigma = Evd.from_env env in
-  let () = Classes.declare_instance env sigma None true r in
-  let () = if is_coe then ComCoercion.try_add_new_coercion r ~local:true ~poly:false in
+  let () = Classes.declare_instance env sigma None Hints.Local r in
+  let () =
+    if is_coe = Vernacexpr.AddCoercion then
+      ComCoercion.try_add_new_coercion
+        r ~local:true ~reversible:true in
   ()
 
 let instance_of_univ_entry = function
-  | Polymorphic_entry (_, univs) -> Univ.UContext.instance univs
-  | Monomorphic_entry _ -> Univ.Instance.empty
+  | UState.Polymorphic_entry univs -> UVars.UContext.instance univs
+  | UState.Monomorphic_entry _ -> UVars.Instance.empty
 
-let declare_axiom is_coe ~poly ~local ~kind typ (univs, pl) imps nl {CAst.v=name} =
-  let do_instance = let open Decls in match kind with
-  | Context -> true
-    (* The typeclass behaviour of Variable and Context doesn't depend
-       on section status *)
-  | Definitional | Logical | Conjectural -> false
-  in
+let declare_axiom is_coe ~local ~kind ?deprecation typ (univs, ubinders) imps nl {CAst.v=name} =
   let inl = let open Declaremods in match nl with
     | NoInline -> None
     | DefaultInline -> Some (Flags.get_inline_level())
     | InlineAt i -> Some i
   in
   let kind = Decls.IsAssumption kind in
-  let decl = Declare.ParameterEntry (None,(typ,univs),inl) in
-  let kn = Declare.declare_constant ~name ~local ~kind decl in
+  let entry = Declare.parameter_entry ~univs:(univs, ubinders) ?inline:inl typ in
+  let decl = Declare.ParameterEntry entry in
+  let kn = Declare.declare_constant ~name ~local ~kind ?deprecation decl in
   let gr = GlobRef.ConstRef kn in
   let () = maybe_declare_manual_implicits false gr imps in
-  let () = DeclareUniv.declare_univ_binders gr pl in
   let () = Declare.assumption_message name in
-  let env = Global.env () in
-  let sigma = Evd.from_env env in
-  let () = if do_instance then Classes.declare_instance env sigma None false gr in
   let local = match local with
-    | Declare.ImportNeedQualified -> true
-    | Declare.ImportDefaultBehavior -> false
+    | Locality.ImportNeedQualified -> true
+    | Locality.ImportDefaultBehavior -> false
   in
-  let () = if is_coe then ComCoercion.try_add_new_coercion gr ~local ~poly in
+  let () =
+    if is_coe = Vernacexpr.AddCoercion then
+      ComCoercion.try_add_new_coercion
+        gr ~local ~reversible:true in
   let inst = instance_of_univ_entry univs in
   (gr,inst)
 
-let interp_assumption ~program_mode sigma env impls c =
+let interp_assumption ~program_mode env sigma impl_env bl c =
   let flags = { Pretyping.all_no_fail_flags with program_mode } in
-  let sigma, (ty, impls) = interp_type_evars_impls ~flags env sigma ~impls c in
-  sigma, (ty, impls)
+  let sigma, (impls, ((env_bl, ctx), impls1)) = interp_context_evars ~program_mode ~impl_env env sigma bl in
+  let sigma, (ty, impls2) = interp_type_evars_impls ~flags env_bl sigma ~impls c in
+  let ty = EConstr.it_mkProd_or_LetIn ty ctx in
+  sigma, ty, impls1@impls2
 
-(* When monomorphic the universe constraints and universe names are
-   declared with the first declaration only. *)
-let next_univs =
-  let empty_univs = Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders in
-  function
-  | Polymorphic_entry _, _ as univs -> univs
-  | Monomorphic_entry _, _ -> empty_univs
+let empty_poly_univ_entry = UState.Polymorphic_entry UVars.UContext.empty, UnivNames.empty_binders
+let empty_mono_univ_entry = UState.Monomorphic_entry Univ.ContextSet.empty, UnivNames.empty_binders
+let empty_univ_entry ~poly = if poly then empty_poly_univ_entry else empty_mono_univ_entry
 
-let context_set_of_entry = function
-  | Polymorphic_entry (_,uctx) -> Univ.ContextSet.of_context uctx
-  | Monomorphic_entry uctx -> uctx
+(* When declarations are monomorphic (which is always the case in
+   sections, even when universes are treated as polymorphic variables)
+   the universe constraints and universe names are declared with the
+   first declaration only. *)
 
-let declare_assumptions ~poly ~scope ~kind univs nl l =
-  let () = let open Declare in match scope with
-    | Discharge ->
-      (* declare universes separately for variables *)
-      DeclareUctx.declare_universe_context ~poly (context_set_of_entry (fst univs))
-    | Global _ -> ()
-  in
+let clear_univs scope univ =
+  match scope, univ with
+  | Locality.Global _, (UState.Polymorphic_entry _, _ as univs) -> univs
+  | _, (UState.Monomorphic_entry _, _) -> empty_univ_entry ~poly:false
+  | Locality.Discharge, (UState.Polymorphic_entry _, _) -> empty_univ_entry ~poly:true
+
+let declare_assumptions ~scope ~kind ?deprecation univs nl l =
   let _, _ = List.fold_left (fun (subst,univs) ((is_coe,idl),typ,imps) ->
       (* NB: here univs are ignored when scope=Discharge *)
       let typ = replace_vars subst typ in
       let univs,subst' =
         List.fold_left_map (fun univs id ->
             let refu = match scope with
-              | Declare.Discharge ->
-                declare_variable is_coe ~kind typ imps Glob_term.Explicit id;
-                GlobRef.VarRef id.CAst.v, Univ.Instance.empty
-              | Declare.Global local ->
-                declare_axiom is_coe ~local ~poly ~kind typ univs imps nl id
+              | Locality.Discharge ->
+                declare_variable is_coe ~kind typ univs imps Glob_term.Explicit id;
+                GlobRef.VarRef id.CAst.v, UVars.Instance.empty
+              | Locality.Global local ->
+                declare_axiom is_coe ~local ~kind ?deprecation typ univs imps nl id
             in
-            next_univs univs, (id.CAst.v, Constr.mkRef refu))
+            clear_univs scope univs, (id.CAst.v, Constr.mkRef refu))
           univs idl
       in
-      subst'@subst, next_univs univs)
+      subst'@subst, clear_univs scope univs)
       ([], univs) l
   in
   ()
 
 let maybe_error_many_udecls = function
   | ({CAst.loc;v=id}, Some _) ->
-    user_err ?loc ~hdr:"many_universe_declarations"
-      Pp.(str "When declaring multiple axioms in one command, " ++
-          str "only the first is allowed a universe binder " ++
-          str "(which will be shared by the whole block).")
+    user_err ?loc
+      Pp.(strbrk "When declaring multiple axioms in one command, " ++
+          strbrk "only the first is allowed a universe binder " ++
+          strbrk "(which will be shared by the whole block).")
   | (_, None) -> ()
 
 let process_assumptions_udecls ~scope l =
@@ -129,9 +122,8 @@ let process_assumptions_udecls ~scope l =
       udecl, id
     | (_, ([], _))::_ | [] -> assert false
   in
-  let open Declare in
   let () = match scope, udecl with
-    | Discharge, Some _ ->
+    | Locality.Discharge, Some _ ->
       let loc = first_id.CAst.loc in
       let msg = Pp.str "Section variables cannot be polymorphic." in
       user_err ?loc  msg
@@ -139,7 +131,7 @@ let process_assumptions_udecls ~scope l =
   in
   udecl, List.map (fun (coe, (idl, c)) -> coe, (List.map fst idl, c)) l
 
-let do_assumptions ~program_mode ~poly ~scope ~kind nl l =
+let do_assumptions ~program_mode ~poly ~scope ~kind ?deprecation nl l =
   let open Context.Named.Declaration in
   let env = Global.env () in
   let udecl, l = process_assumptions_udecls ~scope l in
@@ -155,7 +147,7 @@ let do_assumptions ~program_mode ~poly ~scope ~kind nl l =
   in
   (* We interpret all declarations in the same evar_map, i.e. as a telescope. *)
   let (sigma,_,_),l = List.fold_left_map (fun (sigma,env,ienv) (is_coe,(idl,c)) ->
-    let sigma,(t,imps) = interp_assumption ~program_mode sigma env ienv c in
+    let sigma,t,imps = interp_assumption ~program_mode env sigma ienv [] c in
     let r = Retyping.relevance_of_type env sigma t in
     let env =
       EConstr.push_named_context (List.map (fun {CAst.v=id} -> LocalAssum (make_annot id r,t)) idl) env in
@@ -171,96 +163,83 @@ let do_assumptions ~program_mode ~poly ~scope ~kind nl l =
   let nf_evar c = EConstr.to_constr sigma c in
   let uvars, l = List.fold_left_map (fun uvars (coe,t,imps) ->
       let t = nf_evar t in
-      let uvars = Univ.LSet.union uvars (Vars.universes_of_constr t) in
+      let uvars = Univ.Level.Set.union uvars (Vars.universes_of_constr t) in
       uvars, (coe,t,imps))
-      Univ.LSet.empty l
+      Univ.Level.Set.empty l
   in
-  (* XXX: Using `DeclareDef.prepare_parameter` here directly is not
+  (* XXX: Using `Declare.prepare_parameter` here directly is not
      possible as we indeed declare several parameters; however,
      restrict_universe_context should be called in a centralized place
      IMO, thus I think we should adapt `prepare_parameter` to handle
      this case too. *)
   let sigma = Evd.restrict_universe_context sigma uvars in
   let univs = Evd.check_univ_decl ~poly sigma udecl in
-  let ubinders = Evd.universe_binders sigma in
-  declare_assumptions ~poly ~scope ~kind (univs,ubinders) nl l
+  declare_assumptions ~scope ~kind ?deprecation univs nl l
 
 let context_subst subst (name,b,t,impl) =
   name, Option.map (Vars.substl subst) b, Vars.substl subst t, impl
 
 let context_insection sigma ~poly ctx =
-  let uctx = Evd.universe_context_set sigma in
-  let () = DeclareUctx.declare_universe_context ~poly uctx in
-  let fn subst (name,_,_,_ as d) =
+  let uctx = Evd.evar_universe_context sigma in
+  let univs = UState.univ_entry ~poly uctx in
+  let fn i subst (name,_,_,_ as d) =
     let d = context_subst subst d in
+    let univs = if i = 0 then univs else empty_univ_entry ~poly in
     let () = match d with
       | name, None, t, impl ->
         let kind = Decls.Context in
-        declare_variable false ~kind t [] impl (CAst.make name)
+        declare_variable NoCoercion ~kind t univs [] impl (CAst.make name)
       | name, Some b, t, impl ->
-        (* We need to get poly right for check_same_poly *)
-        let univs = if poly then Polymorphic_entry ([| |], Univ.UContext.empty)
-          else Monomorphic_entry Univ.ContextSet.empty
-        in
         let entry = Declare.definition_entry ~univs ~types:t b in
-        (* XXX Fixme: Use DeclareDef.prepare_definition *)
-        let uctx = Evd.evar_universe_context sigma in
+        (* XXX Fixme: Use Declare.prepare_definition *)
         let kind = Decls.(IsDefinition Definition) in
         let _ : GlobRef.t =
-          Declare.declare_entry ~name ~scope:Declare.Discharge
+          Declare.declare_entry ~name ~scope:Locality.Discharge
             ~kind ~impargs:[] ~uctx entry
         in
         ()
     in
     Constr.mkVar name :: subst
   in
-  let _ : Vars.substl = List.fold_left fn [] ctx in
+  let _ : Vars.substl = List.fold_left_i fn 0 [] ctx in
   ()
 
 let context_nosection sigma ~poly ctx =
-  let univs =
-    match ctx, poly with
-    | [_], _ | _, true -> Evd.univ_entry ~poly sigma
-    | _, false ->
-      (* Multiple monomorphic axioms: declare universes separately to
-         avoid redeclaring them. *)
-      let uctx = Evd.universe_context_set sigma in
-      let () = DeclareUctx.declare_universe_context ~poly uctx in
-      Monomorphic_entry Univ.ContextSet.empty
-  in
-  let fn subst d =
+  let (univ_entry,ubinders as univs) = Evd.univ_entry ~poly sigma in
+  let fn i subst d =
     let (name,b,t,_impl) = context_subst subst d in
     let kind = Decls.(IsAssumption Logical) in
+    let local = if Lib.is_modtype () then Locality.ImportDefaultBehavior
+      else Locality.ImportNeedQualified
+    in
+    (* Multiple monomorphic axioms: declare universes only on the first declaration *)
+    let univs = if i = 0 then univs else clear_univs (Locality.Global local) univs in
     let decl = match b with
       | None ->
-        Declare.ParameterEntry (None,(t,univs),None)
+        let entry = Declare.parameter_entry ~univs:(univ_entry, ubinders) t in
+        Declare.ParameterEntry entry
       | Some b ->
         let entry = Declare.definition_entry ~univs ~types:t b in
         Declare.DefinitionEntry entry
     in
-    let local = if Lib.is_modtype () then Declare.ImportDefaultBehavior
-      else Declare.ImportNeedQualified
-    in
     let cst = Declare.declare_constant ~name ~kind ~local decl in
     let () = Declare.assumption_message name in
     let env = Global.env () in
-    (* why local when is_modtype? *)
-    let () = if Lib.is_modtype() || Option.is_empty b then
-        Classes.declare_instance env sigma None (Lib.is_modtype()) (GlobRef.ConstRef cst)
+    let () = if Option.is_empty b then
+        Classes.declare_instance env sigma None Hints.SuperGlobal (GlobRef.ConstRef cst)
     in
-    Constr.mkConstU (cst,instance_of_univ_entry univs) :: subst
+    Constr.mkConstU (cst,instance_of_univ_entry univ_entry) :: subst
   in
-  let _ : Vars.substl = List.fold_left fn [] ctx in
+  let _ : Vars.substl = List.fold_left_i fn 0 [] ctx in
   ()
 
-let context ~poly l =
-  let env = Global.env() in
-  let sigma = Evd.from_env env in
+let interp_context env sigma l =
   let sigma, (_, ((_env, ctx), impls)) = interp_context_evars ~program_mode:false env sigma l in
   (* Note, we must use the normalized evar from now on! *)
-  let sigma = Evd.minimize_universes sigma in
   let ce t = Pretyping.check_evars env sigma t in
   let () = List.iter (fun decl -> Context.Rel.Declaration.iter_constr ce decl) ctx in
+  let sigma, ctx = Evarutil.finalize
+      sigma (fun nf -> List.map (RelDecl.map_constr_het nf) ctx) in
   (* reorder, evar-normalize and add implicit status *)
   let ctx = List.rev_map (fun d ->
       let {binder_name=name}, b, t = RelDecl.to_tuple d in
@@ -268,19 +247,23 @@ let context ~poly l =
         | Anonymous -> user_err Pp.(str "Anonymous variables not allowed in contexts.")
         | Name id -> id
       in
-      let b = Option.map (EConstr.to_constr sigma) b in
-      let t = EConstr.to_constr sigma t in
       let impl = let open Glob_term in
       let search x = match x.CAst.v with
         | Some (Name id',max) when Id.equal name id' ->
           Some (if max then MaxImplicit else NonMaxImplicit)
         | _ -> None
         in
-        try CList.find_map search impls with Not_found -> Explicit
+        Option.default Explicit (CList.find_map search impls)
       in
       name,b,t,impl)
       ctx
   in
-  if Global.sections_are_opened ()
+   sigma, ctx
+
+let do_context ~poly l =
+  let env = Global.env() in
+  let sigma = Evd.from_env env in
+  let sigma, ctx = interp_context env sigma l in
+  if Lib.sections_are_opened ()
   then context_insection sigma ~poly ctx
   else context_nosection sigma ~poly ctx
