@@ -47,7 +47,7 @@ type family = SPropF | PropF | TypeF
 let family_of_sort_family = let open Sorts in function
     | InSProp -> SPropF
     | InProp -> PropF
-    | InSet | InType -> TypeF
+    | InSet | InType | InQSort -> TypeF
 
 let get_sigmatypes sigma ~sort ~predsort =
   let open EConstr in
@@ -106,9 +106,15 @@ let telescope env sigma l =
   telescope sigma l
 
 let nf_evar_context sigma ctx =
-  List.map (map_constr (fun c -> Evarutil.nf_evar sigma c)) ctx
+  let nf c = Evarutil.nf_evar sigma c in
+  let nfa na = UnivSubst.nf_binder_annot (fun r -> Evarutil.nf_relevance sigma r) na in
+  let map = function
+  | RelDecl.LocalAssum (na, t) -> RelDecl.LocalAssum (nfa na, nf t)
+  | RelDecl.LocalDef (na, c, t) -> RelDecl.LocalDef (nfa na, nf c, nf t)
+  in
+  List.map map ctx
 
-let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r measure notation =
+let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?deprecation ?using r measure notation =
   let open EConstr in
   let open Vars in
   Coqlib.check_required_library ["Coq";"Program";"Wf"];
@@ -131,13 +137,13 @@ let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r
       user_err ?loc:(constr_loc r)
         (Printer.pr_econstr_env env sigma rel ++ str " is not an homogeneous binary relation.")
     in
-    try
-      let ctx, ar = Reductionops.splay_prod_n env sigma 2 relty in
-      match ctx, EConstr.kind sigma ar with
-      | [LocalAssum (_,t); LocalAssum (_,u)], Sort s
+    let ctx, ar =
+      try Reductionops.whd_decompose_prod_n env sigma 2 relty
+      with Invalid_argument _ -> error () in
+    match ctx, EConstr.kind sigma ar with
+      | [(_,t); (_,u)], Sort s
         when Sorts.is_prop (ESorts.kind sigma s) && Reductionops.is_conv env sigma t u -> t
       | _, _ -> error ()
-    with e when CErrors.noncritical e -> error ()
   in
   let sigma, measure = interp_casted_constr_evars ~program_mode:true binders_env sigma measure relargty in
   let sigma, wf_rel, wf_rel_fun, measure_fn =
@@ -196,10 +202,14 @@ let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r
       Constrintern.compute_internalization_data env sigma recname
         Constrintern.Recursive full_arity impls
     in
+    let impl = Some Impargs.{
+      impl_pos = (Name (Id.of_string "recproof"), 1, None);
+      impl_expl = Manual;
+      impl_max = true;
+      impl_force = false;
+    } in
     let newimpls = Id.Map.singleton recname
-        (Constrintern.extend_internalization_data interning_data
-           (Some ((Name (Id.of_string "recproof"),1,None), Impargs.Manual, (true, false)))
-           []) in
+        (Constrintern.extend_internalization_data interning_data impl []) in
     interp_casted_constr_evars ~program_mode:true (push_rel_context ctx env) sigma
       ~impls:newimpls body (lift 1 top_arity)
   in
@@ -217,6 +227,7 @@ let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r
     sigma, mkApp (h_a_term, [| argtyp ; wf_rel ; h_e_term; prop |])
   in
   let sigma, def = Typing.solve_evars env sigma def in
+  let sigma = Evd.collapse_sort_variables sigma in
   let sigma = Evarutil.nf_evar_map sigma in
   let def = mkApp (def, [|intern_body_lam|]) in
   let binders_rel = nf_evar_context sigma binders_rel in
@@ -258,11 +269,11 @@ let build_wellfounded pm (recname,pl,bl,arityc,body) poly ?typing_flags ?using r
   in
   let using =
     let terms = List.map EConstr.of_constr [evars_def; evars_typ] in
-    Option.map (fun using -> Proof_using.definition_using env sigma ~using ~terms) using
+    Option.map (fun using -> Proof_using.definition_using env sigma ~fixnames:[] ~using ~terms) using
   in
   let uctx = Evd.evar_universe_context sigma in
   let cinfo = Declare.CInfo.make ~name:recname ~typ:evars_typ ?using () in
-  let info = Declare.Info.make ~udecl ~poly ~hook ?typing_flags () in
+  let info = Declare.Info.make ~udecl ~poly ~hook ?typing_flags ?deprecation () in
   let pm, _ =
     Declare.Obls.add_definition ~pm ~cinfo ~info ~term:evars_def ~uctx evars in
   pm
@@ -274,7 +285,7 @@ let out_def = function
 let collect_evars_of_term evd c ty =
   Evar.Set.union (Evd.evars_of_term evd c) (Evd.evars_of_term evd ty)
 
-let do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind fixl =
+let do_program_recursive ~pm ~scope ?clearbody ~poly ?typing_flags ?deprecation ?using fixkind fixl =
   let cofix = fixkind = Declare.Obls.IsCoFixpoint in
   let (env, rec_sign, udecl, evd), fix, info =
     let env = Global.env () in
@@ -286,10 +297,11 @@ let do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind fixl =
   let evd = Typeclasses.resolve_typeclasses ~filter:Typeclasses.no_goals ~fail:true env evd in
     (* Solve remaining evars *)
   let evd = nf_evar_map_undefined evd in
+  let (fixnames,fixrs,fixdefs,fixtypes) = fix in
   let collect_evars name def typ impargs =
     (* Generalize by the recursive prototypes  *)
     let terms = [def; typ] in
-    let using = Option.map (fun using -> Proof_using.definition_using env evd ~using ~terms) using in
+    let using = Option.map (fun using -> Proof_using.definition_using env evd ~fixnames ~using ~terms) using in
     let def = nf_evar evd (EConstr.it_mkNamedLambda_or_LetIn evd def rec_sign) in
     let typ = nf_evar evd (EConstr.it_mkNamedProd_or_LetIn evd typ rec_sign) in
     let deps = collect_evars_of_term evd def typ in
@@ -299,7 +311,6 @@ let do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind fixl =
     let cinfo = Declare.CInfo.make ~name ~typ ~impargs ?using () in
     (cinfo, def, evars)
   in
-  let (fixnames,fixrs,fixdefs,fixtypes) = fix in
   let fiximps = List.map pi2 info in
   let fixdefs = List.map out_def fixdefs in
   let defs = List.map4 collect_evars fixnames fixdefs fixtypes fiximps in
@@ -329,16 +340,16 @@ let do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind fixl =
   | Declare.Obls.IsCoFixpoint -> Decls.(IsDefinition CoFixpoint)
   in
   let ntns = List.map_append (fun { Vernacexpr.notations } -> List.map Metasyntax.prepare_where_notation notations ) fixl in
-  let info = Declare.Info.make ~poly ~scope ~kind ~udecl ?typing_flags () in
+  let info = Declare.Info.make ~poly ~scope ?clearbody ~kind ~udecl ?typing_flags ?deprecation () in
   Declare.Obls.add_mutual_definitions ~pm defs ~info ~uctx ~ntns fixkind
 
-let do_fixpoint ~pm ~scope ~poly ?typing_flags ?using l =
+let do_fixpoint ~pm ~scope ?clearbody ~poly ?typing_flags ?deprecation ?using l =
   let g = List.map (fun { Vernacexpr.rec_order } -> rec_order) l in
     match g, l with
     | [Some { CAst.v = CWfRec (n,r) }],
       [ Vernacexpr.{fname={CAst.v=id}; univs; binders; rtype; body_def; notations} ] ->
         let recarg = mkIdentC n.CAst.v in
-        build_wellfounded pm (id, univs, binders, rtype, out_def body_def) poly ?typing_flags ?using r recarg notations
+        build_wellfounded pm (id, univs, binders, rtype, out_def body_def) poly ?typing_flags ?deprecation ?using r recarg notations
 
     | [Some { CAst.v = CMeasureRec (n, m, r) }],
       [Vernacexpr.{fname={CAst.v=id}; univs; binders; rtype; body_def; notations }] ->
@@ -359,11 +370,11 @@ let do_fixpoint ~pm ~scope ~poly ?typing_flags ?using l =
           Vernacexpr.(ComFixpoint.adjust_rec_order ~structonly:true fix.binders fix.rec_order)) l in
       let fixkind = Declare.Obls.IsFixpoint annots in
       let l = List.map2 (fun fix rec_order -> { fix with Vernacexpr.rec_order }) l annots in
-      do_program_recursive ~pm ~scope ~poly ?typing_flags ?using fixkind l
+      do_program_recursive ~pm ~scope ?clearbody ~poly ?typing_flags ?using fixkind l
     | _, _ ->
       CErrors.user_err
         (str "Well-founded fixpoints not allowed in mutually recursive blocks.")
 
-let do_cofixpoint ~pm ~scope ~poly ?using fixl =
+let do_cofixpoint ~pm ~scope ~poly ?deprecation ?using fixl =
   let fixl = List.map (fun fix -> { fix with Vernacexpr.rec_order = None }) fixl in
-  do_program_recursive ~pm ~scope ~poly ?using Declare.Obls.IsCoFixpoint fixl
+  do_program_recursive ~pm ~scope ~poly ?deprecation ?using Declare.Obls.IsCoFixpoint fixl
