@@ -17,7 +17,6 @@ open CAst
 
 open Extend
 open Constrexpr
-open Constrexpr_ops
 open Vernacexpr
 open Declaremods
 open Pputils
@@ -63,13 +62,15 @@ let pr_uconstraint (l, d, r) =
   pr_sort_name_expr l ++ spc () ++ Univ.pr_constraint_type d ++ spc () ++
   pr_sort_name_expr r
 
-let pr_univ_name_list = function
-  | None -> mt ()
-  | Some l ->
-    str "@{" ++ prlist_with_sep spc pr_lname l ++ str"}"
+let pr_full_univ_name_list = function
+  | None -> mt()
+  | Some (ql, ul) ->
+    str "@{" ++ prlist_with_sep spc pr_lname ql ++
+    (if List.is_empty ql then mt() else strbrk " | ") ++
+    prlist_with_sep spc pr_lname ul ++ str "}"
 
 let pr_variance_lident (lid,v) =
-  let v = Option.cata Univ.Variance.pr (mt()) v in
+  let v = Option.cata UVars.Variance.pr (mt()) v in
   v ++ pr_lident lid
 
 let pr_univdecl_instance l extensible =
@@ -82,7 +83,7 @@ let pr_cumul_univdecl_instance l extensible =
 
 let pr_univdecl_constraints l extensible =
   if List.is_empty l && extensible then mt ()
-  else str"|" ++ spc () ++ prlist_with_sep (fun () -> str",") pr_uconstraint l ++
+  else str"|" ++ spc () ++ prlist_with_sep (fun () -> strbrk" | ") pr_uconstraint l ++
        (if extensible then str"+" else mt())
 
 let pr_universe_decl l =
@@ -142,9 +143,7 @@ let pr_import_module (m,f) =
   | ImportNames ns -> surround (prlist_with_sep pr_comma pr_one_import_filter_name ns)
 
 let sep_end = function
-  | VernacBullet _
-  | VernacSubproof _
-  | VernacEndSubproof -> str""
+  | VernacSynPure (VernacBullet _ | VernacSubproof _ | VernacEndSubproof) -> str""
   | _ -> str"."
 
 let sep = fun _ -> spc()
@@ -199,7 +198,7 @@ let level_of_pattern_level = function None -> DefaultLevel | Some n -> NumLevel
 let pr_constr_as_binder_kind = let open Notation_term in function
     | AsIdent -> spc () ++ keyword "as ident"
     | AsName -> spc () ++ keyword "as name"
-    | AsNameOrPattern -> spc () ++ keyword "as pattern"
+    | AsAnyPattern -> spc () ++ keyword "as pattern"
     | AsStrictPattern -> spc () ++ keyword "as strict pattern"
 
 let pr_strict b = if b then str "strict " else mt ()
@@ -235,10 +234,16 @@ let pr_search_where = function
   | InConcl, true -> str "headconcl:"
   | InConcl, false -> str "concl:"
 
+let pr_delimiter_depth = function
+  | DelimOnlyTmpScope -> str "%_"
+  | DelimUnboundedScope -> str "%"
+
+let pr_scope_delimiter (d, sc) = pr_delimiter_depth d ++ str sc
+
 let pr_search_item = function
   | SearchSubPattern (where,p) ->
     pr_search_where where ++ pr_constr_pattern_expr p
-  | SearchString (where,s,sc) -> pr_search_where where ++ qs s ++ pr_opt (fun sc -> str "%" ++ str sc) sc
+  | SearchString (where,s,sc) -> pr_search_where where ++ qs s ++ pr_opt pr_scope_delimiter sc
   | SearchKind kind -> str "is:" ++ str (string_of_logical_kind kind)
 
 let rec pr_search_request = function
@@ -388,7 +393,8 @@ let pr_module_binders l pr_c =
   prlist_strict (pr_module_vardecls pr_c) l
 
 let pr_type_option pr_c = function
-  | { v = CHole (k, Namegen.IntroAnonymous, _) } -> mt()
+  | { v = CHole (Some GNamedHole _) } as c -> brk(0,2) ++ str" :" ++ pr_c c
+  | { v = CHole _ } -> mt()
   | _ as c -> brk(0,2) ++ str" :" ++ pr_c c
 
 let pr_binders_arg =
@@ -489,17 +495,20 @@ let pr_syntax_modifiers = function
   | l -> spc() ++
          hov 1 (str"(" ++ prlist_with_sep sep_v2 pr_syntax_modifier l ++ str")")
 
-let pr_decl_notation prc decl_ntn =
+let pr_notation_declaration ntn_decl =
   let open Vernacexpr in
   let
-    { decl_ntn_string = {CAst.loc;v=ntn};
-      decl_ntn_interp = c;
-      decl_ntn_modifiers = modifiers;
-      decl_ntn_scope = scopt } = decl_ntn in
-  fnl () ++ keyword "where " ++ qs ntn ++ str " := "
-  ++ Flags.without_option Flags.beautify prc c
+    { ntn_decl_string = {CAst.loc;v=ntn};
+      ntn_decl_interp = c;
+      ntn_decl_modifiers = modifiers;
+      ntn_decl_scope = scopt } = ntn_decl in
+  qs ntn ++ spc () ++ str ":=" ++ spc ()
+  ++ Flags.without_option Flags.beautify pr_constr c
   ++ pr_syntax_modifiers modifiers
-  ++ pr_opt (fun sc -> str ": " ++ str sc) scopt
+  ++ pr_opt (fun sc -> spc () ++ str ":" ++ spc () ++ str sc) scopt
+
+let pr_where_notation decl_ntn =
+  fnl () ++ keyword "where " ++ pr_notation_declaration decl_ntn
 
 let pr_rec_definition { fname; univs; rec_order; binders; rtype; body_def; notations } =
   let pr_pure_lconstr c = Flags.without_option Flags.beautify pr_lconstr c in
@@ -507,7 +516,7 @@ let pr_rec_definition { fname; univs; rec_order; binders; rtype; body_def; notat
   pr_ident_decl (fname,univs) ++ pr_binders_arg binders ++ annot
   ++ pr_type_option (fun c -> spc() ++ pr_lconstr_expr c) rtype
   ++ pr_opt (fun def -> str":=" ++ brk(1,2) ++ pr_pure_lconstr def) body_def
-  ++ prlist (pr_decl_notation @@ pr_constr) notations
+  ++ prlist pr_where_notation notations
 
 let pr_statement head (idpl,(bl,c)) =
   hov 2
@@ -525,6 +534,11 @@ let pr_lconstrarg c =
   spc () ++ pr_lconstr c
 let pr_intarg n = spc () ++ int n
 
+let pr_vernac_attributes =
+  function
+  | [] -> mt ()
+  | flags ->  str "#[" ++ prlist_with_sep pr_comma Attributes.pr_vernac_flag flags ++ str "]" ++ cut ()
+
 let pr_oc coe ins = match coe, ins with
   | NoCoercion, NoInstance -> str" :"
   | AddCoercion, NoInstance -> str" :>"
@@ -532,24 +546,27 @@ let pr_oc coe ins = match coe, ins with
   | AddCoercion, BackInstance -> str" ::>"
   | _, BackInstanceWarning -> str" :>"  (* remove this line at end of deprecation phase *)
 
-let pr_record_field (x, { rf_coercion = coe ; rf_instance = ins ; rf_priority = pri ; rf_notation = ntn }) =
+let pr_record_field (x, { rfu_attrs = attr ; rfu_coercion = coe ; rfu_instance = ins ; rfu_priority = pri ; rfu_notation = ntn }) =
   let prx = match x with
     | AssumExpr (id,binders,t) ->
-      hov 1 (pr_lname id ++
+      hov 1 (pr_vernac_attributes attr ++
+             pr_lname id ++
              pr_binders_arg binders ++ spc() ++
              pr_oc coe ins ++ spc() ++
              pr_lconstr_expr t)
     | DefExpr(id,binders,b,opt) -> (match opt with
         | Some t ->
-          hov 1 (pr_lname id ++
+          hov 1 (pr_vernac_attributes attr ++
+                 pr_lname id ++
                  pr_binders_arg binders ++ spc() ++
                  pr_oc coe ins ++ spc() ++
                  pr_lconstr_expr t ++ str" :=" ++ pr_lconstr b)
         | None ->
-          hov 1 (pr_lname id ++ str" :=" ++ spc() ++
+          hov 1 (pr_vernac_attributes attr ++
+                 pr_lname id ++ str" :=" ++ spc() ++
                  pr_lconstr b)) in
   let prpri = match pri with None -> mt() | Some i -> str "| " ++ int i in
-  prx ++ prpri ++ prlist (pr_decl_notation @@ pr_constr) ntn
+  prx ++ prpri ++ prlist pr_where_notation ntn
 
 let pr_record_decl c fs obinder =
   pr_opt pr_lident c ++ pr_record "{" "}" pr_record_field fs ++
@@ -613,7 +630,7 @@ let pr_printable = function
     let pr_subgraph = prlist_with_sep spc pr_qualid in
     keyword cmd ++ pr_opt pr_subgraph g ++ pr_opt str fopt
   | PrintName (qid,udecl) ->
-    keyword "Print" ++ spc()  ++ pr_smart_global qid ++ pr_univ_name_list udecl
+    keyword "Print" ++ spc()  ++ pr_smart_global qid ++ pr_full_univ_name_list udecl
   | PrintModuleType qid ->
     keyword "Print Module Type" ++ spc() ++ pr_qualid qid
   | PrintModule qid ->
@@ -628,7 +645,7 @@ let pr_printable = function
     keyword "Print Visibility" ++ pr_opt str s
   | PrintAbout (qid,l,gopt) ->
     pr_opt (fun g -> Goal_select.pr_goal_selector g ++ str ":"++ spc()) gopt
-    ++ keyword "About" ++ spc()  ++ pr_smart_global qid ++ pr_univ_name_list l
+    ++ keyword "About" ++ spc()  ++ pr_smart_global qid ++ pr_full_univ_name_list l
   | PrintImplicit qid ->
     keyword "Print Implicit" ++ spc()  ++ pr_smart_global qid
   (* spiwack: command printing all the axioms and section variables used in a
@@ -668,7 +685,7 @@ let pr_using e =
 let pr_extend s cl =
   let pr_arg a =
     try pr_gen a
-    with Failure _ -> str "<error in " ++ str (fst s) ++ str ">" in
+    with Failure _ -> str "<error in " ++ str s.ext_entry ++ str ">" in
   try
     let rl = Egramml.get_extend_vernac_rule s in
     let rec aux rl cl =
@@ -679,19 +696,11 @@ let pr_extend s cl =
       | _ -> assert false in
     hov 1 (pr_sequence identity (aux rl cl))
   with Not_found ->
-    hov 1 (str "TODO(" ++ str (fst s) ++ spc () ++ prlist_with_sep sep pr_arg cl ++ str ")")
+    hov 1 (str "TODO(" ++ str s.ext_entry ++ spc () ++ prlist_with_sep sep pr_arg cl ++ str ")")
 
-let pr_vernac_expr v =
+let pr_synpure_vernac_expr v =
   let return = tag_vernac v in
   match v with
-  | VernacLoad (f,s) ->
-    return (
-      keyword "Load"
-      ++ if f then
-        (spc() ++ keyword "Verbose" ++ spc())
-      else
-        spc() ++ qs s
-    )
 
   (* Proof management *)
   | VernacAbortAll ->
@@ -731,6 +740,8 @@ let pr_vernac_expr v =
   | VernacCheckGuard ->
     return (keyword "Guarded")
 
+  | VernacValidateProof -> return (keyword "Validate Proof")
+
   (* Resetting *)
   | VernacResetName id ->
     return (keyword "Reset" ++ spc() ++ pr_lident id)
@@ -765,24 +776,6 @@ let pr_vernac_expr v =
       keyword "Bind Scope" ++ spc () ++ str sc ++
       spc() ++ keyword "with" ++ spc () ++ prlist_with_sep spc pr_class_rawexpr cll
     )
-  | VernacNotation (infix,c,({v=s},l),opt) ->
-    return (
-      hov 2 (hov 0 (keyword (if infix then "Infix" else "Notation") ++ spc() ++ qs s ++
-             str " :=" ++ Flags.without_option Flags.beautify pr_constrarg c) ++
-             pr_syntax_modifiers l ++
-             (match opt with
-              | None -> mt()
-              | Some sc -> spc() ++ str":" ++ spc() ++ str sc))
-    )
-  | VernacReservedNotation (_, (s, l)) ->
-    return (
-      keyword "Reserved Notation" ++ spc() ++ pr_ast qs s ++
-      pr_syntax_modifiers l
-    )
-  | VernacDeclareCustomEntry s ->
-    return (
-      keyword "Declare Custom Entry " ++ str s
-    )
   | VernacEnableNotation (on,rule,interp,flags,scope) ->
     let pr_flag = function
       | EnableNotationEntry CAst.{v=InConstrEntry} -> str "in constr"
@@ -809,11 +802,12 @@ let pr_vernac_expr v =
 
   (* Gallina *)
   | VernacDefinition ((discharge,kind),id,b) -> (* A verifier... *)
-    let pr_def_token dk =
+    let isgoal = Name.is_anonymous (fst id).v in
+    let pr_def_token =
       keyword (
-        if Name.is_anonymous (fst id).v
+        if isgoal
         then "Goal"
-        else string_of_definition_object_kind dk)
+        else string_of_definition_object_kind kind)
     in
     let pr_reduce = function
       | None -> mt()
@@ -822,24 +816,22 @@ let pr_vernac_expr v =
         pr_red_expr r ++
         keyword " in" ++ spc()
     in
-    let pr_def_body = function
+    let pr_def_body = match b with
       | DefineBody (bl,red,body,d) ->
         let ty = match d with
           | None -> mt()
           | Some ty -> spc() ++ str":" ++ pr_spc_lconstr ty
         in
-        (pr_binders_arg bl,ty,Some (pr_reduce red ++ pr_lconstr body))
+        pr_binders_arg bl  ++ ty ++ str " :=" ++ spc() ++ pr_reduce red ++ pr_lconstr body
       | ProveBody (bl,t) ->
-        let typ u = if (fst id).v = Anonymous then (assert (bl = []); u) else (str" :" ++ u) in
-        (pr_binders_arg bl, typ (pr_spc_lconstr t), None) in
-    let (binds,typ,c) = pr_def_body b in
+        let typ u = if isgoal then (assert (bl = []); u) else (str" :" ++ u) in
+        pr_binders_arg bl ++ typ (pr_spc_lconstr t)
+    in
     return (
       hov 2 (
-        pr_def_token kind ++ spc()
-        ++ pr_lname_decl id ++ binds ++ typ
-        ++ (match c with
-            | None -> mt()
-            | Some cc -> str" :=" ++ spc() ++ cc))
+        pr_def_token
+        ++ (if isgoal then mt() else spc() ++ pr_lname_decl id)
+        ++ pr_def_body)
     )
 
   | VernacStartTheoremProof (ki,l) ->
@@ -869,8 +861,8 @@ let pr_vernac_expr v =
     return (hov 2 (pr_assumption_token (n > 1) discharge kind ++
                    pr_non_empty_arg pr_assumption_inline t ++ spc() ++ assumptions))
   | VernacInductive (f,l) ->
-    let pr_constructor ((coe,ins),(id,c)) =
-      hov 2 (pr_lident id ++ pr_oc coe ins ++
+    let pr_constructor ((attr,coe,ins),(id,c)) =
+      hov 2 (pr_vernac_attributes attr ++ pr_lident id ++ pr_oc coe ins ++
              Flags.without_option Flags.beautify pr_spc_lconstr c)
     in
     let pr_constructor_list l = match l with
@@ -891,7 +883,7 @@ let pr_vernac_expr v =
         pr_opt (fun p -> str "|" ++ spc() ++ pr_and_type_binders_arg p) indpar ++
         pr_opt (fun s -> str":" ++ spc() ++ pr_lconstr_expr s) s ++
         str" :=") ++ pr_constructor_list lc ++
-      prlist (pr_decl_notation @@ pr_constr) ntn
+      prlist pr_where_notation ntn
     in
     let kind =
       match f with
@@ -924,7 +916,7 @@ let pr_vernac_expr v =
       pr_ident_decl (fname,univs) ++ spc() ++ pr_binders binders ++ spc() ++ str":" ++
       spc() ++ pr_lconstr_expr rtype ++
       pr_opt (fun def -> str":=" ++ brk(1,2) ++ pr_lconstr def) body_def ++
-      prlist (pr_decl_notation @@ pr_constr) notations
+      prlist pr_where_notation notations
     in
     return (
       hov 0 (local ++ keyword "CoFixpoint" ++ spc() ++
@@ -957,35 +949,9 @@ let pr_vernac_expr v =
     )
 
   (* Gallina extensions *)
-  | VernacBeginSection id ->
-    return (hov 2 (keyword "Section" ++ spc () ++ pr_lident id))
-  | VernacEndSegment id ->
-    return (hov 2 (keyword "End" ++ spc() ++ pr_lident id))
   | VernacNameSectionHypSet (id,set) ->
     return (hov 2 (keyword "Collection" ++ spc() ++ pr_lident id ++ spc()++
                    str ":="++spc()++pr_using set))
-  | VernacExtraDependency(from,file,id) ->
-    return (
-      hov 2
-        (keyword "From" ++ spc () ++ pr_module from ++ spc () ++
-         keyword "Extra" ++ spc() ++ keyword "Dependency" ++ spc() ++ qs file ++
-         pr_opt (fun x -> spc() ++ keyword "as" ++ spc () ++ pr_id x) id)
-    )
-  | VernacRequire (from, exp, l) ->
-    let from = match from with
-      | None -> mt ()
-      | Some r -> keyword "From" ++ spc () ++ pr_module r ++ spc ()
-    in
-    return (
-      hov 2
-        (from ++ keyword "Require" ++ spc() ++ pr_require_token exp ++
-         prlist_with_sep sep pr_import_module l)
-    )
-  | VernacImport (f,l) ->
-    return (
-      pr_export_with_cats f ++ spc() ++
-      prlist_with_sep sep pr_import_module l
-    )
   | VernacCanonical q ->
     return (
       keyword "Canonical Structure" ++ spc() ++ pr_smart_global q
@@ -1058,63 +1024,6 @@ let pr_vernac_expr v =
       hov 1 (keyword "Existing" ++ spc () ++ keyword "Class" ++ spc () ++ pr_qualid id)
     )
 
-  (* Modules and Module Types *)
-  | VernacDefineModule (export,m,bl,tys,bd) ->
-    let b = pr_module_binders bl pr_lconstr in
-    return (
-      hov 2 (keyword "Module" ++ spc() ++ pr_require_token export ++
-             pr_lident m ++ b ++
-             pr_of_module_type pr_lconstr tys ++
-             (if List.is_empty bd then mt () else str ":= ") ++
-             prlist_with_sep (fun () -> str " <+")
-               (pr_module_ast_inl true pr_lconstr) bd)
-    )
-  | VernacDeclareModule (export,id,bl,m1) ->
-    let b = pr_module_binders bl pr_lconstr in
-    return (
-      hov 2 (keyword "Declare Module" ++ spc() ++ pr_require_token export ++
-             pr_lident id ++ b ++ str " :" ++
-             pr_module_ast_inl true pr_lconstr m1)
-    )
-  | VernacDeclareModuleType (id,bl,tyl,m) ->
-    let b = pr_module_binders bl pr_lconstr in
-    let pr_mt = pr_module_ast_inl true pr_lconstr in
-    return (
-      hov 2 (keyword "Module Type " ++ pr_lident id ++ b ++
-             prlist_strict (fun m -> str " <:" ++ pr_mt m) tyl ++
-             (if List.is_empty m then mt () else str ":= ") ++
-             prlist_with_sep (fun () -> str " <+ ") pr_mt m)
-    )
-  | VernacInclude (mexprs) ->
-    let pr_m = pr_module_ast_inl false pr_lconstr in
-    return (
-      hov 2 (keyword "Include" ++ spc() ++
-             prlist_with_sep (fun () -> str " <+ ") pr_m mexprs)
-    )
-
-  (* Auxiliary file and library management *)
-  | VernacAddLoadPath { implicit; physical_path; logical_path } ->
-    return (
-      hov 2
-        (keyword "Add" ++
-         (if implicit then spc () ++ keyword "Rec" ++ spc () else spc()) ++
-         keyword "LoadPath" ++ spc() ++ qs physical_path ++
-         spc() ++ keyword "as" ++ spc() ++ DirPath.print logical_path))
-  | VernacRemoveLoadPath s ->
-    return (keyword "Remove LoadPath" ++ qs s)
-  | VernacAddMLPath (s) ->
-    return (
-      keyword "Add"
-      ++ keyword "ML Path"
-      ++ qs s
-    )
-  | VernacDeclareMLModule (l) ->
-    return (
-      hov 2 (keyword "Declare ML Module" ++ spc() ++ prlist_with_sep sep qs l)
-    )
-  | VernacChdir s ->
-    return (keyword "Cd" ++ pr_opt qs s)
-
   (* Commands *)
   | VernacCreateHintDb (dbname,b) ->
     return (
@@ -1124,7 +1033,7 @@ let pr_vernac_expr v =
   | VernacRemoveHints (dbnames, ids) ->
     return (
       hov 1 (keyword "Remove Hints" ++ spc () ++
-             prlist_with_sep spc (fun r -> pr_id (coerce_reference_to_id r)) ids ++
+             prlist_with_sep spc (fun r -> pr_qualid r) ids ++
              pr_opt_hintbases dbnames)
     )
   | VernacHints (dbnames,h) ->
@@ -1140,7 +1049,7 @@ let pr_vernac_expr v =
       hov 2 (
         keyword "Arguments" ++ spc() ++
         pr_smart_global q ++
-        let pr_s = prlist (fun {v=s} -> str "%" ++ str s) in
+        let pr_s = prlist (fun {v=s} -> pr_scope_delimiter s) in
         let pr_if b x = if b then x else str "" in
         let pr_one_arg (x,k) = pr_if k (str"!") ++ Name.print x in
         let pr_br imp force x =
@@ -1157,7 +1066,9 @@ let pr_vernac_expr v =
           else
             let rec fold extra = function
               | RealArg arg :: tl when
-                  List.equal (fun a b -> String.equal a.CAst.v b.CAst.v) arg.notation_scope s
+                  List.equal
+                    (fun a b -> let da, a = a.CAst.v in let db, b = b.CAst.v in
+                     da = db && String.equal a b) arg.notation_scope s
                   && arg.implicit_status = imp ->
                 fold ((arg.name,arg.recarg_like) :: extra) tl
               | args -> List.rev extra, args
@@ -1243,12 +1154,6 @@ let pr_vernac_expr v =
       hov 1 (keyword "Strategy" ++ spc() ++
              hv 0 (prlist_with_sep sep pr_line l))
     )
-  | VernacSetOption (export, na,v) ->
-    let export = if export then keyword "Export" ++ spc () else mt () in
-    let set = if v == OptionUnset then "Unset" else "Set" in
-    return (
-      hov 2 (export ++ keyword set ++ spc() ++ pr_set_option na v)
-    )
   | VernacAddOption (na,l) ->
     return (
       hov 2 (keyword "Add" ++ spc() ++ pr_printoption na (Some l))
@@ -1320,10 +1225,12 @@ let pr_vernac_expr v =
         (keyword "Comments" ++ spc()
          ++ prlist_with_sep sep (pr_comment pr_constr) l)
     )
-
-  (* For extension *)
-  | VernacExtend (s,c) ->
-    return (pr_extend s c)
+  | VernacAttributes attrs ->
+    return (
+      hov 2
+        (keyword "Attributes" ++ spc () ++
+         pr_vernac_attributes attrs)
+    )
   | VernacProof (None, None) ->
     return (keyword "Proof")
   | VernacProof (None, Some e) ->
@@ -1337,8 +1244,6 @@ let pr_vernac_expr v =
       keyword "using" ++ spc() ++ pr_using e ++ spc() ++
       keyword "with" ++ spc() ++ pr_gen te
     )
-  | VernacProofMode s ->
-    return (keyword "Proof Mode" ++ str s)
   | VernacBullet b ->
     (* XXX: Redundant with Proof_bullet.print *)
     return (let open Proof_bullet in begin match b with
@@ -1353,9 +1258,114 @@ let pr_vernac_expr v =
   | VernacEndSubproof ->
     return (str "}")
 
+let pr_synterp_vernac_expr v =
+  let return = tag_vernac v in
+  match v with
+  | VernacLoad (f,s) ->
+    return (
+      keyword "Load"
+      ++ if f then
+        (spc() ++ keyword "Verbose" ++ spc())
+      else
+        spc() ++ qs s
+    )
+
+  | VernacBeginSection id ->
+    return (hov 2 (keyword "Section" ++ spc () ++ pr_lident id))
+  | VernacEndSegment id ->
+    return (hov 2 (keyword "End" ++ spc() ++ pr_lident id))
+  | VernacNotation (infix,ntn_decl) ->
+    return (
+      hov 2 (hov 0 (keyword (if infix then "Infix" else "Notation") ++ spc() ++
+      pr_notation_declaration ntn_decl))
+    )
+  | VernacReservedNotation (_, (s, l)) ->
+    return (
+      keyword "Reserved Notation" ++ spc() ++ pr_ast qs s ++
+      pr_syntax_modifiers l
+    )
+  | VernacDeclareCustomEntry s ->
+    return (
+      keyword "Declare Custom Entry " ++ str s
+    )
+  | VernacRequire (from, exp, l) ->
+    let from = match from with
+      | None -> mt ()
+      | Some r -> keyword "From" ++ spc () ++ pr_module r ++ spc ()
+    in
+    return (
+      hov 2
+        (from ++ keyword "Require" ++ spc() ++ pr_require_token exp ++
+         prlist_with_sep sep pr_import_module l)
+    )
+  | VernacImport (f,l) ->
+    return (
+      pr_export_with_cats f ++ spc() ++
+      prlist_with_sep sep pr_import_module l
+    )
+  (* Modules and Module Types *)
+  | VernacDefineModule (export,m,bl,tys,bd) ->
+    let b = pr_module_binders bl pr_lconstr in
+    return (
+      hov 2 (keyword "Module" ++ spc() ++ pr_require_token export ++
+             pr_lident m ++ b ++
+             pr_of_module_type pr_lconstr tys ++
+             (if List.is_empty bd then mt () else str ":= ") ++
+             prlist_with_sep (fun () -> str " <+")
+               (pr_module_ast_inl true pr_lconstr) bd)
+    )
+  | VernacDeclareModule (export,id,bl,m1) ->
+    let b = pr_module_binders bl pr_lconstr in
+    return (
+      hov 2 (keyword "Declare Module" ++ spc() ++ pr_require_token export ++
+             pr_lident id ++ b ++ str " :" ++
+             pr_module_ast_inl true pr_lconstr m1)
+    )
+  | VernacDeclareModuleType (id,bl,tyl,m) ->
+    let b = pr_module_binders bl pr_lconstr in
+    let pr_mt = pr_module_ast_inl true pr_lconstr in
+    return (
+      hov 2 (keyword "Module Type " ++ pr_lident id ++ b ++
+             prlist_strict (fun m -> str " <:" ++ pr_mt m) tyl ++
+             (if List.is_empty m then mt () else str ":= ") ++
+             prlist_with_sep (fun () -> str " <+ ") pr_mt m)
+    )
+  | VernacInclude (mexprs) ->
+    let pr_m = pr_module_ast_inl false pr_lconstr in
+    return (
+      hov 2 (keyword "Include" ++ spc() ++
+             prlist_with_sep (fun () -> str " <+ ") pr_m mexprs)
+    )
+
+  (* Auxiliary file and library management *)
+  | VernacDeclareMLModule (l) ->
+    return (
+      hov 2 (keyword "Declare ML Module" ++ spc() ++ prlist_with_sep sep qs l)
+    )
+  | VernacChdir s ->
+    return (keyword "Cd" ++ pr_opt qs s)
+  | VernacSetOption (export, na,v) ->
+    let export = if export then keyword "Export" ++ spc () else mt () in
+    let set = if v == OptionUnset then "Unset" else "Set" in
+    return (
+      hov 2 (export ++ keyword set ++ spc() ++ pr_set_option na v)
+    )
+  | VernacExtraDependency(from,file,id) ->
+    return (
+      hov 2
+        (keyword "From" ++ spc () ++ pr_module from ++ spc () ++
+         keyword "Extra" ++ spc() ++ keyword "Dependency" ++ spc() ++ qs file ++
+         pr_opt (fun x -> spc() ++ keyword "as" ++ spc () ++ pr_id x) id)
+    )
+  | VernacExtend (s,c) ->
+    return (pr_extend s c)
+  | VernacProofMode s ->
+    return (keyword "Proof Mode" ++ str s)
+
 let pr_control_flag (p : control_flag) =
   let w = match p with
-    | ControlTime _ -> keyword "Time"
+    | ControlTime -> keyword "Time"
+    | ControlInstructions -> keyword "Instructions"
     | ControlRedirect s -> keyword "Redirect" ++ spc() ++ qs s
     | ControlTimeout n -> keyword "Timeout " ++ int n
     | ControlFail -> keyword "Fail"
@@ -1365,10 +1375,10 @@ let pr_control_flag (p : control_flag) =
 
 let pr_vernac_control flags = Pp.prlist pr_control_flag flags
 
-let pr_vernac_attributes =
-  function
-  | [] -> mt ()
-  | flags ->  str "#[" ++ prlist_with_sep pr_comma Attributes.pr_vernac_flag flags ++ str "]" ++ cut ()
+let pr_vernac_expr v =
+  match v with
+  | VernacSynPure e -> pr_synpure_vernac_expr e
+  | VernacSynterp e -> pr_synterp_vernac_expr e
 
 let pr_vernac ({v = {control; attrs; expr}} as v) =
   tag_vernac v
