@@ -25,28 +25,13 @@ open Libobject
 open Libnames
 open Globnames
 open Printer
-open Printmod
 open Context.Rel.Declaration
 
 module RelDecl = Context.Rel.Declaration
 module NamedDecl = Context.Named.Declaration
 
-type object_pr = {
-  print_inductive           : MutInd.t -> UnivNames.univ_name_list option -> Pp.t;
-  print_constant_with_infos : Constant.t -> UnivNames.univ_name_list option -> Pp.t;
-  print_section_variable    : env -> Evd.evar_map -> variable -> Pp.t;
-  print_abbreviation        : env -> abbreviation -> Pp.t;
-  print_module              : ModPath.t -> Pp.t;
-  print_modtype             : ModPath.t -> Pp.t;
-  print_named_decl          : env -> Evd.evar_map -> Constr.named_declaration -> Pp.t;
-  print_library_leaf       : env -> Evd.evar_map -> bool -> ModPath.t -> Libobject.t -> Pp.t option;
-  print_context             : env -> Evd.evar_map -> bool -> int option -> Lib.library_segment -> Pp.t;
-  print_typed_value_in_env  : Environ.env -> Evd.evar_map -> EConstr.constr * EConstr.types -> Pp.t;
-  print_eval                : Reductionops.reduction_function -> env -> Evd.evar_map -> Constrexpr.constr_expr -> EConstr.unsafe_judgment -> Pp.t;
-}
-
-let gallina_print_module mp = print_module ~with_body:true mp
-let gallina_print_modtype = print_modtype
+let print_module mp = Printmod.print_module ~with_body:true mp
+let print_modtype = Printmod.print_modtype
 
 (**************)
 (** Utilities *)
@@ -71,12 +56,12 @@ let print_basename sp = pr_global (GlobRef.ConstRef sp)
 let print_ref reduce ref udecl =
   let env = Global.env () in
   let typ, univs = Typeops.type_of_global_in_context env ref in
-  let inst = Univ.make_abstract_instance univs in
+  let inst = UVars.make_abstract_instance univs in
   let bl = Printer.universe_binders_with_opt_names (Environ.universes_of_global env ref) udecl in
   let sigma = Evd.from_ctx (UState.of_names bl) in
   let typ =
     if reduce then
-      let ctx,ccl = Reductionops.splay_prod_assum env sigma (EConstr.of_constr typ)
+      let ctx,ccl = Reductionops.whd_decompose_prod_decls env sigma (EConstr.of_constr typ)
       in EConstr.to_constr sigma (EConstr.it_mkProd_or_LetIn ccl ctx)
     else typ in
   let typ = Arguments_renaming.rename_type typ ref in
@@ -140,14 +125,13 @@ let print_impargs_list prefix l =
 
 let need_expansion impl ref =
   let typ, _ = Typeops.type_of_global_in_context (Global.env ()) ref in
-  let ctx = Term.prod_assum typ in
+  let ctx = Term.prod_decls typ in
   let nprods = List.count is_local_assum ctx in
   not (List.is_empty impl) && List.length impl >= nprods &&
     let _,lastimpl = List.chop nprods impl in
       List.exists is_status_implicit lastimpl
 
 let print_impargs ref =
-  let ref = Smartlocate.smart_global ref in
   let impl = implicits_of_global ref in
   let has_impl = not (List.is_empty impl) in
   (* Need to reduce since implicits are computed with products flattened *)
@@ -209,7 +193,7 @@ let print_if_is_coercion ref =
 
 let pr_template_variables = function
   | [] -> mt ()
-  | vars -> str " on " ++ prlist_with_sep spc UnivNames.(pr_with_global_universes empty_binders) vars
+  | vars -> str " on " ++ prlist_with_sep spc UnivNames.pr_level_with_global_universes vars
 
 let print_polymorphism ref =
   let poly = Global.is_polymorphic ref in
@@ -258,7 +242,7 @@ let needs_extra_scopes ref scopes =
 
 let implicit_kind_of_status = function
   | None -> Anonymous, Glob_term.Explicit
-  | Some ((na,_,_),_,(maximal,_)) -> na, if maximal then Glob_term.MaxImplicit else Glob_term.NonMaxImplicit
+  | Some imp -> pi1 imp.impl_pos, if imp.impl_max then Glob_term.MaxImplicit else Glob_term.NonMaxImplicit
 
 let extra_implicit_kind_of_status imp =
   let _,imp = implicit_kind_of_status imp in
@@ -290,7 +274,7 @@ let rec main_implicits i renames recargs scopes impls =
       | [], (None::_ | []) -> (Anonymous, Glob_term.Explicit)
     in
     let notation_scope = match scopes with
-      | scope :: _ -> List.map CAst.make scope
+      | scope :: _ -> List.map (fun s -> CAst.make (Constrexpr.DelimUnboundedScope, s)) scope
       | [] -> []
     in
     let status = {Vernacexpr.implicit_status; name; recarg_like; notation_scope} in
@@ -316,12 +300,15 @@ let rec insert_fake_args volatile bidi impls =
 let print_arguments ref =
   let qid = Nametab.shortest_qualid_of_global Id.Set.empty ref in
   let flags, recargs, nargs_for_red =
-    let open Reductionops.ReductionBehaviour in
-    match get ref with
-    | None -> [], [], None
-    | Some NeverUnfold -> [`ReductionNeverUnfold], [], None
-    | Some (UnfoldWhen { nargs; recargs }) -> [], recargs, nargs
-    | Some (UnfoldWhenNoMatch { nargs; recargs }) -> [`ReductionDontExposeCase], recargs, nargs
+    match ref with
+    | ConstRef ref ->
+      begin match Reductionops.ReductionBehaviour.get ref with
+      | None -> [], [], None
+      | Some NeverUnfold -> [`ReductionNeverUnfold], [], None
+      | Some (UnfoldWhen { nargs; recargs }) -> [], recargs, nargs
+      | Some (UnfoldWhenNoMatch { nargs; recargs }) -> [`ReductionDontExposeCase], recargs, nargs
+      end
+    | _ -> [], [], None
   in
   let names, not_renamed =
     try Arguments_renaming.arguments_names ref, false
@@ -345,7 +332,7 @@ let print_arguments ref =
     let open Constrexpr in
     let open Vernacexpr in
     [Ppvernac.pr_vernac_expr
-       (VernacArguments (CAst.make (AN qid), impls, moreimpls, flags)) ++
+       (VernacSynPure (VernacArguments (CAst.make (AN qid), impls, moreimpls, flags))) ++
      (if not_renamed then mt () else
       fnl () ++ str "  (where some original arguments have been renamed)")]
 
@@ -443,25 +430,34 @@ let locate_any_name qid =
     try String.Map.iter iter !locatable_map; Undefined qid
     with ObjFound obj -> obj
 
+let canonical_info ref =
+  let cref = canonical_gr ref in
+  if GlobRef.UserOrd.equal ref cref then mt ()
+  else match Nametab.path_of_global cref with
+    | path -> spc() ++ str "(syntactically equal to" ++ spc() ++ pr_path path ++ str ")"
+    | exception Not_found -> spc() ++ str "(missing canonical, bug?)"
+
 let pr_located_qualid = function
   | Term ref ->
       let ref_str = let open GlobRef in match ref with
           ConstRef _ -> "Constant"
         | IndRef _ -> "Inductive"
         | ConstructRef _ -> "Constructor"
-        | VarRef _ -> "Variable" in
-      str ref_str ++ spc () ++ pr_path (Nametab.path_of_global ref)
+        | VarRef _ -> "Variable"
+      in
+      let extra = canonical_info ref in
+      str ref_str ++ spc () ++ pr_path (Nametab.path_of_global ref) ++ extra
   | Abbreviation kn ->
       str "Notation" ++ spc () ++ pr_path (Nametab.path_of_abbreviation kn)
   | Dir dir ->
       let s,mp =
         let open Nametab in
         let open GlobDirRef in match dir with
-        | DirOpenModule mp -> "Open Module", mp
-        | DirOpenModtype mp -> "Open Module Type", mp
-        | DirOpenSection _ -> anomaly (Pp.str"Sections are not locatable")
+        | DirOpenModule mp -> "Open Module", ModPath.print mp
+        | DirOpenModtype mp -> "Open Module Type", ModPath.print mp
+        | DirOpenSection dir -> "Open Section", DirPath.print dir
       in
-      str s ++ spc () ++ str (ModPath.to_string mp)
+      str s ++ spc () ++ mp
   | Module mp ->
     str "Module" ++ spc () ++ DirPath.print (Nametab.dirpath_of_module mp)
   | ModuleType mp ->
@@ -577,7 +573,7 @@ let print_located_qualid ref = print_located_qualid "object" LocAny ref
 (**** Printing declarations and judgments *)
 (****  Gallina layer                  *****)
 
-let gallina_print_typed_value_in_env env sigma (trm,typ) =
+let print_typed_value_in_env env sigma (trm,typ) =
   (pr_leconstr_env ~inctx:true env sigma trm ++ fnl () ++
      str "     : " ++ pr_letype_env env sigma typ)
 
@@ -598,7 +594,7 @@ let print_named_def env sigma name body typ =
 let print_named_assum env sigma name typ =
   str "*** [" ++ str name ++ str " : " ++ pr_ltype_env env sigma typ ++ str "]"
 
-let gallina_print_named_decl env sigma =
+let print_named_decl_gen env sigma =
   let open Context.Named.Declaration in
   function
   | LocalAssum (id, typ) ->
@@ -612,19 +608,19 @@ let assumptions_for_print lna =
 (*********************)
 (* *)
 
-let gallina_print_inductive sp udecl =
+let print_inductive sp udecl =
   let env = Global.env() in
   let mib = Environ.lookup_mind sp env in
   let mipv = mib.mind_packets in
-  pr_mutual_inductive_body env sp mib udecl ++
+  Printmod.pr_mutual_inductive_body env sp mib udecl ++
   with_line_skip
     (print_primitive_record mib.mind_finite mipv mib.mind_record @
      print_inductive_args sp mipv)
 
 let print_named_decl env sigma id =
-  gallina_print_named_decl env sigma (Global.lookup_named id) ++ fnl ()
+  print_named_decl_gen env sigma (Global.lookup_named id) ++ fnl ()
 
-let gallina_print_section_variable env sigma id =
+let print_section_variable env sigma id =
   print_named_decl env sigma id ++
   with_line_skip (print_name_infos (GlobRef.VarRef id))
 
@@ -638,7 +634,7 @@ let print_typed_body env evd (val_0,typ) =
 let print_instance sigma cb =
   if Declareops.constant_is_polymorphic cb then
     let univs = Declareops.constant_polymorphic_context cb in
-    let inst = Univ.make_abstract_instance univs in
+    let inst = UVars.make_abstract_instance univs in
     pr_universe_instance sigma inst
   else mt()
 
@@ -670,11 +666,11 @@ let print_constant with_values sp udecl =
       (if with_values then print_typed_body env sigma (Some c,typ) else pr_ltype typ)++
       Printer.pr_universes sigma univs ?priv)
 
-let gallina_print_constant_with_infos sp udecl =
+let print_constant_with_infos sp udecl =
   print_constant true sp udecl ++
   with_line_skip (print_name_infos (GlobRef.ConstRef sp))
 
-let gallina_print_abbreviation env kn =
+let print_abbreviation env kn =
   let qid = Nametab.shortest_qualid_of_abbreviation Id.Set.empty kn
   and (vars,a) = Abbreviation.search_abbreviation kn in
   let c = Notation_ops.glob_constr_of_notation_constr a in
@@ -698,7 +694,7 @@ let handle h (Libobject.Dyn.Dyn (tag, o)) = match DynHandle.find tag h with
    no reason that an object in the stack corresponds to a user-facing
    declaration. It may have been so at the time this was written, but this
    needs to be done in a more principled way. *)
-let gallina_print_library_leaf env sigma with_values mp lobj =
+let print_library_leaf env sigma with_values mp lobj =
   match lobj with
   | AtomicObject o ->
     let handler =
@@ -713,13 +709,13 @@ let gallina_print_library_leaf env sigma with_values mp lobj =
       end @@
       DynHandle.add DeclareInd.Internal.objInductive begin fun (id,_) ->
         let kn = MutInd.make2 mp (Label.of_id id) in
-        Some (gallina_print_inductive kn None)
+        Some (print_inductive kn None)
       end @@
       DynHandle.empty
     in
     handle handler o
   | ModuleObject (id,_) ->
-    Some (print_module ~with_body:with_values (MPdot (mp,Label.of_id id)))
+    Some (Printmod.print_module ~with_body:with_values (MPdot (mp,Label.of_id id)))
   | ModuleTypeObject (id,_) ->
     Some (print_modtype (MPdot (mp, Label.of_id id)))
   | IncludeObject _ | KeepObject _ | ExportObject _ -> None
@@ -733,7 +729,7 @@ let print_leaves env sigma with_values mp =
     | [] -> n, mt()
     | o :: rest ->
       if is_done n then n, mt()
-      else begin match gallina_print_library_leaf env sigma with_values mp o with
+      else begin match print_library_leaf env sigma with_values mp o with
         | Some pp ->
           let n, prest = prec (decr n) rest in
           n, prest ++ pp
@@ -742,7 +738,7 @@ let print_leaves env sigma with_values mp =
   in
   prec
 
-let gallina_print_context env sigma with_values =
+let print_context env sigma with_values =
   let rec prec n = function
     | [] -> mt()
     | (node, leaves) :: rest ->
@@ -765,42 +761,6 @@ let print_library_node = function
   | Lib.CompilingLibrary { Nametab.obj_dir; _ } ->
     str " >>>>>>> Library " ++ DirPath.print obj_dir
 
-let gallina_print_eval red_fun env sigma _ {uj_val=trm;uj_type=typ} =
-  let ntrm = red_fun env sigma trm in
-  (str "     = " ++ gallina_print_typed_value_in_env env sigma (ntrm,typ))
-
-(******************************************)
-(**** Printing abstraction layer          *)
-
-let default_object_pr = {
-  print_inductive           = gallina_print_inductive;
-  print_constant_with_infos = gallina_print_constant_with_infos;
-  print_section_variable    = gallina_print_section_variable;
-  print_abbreviation        = gallina_print_abbreviation;
-  print_module              = gallina_print_module;
-  print_modtype             = gallina_print_modtype;
-  print_named_decl          = gallina_print_named_decl;
-  print_library_leaf        = gallina_print_library_leaf;
-  print_context             = gallina_print_context;
-  print_typed_value_in_env  = gallina_print_typed_value_in_env;
-  print_eval                = gallina_print_eval;
-}
-
-let object_pr = ref default_object_pr
-let set_object_pr = (:=) object_pr
-
-let print_inductive x = !object_pr.print_inductive x
-let print_constant_with_infos c = !object_pr.print_constant_with_infos c
-let print_section_variable c = !object_pr.print_section_variable c
-let print_abbreviation x = !object_pr.print_abbreviation x
-let print_module x  = !object_pr.print_module  x
-let print_modtype x = !object_pr.print_modtype x
-let print_named_decl x = !object_pr.print_named_decl x
-let print_library_leaf x = !object_pr.print_library_leaf x
-let print_context x = !object_pr.print_context x
-let print_typed_value_in_env x = !object_pr.print_typed_value_in_env x
-let print_eval x = !object_pr.print_eval x
-
 (******************************************)
 (**** Printing declarations and judgments *)
 (****  Abstract layer                 *****)
@@ -808,9 +768,9 @@ let print_eval x = !object_pr.print_eval x
 let print_judgment env sigma {uj_val=trm;uj_type=typ} =
   print_typed_value_in_env env sigma (trm, typ)
 
-let print_safe_judgment env sigma j =
-  let trm = Safe_typing.j_val j in
-  let typ = Safe_typing.j_type j in
+let print_safe_judgment {Safe_typing.jdg_env=senv; jdg_val=trm; jdg_type=typ} =
+  let env = Safe_typing.env_of_safe_env senv in
+  let sigma = Evd.from_env env in
   let trm = EConstr.of_constr trm in
   let typ = EConstr.of_constr typ in
   print_typed_value_in_env env sigma (trm, typ)
@@ -829,7 +789,7 @@ let handleF h (Libobject.Dyn.Dyn (tag, o)) = match DynHandleF.find tag h with
 | f -> f o
 | exception Not_found -> mt ()
 
-(* TODO: see the comment for {!gallina_print_leaf_entry} *)
+(* TODO: see the comment for {!print_leaf_entry} *)
 let print_full_pure_atomic env sigma mp lobj =
   let handler =
     DynHandleF.add Declare.Internal.Constant.tag begin fun (id,_) ->
@@ -860,7 +820,7 @@ let print_full_pure_atomic env sigma mp lobj =
       let kn = KerName.make mp (Label.of_id id) in
       let mind = Global.mind_of_delta_kn kn in
       let mib = Global.lookup_mind mind in
-      pr_mutual_inductive_body (Global.env()) mind mib None ++
+      Printmod.pr_mutual_inductive_body (Global.env()) mind mib None ++
       str "." ++ fnl () ++ fnl ()
     end @@
     DynHandleF.empty
@@ -938,7 +898,7 @@ let print_any_name env sigma na udecl =
   try  (* Var locale de but, pas var de section... donc pas d'implicits *)
     let dir,str = repr_qualid qid in
     if not (DirPath.is_empty dir) then raise Not_found;
-    str |> Global.lookup_named |> print_named_decl env sigma
+    str |> Global.lookup_named |> print_named_decl_gen env sigma
 
   with Not_found -> user_err ?loc:qid.loc (pr_qualid qid ++ spc () ++ str "not a defined object.")
 
@@ -986,37 +946,21 @@ let print_notation env sigma entry raw_ntn =
   let ntn = (entry, interp_ntn) in
   try
     let lvl = Notation.level_of_notation ntn in
-    let args = Notgram_ops.subentries_of_notation ntn in
+    let args = Notgram_ops.non_terminals_of_notation ntn in
     let pplvl = Metasyntax.pr_level ntn lvl args in
     Pp.(str "Notation \"" ++ str interp_ntn ++ str "\"" ++ spc ()
       ++ pplvl ++ pr_comma () ++ print_notation_grammar env sigma ntn
       ++ str ".")
   with Not_found -> error_print_notation_not_found entry raw_ntn
 
-let print_opaque_name env sigma qid =
-  let open GlobRef in
-  match Nametab.global qid with
-    | ConstRef cst ->
-      let cb = Global.lookup_constant cst in
-      if Declareops.constant_has_body cb then
-        print_constant_with_infos cst None
-      else
-        user_err Pp.(str "Not a defined constant.")
-    | IndRef (sp,_) ->
-      print_inductive sp None
-    | ConstructRef cstr as gr ->
-      let ty, ctx = Typeops.type_of_global_in_context env gr in
-      let ty = EConstr.of_constr ty in
-      let open EConstr in
-      print_typed_value_in_env env sigma (mkConstruct cstr, ty)
-    | VarRef id ->
-      env |> lookup_named id |> print_named_decl env sigma
-
 let print_about_any ?loc env sigma k udecl =
   maybe_error_reject_univ_decl k udecl;
   match k with
   | Term ref ->
-    let rb = Reductionops.ReductionBehaviour.print ref in
+    let rb = match ref with
+      | ConstRef ref -> Reductionops.ReductionBehaviour.print ref
+      | _ -> mt()
+    in
     Dumpglob.add_glob ?loc ref;
       pr_infos_list
        (print_ref false ref udecl :: blankline ::
@@ -1027,9 +971,6 @@ let print_about_any ?loc env sigma k udecl =
   print_bidi_hints ref @
   [hov 0 (str "Expands to: " ++ pr_located_qualid k)])
   | Abbreviation kn ->
-    let () = match Abbreviation.search_abbreviation kn with
-    | [],Notation_term.NRef (ref,_) -> Dumpglob.add_glob ?loc ref
-    | _ -> () in
       v 0 (
       print_abbreviation env kn ++ fnl () ++
       hov 0 (str "Expands to: " ++ pr_located_qualid k))
@@ -1089,12 +1030,12 @@ let print_path_between cls clt =
 let print_canonical_projections env sigma grefs =
   let open Structures in
   let match_proj_gref { CSTable.projection; value; solution } gr =
-    GlobRef.equal projection gr ||
+    QGlobRef.equal env projection gr ||
     begin match value with
-      | ValuePattern.Const_cs y -> GlobRef.equal y gr
+      | ValuePattern.Const_cs y -> GlobRef.CanOrd.equal y gr
       | _ -> false
     end ||
-    GlobRef.equal solution gr
+    QGlobRef.equal env solution gr
   in
   let projs =
     List.filter (fun p -> List.for_all (match_proj_gref p) grefs)
@@ -1122,7 +1063,7 @@ let print_typeclasses () =
     prlist_with_sep fnl (pr_typeclass env) (typeclasses ())
 
 let pr_instance env i =
-  (*   gallina_print_constant_with_infos i.is_impl *)
+  (*   print_constant_with_infos i.is_impl *)
   (* lighter *)
   print_ref false (instance_impl i) None ++
   begin match hint_priority i with
@@ -1137,6 +1078,5 @@ let print_all_instances () =
 
 let print_instances r =
   let env = Global.env () in
-  let sigma = Evd.from_env env in
-  let inst = instances env sigma r in
-    prlist_with_sep fnl (pr_instance env) inst
+  let inst = instances_exn env (Evd.from_env env) r in
+  prlist_with_sep fnl (pr_instance env) inst
