@@ -1,7 +1,7 @@
 (************************************************************************)
 (*         *   The Coq Proof Assistant / The Coq Development Team       *)
-(*  v      *   INRIA, CNRS and contributors - Copyright 1999-2019       *)
-(* <O___,, *       (see CREDITS file for the list of authors)           *)
+(*  v      *         Copyright INRIA, CNRS and contributors             *)
+(* <O___,, * (see version control and CREDITS file for authors & dates) *)
 (*   \VV/  **************************************************************)
 (*    //   *    This file is distributed under the terms of the         *)
 (*         *     GNU Lesser General Public License Version 2.1          *)
@@ -109,7 +109,7 @@ module Tag = struct
 
 end
 
-let msgnl_with ?pre_hdr fmt strm =
+let msgnl_with fmt strm =
   pp_with fmt (strm ++ fnl ());
   Format.pp_print_flush fmt ()
 
@@ -168,6 +168,7 @@ let default_tag_map () = let open Terminal in [
     "message.error"    , make ~bold:true ~fg_color:`WHITE ~bg_color:`RED ()
   ; "message.warning"  , make ~bold:true ~fg_color:`WHITE ~bg_color:`YELLOW ()
   ; "message.debug"    , make ~bold:true ~fg_color:`WHITE ~bg_color:`MAGENTA ()
+  ; "message.prompt"   , make ~fg_color:`GREEN ()
   (* Coming from the printer *)
   ; "constr.evar"      , make            ~fg_color:`LIGHT_BLUE ()
   ; "constr.keyword"   , make ~bold:true ()
@@ -240,50 +241,58 @@ let make_style_stack () =
   | []      -> default_style  (* Anomalous case, but for robustness *)
   | st :: _ -> st
   in
-  let open_tag tag =
-    let (tpfx, ttag) = split_tag tag in
-    if tpfx = end_pfx then "" else
-      let style = get_style ttag in
-      (* Merge the current settings and the style being pushed.  This
-         allows restoring the previous settings correctly in a pop
-         when both set the same attribute.  Example: current settings
-         have red FG, the pushed style has green FG.  When popping the
-         style, we should set red FG, not default FG. *)
-    let style = Terminal.merge (peek ()) style in
-    let diff = Terminal.diff (peek ()) style in
-    style_stack := style :: !style_stack;
-    if tpfx = start_pfx then diff_tag_stack := ttag :: !diff_tag_stack;
-    Terminal.eval diff
+  let open_tag = function
+    | Format.String_tag tag ->
+      let (tpfx, ttag) = split_tag tag in
+      if tpfx = end_pfx then "" else
+        let style = get_style ttag in
+        (* Merge the current settings and the style being pushed.  This
+           allows restoring the previous settings correctly in a pop
+           when both set the same attribute.  Example: current settings
+           have red FG, the pushed style has green FG.  When popping the
+           style, we should set red FG, not default FG. *)
+        let style = Terminal.merge (peek ()) style in
+        let diff = Terminal.diff (peek ()) style in
+        style_stack := style :: !style_stack;
+        if tpfx = start_pfx then diff_tag_stack := ttag :: !diff_tag_stack;
+        Terminal.eval diff
+    | _ -> Terminal.eval default_style
   in
-  let close_tag tag =
-    let (tpfx, _) = split_tag tag in
+  let close_tag = function
+    | Format.String_tag tag ->
+      let (tpfx, _) = split_tag tag in
       if tpfx = start_pfx then "" else begin
         if tpfx = end_pfx then diff_tag_stack := (try List.tl !diff_tag_stack with tl -> []);
         match !style_stack with
         | []       -> (* Something went wrong, we fallback *)
-                      Terminal.eval default_style
+          Terminal.eval default_style
         | cur :: rem -> style_stack := rem;
-                      if cur = (peek ()) then "" else
-                        if rem = [] then Terminal.reset else
-                          Terminal.eval (Terminal.diff cur (peek ()))
+          if cur = (peek ()) then "" else
+          if rem = [] then Terminal.reset else
+            Terminal.eval (Terminal.diff cur (peek ()))
       end
+    | _ -> Terminal.eval default_style
   in
   let clear () = style_stack := [] in
   open_tag, close_tag, clear
 
 let make_printing_functions () =
-  let print_prefix ft tag =
-    let (tpfx, ttag) = split_tag tag in
-    if tpfx <> end_pfx then
-      let style = get_style ttag in
-      match style.Terminal.prefix with Some s -> Format.pp_print_as ft 0 s | None -> () in
-
-  let print_suffix ft tag =
-    let (tpfx, ttag) = split_tag tag in
-    if tpfx <> start_pfx then
-      let style = get_style ttag in
-      match style.Terminal.suffix with Some s -> Format.pp_print_as ft 0 s | None -> () in
-
+  let print_prefix ft = function
+    | Format.String_tag tag ->
+      let (tpfx, ttag) = split_tag tag in
+      if tpfx <> end_pfx then
+        let style = get_style ttag in
+        (match style.Terminal.prefix with Some s -> Format.pp_print_as ft 0 s | None -> ())
+    | _ -> ()
+  in
+  let print_suffix ft = function
+    | Format.String_tag tag ->
+      let (tpfx, ttag) = split_tag tag in
+      if tpfx <> start_pfx then
+        let style = get_style ttag in
+        (match style.Terminal.suffix with Some s -> Format.pp_print_as ft 0 s | None -> ())
+    | _ -> ()
+  in
   print_prefix, print_suffix
 
 let init_output_fns () =
@@ -310,10 +319,10 @@ let init_terminal_output ~color =
   let open_tag, close_tag, clear_tag = make_style_stack () in
   let print_prefix, print_suffix = make_printing_functions () in
   let tag_handler ft = {
-    Format.mark_open_tag   = open_tag;
-    Format.mark_close_tag  = close_tag;
-    Format.print_open_tag  = print_prefix ft;
-    Format.print_close_tag = print_suffix ft;
+    Format.mark_open_stag   = open_tag;
+    Format.mark_close_stag  = close_tag;
+    Format.print_open_stag  = print_prefix ft;
+    Format.print_close_stag = print_suffix ft;
   } in
   if color then
     (* Use 0-length markers *)
@@ -329,8 +338,8 @@ let init_terminal_output ~color =
       Format.pp_set_print_tags !std_ft true;
       Format.pp_set_print_tags !err_ft true
     end;
-  Format.pp_set_formatter_tag_functions !std_ft (tag_handler !std_ft) [@warning "-3"];
-  Format.pp_set_formatter_tag_functions !err_ft (tag_handler !err_ft) [@warning "-3"]
+  Format.pp_set_formatter_stag_functions !std_ft (tag_handler !std_ft);
+  Format.pp_set_formatter_stag_functions !err_ft (tag_handler !err_ft)
 
 (* Rules for emacs:
    - Debug/info: emacs_quote_info
@@ -359,21 +368,12 @@ let in_phase ~phase f x =
     default_phase := op;
     res
   with exn ->
-    let iexn = Backtrace.add_backtrace exn in
+    let iexn = Exninfo.capture exn in
     default_phase := op;
-    Util.iraise iexn
+    Exninfo.iraise iexn
 
 let pr_loc loc =
-    let fname = loc.Loc.fname in
-    match fname with
-    | Loc.ToplevelInput ->
-      Loc.(str"Toplevel input, characters " ++ int loc.bp ++
-           str"-" ++ int loc.ep ++ str":")
-    | Loc.InFile fname ->
-      Loc.(str"File " ++ str "\"" ++ str fname ++ str "\"" ++
-           str", line " ++ int loc.line_nb ++ str", characters " ++
-           int (loc.bp-loc.bol_pos) ++ str"-" ++ int (loc.ep-loc.bol_pos) ++
-           str":")
+  Loc.pr loc ++ str ":"
 
 let pr_phase ?loc () =
   match !default_phase, loc with
@@ -394,7 +394,7 @@ let pr_phase ?loc () =
      None
 
 let print_err_exn any =
-  let (e, info) = CErrors.push any in
+  let (e, info) = Exninfo.capture any in
   let loc = Loc.get_loc info in
   let pre_hdr = pr_phase ?loc () in
   let msg = CErrors.iprint (e, info) ++ fnl () in
@@ -404,6 +404,7 @@ let with_output_to_file fname func input =
   let channel = open_out (String.concat "." [fname; "out"]) in
   let old_fmt = !std_ft, !err_ft, !deep_ft in
   let new_ft = Format.formatter_of_out_channel channel in
+  set_gp new_ft (get_gp !std_ft);
   std_ft := new_ft;
   err_ft := new_ft;
   deep_ft := new_ft;
@@ -412,13 +413,15 @@ let with_output_to_file fname func input =
     std_ft := Util.pi1 old_fmt;
     err_ft := Util.pi2 old_fmt;
     deep_ft := Util.pi3 old_fmt;
+    Format.pp_print_flush new_ft ();
     close_out channel;
     output
   with reraise ->
-    let reraise = Backtrace.add_backtrace reraise in
+    let reraise = Exninfo.capture reraise in
     std_ft := Util.pi1 old_fmt;
     err_ft := Util.pi2 old_fmt;
     deep_ft := Util.pi3 old_fmt;
+    Format.pp_print_flush new_ft ();
     close_out channel;
     Exninfo.iraise reraise
 
